using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace Blu
{
    public class Unicode
    {
        public Lang[] langs = new Lang[8];

        public Unicode(BinaryReader br)
        {
            int matgOffset400 = Map.tags.tagsList[0].MetaOffset + 400;

            //go through 8 times as there are 8 languages
            for (int x = 0; x < 8; x++)
            {
                //chunk size is 28 bytes
                br.BaseStream.Position = matgOffset400 + (28 * x);

                //store info
                langs[x].language = (Language)x;

                langs[x].stringCount = br.ReadInt32();
                langs[x].tableSize = br.ReadInt32(); //size of all strings along with null terminations

                langs[x].indexPointerOffset = br.BaseStream.Position;
                langs[x].indexOffset = br.ReadInt32();
                langs[x].tableOffset = br.ReadInt32();

                langs[x].IDSizes = new IDSize[langs[x].stringCount];
                langs[x].UniStrings = new UniString[langs[x].stringCount];

                br.BaseStream.Position = langs[x].indexOffset;
                for (int y = 0; y < langs[x].stringCount; y++)
                {
                    langs[x].IDSizes[y] = new IDSize(br.ReadInt16(), br.ReadInt16());
                    langs[x].UniStrings[y] = new UniString(br.ReadInt32());
                }

                br.BaseStream.Position = langs[x].tableOffset;

                for (int y = 0; y < langs[x].stringCount; y++)
                {
                    langs[x].UniStrings[y].Length = y + 1 == langs[x].stringCount ?
                        langs[x].tableSize - langs[x].UniStrings[y].Offset - 1 :
                        langs[x].UniStrings[y + 1].Offset - langs[x].UniStrings[y].Offset - 1;

                    for (int z = 0; z < langs[x].UniStrings[y].Length; z++)
                    {
                        byte b = br.ReadByte();
                        switch (b)
                        {
                            case 226:
                                {
                                    short punc = br.ReadInt16();
                                    if (punc == -26240)
                                    {
                                        langs[x].UniStrings[y].uString += "'";
                                        z += 2;
                                    }
                                    else
                                    {
                                        langs[x].UniStrings[y].uString += (char)b;
                                        br.BaseStream.Position -= 2;
                                    }
                                    break;
                                }
                            case 238: //start of a sumerfin
                                {
                                    short img = br.ReadInt16();
                                    z += 2;
                                    int indx = Array.IndexOf(Globals.uniDwords, img);

                                    if (indx == -1) langs[x].UniStrings[y].uString += "[?]"; // :-D
                                    else langs[x].UniStrings[y].uString += Globals.uniTranslations[indx];

                                    break;
                                }

                            case 13: //newline
                                {
                                    br.BaseStream.Position++;
                                    z++;
                                    langs[x].UniStrings[y].uString += @" \r\n ";// " [ Newline ] ";
                                    break;
                                }

                            default:
                                {
                                    langs[x].UniStrings[y].uString += (char)b;
                                    break;
                                }
                        }
                    }
                    //skip null
                    br.BaseStream.Position++;
                }
            }
        }

        public void ReReadStrings()
        {
            BinaryReader br = new BinaryReader(new FileStream(Map.path, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            //go through 8 times as there are 8 languages
            for (int x = 0; x < 8; x++)
            {
                br.BaseStream.Position = langs[x].tableOffset;

                for (int y = 0; y < langs[x].stringCount; y++)
                {
                    langs[x].UniStrings[y].uString = "";
                    for (int z = 0; z < langs[x].UniStrings[y].Length; z++)
                    {
                        byte b = br.ReadByte();
                        switch (b)
                        {
                            case 238: //start of a sumerfin
                                {
                                    short img = br.ReadInt16();
                                    z += 2;
                                    int indx = Array.IndexOf(Globals.uniDwords, img);

                                    if (indx == -1) langs[x].UniStrings[y].uString += "[?]"; // :-D
                                    else langs[x].UniStrings[y].uString += Globals.uniTranslations[indx];

                                    break;
                                }

                            case 13: //newline
                                {
                                    br.BaseStream.Position++;
                                    z++;
                                    langs[x].UniStrings[y].uString += @" \r\n ";// " [ Newline ] ";
                                    break;
                                }

                            default:
                                {
                                    langs[x].UniStrings[y].uString += (char)b;
                                    break;
                                }
                        }
                    }
                    //skip null
                    br.BaseStream.Position++;
                }
            }
            br.Close();
        }

        public void WriteString(int langIndx, int strIndx, string newText)
        {
            BinaryWriter bw = new BinaryWriter(new FileStream(Map.path, FileMode.Open,
              FileAccess.ReadWrite, FileShare.ReadWrite));

            int strOffset = langs[langIndx].tableOffset + langs[langIndx].UniStrings[strIndx].Offset;
            bw.BaseStream.Position = strOffset;

            MemoryStream ms = new MemoryStream();

            //go throu newtext
            for (int x = 0; x < newText.Length; x++)
            {
                if (newText[x] == '\'')
                {
                    ms.WriteByte(226);
                    ms.WriteByte(128);
                    ms.WriteByte(153);
                }

                bool bool1 = false;
                for (int y = 0; y < Globals.uniTranslations.Length; y++)
                {
                    if (Globals.uniTranslations[y].Length <= newText.Length - x)
                    {
                        if (newText.Substring(x, Globals.uniTranslations[y].Length) == Globals.uniTranslations[y])
                        {
                            ms.WriteByte(238);
                            ms.Write(BitConverter.GetBytes(Globals.uniDwords[y]), 0, 2);

                            bool1 = true;
                            x += Globals.uniTranslations[y].Length - 1;
                            break; //foreach loop
                        }
                    }
                }
                if (!bool1) ms.WriteByte((byte)newText[x]);
            }

            byte[] txtBytes = ms.ToArray();

            //hmm
            //for (int x = 0; x < Globals.uniTranslations.Length; x++)
            //{
            //    newText = newText.Replace(Globals.uniTranslations[x],
            //        BitConverter.ToString(BitConverter.GetBytes(Globals.uniDwords[x])));
            //}

            if (txtBytes.Length == langs[langIndx].UniStrings[strIndx].Length) bw.Write(txtBytes);
            else
            {
                //open in reader
                BinaryReader br = new BinaryReader(bw.BaseStream);

                //goto end of string
                br.BaseStream.Position += langs[langIndx].UniStrings[strIndx].Length;

                //store eof table
                byte[] eofLangTable = br.ReadBytes((langs[langIndx].tableOffset
                    + langs[langIndx].tableSize) - (int)br.BaseStream.Position);

                //next thing after padding of table
                int afterTablePadding = langIndx == langs.Length - 1 ?
                    Map.header.strangeFileStringsOffset : langs[langIndx + 1].indexOffset;

                int oldPaddingLen = afterTablePadding - (int)br.BaseStream.Position;
                br.BaseStream.Position = afterTablePadding;

                //store Map downward from pos
                byte[] mapHalf = br.ReadBytes((int)(br.BaseStream.Length - br.BaseStream.Position));

                //goto start of string
                bw.BaseStream.Position = strOffset;

                //write new text and eof
                bw.Write(txtBytes);
                bw.Write(eofLangTable);

                long langTableEnd = bw.BaseStream.Position;

                //shift
                int shift = txtBytes.Length - langs[langIndx].UniStrings[strIndx].Length;

                //--[String offsets]

                //go 1 after the string in the index
                bw.BaseStream.Position = langs[langIndx].indexOffset + (strIndx * 8) + 8;
                for (int x = strIndx + 1; x < langs[langIndx].stringCount; x++)
                {
                    bw.BaseStream.Position += 4;
                    bw.Write(langs[langIndx].UniStrings[x].Offset += shift);
                }

                //get padding length
                langs[langIndx].tableSize += shift;
                int newPaddingLen = 512 - (langs[langIndx].tableSize % 512);

                bw.BaseStream.Position = langTableEnd;
                if (newPaddingLen == 512) //remainder was 0
                {
                    newPaddingLen = 0;
                    //write stored half
                    bw.Write(mapHalf);
                }
                else //means remainder wasnt 0
                {
                    byte[] padding = new byte[newPaddingLen];
                    bw.Write(padding);
                    bw.Write(mapHalf);
                }

                shift += newPaddingLen - oldPaddingLen;

                //now Map hlf is there, write size
                bw.BaseStream.Position = (Map.tags.tagsList[0].MetaOffset + 400 + shift) + (28 * langIndx) + 4;
                bw.Write(langs[langIndx].tableSize);

                if (shift != 0) //the padding hasnt realigned the rest of the Map.. so we have to
                {
                    //--[Header]
                    bw.BaseStream.Position = 16;
                    bw.Write(Map.header.indexOffset += shift);

                    Map.header.metaStart += shift;

                    br.BaseStream.Position = 344;
                    bw.Write(Map.header.strangeFileStringsOffset += shift);

                    //--[Index]
                    Map.index.primaryMagic -= shift;
                    Map.index.offsetOfObjectIndex += shift;
                    Map.index.secondaryMagic -= shift;

                    //--[Unicode]
                    int matgOffset400 = Map.tags.tagsList[0].MetaOffset + 400 + shift;
                    for (int x = langIndx + 1; x < 8; x++)
                    {
                        bw.BaseStream.Position = matgOffset400 + (28 * x) + 8;
                        langs[x].indexPointerOffset = bw.BaseStream.Position;

                        bw.Write(langs[x].indexOffset += shift);
                        bw.Write(langs[x].tableOffset += shift);
                    }

                    //--[Bitmaps]
                    List<int> bitmTags = new List<int>();
                    for (int x = 0; x < Map.tags.tagsList.Length; x++)
                    {
                        if (Map.tags.tagsList[x].Type == "bitm")
                            bitmTags.Add(x);
                    }

                    for (int z = 0; z < bitmTags.Count; z++)
                    {
                        br.BaseStream.Position = Map.tags.tagsList[bitmTags[z]].MetaOffset + 68 + shift;

                        int tempc = br.ReadInt32();
                        int tempr = br.ReadInt32() - Map.index.secondaryMagic;

                        for (int x = 0; x < tempc; x++)
                        {
                            for (int y = 0; y < 6; y++)
                            {
                                int PointerOffset = tempr + (x * 116) + 28 + (y * 4);
                                br.BaseStream.Position = PointerOffset;

                                int Offset = br.ReadInt32();
                                if (Offset == -1) break;

                                if (Map.IsInternal(Offset))
                                {
                                    bw.BaseStream.Position = PointerOffset;
                                    bw.Write(Offset + shift);
                                }
                            }
                        }
                    }

                    //--[Meta Offsets]
                    for (int x = 0; x < Map.tags.tagsList.Length; x++)
                    {
                        if (Map.tags.tagsList[x].Type != "sbsp" && Map.tags.tagsList[x].Type != "ltmp")
                            Map.tags.tagsList[x].MetaOffset += shift;
                    }

                    //--[End]
                    bw.BaseStream.Position = Map.tags.tagsList[Map.tags.tagsList.Length - 1].MetaOffset +
                        Map.tags.tagsList[Map.tags.tagsList.Length - 1].MetaSize;

                    newPaddingLen = 512 - ((int)bw.BaseStream.Position % 512);
                    newPaddingLen += newPaddingLen == 512 ? 512 : 1024;

                    int eof = (int)bw.BaseStream.Position + newPaddingLen;

                    while (bw.BaseStream.Position < eof)
                        bw.Write((byte)205); //CD

                    bw.BaseStream.SetLength(eof);

                    //--[Header]
                    bw.BaseStream.Position = 8;
                    bw.Write(eof);
                }
                br.Close();
            }

            bw.Close();

            langs[langIndx].UniStrings[strIndx].uString = newText;
            langs[langIndx].UniStrings[strIndx].Length = txtBytes.Length;
            //Map.rawData.rdh = ud; //not needed
        }
    }

    public enum Language
    {
        English,
        Japanese,
        German,
        French,
        Spanish,
        Italian,
        Korean,
        Chinese
    }

    public struct DwordTranslation
    {
        public short Dword;
        public string Translation;

        public DwordTranslation(short Dword, string Translation)
        {
            this.Dword = Dword;
            this.Translation = Translation;
        }
    }

    public struct IDSize
    {
        public int ID;
        //public byte SizeMayb;

        //public IDSize(int iID, byte nufin, byte iSize)
        public IDSize(int iID, short unk)
        {
            this.ID = iID;
            //this.SizeMayb = iSize;
        }
    }

    public struct UniString
    {
        public int Offset;
        public string uString;
        public int Length;

        public UniString(int iOffset)
        {
            this.Offset = iOffset;
            this.uString = "";
            this.Length = 0;
        }
    }

    public struct Lang
    {
        public Language language;

        public int
        stringCount,
        tableSize,
        indexOffset,
        tableOffset;

        public long
        indexPointerOffset,
        tablePointerOffset;

        public IDSize[] IDSizes;
        public UniString[] UniStrings;
    }
}
