using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace Menu_Menu_Menu
{
    public partial class Form1 : Form
    {
        public Map map = null;
        public MainMenu menu = null;

        public Form1()
        {
            InitializeComponent();
            map = new Map(pb1, lab1);
            menu = new MainMenu(map);
        }

        #region Toolbar
        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            o.FileName = "";
            if (o.ShowDialog() == DialogResult.Cancel) return;
            map.Load(o.FileName);

            map.UpdateProgress("Displaying Map Info...", 6);

            menu.Load();
            RefreshForm();
            this.Text = "Menu Menu Menu - " + Path.GetFileName(o.FileName);

            map.EndProgress();

            if (!menu.Loaded)
            {
                saveChangesToolStripMenuItem.Enabled = false;
                lv1.ContextMenuStrip = null;
                MessageBox.Show("Invalid Mainmenu.", "",
                MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
            else
            {
                saveChangesToolStripMenuItem.Enabled = true;
                lv1.ContextMenuStrip = contextMenuStrip2;
            }

            //temp
            //int mPad = 512 - (map.header.metaSize % 512);
            //int mapPad = 512 - (map.header.filesize % 512);
            //this.Text = map.header.metaStart.ToString() + ";; " + map.header.metaSize.ToString();
            //mPad.ToString() + " :: " + mapPad.ToString();
        }

        private void resignToolStripMenuItem_Click(object sender, EventArgs e)
        {
            FileStream fs = new FileStream(map.path,
                FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite);

            BinaryReader br = new BinaryReader(fs);
            BinaryWriter bw = new BinaryWriter(fs);

            int sig = 0;
            long length = br.BaseStream.Length;

            br.BaseStream.Position = 2048;

            for (long i = 2048; i < length; i += 4)
                sig ^= br.ReadInt32();

            br.BaseStream.Position = 720;
            bw.Write(sig);

            br.Close();
            bw.Close();

            map.header.checksum = sig;
            lab1.Text = "Done! New Signature is 0x" + sig.ToString("X");
        }

        private void saveChangesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            menu.SaveAllChanges();
            lab1.Text = "Changes Saved";
        }
        #endregion

        #region Misc
        public void RefreshForm()
        {
            menu.Ready = false;
            panel1.Enabled = false;
            radioButton4.Enabled = radioButton3.Enabled = false;
            resignToolStripMenuItem.Enabled = false;
            extrasToolStripMenuItem.Enabled = false;

            imageList1.Images.Clear();
            lv1.Items.Clear();

            prev.Image = null;
            radioButton1.Checked = radioButton2.Checked = false;

            textBox5.Text = textBox6.Text =
                textBox7.Text = textBox1.Text =
                textBox2.Text = textBox3.Text =
                textBox4.Text = "";

            comboBox1.SelectedItem = comboBox2.SelectedItem = null;

            if (menu.Loaded)
            {
                for (int x = 0; x < menu.Slots; x++)
                {
                    imageList1.Images.Add(menu.mapSlots[x].Img);
                    lv1.Items.Add(menu.mapSlots[x].mapName, x);
                }

                menu.Ready = true;
                resignToolStripMenuItem.Enabled = true;
                extrasToolStripMenuItem.Enabled = true;
                radioButton4.Enabled = radioButton3.Enabled = true;
            }
        }

        private void radioButton4_CheckedChanged(object sender, EventArgs e)
        {
            menu.mapFilter = radioButton4.Checked ? 0 : 1;
            menu.Load();
            RefreshForm();
        }
        #endregion

        #region Extras
        long bluFilterOffset;
        long camMoveOffset;
        long noWeapOffset;
        long textMarqueeOffset1;
        long textMarqueeOffset2;

        private void extrasToolStripMenuItem_DropDownOpening(object sender, EventArgs e)
        {
            BinaryReader br = new BinaryReader(new FileStream(map.path, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            //blue filter
            br.BaseStream.Position = map.tags.tagsList[3].MetaOffset + 980;
            bluFilterOffset = br.ReadInt32() - map.index.secondaryMagic;
            br.BaseStream.Position = bluFilterOffset + 20;
            bluFilterTool.Checked = (br.ReadInt32() != -1);

            //cam move
            camMoveOffset = map.tags.tagsList[3].MetaOffset + 440;
            br.BaseStream.Position = camMoveOffset;
            camMoveTool.Checked = (br.ReadInt32() != 1);

            //'No Weapon' Quick Option
            int silyStartWeapIndx = map.tags.FindTagIndexByID(-432143330);
            noWeapOffset = map.tags.tagsList[silyStartWeapIndx].MetaOffset + 32;
            br.BaseStream.Position = noWeapOffset;
            NoWeapTool.Checked = (br.ReadInt32() != -2145249436);

            //Mainmenu Text Marquee Speed
            br.BaseStream.Position = map.tags.tagsList[map.tags.FindTagIndexByID(-450231542)].MetaOffset + 36;
            br.BaseStream.Position = (br.ReadInt32() - map.index.secondaryMagic) + 40;
            textMarqueeOffset1 = (br.ReadInt32() - map.index.secondaryMagic) + 16;
            br.BaseStream.Position = textMarqueeOffset1;
            toolStripTextBox1.Text = br.ReadInt32().ToString();

            br.BaseStream.Position = map.tags.tagsList[map.tags.FindTagIndexByID(-442629250)].MetaOffset + 36;
            br.BaseStream.Position = (br.ReadInt32() - map.index.secondaryMagic) + 40;
            textMarqueeOffset2 = (br.ReadInt32() - map.index.secondaryMagic) + 16;

            br.Close();
        }

        private void bluFilterTool_Click(object sender, EventArgs e)
        {
            BinaryWriter bw = new BinaryWriter(new FileStream(map.path, FileMode.Open,
               FileAccess.ReadWrite, FileShare.ReadWrite));

            if (bluFilterTool.Checked) //remove it
            {
                bw.BaseStream.Position = bluFilterOffset + 20;
                bw.Write(-1);
                bw.BaseStream.Position = bluFilterOffset + 36 + 20;
                bw.Write(-1);
                lab1.Text = "Blue Filter Removed";
            }
            else //add it
            {
                bw.BaseStream.Position = bluFilterOffset + 16;
                bw.Write("roge".ToCharArray());
                bw.Write(-466484718);
                bw.BaseStream.Position = bluFilterOffset + 36 + 16;
                bw.Write("roge".ToCharArray());
                bw.Write(-466091496);
                lab1.Text = "Blue Filter Added";
            }

            bw.Close();
        }

        private void camMoveTool_Click(object sender, EventArgs e)
        {
            BinaryWriter bw = new BinaryWriter(new FileStream(map.path, FileMode.Open,
              FileAccess.ReadWrite, FileShare.ReadWrite));

            if (camMoveTool.Checked) //remove it
            {
                bw.BaseStream.Position = camMoveOffset;
                bw.Write(1);
                lab1.Text = "Moving Camera Removed";
            }
            else //add it
            {
                bw.BaseStream.Position = camMoveOffset;
                bw.Write(2);
                lab1.Text = "Moving Camera Added";
            }

            bw.Close();
        }

        private void NoWeapTool_Click(object sender, EventArgs e)
        {
            BinaryWriter bw = new BinaryWriter(new FileStream(map.path, FileMode.Open,
              FileAccess.ReadWrite, FileShare.ReadWrite));

            if (NoWeapTool.Checked) //remove it
            {
                bw.BaseStream.Position = noWeapOffset;
                bw.Write(-2145249436);
                lab1.Text = "'No Weapon' Quick Option Removed";
            }
            else //add it
            {
                bw.BaseStream.Position = noWeapOffset;
                bw.Write(-2145248732);
                lab1.Text = "'No Weapon' Quick Option Added";
            }

            bw.Close();
        }

        private void toolStripTextBox1_TextChanged(object sender, EventArgs e)
        {
            int tmp;
            if (!int.TryParse(toolStripTextBox1.Text, out tmp))
            {
                toolStripTextBox1.Text = "0";
                toolStripTextBox1.SelectionStart = 1;
            }
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            BinaryWriter bw = new BinaryWriter(new FileStream(map.path, FileMode.Open,
              FileAccess.ReadWrite, FileShare.ReadWrite));

            bw.BaseStream.Position = textMarqueeOffset1;
            bw.Write(int.Parse(toolStripTextBox1.Text));

            bw.BaseStream.Position = textMarqueeOffset2;
            bw.Write(int.Parse(toolStripTextBox1.Text));

            lab1.Text = "Mainmenu Text Marquee Speed Set To " + toolStripTextBox1.Text;

            bw.Close();

        }
        #endregion

        #region Listview
        private void lv1_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (menu.Ready && lv1.FocusedItem == null)
            {
                prev.Image = null;
                radioButton1.Checked = radioButton2.Checked = false;

                textBox5.Text = textBox6.Text =
                    textBox7.Text = textBox1.Text =
                    textBox2.Text = textBox3.Text =
                    textBox4.Text = "";

                lab1.Text = "Idle";
            }
            else
            {
                prev.Image = menu.mapSlots[lv1.FocusedItem.Index].Img;
                textBox5.Text = menu.mapSlots[lv1.FocusedItem.Index].mapName;
                textBox6.Text = menu.mapSlots[lv1.FocusedItem.Index].Scenario;
                textBox7.Text = menu.mapSlots[lv1.FocusedItem.Index].mapDescription;

                comboBox1_SelectedIndexChanged(null, null);

                textBox3.Tag = menu.mapSlots[lv1.FocusedItem.Index].Start.ToString();
                textBox3.Text = menu.mapSlots[lv1.FocusedItem.Index].Start.ToString();

                textBox4.Tag = menu.mapSlots[lv1.FocusedItem.Index].End.ToString();
                textBox4.Text = menu.mapSlots[lv1.FocusedItem.Index].End.ToString();

                switch (menu.mapSlots[lv1.FocusedItem.Index].Locked)
                {
                    case 0:
                        radioButton2.Checked = true;
                        break;

                    case 1:
                        radioButton1.Checked = true;
                        break;
                }

                switch (menu.mapSlots[lv1.FocusedItem.Index].TeamOptions)
                {
                    case 0:
                        comboBox2.SelectedIndex = 0;
                        break;

                    case 134742528:
                        comboBox2.SelectedIndex = 1;
                        break;

                    case 134743040:
                        comboBox2.SelectedIndex = 2;
                        break;

                    case 134742536:
                        comboBox2.SelectedIndex = 3;
                        break;

                    default:
                        comboBox2.SelectedItem = null;
                        break;
                }

                if (menu.mapFilter == 0)
                {
                    groupBox2.Enabled =
                    label6.Enabled = textBox6.Enabled =
                    label9.Enabled = comboBox2.Enabled = false;

                }
                else
                {
                    groupBox2.Enabled =
                    label6.Enabled = textBox6.Enabled =
                    label9.Enabled = comboBox2.Enabled = true;
                }

                panel1.Enabled = true;
                lab1.Text = "Selected Map Slot: " + lv1.FocusedItem.Index.ToString();
            }
        }

        private void contextMenuStrip2_Opening(object sender, CancelEventArgs e)
        {
            if (lv1.FocusedItem == null)
            {
                dupTool.Text = "Duplicate...";
                delTool.Text = "Delete...";
                ReadMapFileTool.Enabled = clearSlotTool.Enabled =
                    dupTool.Enabled = delTool.Enabled = false;
            }
            else
            {
                dupTool.Text = "Duplicate " + lv1.FocusedItem.Text;
                delTool.Text = "Delete " + lv1.FocusedItem.Text;
                ReadMapFileTool.Enabled = clearSlotTool.Enabled =
                    dupTool.Enabled = delTool.Enabled = true;
            }
        }

        private void ReadMapFileTool_Click(object sender, EventArgs e)
        {
            o.FileName = "";
            if (o.ShowDialog() == DialogResult.Cancel) return;

            BinaryReader br = new BinaryReader(new FileStream(o.FileName, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            Encoding en = Encoding.UTF8;

            //map internal
            br.BaseStream.Position = 408;
            string internalName = en.GetString(br.ReadBytes(32)).Split('\0')[0];

            //langs
            for (int x = 0; x < menu.mapSlots[lv1.FocusedItem.Index].Names.Length; x++)
                menu.mapSlots[lv1.FocusedItem.Index].Names[x] = internalName;

            //mm internal
            menu.mapSlots[lv1.FocusedItem.Index].mapName = internalName;

            if (menu.mapFilter == 1) //mp
            {
                //scenario
                br.BaseStream.Position = 444;
                string scenarioName = en.GetString(br.ReadBytes(256)).Split('\0')[0];

                menu.mapSlots[lv1.FocusedItem.Index].Scenario = scenarioName;
            }

            br.Close();

            lv1.FocusedItem.Text = internalName;
            lv1_SelectedIndexChanged(null, null);
            menu.SaveMapSlot(lv1.FocusedItem.Index);
        }

        private void clearSlotTool_Click(object sender, EventArgs e)
        {
            if (menu.mapFilter == 1) //mp
                menu.mapSlots[lv1.FocusedItem.Index].Locked = 1;

            //langs
            for (int x = 0; x < menu.mapSlots[lv1.FocusedItem.Index].Names.Length; x++)
            {
                menu.mapSlots[lv1.FocusedItem.Index].Names[x] = "";
                menu.mapSlots[lv1.FocusedItem.Index].Descriptions[x] = "";
            }

            //internal
            menu.mapSlots[lv1.FocusedItem.Index].mapName = "";

            if (menu.mapFilter == 1) //mp
                menu.mapSlots[lv1.FocusedItem.Index].Scenario = "";

            menu.mapSlots[lv1.FocusedItem.Index].mapDescription = "";

            lv1.FocusedItem.Text = "";
            lv1_SelectedIndexChanged(null, null);
            menu.SaveMapSlot(lv1.FocusedItem.Index);
        }

        private void dupTool_Click(object sender, EventArgs e)
        {
            menu.DuplicateMapSlot(lv1.FocusedItem.Index);
            menu.Load();
            RefreshForm();
        }

        private void delTool_Click(object sender, EventArgs e)
        {
            menu.DeleteMapSlot(lv1.FocusedItem.Index);
            menu.Load();
            RefreshForm();
        }
        #endregion

        #region Image

        private void exportToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SaveFileDialog sf = new SaveFileDialog();

            sf.FileName = (menu.mapSlots[lv1.FocusedItem.Index].mapName == String.Empty ? "null"
                : menu.mapSlots[lv1.FocusedItem.Index].mapName.ToLower().Replace(' ', '_')) + "_preview";

            sf.Filter = "DDS Files (*.dds)|*.dds";
            if (sf.ShowDialog() == DialogResult.Cancel) return;

            map.bitm.dds.ExtractDDS(menu.mapSlots[lv1.FocusedItem.Index].bi,
                menu.mapSlots[lv1.FocusedItem.Index].ImgBuffer, sf.FileName);
        }

        private void importToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFileDialog of = new OpenFileDialog();

            of.Filter = "DDS Files (*.dds)|*.dds";
            if (of.ShowDialog() == DialogResult.Cancel) return;

            if (map.bitm.dds.InjectDDS(menu.mapSlots[lv1.FocusedItem.Index].bi,
                of.FileName, menu, lv1.FocusedItem.Index))
            {
                BinaryReader br = new BinaryReader(new FileStream(map.path, FileMode.Open,
                    FileAccess.ReadWrite, FileShare.ReadWrite));

                imageList1.Images.Clear();

                for (int x = 0; x < menu.Slots; x++)
                {
                    menu.readMapSlotImage(br, x);
                    imageList1.Images.Add(menu.mapSlots[x].Img);
                }

                prev.Image = menu.mapSlots[lv1.FocusedItem.Index].Img;
                br.Close();
            }
        }

        private void selectDifferentBitmTagToolStripMenuItem_Click(object sender, EventArgs e)
        {
            new Form2(this, lv1.FocusedItem.Index, map.tags.tagsList[
                map.tags.FindTagIndexByID(menu.mapSlots[lv1.FocusedItem.Index].TagID)
                ].Path).ShowDialog();
        }

        private void createNewRawForBitmapToolStripMenuItem_Click(object sender, EventArgs e)
        {
            //open map in reader and writer
            FileStream fs = new FileStream(map.path, FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite);
            BinaryReader br = new BinaryReader(fs);
            BinaryWriter bw = new BinaryWriter(fs);

            //goto index in map (stright after bitmaps)
            br.BaseStream.Position = map.header.indexOffset;

            //store map downward from pos
            byte[] bottomHalf = br.ReadBytes((int)(br.BaseStream.Length - br.BaseStream.Position));

            //goto first reflexive of bitm tag
            br.BaseStream.Position = map.tags.tagsList[map.tags.FindTagIndexByID(
                menu.mapSlots[lv1.FocusedItem.Index].TagID)].MetaOffset + 68;

            int tempc = br.ReadInt32(); //chunk cnt
            int tempr = br.ReadInt32() - map.index.secondaryMagic; //translation

            //go back to index offset
            bw.BaseStream.Position = map.header.indexOffset;

            List<long> Pointers = new List<long>();
            List<int> NewOffsets = new List<int>();

            //foreach chunk
            for (int x = 0; x < tempc; x++)
            {
                //foreach LOD
                for (int y = 0; y < 6; y++)
                {
                    int newbOffset = (int)bw.BaseStream.Position;

                    //long PointerOffset = tempr + (x * 116) + 28 + (y * 4);
                    br.BaseStream.Position = tempr + (x * 116) + 28 + (y * 4);

                    int bOffset = br.ReadInt32();
                    if (bOffset == -1)
                    {
                        bw.BaseStream.Position = newbOffset;
                        break;
                    }

                    Pointers.Add(br.BaseStream.Position - 4);
                    NewOffsets.Add(newbOffset);

                    br.BaseStream.Position += 20;
                    int bSize = br.ReadInt32();

                    //goto bitmap in map
                    br.BaseStream.Position = bOffset;
                    byte[] bArray = br.ReadBytes(bSize);

                    bw.BaseStream.Position = newbOffset;
                    bw.Write(bArray);

                    //write bitmap padding
                    byte[] padding = new byte[512 - (bSize % 512)];
                    if (padding.Length != 512) bw.Write(padding);
                }
            }

            //now we need to shift
            int shift = (int)bw.BaseStream.Position - map.header.indexOffset;

            map.header.indexOffset = (int)bw.BaseStream.Position;
            bw.Write(bottomHalf);

            //--[Header]
            bw.BaseStream.Position = 16;
            bw.Write(map.header.indexOffset);

            //fix offsets
            map.header.metaStart += shift;

            map.index.primaryMagic -= shift;
            map.index.offsetOfObjectIndex += shift;
            map.index.secondaryMagic -= shift;

            //foreach chunk
            for (int x = 0; x < Pointers.Count; x++)
            {
                bw.BaseStream.Position = (Pointers[x] + shift);
                bw.Write(NewOffsets[x]);
            }
            menu.mapSlots[lv1.FocusedItem.Index].ImgOffset = NewOffsets[0];

            //fix meta offsets
            for (int x = 0; x < map.tags.tagsList.Length; x++)
            {
                if (map.tags.tagsList[x].Type != "sbsp" && map.tags.tagsList[x].Type != "ltmp")
                    map.tags.tagsList[x].MetaOffset += shift;
            }

            //--[End]
            bw.BaseStream.Position = map.tags.tagsList[map.tags.tagsList.Length - 1].MetaOffset +
                map.tags.tagsList[map.tags.tagsList.Length - 1].MetaSize;

            int newPaddingLen = 512 - ((int)bw.BaseStream.Position % 512);
            newPaddingLen += newPaddingLen == 512 ? 2560 : 3072;

            int eof = (int)bw.BaseStream.Position + newPaddingLen;

            while (bw.BaseStream.Position < eof)
                bw.Write((byte)205); //CD

            bw.BaseStream.SetLength(eof);

            //--[Header]
            bw.BaseStream.Position = 8;
            bw.Write(eof);

            br.Close();
            bw.Close();
            lab1.Text = "New Bitmap Raw Added";
        }

        #endregion

        #region Editor

        private void radioButton1_CheckedChanged(object sender, EventArgs e)
        {
            if (menu.Ready)
            {
                switch (radioButton1.Checked)
                {
                    case true:
                        menu.mapSlots[lv1.FocusedItem.Index].Locked = 1;
                        break;

                    case false:
                        menu.mapSlots[lv1.FocusedItem.Index].Locked = 0;
                        break;
                }
            }
        }

        private void comboBox1_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (menu.Ready)
            {
                if (comboBox1.SelectedItem == null) comboBox1.SelectedIndex = 0;
                else
                {
                    textBox1.Text = menu.mapSlots[lv1.FocusedItem.Index].Names[comboBox1.SelectedIndex];
                    textBox2.Text = menu.mapSlots[lv1.FocusedItem.Index].Descriptions[comboBox1.SelectedIndex];
                }
            }
        }

        private void textBox1_TextChanged(object sender, EventArgs e)
        {
            if (menu.Ready)
            {
                menu.mapSlots[lv1.FocusedItem.Index].Names
                    [comboBox1.SelectedIndex] = textBox1.Text;
            }
        }

        private void textBox2_TextChanged(object sender, EventArgs e)
        {
            if (menu.Ready)
            {
                menu.mapSlots[lv1.FocusedItem.Index].Descriptions
                    [comboBox1.SelectedIndex] = textBox2.Text;
            }
        }

        private void textBox5_TextChanged(object sender, EventArgs e)
        {
            if (menu.Ready)
            {
                menu.mapSlots[lv1.FocusedItem.Index].mapName = textBox5.Text;
                lv1.FocusedItem.Text = textBox5.Text;
            }
        }

        private void textBox6_TextChanged(object sender, EventArgs e)
        {
            if (menu.Ready)
            {
                menu.mapSlots[lv1.FocusedItem.Index].Scenario = textBox6.Text;
            }
        }

        private void textBox7_TextChanged(object sender, EventArgs e)
        {
            if (menu.Ready)
            {
                menu.mapSlots[lv1.FocusedItem.Index].mapDescription = textBox7.Text;
            }
        }

        private void textBox3_TextChanged(object sender, EventArgs e)
        {
            if (menu.Ready)
            {
                int i;
                bool b = int.TryParse(textBox3.Text, out i);
                if (b)
                {
                    menu.mapSlots[lv1.FocusedItem.Index].Start = i;
                    textBox3.Tag = textBox3.Text;
                }
                else textBox3.Text = textBox3.Tag.ToString();
            }
        }

        private void textBox4_TextChanged(object sender, EventArgs e)
        {
            if (menu.Ready)
            {
                int i;
                bool b = int.TryParse(textBox4.Text, out i);
                if (b)
                {
                    menu.mapSlots[lv1.FocusedItem.Index].End = i;
                    textBox4.Tag = textBox4.Text;
                }
                else textBox4.Text = textBox4.Tag.ToString();
            }
        }

        private void button1_Click(object sender, EventArgs e)
        {
            //find highest value
            int highval = 0;
            for (int x = 0; x < menu.mapSlots.Length; x++)
            {
                if (x == lv1.FocusedItem.Index) continue;
                if (menu.mapSlots[x].Start > highval) highval = menu.mapSlots[x].Start;
                if (menu.mapSlots[x].End > highval) highval = menu.mapSlots[x].End;
            }

            //new val
            highval += 10;

            textBox3.Text = (menu.mapSlots[lv1.FocusedItem.Index].Start = highval).ToString();
            textBox4.Text = (menu.mapSlots[lv1.FocusedItem.Index].End = highval).ToString();

            menu.SaveMapSlot(lv1.FocusedItem.Index);
        }

        private void comboBox2_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (menu.Ready)
            {
                switch (comboBox2.SelectedIndex)
                {
                    case 0:
                        menu.mapSlots[lv1.FocusedItem.Index].TeamOptions = 0;
                        break;

                    case 1:
                        menu.mapSlots[lv1.FocusedItem.Index].TeamOptions = 134742528;
                        break;

                    case 2:
                        menu.mapSlots[lv1.FocusedItem.Index].TeamOptions = 134743040;
                        break;

                    case 3:
                        menu.mapSlots[lv1.FocusedItem.Index].TeamOptions = 134742536;
                        break;
                }
            }
        }

        #endregion
    }
}