using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Windows.Forms;

namespace Menu_Menu_Menu
{
    public struct Tag
    {
        public int Ident;
        public string Type;

        public string Path;
        public int PathOffset;
        public int PathLength;

        public int RawMetaOffset;
        public int MetaOffset;
        public int MetaSize;

        public string MetaExtractPath;
    }

    public enum SortTagsBy
    {
        TagType,
        FolderHierarchy
    }

    public class Tags
    {
        public Map map;

        public Tag[] tagsList;
        public int selectedIndex = -1;

        public SortTagsBy sortTagsBy = SortTagsBy.TagType;
        public List<string> tagTypes = new List<string>();

        public Tags(Map map, BinaryReader br)
        {
            this.map = map;

            br.BaseStream.Position = map.index.offsetOfObjectIndex;

            //will store the tags of the map
            tagsList = new Tag[map.index.objectCount];
            tagTypes.Clear();

            //count through the 'said' tag amount
            for (int x = 0; x < map.index.objectCount; x++)
            {
                //this is the tagtype, they are backward in map so we reverse
                string tt = map.Reverse(new string(br.ReadChars(4)));

                //if its not in the list add it
                if (!tagTypes.Contains(tt)) tagTypes.Add(tt);

                //and this is the rest of the info
                tagsList[x].Type = tt;
                tagsList[x].Ident = br.ReadInt32();
                tagsList[x].RawMetaOffset = br.ReadInt32();
                tagsList[x].MetaOffset = tagsList[x].RawMetaOffset - map.index.secondaryMagic;
                tagsList[x].MetaSize = br.ReadInt32();
            }

            tagTypes.Sort();

            //go to the filetable which contains all the strings - globals/globals etc
            br.BaseStream.Position = (int)map.header.fileTableOffset;

            //go through the tags and add in their paths
            char chr;
            for (int i = 0; i < map.index.objectCount; i++)
            {
                tagsList[i].PathOffset = (int)br.BaseStream.Position;

                while ((chr = br.ReadChar()) != '\0')
                    tagsList[i].Path += chr;

                tagsList[i].PathLength = tagsList[i].Path.Length;
            }
        }

        public int FindTagIndexByID(int ID)
        {
            //go through tags
            for (int x = 0; x < tagsList.Length; x++)
            {
                //if theres an ID match
                if (tagsList[x].Ident == ID)
                    return x; //return the tags index in the list
            }
            return -1; //no matches, so return -1
        }

        public int FindTagIndexByTypeAndPath(string type, string path)
        {
            //go through tags
            for (int x = 0; x < tagsList.Length; x++)
            {
                //if theres a type and path match 
                if (tagsList[x].Type == type && tagsList[x].Path == path)
                    return x; //return the tags index in the list
            }
            return -1; //no matches, so return -1
        }

        public TreeNode[] LoadTagsIntoTreenodes(string tagType)
        {
            if (tagType == String.Empty) // then we load all the tags
            {
                switch (sortTagsBy)
                {
                    case SortTagsBy.TagType:
                        //the list that will contain the tagtypes that will
                        //contain the tags respectively
                        TreeNode[] tnList1 = new TreeNode[tagTypes.Count];

                        //go through tagtypes
                        for (int x = 0; x < tagTypes.Count; x++)
                        {
                            //make a new node for the type
                            tnList1[x] = new TreeNode(tagTypes[x], 1, 1);

                            //go through the tags
                            for (int y = 0; y < tagsList.Length; y++)
                            {
                                //if tag[y]'s type is a match to the current type
                                if (tagsList[y].Type == tagTypes[x])
                                    tnList1[x].Nodes.Add("", tagsList[y].Path, 2, 2); //add it
                            }
                        }
                        return tnList1; //return the now pop. list

                    case SortTagsBy.FolderHierarchy:
                        //for storing the tags
                        List<TreeNode> tnList2 = new List<TreeNode>();

                        //go through the tags
                        for (int x = 0; x < tagsList.Length; x++)
                        {
                            string[] str = tagsList[x].Path.Split('\\');

                            if (str.Length == 1) tnList2.Add(new TreeNode(str[0] + "." + tagsList[x].Type, 2, 2));
                            else
                            {
                                TreeNode node = null;
                                foreach (TreeNode tn in tnList2)
                                {
                                    if (tn.Text == str[0])
                                    {
                                        node = tn;
                                        break;
                                    }
                                }

                                if (node == null)
                                {
                                    node = new TreeNode(str[0], 1, 1);
                                    tnList2.Add(node);
                                }

                                for (int y = 1; y < str.Length - 1; y++)
                                {
                                    bool exists = false;
                                    foreach (TreeNode tn in node.Nodes)
                                    {
                                        if (tn.Text == str[y])
                                        {
                                            node = tn;
                                            exists = true;
                                            break;
                                        }
                                    }
                                    if (!exists) node = node.Nodes.Add("", str[y], 1, 1);
                                }
                                node.Nodes.Add("", str[str.Length - 1] + "." + tagsList[x].Type, 2, 2);
                            }
                        }
                        return tnList2.ToArray();

                    default:
                        return null;
                }
            }
            else // we load the tag the user wants
            {
                //size of 1 as we'll only be loading one tag
                TreeNode[] tnList3 = new TreeNode[1];

                //give the node the name that the user has given
                tnList3[0] = new TreeNode(tagType);

                //go through the tags
                for (int y = 0; y < tagsList.Length; y++)
                {
                    //if tag[y]'s type is a match to the type
                    if (tagsList[y].Type == tagType)
                        tnList3[0].Nodes.Add(tagsList[y].Path); //add it
                }
                return tnList3; //return the now pop. list
            }
        }

        public string[] LoadTagsFromTypeIntoStrings(string type)
        {
            if (type == String.Empty) // then we load all the tags
            {
                //list for tags
                string[] tmpList = new string[tagsList.Length];

                //go through and add all tags to the list
                for (int i = 0; i < tagsList.Length; i++)
                    tmpList[i] = tagsList[i].Path;

                return tmpList; //return the list

            }
            else // we load the tag the user wants
            {
                //we use List<> as we dont know how many tags
                //we'll be loading
                List<string> tmplist = new List<string>();

                //go through the tags
                for (int i = 0; i < tagsList.Length; i++)
                {
                    //if tag[i]'s type matches what we want
                    if (tagsList[i].Type == type)
                        tmplist.Add(tagsList[i].Path); //add it
                }

                return tmplist.ToArray(); //convert and return the array
            }
        }
    }
}