using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Drawing;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Drawing.Imaging;

namespace Menu_Menu_Menu
{
    public struct MapSlot
    {
        public long metaOffset;

        public int Start;
        public byte[] bitmTag;
        public int TagID;

        public string[] Names;
        public string mapName;

        public string[] Descriptions;
        public string mapDescription;

        public string Scenario;
        public int End;
        public int Locked;
        public int TeamOptions;

        public BitmapInfo bi;
        public int ImgOffset;
        public long ImgSizePointer;
        public int ImgSize;

        public byte[] ImgBuffer;
        public Bitmap Img;
    }

    public class MainMenu
    {
        public Map map;
        public bool Loaded;
        public bool Ready;

        public int Offset;
        public int Slots;
        public MapSlot[] mapSlots;
        public int mapSlotSize;

        public int mapFilter = 0;

        public MainMenu(Map map)
        {
            this.map = map;
        }

        public void Load()
        {
            BinaryReader br = new BinaryReader(new FileStream(map.path,
               FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite));

            int plus = mapFilter == 0 ? 15812 : 15820;
            mapSlotSize = mapFilter == 0 ? 2896 : 3172;
            br.BaseStream.Position = map.tags.tagsList[0].MetaOffset + plus;

            Slots = br.ReadInt32();
            Offset = br.ReadInt32() - map.index.secondaryMagic;

            if (Offset < 0)
            {
                Loaded = false;
            }
            else
            {
                mapSlots = new MapSlot[Slots];

                br.BaseStream.Position = Offset;

                //populate info
                for (int x = 0; x < Slots; x++)
                    mapSlots[x] = returnMapSlot(br);

                //populate images
                for (int x = 0; x < Slots; x++)
                    readMapSlotImage(br, x);

                Loaded = true;
            }

            br.Close();
        }

        public MapSlot returnMapSlot(BinaryReader br)
        {
            MapSlot ms = new MapSlot();

            ms.metaOffset = br.BaseStream.Position;
            if (mapFilter == 0) br.BaseStream.Position += 4;

            ms.Start = br.ReadInt32();
            ms.bitmTag = br.ReadBytes(4);
            ms.TagID = br.ReadInt32();

            Encoding en = Encoding.Unicode;

            //diff lang names
            ms.Names = new string[8];
            for (int x = 0; x < ms.Names.Length; x++)
                ms.Names[x] = en.GetString(br.ReadBytes(64)).Split('\0')[0];

            //internal name
            ms.mapName = en.GetString(br.ReadBytes(64)).Split('\0')[0];

            //diff lang descriptions
            ms.Descriptions = new string[8];
            for (int x = 0; x < ms.Descriptions.Length; x++)
                ms.Descriptions[x] = en.GetString(br.ReadBytes(256)).Split('\0')[0];

            //internal description
            ms.mapDescription = en.GetString(br.ReadBytes(256)).Split('\0')[0];

            if (mapFilter == 1)
            {
                //scenario
                char chr;
                ms.Scenario = "";
                while (ms.Scenario.Length != 256 && (chr = br.ReadChar()) != '\0')
                    ms.Scenario += chr;

                if (ms.Scenario.Length != 256) br.BaseStream.Position += 255 - ms.Scenario.Length;
                ms.End = br.ReadInt32();

                ms.Locked = br.ReadInt32();
                ms.TeamOptions = br.ReadInt32();

                br.BaseStream.Position += 12;
            }

            return ms;
        }

        public void readMapSlotImage(BinaryReader br, int MapSlotIndx)
        {
            int tagIndx = map.tags.FindTagIndexByID(mapSlots[MapSlotIndx].TagID);
            br.BaseStream.Position = map.tags.tagsList[tagIndx].MetaOffset + 68 + 4;

            int tempr = br.ReadInt32() - map.index.secondaryMagic;
            br.BaseStream.Position = tempr;

            mapSlots[MapSlotIndx].bi = new BitmapInfo(br);
            br.BaseStream.Position += 4;

            mapSlots[MapSlotIndx].ImgOffset = br.ReadInt32();
            br.BaseStream.Position += 20;
            mapSlots[MapSlotIndx].ImgSizePointer = br.BaseStream.Position;
            mapSlots[MapSlotIndx].ImgSize = br.ReadInt32();

            br.BaseStream.Position = mapSlots[MapSlotIndx].ImgOffset;
            mapSlots[MapSlotIndx].ImgBuffer = br.ReadBytes(mapSlots[MapSlotIndx].ImgSize);

            mapSlots[MapSlotIndx].Img = map.bitm.DecodeByteArrayAndReturnBitmap(
                mapSlots[MapSlotIndx].ImgBuffer, mapSlots[MapSlotIndx].bi, 0, 0);
        }

        public void SaveMapSlot(int MapSlotIndx)
        {
            BinaryWriter bw = new BinaryWriter(new FileStream(map.path,
              FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite));

            bw.BaseStream.Position = mapSlots[MapSlotIndx].metaOffset;
            if (mapFilter == 0) bw.BaseStream.Position += 4;

            bw.Write(mapSlots[MapSlotIndx].Start);
            bw.Write(mapSlots[MapSlotIndx].bitmTag);
            bw.Write(mapSlots[MapSlotIndx].TagID);

            Encoding en = Encoding.Unicode;
            byte[] bUnicode, nul;

            //diff lang names
            for (int x = 0; x < mapSlots[MapSlotIndx].Names.Length; x++)
            {
                bUnicode = en.GetBytes(mapSlots[MapSlotIndx].Names[x]);
                bw.Write(bUnicode);
                nul = new byte[64 - bUnicode.Length];
                bw.Write(nul);
            }

            //internal name
            bUnicode = en.GetBytes(mapSlots[MapSlotIndx].mapName);
            bw.Write(bUnicode);
            nul = new byte[64 - bUnicode.Length];
            bw.Write(nul);

            //diff lang descriptions
            for (int x = 0; x < mapSlots[MapSlotIndx].Descriptions.Length; x++)
            {
                bUnicode = en.GetBytes(mapSlots[MapSlotIndx].Descriptions[x]);
                bw.Write(bUnicode);
                nul = new byte[256 - bUnicode.Length];
                bw.Write(nul);
            }

            //internal description
            bUnicode = en.GetBytes(mapSlots[MapSlotIndx].mapDescription);
            bw.Write(bUnicode);
            nul = new byte[256 - bUnicode.Length];
            bw.Write(nul);

            if (mapFilter == 1)
            {
                //scenario
                bw.Write(mapSlots[MapSlotIndx].Scenario.ToCharArray());
                nul = new byte[256 - mapSlots[MapSlotIndx].Scenario.Length];
                bw.Write(nul);
            }

            bw.Write(mapSlots[MapSlotIndx].End);

            if (mapFilter == 1)
            {
                bw.Write(mapSlots[MapSlotIndx].Locked);
                bw.Write(mapSlots[MapSlotIndx].TeamOptions);
            }

            //img
            bw.BaseStream.Position = mapSlots[MapSlotIndx].ImgOffset;
            bw.Write(mapSlots[MapSlotIndx].ImgBuffer);

            bw.Close();
        }

        public void SaveAllChanges()
        {
            BinaryWriter bw = new BinaryWriter(new FileStream(map.path,
              FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite));

            for (int m = 0; m < mapSlots.Length; m++)
            {
                bw.BaseStream.Position = mapSlots[m].metaOffset;
                if (mapFilter == 0) bw.BaseStream.Position += 4;

                bw.Write(mapSlots[m].Start);
                bw.Write(mapSlots[m].bitmTag);
                bw.Write(mapSlots[m].TagID);

                Encoding en = Encoding.Unicode;
                byte[] bUnicode, nul;

                //diff lang names
                for (int x = 0; x < mapSlots[m].Names.Length; x++)
                {
                    bUnicode = en.GetBytes(mapSlots[m].Names[x]);
                    bw.Write(bUnicode);
                    nul = new byte[64 - bUnicode.Length];
                    bw.Write(nul);
                }

                //internal name
                bUnicode = en.GetBytes(mapSlots[m].mapName);
                bw.Write(bUnicode);
                nul = new byte[64 - bUnicode.Length];
                bw.Write(nul);

                //diff lang descriptions
                for (int x = 0; x < mapSlots[m].Descriptions.Length; x++)
                {
                    bUnicode = en.GetBytes(mapSlots[m].Descriptions[x]);
                    bw.Write(bUnicode);
                    nul = new byte[256 - bUnicode.Length];
                    bw.Write(nul);
                }

                //internal description
                bUnicode = en.GetBytes(mapSlots[m].mapDescription);
                bw.Write(bUnicode);
                nul = new byte[256 - bUnicode.Length];
                bw.Write(nul);

                if (mapFilter == 1)
                {
                    //scenario
                    bw.Write(mapSlots[m].Scenario.ToCharArray());
                    nul = new byte[256 - mapSlots[m].Scenario.Length];
                    bw.Write(nul);
                }

                bw.Write(mapSlots[m].End);

                if (mapFilter == 1)
                {
                    bw.Write(mapSlots[m].Locked);
                    bw.Write(mapSlots[m].TeamOptions);
                }

                //img
                bw.BaseStream.Position = mapSlots[m].ImgOffset;
                bw.Write(mapSlots[m].ImgBuffer);
            }

            bw.Close();
        }

        public void DuplicateMapSlot(int MapSlotIndx)
        {
            FileStream fs = new FileStream(map.path,
                FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite);

            BinaryReader br = new BinaryReader(fs);
            BinaryWriter bw = new BinaryWriter(fs);

            //--[Add new slot to matg]
            br.BaseStream.Position = mapSlots[MapSlotIndx].metaOffset;
            byte[] dupArray = br.ReadBytes(mapSlotSize);

            br.BaseStream.Position = mapSlots[mapSlots.Length - 1].metaOffset + mapSlotSize;
            byte[] mapHalf = br.ReadBytes((int)(br.BaseStream.Length - br.BaseStream.Position));

            bw.BaseStream.Position = mapSlots[mapSlots.Length - 1].metaOffset + mapSlotSize;
            bw.Write(dupArray);
            bw.Write(mapHalf);

            int plus = mapFilter == 0 ? 15812 : 15820;
            bw.BaseStream.Position = map.tags.tagsList[0].MetaOffset + plus;
            bw.Write(Slots += 1);

            int shift = dupArray.Length;

            if (mapFilter == 0) //have to fix mp reflex
            {
                br.BaseStream.Position = map.tags.tagsList[0].MetaOffset + 15820 + 4;

                int mpOffset = br.ReadInt32() + shift;
                bw.BaseStream.Position -= 4;
                bw.Write(mpOffset);
            }

            //--[Reflexives]
            for (int y = 1; y < map.index.objectCount; y++)
            {
                //these tags are way above the norm meta
                if (map.tags.tagsList[y].Type == "sbsp" || map.tags.tagsList[y].Type == "ltmp") continue;

                //goto new metaoffset of tag + 4 (skip chunkcnt if ther is one)
                br.BaseStream.Position = (map.tags.tagsList[y].MetaOffset + shift) + 4;

                for (int x = 4; x < map.tags.tagsList[y].MetaSize; x += 4)
                {
                    int rawReflx = br.ReadInt32();
                    int transReflx = rawReflx - map.index.secondaryMagic;

                    //if the offset is within the current tags meta
                    if (transReflx >= map.tags.tagsList[y].MetaOffset &&
                        transReflx <= map.tags.tagsList[y].MetaOffset + map.tags.tagsList[y].MetaSize)
                    {
                        //we have a reflexive
                        bw.BaseStream.Position -= 4;
                        bw.Write(rawReflx + shift);

                        bw.BaseStream.Position += 4;
                        x += 4;
                    }
                }
            }

            //--[Meta Offsets]
            for (int x = 1; x < map.tags.tagsList.Length; x++)
            {
                if (map.tags.tagsList[x].Type != "sbsp" && map.tags.tagsList[x].Type != "ltmp")
                    map.tags.tagsList[x].MetaOffset += shift;
            }

            //--[Tags]
            bw.BaseStream.Position = map.index.offsetOfObjectIndex + 12;

            bw.Write(map.tags.tagsList[0].MetaSize += shift);
            bw.BaseStream.Position += 8;

            for (int x = 1; x < map.index.objectCount; x++)
            {
                if (map.tags.tagsList[x].Type == "sbsp" || map.tags.tagsList[x].Type == "ltmp") bw.BaseStream.Position += 16;
                else
                {
                    bw.Write(map.tags.tagsList[x].RawMetaOffset += shift);
                    bw.BaseStream.Position += 12;
                }
            }

            //--[End]
            bw.BaseStream.Position = map.tags.tagsList[map.tags.tagsList.Length - 1].MetaOffset +
                map.tags.tagsList[map.tags.tagsList.Length - 1].MetaSize;

            int newPaddingLen = 512 - ((int)bw.BaseStream.Position % 512);
            newPaddingLen += newPaddingLen == 512 ? 2560 : 3072;

            int eof = (int)bw.BaseStream.Position + newPaddingLen;

            while (bw.BaseStream.Position < eof)
                bw.Write((byte)205); //CD

            bw.BaseStream.SetLength(eof);

            //--[Header]
            bw.BaseStream.Position = 8;
            bw.Write(eof);

            br.BaseStream.Position = 24;
            int diff = (eof - map.header.metaStart) - map.header.metaSize;
            bw.Write(map.header.metaSize += diff);
            bw.Write(map.header.nonRawSize += diff);

            br.Close();
            bw.Close();
        }

        public void DeleteMapSlot(int MapSlotIndx)
        {
            FileStream fs = new FileStream(map.path,
                FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite);

            BinaryReader br = new BinaryReader(fs);
            BinaryWriter bw = new BinaryWriter(fs);

            //sp and mp have diff sizes
            int shift = -mapSlotSize;

            br.BaseStream.Position = mapSlots[MapSlotIndx].metaOffset + mapSlotSize;
            byte[] mapHalf = br.ReadBytes((int)(br.BaseStream.Length - br.BaseStream.Position));

            bw.BaseStream.Position = mapSlots[MapSlotIndx].metaOffset;
            bw.Write(mapHalf);

            int plus = mapFilter == 0 ? 15812 : 15820;
            bw.BaseStream.Position = map.tags.tagsList[0].MetaOffset + plus;
            bw.Write(Slots -= 1);

            if (mapFilter == 0) //have to fix mp reflex
            {
                br.BaseStream.Position = map.tags.tagsList[0].MetaOffset + 15820 + 4;

                int mpOffset = br.ReadInt32() + shift;
                bw.BaseStream.Position -= 4;
                bw.Write(mpOffset);
            }

            //--[Reflexives]
            for (int y = 1; y < map.index.objectCount; y++)
            {
                //these tags are way above the norm meta
                if (map.tags.tagsList[y].Type == "sbsp" || map.tags.tagsList[y].Type == "ltmp") continue;

                //goto new metaoffset of tag + 4 (skip chunkcnt if ther is one)
                br.BaseStream.Position = (map.tags.tagsList[y].MetaOffset + shift) + 4;

                for (int x = 4; x < map.tags.tagsList[y].MetaSize; x += 4)
                {
                    int rawReflx = br.ReadInt32();
                    int transReflx = rawReflx - map.index.secondaryMagic;

                    //if the offset is within the current tags meta
                    if (transReflx >= map.tags.tagsList[y].MetaOffset &&
                        transReflx <= map.tags.tagsList[y].MetaOffset + map.tags.tagsList[y].MetaSize)
                    {
                        //we have a reflexive
                        bw.BaseStream.Position -= 4;
                        bw.Write(rawReflx + shift);

                        bw.BaseStream.Position += 4;
                        x += 4;
                    }
                }
            }

            //--[Meta Offsets]
            for (int x = 1; x < map.tags.tagsList.Length; x++)
            {
                if (map.tags.tagsList[x].Type != "sbsp" && map.tags.tagsList[x].Type != "ltmp")
                    map.tags.tagsList[x].MetaOffset += shift;
            }

            //--[Tags]
            bw.BaseStream.Position = map.index.offsetOfObjectIndex + 12;

            bw.Write(map.tags.tagsList[0].MetaSize += shift);
            bw.BaseStream.Position += 8;

            for (int x = 1; x < map.index.objectCount; x++)
            {
                if (map.tags.tagsList[x].Type == "sbsp" || map.tags.tagsList[x].Type == "ltmp") bw.BaseStream.Position += 16;
                else
                {
                    bw.Write(map.tags.tagsList[x].RawMetaOffset += shift);
                    bw.BaseStream.Position += 12;
                }
            }

            //--[End]
            bw.BaseStream.Position = map.tags.tagsList[map.tags.tagsList.Length - 1].MetaOffset +
                map.tags.tagsList[map.tags.tagsList.Length - 1].MetaSize;

            int newPaddingLen = 512 - ((int)bw.BaseStream.Position % 512);
            newPaddingLen += newPaddingLen == 512 ? 2560 : 3072;

            int eof = (int)bw.BaseStream.Position + newPaddingLen;

            while (bw.BaseStream.Position < eof)
                bw.Write((byte)205); //CD

            bw.BaseStream.SetLength(eof);

            //--[Header]
            bw.BaseStream.Position = 8;
            bw.Write(eof);

            br.BaseStream.Position = 24;
            int diff = (eof - map.header.metaStart) - map.header.metaSize;
            bw.Write(map.header.metaSize += diff);
            bw.Write(map.header.nonRawSize += diff);

            br.Close();
            bw.Close();
        }
    }
}
