using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.IO;
using System.Windows.Forms;

namespace Dark
{
    public class EntReader
    {
        public Map map;

        public XmlTextReader Reader;
        public List<BaseNode> Nodes;

        public string entPath;
        public string entAuthor;
        public string entVersion;
        public string entType;

        public List<Control> RControls;

        private BitmaskNode atmBitmask;
        private EnumNode atmEnum;
        private bool TagJustRead;
        private StructNode atmStruct;

        public EntReader(Map map)
        {
            this.map = map;
        }

        public bool Load(string iType)
        {
            entPath = Application.StartupPath + @"\Plugins\Ent\" + iType + ".ent";

            if (!File.Exists(entPath))
            {
                return false;
            }
            else
            {
                Reader = new XmlTextReader(new FileStream(entPath, FileMode.Open,
                    FileAccess.ReadWrite, FileShare.ReadWrite));

                Reader.WhitespaceHandling = WhitespaceHandling.None;

                Nodes = new List<BaseNode>();

                atmBitmask = null;
                atmEnum = null;
                atmStruct = null;

                //Loops through the entire ent file
                while (Reader.Read())
                {
                    switch (Reader.NodeType.ToString())
                    {
                        case "Element":
                            {
                                switch (Reader.Name.ToLower())
                                {
                                    case "plugin":
                                        {
                                            entAuthor = Reader.GetAttribute("author");
                                            entVersion = Reader.GetAttribute("version");
                                            entType = Reader.GetAttribute("class");
                                            break;
                                        }
                                    case "revision":
                                        {
                                            //todo..
                                            break;
                                        }

                                    default:
                                        {
                                            if (atmBitmask != null)
                                            {
                                                BaseNode MaybeOption = ReadLine();

                                                if (MaybeOption.NodeType == NodeType.Option) atmBitmask.Nodes.Add((OptionNode)MaybeOption);
                                                else
                                                {
                                                    atmBitmask = null;
                                                    Nodes.Add(MaybeOption);
                                                }
                                            }
                                            else if (atmEnum != null)
                                            {
                                                BaseNode MaybeOption = ReadLine();

                                                if (MaybeOption.NodeType == NodeType.Option) atmEnum.Nodes.Add((OptionNode)MaybeOption);
                                                else
                                                {
                                                    atmEnum = null;
                                                    Nodes.Add(MaybeOption);
                                                }
                                            }
                                            else if (TagJustRead)
                                            {
                                                BaseNode MaybeID = ReadLine();
                                                if (MaybeID.NodeType != NodeType.ID) Nodes.Add(MaybeID);
                                                TagJustRead = false;

                                                //if (MaybeID.NodeType == NodeType.ID) atmTag.IDnode = (IDNode)MaybeID;
                                                //else
                                                //{
                                                //    atmTag.IDnode = null;
                                                //    Nodes.Add(MaybeID);
                                                //}
                                                //atmTag = null;
                                            }
                                            else
                                            {
                                                Nodes.Add(ReadLine());
                                            }
                                            break;
                                        }
                                }
                                break;
                            }
                        case "EndElement":
                            {
                                switch (Reader.Name.ToLower())
                                {
                                    case "bitmask":
                                    case "bitmask8":
                                    case "bitmask16":
                                    case "bitmask32":
                                        {
                                            atmBitmask = null;
                                            break;
                                        }

                                    case "enum":
                                    case "enum8":
                                    case "enum16":
                                    case "enum32":
                                        {
                                            atmEnum = null;
                                            break;
                                        }

                                    case "struct":
                                        {
                                            atmStruct = atmStruct.Parent;
                                            break;
                                        }
                                    case "plugin":
                                        {
                                            break;
                                        }
                                }
                                break;
                            }
                    }
                }
                Reader.Close();
                return true;
            }
        }

        private BaseNode ReadLine()
        {
            switch (Reader.Name.ToLower())
            {
                case "struct":
                    {
                        return (atmStruct = new StructNode(
                            map, 
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"),
                            Reader.GetAttribute("size")));
                    }
                case "tag":
                    {
                        TagJustRead = true;
                        return new TagNode(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"));
                    }
                case "id":
                    {
                        return new IDNode(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"));
                    }
                case "script":
                case "stringid":
                    {
                        return new StringIDNode(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"));
                    }
                case "unused":
                    {
                        return new UnusedNode(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"),
                            Reader.GetAttribute("size"));
                    }
                case "undefined":
                    {
                        return new UndefinedNode(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"));
                    }
                case "float":
                    {
                        return new FloatNode(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"));
                    }
                case "int":
                    {
                        return new Int32Node(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"));
                    }
                case "uint":
                    {
                        return new UInt32Node(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"));
                    }
                case "short":
                    {
                        return new ShortNode(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"));
                    }
                case "ushort":
                    {
                        return new UShortNode(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"));
                    }
                case "byte":
                    {
                        return new ByteNode(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"));
                    }
                case "bitmask":
                case "bitmask32":
                    {
                        return (atmBitmask = new BitmaskNode(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"),
                            32));
                    }
                case "bitmask16":
                    {
                        return (atmBitmask = new BitmaskNode(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"),
                            16));
                    }
                case "bitmask8":
                    {
                        return (atmBitmask = new BitmaskNode(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"),
                            8));
                    }
                case "option":
                    {
                        return new OptionNode(
                           Reader.GetAttribute("name"),
                           Reader.GetAttribute("value"));
                    }
                case "enum":
                case "enum32":
                    {
                        return (atmEnum = new EnumNode(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"),
                            32));
                    }
                case "enum16":
                    {
                        return (atmEnum = new EnumNode(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"),
                            16));
                    }
                case "enum8":
                    {
                        return (atmEnum = new EnumNode(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"),
                            8));
                    }
                case "string32":
                    {
                        return new StringNode(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"),
                            32);
                    }
                case "string256":
                    {
                        return new StringNode(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"),
                            256);
                    }
                case "unicode64":
                    {
                        return new UnicodeNode(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"),
                            64);
                    }
                case "unicode256":
                    {
                        return new UnicodeNode(
                            map,
                            atmStruct,
                            Reader.GetAttribute("name"),
                            Reader.GetAttribute("offset"),
                            Reader.GetAttribute("visible"),
                            256);
                    }

                default:
                    {
                        return new BaseNode();
                    }
            }
        }

        public void UpdateMetaVisibility()
        {
            for (int x = 0; x < Nodes.Count; x++)
                Nodes[x].UpdateVisibility();
        }

        public void InitializeMetaControls()
        {
            //loop through nodes and initialize their ctrls
            for (int x = 0; x < Nodes.Count; x++)
            {
                if (Nodes[x].IsVisible) Nodes[x].InitializeControls();
            }
        }

        public void ReadMetaValues()
        {
            BinaryReader br = new BinaryReader(new FileStream(map.path, FileMode.Open,
              FileAccess.ReadWrite, FileShare.ReadWrite));

            //loop through nodes
            for (int x = 0; x < Nodes.Count; x++)
            {
                if (Nodes[x].IsVisible)
                {
                    //means this isnt in a reflexive, so pass -1 meaning just use meta offset
                    if (Nodes[x].Parent == null)
                    {
                        Nodes[x].ReadValue(br, -1);
                    }
                    //pass on the translated reflexive
                    else if (Nodes[x].Parent.IsVisible && Nodes[x].Parent.Enabled)
                    {
                        Nodes[x].ReadValue(br, Nodes[x].Parent.atmChunk);
                    }
                }
            }
            br.Close();
        }

        public void DisplayMeta(Panel pnl)
        {
            //suspend
            pnl.SuspendLayout();

            //clear
            pnl.Controls.Clear();

             //set progbar pos
            pnl.AutoScrollPosition = new System.Drawing.Point(0, 0);

            //will store ctrls that need resizing
            RControls = new List<Control>();

            //our pos vars
            int xpos = 12, ypos = 12;

            //loop through nodes
            for (int x = 0; x < Nodes.Count; x++)
            {
                if (Nodes[x].IsVisible)
                {
                    //show if invisible?
                    if (!Nodes[x].Visible && !Globals.MetaEditorShownItems[4]) continue;
                    else
                    {
                        if (Nodes[x].Parent == null) //adding to pluginpanel
                        {
                            ypos = Nodes[x].DisplayControls(pnl, xpos, ypos);
                        }
                        //adding to a reflexive panel, 1st check if reflxs are visible
                        else if (Globals.MetaEditorShownItems[0] && Nodes[x].Parent.Visible)
                        {
                            //add to parent panel
                            Nodes[x].Parent.yposPanel = Nodes[x].DisplayControls(
                                Nodes[x].Parent.panel, xpos, Nodes[x].Parent.yposPanel);

                            //correct parent panel height
                            Nodes[x].Parent.panel.Height = Nodes[x].Parent.yposPanel;

                            StructNode tmp = Nodes[x].Parent;
                            while (tmp.Parent != null)
                            {
                                tmp.Parent.yposPanel = tmp.panel.Bottom + 12;
                                tmp.Parent.panel.Height = tmp.Parent.yposPanel;
                                tmp = tmp.Parent;
                            }
                            ypos = tmp.panel.Bottom + 12;
                        }
                    }
                }
            }

            Panel panel1 = new Panel();
            panel1.Top = ypos;
            panel1.Height = 6;
            pnl.Controls.Add(panel1);

            //resume
            pnl.ResumeLayout(true);
        }
    }
}