using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using TD.SandDock;
using System.Xml;
using System.Security.Cryptography;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace Dark
{
    public partial class MapForm : Form
    {
        //class for storing the info of the map
        public Map map;

        public MapForm(string path, ToolStripProgressBar progbar,
            ToolStripStatusLabel proglabel)
        {
            //load form controls
            InitializeComponent();

            //reduce flicker on ltmp panel, keep?
            //this.SetStyle(ControlStyles.DoubleBuffer | ControlStyles.AllPaintingInWmPaint, true);

            tv1.Sorted = true;
            leftSandDock.Width = splitContainer1.Panel1.Width;
            tv2.Nodes[0].Expand();
            pluginpanel.BackColor = Globals.ReflexiveStartingColor;

            this.Text = path;

            map = new Map(progbar, proglabel);
            LoadMap(path);
        }

        public void LoadMap(string path)
        {
            map.Load(path);
            map.UpdateProgress("Displaying Map Info...", 6);

            tv1.Nodes.Clear();
            tv1.Nodes.Add("", map.header.internalName, 0, 0);
            tv1.Nodes[0].Nodes.AddRange(map.tags.LoadTagsIntoTreenodes(""));
            tv1.Nodes[0].Expand();

            grid1.SelectedObject = new HeaderControl(map);
            grid2.SelectedObject = new IndexControl(map);

            ttbox.Items.Clear();
            ttbox.Items.AddRange(map.tags.tagTypes.ToArray());

            map.EndProgress();
        }

        private void MapForm_Load(object sender, EventArgs e)
        {
            //this.Text = map.header.metaStart.ToString() + ";; " + map.header.metaSize.ToString();
        }

        #region Sidebar

        public void CheckTool(ToolStripButton tool)
        {
            toolStripButton1.CheckState = toolStripButton2.CheckState =
            toolStripButton3.CheckState = toolStripButton10.CheckState =
            CheckState.Unchecked;

            tool.CheckState = CheckState.Checked;
        }

        public void VisiblePanel(Panel panel)
        {
            if (!panel.Visible)
            {
                panel1.Visible = panel2.Visible =
                panel3.Visible = panel6.Visible = false;

                panel.Visible = true;
            }
        }

        private void toolStripButton1_Click(object sender, EventArgs e)
        {
            VisiblePanel(panel1);
            CheckTool(toolStripButton1);
        }

        private void toolStripButton2_Click(object sender, EventArgs e)
        {
            VisiblePanel(panel2);
            CheckTool(toolStripButton2);
        }

        private void toolStripButton3_Click(object sender, EventArgs e)
        {
            textBox1.Text = Globals.mainMenuLoc;
            textBox2.Text = Globals.sharedLoc;
            textBox3.Text = Globals.singlePlayerSharedLoc;

            VisiblePanel(panel3);
            CheckTool(toolStripButton3);
        }

        private void toolStripButton10_Click(object sender, EventArgs e)
        {
            contextMenuStrip3.Show(Cursor.Position);
        }

        private void mapAnalyzerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            VisiblePanel(panel6);
            CheckTool(toolStripButton10);
        }

        #endregion

        #region Misc

        public string GetMd5SumFromStream(string str)
        {
            // Now that we have a byte array we can ask the CSP to hash it
            MD5 md5 = new MD5CryptoServiceProvider();
            md5.Initialize();

            FileStream fs = new FileStream(str, FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite);
            byte[] result = md5.ComputeHash(fs);
            fs.Close();

            // Build the final string by converting each byte
            // into hex and appending it to a StringBuilder
            StringBuilder sb = new StringBuilder();

            for (int i = 0; i < result.Length; i++)
                sb.Append(result[i].ToString("X"));

            // And return it
            return sb.ToString();
        }

        private void splitContainer1_Panel1_Resize(object sender, EventArgs e)
        {
            leftSandDock.Width = splitContainer1.Panel1.Width;
        }

        private void searchBox_TextChanged(object sender, EventArgs e)
        {
            tv2.Nodes[0].Nodes.Clear();

            if (searchBox.TextLength != 0)
            {
                string txt = searchBox.Text;
                for (int x = 0; x != map.tags.tagsList.Length; x++)
                {
                    if (txt != searchBox.Text) return;
                    if (map.tags.tagsList[x].Path.StartsWith(searchBox.Text))
                    {
                        tv2.Nodes[0].Text = "Search Results (" +
                            (tv2.Nodes[0].Nodes.Add(x.ToString(), map.tags.tagsList[x].Path, 2, 2)
                            .Index + 1).ToString()
                            + ")";

                        Application.DoEvents();
                    }
                }
            }

            if (tv2.Nodes[0].Nodes.Count == 0)
            {
                tv2.Nodes[0].Text = "Search Results";
                tv2.Nodes[0].Nodes.Add("", "{NONE}", 4, 4);
            }

            tv2.Nodes[0].Expand();
        }

        public bool ShowSpecificControl(string ctrlName)
        {
            ctrlName = ctrlName.Replace("!", "");

            Control ctrl = null;
            foreach (Control box in dockControl4.Controls)
            {
                if (box.Name == ctrlName)
                {
                    box.Visible = true;
                    ctrl = box;
                }
                else box.Visible = false;
            }
            return (ctrl != null);
        }

        private void documentContainer2_ActiveDocumentChanged(object sender, ActiveDocumentEventArgs e)
        {
            switch (documentContainer2.ActiveDocument.Name)
            {
                case "dockControl4": //specific
                    if (dockControl4.Tag.ToString() == "0") //empty
                    {
                        map.StartProgress(1);
                        map.UpdateProgress("Reading raw data...", 1);

                        map.rawData.LoadSelectedTag();
                        LoadSpecificPanel();

                        dockControl4.Tag = "1";
                        map.EndProgress();
                    }
                    break;

                case "dockControl5": //refs
                    //map.refs.ManualScan(map.tags.selectedIndex,
                    //    map.sbsp.FindIndexByTagIndex(map.tags.selectedIndex));

                    map.refs.EntScan();

                    MakeRefColumns();
                    lv1.Items.AddRange(map.refs.LoadRefsIntoListviewItems());
                    ResizeRefColumns();
                    break;

                case "dockControl6": //editor
                    if (dockControl6.Tag.ToString() == "0") //empty
                    {
                        if (parsedBtn.Checked) parsedBtn_Click(null, null);
                        else if (rawBtn.Checked) rawBtn_Click(null, null);

                        dockControl6.Tag = "1";
                    }
                    break;

                case "dockControl8": //cloner
                    if (dockControl8.Tag.ToString() == "0") //empty
                    {
                        LoadReflexivesAndChunks();
                        dockControl8.Tag = "1";
                    }
                    break;
            }
        }

        #endregion

        #region Treeviews

        private void tv1_AfterSelect(object sender, TreeViewEventArgs e)
        {
            if (e.Node != null)
            {
                if (e.Node.Nodes.Count == 0) //tag
                {
                    switch (map.tags.sortTagsBy)
                    {
                        case SortTagsBy.TagType:

                            TreeviewTagSelected(
                                map.tags.FindTagIndexByTypeAndPath(e.Node.Parent.Text, e.Node.Text));

                            break;

                        case SortTagsBy.FolderHierarchy:

                            string tagPath = e.Node.Text.Remove(e.Node.Text.Length - 5);
                            TreeNode node = e.Node.Parent;

                            while (node.Parent != null)
                            {
                                tagPath = tagPath.Insert(0, node.Text + '\\');
                                node = node.Parent;
                            }

                            TreeviewTagSelected(
                                map.tags.FindTagIndexByTypeAndPath(e.Node.Text.Substring(e.Node.Text.Length - 4), tagPath));

                            break;
                    }
                }
                else TreeviewOtherSelected();
            }
        }

        private void contextMenuStrip1_Opening(object sender, CancelEventArgs e)
        {
            duplicateToolStripMenuItem.Enabled =
                (tv1.SelectedNode != null && tv1.SelectedNode.Nodes.Count == 0);
        }

        private void tv2_AfterSelect(object sender, TreeViewEventArgs e)
        {
            if (e.Node != null && e.Node.Parent != null && e.Node.ImageIndex != 4) TreeviewTagSelected(int.Parse(e.Node.Name));
            else TreeviewOtherSelected();
        }

        private void tv2_DoubleClick(object sender, EventArgs e)
        {
            if (tv2.SelectedNode != null && tv2.SelectedNode.Parent != null && tv2.SelectedNode.ImageIndex != 4)
            {
                switch (map.tags.sortTagsBy)
                {
                    case SortTagsBy.TagType:
                        foreach (TreeNode typeNode in tv1.Nodes[0].Nodes)
                        {
                            if (typeNode.Text == map.tags.tagsList[map.tags.selectedIndex].Type)
                            {
                                foreach (TreeNode tagNode in typeNode.Nodes)
                                {
                                    if (tagNode.Text == map.tags.tagsList[map.tags.selectedIndex].Path)
                                    {
                                        tv1.SelectedNode = tagNode;
                                        tagNode.EnsureVisible();
                                        dockControl10.Activate();
                                        break;
                                    }
                                }
                                break;
                            }
                        }
                        break;

                    case SortTagsBy.FolderHierarchy:
                        string[] spl = map.tags.tagsList[map.tags.selectedIndex].Path.Split('\\');
                        spl[spl.Length - 1] += "." + map.tags.tagsList[map.tags.selectedIndex].Type;

                        TreeNode node = tv1.Nodes[0];
                        for (int x = 0; x < spl.Length; x++)
                        {
                            foreach (TreeNode tn in node.Nodes)
                            {
                                if (tn.Text == spl[x])
                                {
                                    node = tn;
                                    break;
                                }
                            }
                        }
                        tv1.SelectedNode = node;
                        node.EnsureVisible();
                        dockControl10.Activate();
                        break;
                }
            }
        }

        public void TreeviewOtherSelected()
        {
            if (toolStripButton2.Checked) toolStripButton1_Click(null, null);
            toolStripButton2.Enabled = false;
        }

        public void TreeviewTagSelected(int tagIndx)
        {
            string lastType =
                map.tags.selectedIndex == -1 ? "" : map.tags.tagsList[map.tags.selectedIndex].Type;

            //store index of selected tag
            map.tags.selectedIndex = tagIndx;

            if (lastType != map.tags.tagsList[map.tags.selectedIndex].Type)
                map.entReader.Load(map.tags.tagsList[map.tags.selectedIndex].Type);

            //reset tag editor
            foreach (DockControl dc in documentContainer2.Documents)
                dc.Tag = "0";

            //load the details of the tag
            grid3.SelectedObject = new DetailsControl(map);

            if (!ShowSpecificControl(map.tags.tagsList[tagIndx].Type + "panel"))
                ShowSpecificControl("label1");

            //tag editor side bar button
            if (!toolStripButton2.Checked)
            {
                toolStripButton2.Enabled = true;
                dockControl3.Activate(); //Details
                toolStripButton2_Click(null, null); //select
            }
            else
            {
                if (documentContainer2.ActiveDocument == null) dockControl3.Activate(); //Details
                //else update current tab
                else documentContainer2_ActiveDocumentChanged(null, null);
            }
        }

        private void tagTypeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            map.tags.sortTagsBy = SortTagsBy.TagType;

            tv1.Nodes[0].Nodes.Clear();
            tv1.Nodes[0].Nodes.AddRange(map.tags.LoadTagsIntoTreenodes(""));
            tv1.Nodes[0].Expand();

            TreeviewOtherSelected();

            tagTypeToolStripMenuItem.Checked = true;
            folderHierarchyToolStripMenuItem.Checked = false;
        }

        private void folderHierarchyToolStripMenuItem_Click(object sender, EventArgs e)
        {
            map.tags.sortTagsBy = SortTagsBy.FolderHierarchy;

            tv1.Nodes[0].Nodes.Clear();
            tv1.Nodes[0].Nodes.AddRange(map.tags.LoadTagsIntoTreenodes(""));
            tv1.Nodes[0].Expand();

            TreeviewOtherSelected();

            folderHierarchyToolStripMenuItem.Checked = true;
            tagTypeToolStripMenuItem.Checked = false;
        }

        #endregion

        #region Header

        private void grid1_PropertyValueChanged(object s, PropertyValueChangedEventArgs e)
        {
            grid1.Refresh();

            map.StartProgress(1);
            map.UpdateProgress("Writing Header...", 1);
            map.header.WriteHeader();
            map.EndProgress();
        }

        #endregion

        #region Index

        private void grid2_PropertyValueChanged(object s, PropertyValueChangedEventArgs e)
        {
            grid2.Refresh();

            map.StartProgress(1);
            map.UpdateProgress("Writing Index...", 1);
            map.index.WriteIndex();
            map.EndProgress();
        }

        #endregion

        #region Details

        private void grid3_PropertyValueChanged(object s, PropertyValueChangedEventArgs e)
        {
            grid1.Refresh();

            map.StartProgress(1);
            map.UpdateProgress("Writing Details...", 1);
            map.tags.WriteSelectedTag();

            tv1.SelectedNode.Text = map.tags.tagsList[map.tags.selectedIndex].Path;

            if (tv1.SelectedNode.Parent.Text != map.tags.tagsList[map.tags.selectedIndex].Type)
            {
                foreach (TreeNode typeNode in tv1.Nodes[0].Nodes)
                {
                    if (typeNode.Text == map.tags.tagsList[map.tags.selectedIndex].Type)
                    {
                        TreeNode oldNode = tv1.SelectedNode;
                        TreeNode newNode = (TreeNode)tv1.SelectedNode.Clone();
                        typeNode.Nodes.Add(newNode);
                        tv1.SelectedNode = newNode;
                        oldNode.Remove();
                        break;
                    }
                }
            }

            map.EndProgress();
        }

        private void button15_Click(object sender, EventArgs e)
        {
            s.FileName = map.tags.tagsList[map.tags.selectedIndex].Path.Substring(
                map.tags.tagsList[map.tags.selectedIndex].Path.LastIndexOf('\\') + 1);

            s.Filter = "Meta Files (*.meta)|*.meta|All Files (*.*)|*.*";

            if (s.ShowDialog() == DialogResult.Cancel) return;

            BinaryReader br = new BinaryReader(new FileStream(map.path, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            BinaryWriter bw = new BinaryWriter(new FileStream(s.FileName, FileMode.Create,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            if (checkBox3.Checked) //recursive
            {
                //gather tags
            }
            else
            {
                br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset;
                bw.Write(br.ReadBytes(map.tags.tagsList[map.tags.selectedIndex].MetaSize));
            }
            br.Close();
            bw.Close();
        }

        private void button16_Click(object sender, EventArgs e)
        {
            o.Filter = "Meta Files (*.meta)|*.meta|All Files (*.*)|*.*";

            if (o.ShowDialog() == DialogResult.Cancel) return;

            BinaryReader br = new BinaryReader(new FileStream(o.FileName, FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite));
            byte[] newMeta = br.ReadBytes((int)br.BaseStream.Length);
            br.Close();

            BinaryWriter bw = new BinaryWriter(new FileStream(map.path, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            if (newMeta.Length == map.tags.tagsList[map.tags.selectedIndex].MetaSize)
            {
                bw.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset;
                bw.Write(newMeta);
            }
            else
            {
                //reader
                br = new BinaryReader(bw.BaseStream);

                //difference
                int shift = newMeta.Length - map.tags.tagsList[map.tags.selectedIndex].MetaSize;

                //goto just after tag meta 
                br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset
                    + map.tags.tagsList[map.tags.selectedIndex].MetaSize;

                int metaEnd = map.tags.tagsList[map.tags.tagsList.Length - 1].MetaOffset
                    + map.tags.tagsList[map.tags.tagsList.Length - 1].MetaSize;

                byte[] eofMeta = br.ReadBytes(metaEnd - (int)br.BaseStream.Position);

                bw.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset;
                bw.Write(newMeta);
                bw.Write(eofMeta);

                //--[End]
                int paddingLen = 512 - ((int)bw.BaseStream.Position % 512);
                paddingLen += paddingLen == 512 ? 512 : 1024;

                int eof = (int)bw.BaseStream.Position + paddingLen;

                while (bw.BaseStream.Position < eof)
                    bw.Write((byte)205); //CD

                bw.BaseStream.SetLength(eof);

                //--[Header]
                bw.BaseStream.Position = 8;
                bw.Write(eof);

                bw.BaseStream.Position += 12;
                bw.Write(map.header.metaSize += shift);
                bw.Write(map.header.nonRawSize += shift);

                //--[Reflexives]
                for (int y = map.tags.selectedIndex; y < map.index.objectCount; y++)
                {
                    //these tags are way above the norm meta
                    if (map.tags.tagsList[y].Type == "sbsp" || map.tags.tagsList[y].Type == "ltmp") continue;

                    //goto new metaoffset of tag + 4 (skip chunkcnt if ther is one)
                    br.BaseStream.Position = (map.tags.tagsList[y].MetaOffset + shift) + 4;

                    for (int x = 4; x < map.tags.tagsList[y].MetaSize; x += 4)
                    {
                        int rawReflx = br.ReadInt32();
                        int transReflx = rawReflx - map.index.secondaryMagic;

                        //if the offset is within the current tags meta
                        if (transReflx >= map.tags.tagsList[y].MetaOffset &&
                            transReflx <= map.tags.tagsList[y].MetaOffset + map.tags.tagsList[y].MetaSize)
                        {
                            //we have a reflexive
                            bw.BaseStream.Position -= 4;
                            bw.Write(rawReflx + shift);

                            bw.BaseStream.Position += 4;
                            x += 4;
                        }
                    }
                }

                //--[Tags]
                bw.BaseStream.Position = map.index.offsetOfObjectIndex
                    + (map.tags.selectedIndex * 16) + 12;

                bw.Write(map.tags.tagsList[map.tags.selectedIndex].MetaSize += shift);
                bw.BaseStream.Position += 8;

                for (int x = map.tags.selectedIndex + 1; x < map.index.objectCount; x++)
                {
                    if (map.tags.tagsList[x].Type == "sbsp" || map.tags.tagsList[x].Type == "ltmp") bw.BaseStream.Position += 16;
                    else
                    {
                        bw.Write(map.tags.tagsList[x].RawMetaOffset += shift);
                        map.tags.tagsList[x].MetaOffset += shift;
                        bw.BaseStream.Position += 12;
                    }
                }

                br.Close();
            }
            bw.Close();
        }

        #endregion

        #region References

        public void CorrectSwapPanel()
        {
            button6.Enabled = false;
            tbox.Items.Clear();

            switch (map.refs.Filter)
            {
                case NodeType.Reflexive:
                    ttbox.SelectedItem = null;
                    ttbox.Enabled = tbox.Enabled = SwapButton.Enabled = false;
                    break;

                case NodeType.ID:
                    if (ttbox.SelectedIndex == 0) ttbox_SelectedIndexChanged(null, null);
                    else ttbox.SelectedIndex = 0;
                    ttbox.Enabled = tbox.Enabled = SwapButton.Enabled = true;
                    break;

                case NodeType.StringID:
                    ttbox.SelectedItem = null;
                    ttbox.Enabled = tbox.Sorted = false;
                    tbox.Enabled = SwapButton.Enabled = true;
                    break;
            }
        }

        //reflexs
        private void toolStripButton4_Click(object sender, EventArgs e)
        {
            if (!toolStripButton4.Checked)
            {
                map.refs.Filter = NodeType.Reflexive;
                CorrectSwapPanel();

                if (map.refs.Scanned)
                {
                    MakeRefColumns();
                    lv1.Items.AddRange(map.refs.LoadRefsIntoListviewItems());
                    ResizeRefColumns();
                }
            }

            toolStripButton4.CheckState = CheckState.Checked;

            toolStripButton5.CheckState = toolStripButton7.CheckState
                = CheckState.Unchecked;
        }

        //idents
        private void toolStripButton5_Click(object sender, EventArgs e)
        {
            if (!toolStripButton5.Checked)
            {
                map.refs.Filter = NodeType.ID;
                CorrectSwapPanel();

                if (map.refs.Scanned)
                {
                    MakeRefColumns();
                    lv1.Items.AddRange(map.refs.LoadRefsIntoListviewItems());
                    ResizeRefColumns();
                }
            }

            toolStripButton5.CheckState = CheckState.Checked;

            toolStripButton4.CheckState = toolStripButton7.CheckState
                = CheckState.Unchecked;
        }

        //strings
        private void toolStripButton7_Click(object sender, EventArgs e)
        {
            if (!toolStripButton7.Checked)
            {
                map.refs.Filter = NodeType.StringID;
                CorrectSwapPanel();

                if (map.refs.Scanned)
                {
                    MakeRefColumns();
                    lv1.Items.AddRange(map.refs.LoadRefsIntoListviewItems());
                    ResizeRefColumns();
                }
            }

            toolStripButton7.CheckState = CheckState.Checked;

            toolStripButton4.CheckState = toolStripButton5.CheckState
                = CheckState.Unchecked;
        }

        public void MakeRefColumns()
        {
            lv1.Clear();
            switch (map.refs.Filter)
            {
                case NodeType.Reflexive:
                    lv1.Columns.Add("Description");
                    lv1.Columns.Add("Offset");
                    lv1.Columns.Add("Translation");
                    lv1.Columns.Add("Chunk Count");
                    break;

                case NodeType.ID:
                    lv1.Columns.Add("Description");
                    lv1.Columns.Add("Offset");
                    lv1.Columns.Add("Tag Type");
                    lv1.Columns.Add("Tag Path");
                    break;

                case NodeType.StringID:
                    lv1.Columns.Add("Description");
                    lv1.Columns.Add("Offset");
                    lv1.Columns.Add("Name");
                    break;
            }
        }

        public void ResizeRefColumns()
        {
            switch (map.refs.Filter)
            {
                case NodeType.Reflexive:
                    lv1.Columns[0].AutoResize(ColumnHeaderAutoResizeStyle.ColumnContent);
                    lv1.Columns[1].AutoResize(ColumnHeaderAutoResizeStyle.HeaderSize);
                    lv1.Columns[2].AutoResize(ColumnHeaderAutoResizeStyle.ColumnContent);
                    lv1.Columns[3].AutoResize(ColumnHeaderAutoResizeStyle.HeaderSize);
                    break;

                case NodeType.ID:
                    lv1.Columns[0].AutoResize(ColumnHeaderAutoResizeStyle.ColumnContent);
                    lv1.Columns[1].AutoResize(ColumnHeaderAutoResizeStyle.HeaderSize);
                    lv1.Columns[2].AutoResize(ColumnHeaderAutoResizeStyle.HeaderSize);
                    lv1.Columns[3].AutoResize(ColumnHeaderAutoResizeStyle.ColumnContent);
                    break;

                case NodeType.StringID:
                    lv1.Columns[0].AutoResize(ColumnHeaderAutoResizeStyle.ColumnContent);
                    lv1.Columns[1].AutoResize(ColumnHeaderAutoResizeStyle.HeaderSize);
                    lv1.Columns[2].AutoResize(ColumnHeaderAutoResizeStyle.ColumnContent);
                    break;
            }
        }

        private void ttbox_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (ttbox.SelectedItem != null)
            {
                tbox.Items.Clear();
                tbox.Sorted = true;
                tbox.Items.AddRange(map.tags.LoadTagsFromTypeIntoStrings(ttbox.SelectedItem.ToString()));

                if (map.refs.Filter == NodeType.ID)
                {
                    tbox.Sorted = false;
                    tbox.Items.Insert(0, "Nulled Out");
                }
                if (tbox.Items.Count > 0) tbox.SelectedIndex = 0;
            }
        }

        private void tbox_DropDown(object sender, EventArgs e)
        {
            if (map.refs.Filter == NodeType.StringID && tbox.Items.Count == 0)
            {
                map.StartProgress(1);
                map.UpdateProgress("Populating combobox...", 1);
                tbox.Items.AddRange(map.strings.stringsList);
                map.EndProgress();
            }
        }

        private void lv1_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (lv1.FocusedItem != null)
            {
                switch (map.refs.Filter)
                {
                    case NodeType.Reflexive:
                        button6.Enabled = false;
                        break;

                    case NodeType.ID:
                        ttbox.SelectedIndex = ttbox.FindString(map.refs.iList[lv1.FocusedItem.Index].Type);
                        tbox.SelectedIndex = tbox.FindString(map.refs.iList[lv1.FocusedItem.Index].Path);

                        if (map.refs.iList[lv1.FocusedItem.Index].RefType == RefTypes.Dependency) button6.Enabled = (tbox.SelectedIndex != 0); //0 = null
                        else button6.Enabled = true;
                        break;

                    case NodeType.StringID:
                        tbox.SelectedIndex = tbox.FindString(map.refs.sList[lv1.FocusedItem.Index].Name);
                        button6.Enabled = false;
                        break;
                }
            }
            else button6.Enabled = false;
        }

        private void SwapButton_Click(object sender, EventArgs e)
        {
            if (lv1.SelectedItems.Count == 0)
                MessageBox.Show("You must first make a selection.", "",
                    MessageBoxButtons.OK, MessageBoxIcon.Exclamation);

            else
            {
                switch (map.refs.Filter)
                {
                    case NodeType.ID:
                        foreach (ListViewItem item in lv1.SelectedItems)
                        {
                            map.refs.SwapIdent(item.Index, ttbox.SelectedItem.ToString(),
                                tbox.SelectedItem.ToString());

                            Ident ident = map.refs.iList[lv1.FocusedItem.Index];
                            ident.Type = ttbox.SelectedItem.ToString();
                            ident.Path = tbox.SelectedItem.ToString();
                            map.refs.iList[lv1.FocusedItem.Index] = ident;

                            lv1.Items[item.Index].SubItems[2].Text = ttbox.SelectedItem.ToString();
                            lv1.Items[item.Index].SubItems[3].Text = tbox.SelectedItem.ToString();
                        }
                        break;

                    case NodeType.StringID:
                        foreach (ListViewItem item in lv1.SelectedItems)
                        {
                            map.refs.SwapString(item.Index, tbox.SelectedItem.ToString(),
                                tbox.SelectedIndex);

                            StringID str = map.refs.sList[lv1.FocusedItem.Index];
                            str.Name = tbox.SelectedItem.ToString();
                            map.refs.sList[lv1.FocusedItem.Index] = str;

                            lv1.Items[item.Index].SubItems[2].Text = tbox.SelectedItem.ToString();
                        }
                        break;
                }

                ResizeRefColumns();
            }
        }

        private void button6_Click(object sender, EventArgs e)
        {
            //only enabled wen dep/lone is selected
            string type = map.refs.iList[lv1.FocusedItem.Index].Type,
                tag = map.refs.iList[lv1.FocusedItem.Index].Path;

            foreach (TreeNode tnType in tv1.Nodes[0].Nodes)
            {
                if (tnType.Text == type)
                {
                    foreach (TreeNode tnTag in tnType.Nodes)
                    {
                        if (tnTag.Text == tag)
                        {
                            tv1.SelectedNode = tnTag;
                            tv1.SelectedNode.EnsureVisible();

                            break;
                        }
                    }
                    break;
                }
            }
        }

        #endregion

        #region External Maps

        private void button1_Click(object sender, EventArgs e)
        {
            o.FileName = "mainmenu";
            o.Filter = "Map (*.map)|*.map|All Files (*.*)|*.*";
            if (o.ShowDialog() == DialogResult.Cancel) return;
            textBox1.Text = o.FileName;
        }

        private void button2_Click(object sender, EventArgs e)
        {
            o.FileName = "shared";
            o.Filter = "Map (*.map)|*.map|All Files (*.*)|*.*";
            if (o.ShowDialog() == DialogResult.Cancel) return;
            textBox2.Text = o.FileName;
        }

        private void button3_Click(object sender, EventArgs e)
        {
            o.FileName = "single_player_shared";
            o.Filter = "Map (*.map)|*.map|All Files (*.*)|*.*";
            if (o.ShowDialog() == DialogResult.Cancel) return;
            textBox3.Text = o.FileName;
        }

        private void textBox1_TextChanged(object sender, EventArgs e)
        {
            Globals.mainMenuLoc = textBox1.Text;
        }

        private void textBox2_TextChanged(object sender, EventArgs e)
        {
            Globals.sharedLoc = textBox2.Text;
        }

        private void textBox3_TextChanged(object sender, EventArgs e)
        {
            Globals.singlePlayerSharedLoc = textBox3.Text;
        }

        #endregion

        #region Meta Editor

        public void pluginpanel_Resize(object sender, EventArgs e)
        {
            if (pluginpanel.Controls.Count != 0)
            {
                int spacing = pluginpanel.VerticalScroll.Visible ?
                    SystemInformation.VerticalScrollBarWidth : 0;

                for (int x = 0; x < map.entReader.RControls.Count; x++)
                {
                    map.entReader.RControls[x].Width =
                            pluginpanel.Width - 24 - spacing;
                }
            }
        }

        private void parsedBtn_Click(object sender, EventArgs e)
        {
            rawBtn.Checked = false;
            rToolsBtn.Visible = false;
            hexpanel.Visible = false;

            parsedBtn.Checked = true;
            pToolsBtn.Visible = true;
            pluginpanel.Visible = true;

            map.StartProgress(1);
            map.UpdateProgress("Reading Meta...", 1);

            if (pluginpanel.Tag.ToString() == map.tags.tagsList[map.tags.selectedIndex].Type)
            {
                //just renew the values
                map.entReader.ReadMetaValues();
            }
            else
            {
                //renew the whole editor
                map.entReader.UpdateMetaVisibility();
                map.entReader.InitializeMetaControls();
                map.entReader.ReadMetaValues();
                map.entReader.DisplayMeta(pluginpanel);

                //Application.DoEvents();
                pluginpanel_Resize(null, null);
                pluginpanel.Tag = map.tags.tagsList[map.tags.selectedIndex].Type;
            }

            map.EndProgress();
        }

        private void rawBtn_Click(object sender, EventArgs e)
        {
            parsedBtn.Checked = false;
            pToolsBtn.Visible = false;
            pluginpanel.Visible = false;

            rawBtn.Checked = true;
            rToolsBtn.Visible = true;
            hexpanel.Visible = true;

            map.StartProgress(1);
            map.UpdateProgress("Reading Meta...", 1);

            //open map
            BinaryReader br = new BinaryReader(new FileStream(map.path,
                FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite));

            string hexStr = "";
            char tmpChar;

            string uniStr = "";
            string tmpStr = "";

            byte tmpByte;

            //goto meta
            br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset;

            //loop
            for (int x = 0; x < map.tags.tagsList[map.tags.selectedIndex].MetaSize; x++)
            {
                tmpByte = br.ReadByte();

                //get hex
                tmpStr = tmpByte.ToString("X");
                if (tmpStr.Length == 1) tmpStr = "0" + tmpStr;
                hexStr += tmpStr + " ";

                //get unicode
                tmpChar = Convert.ToChar(tmpByte);
                if (tmpChar != '\0' && (char.IsLetterOrDigit(tmpChar) || char.IsPunctuation(tmpChar))) uniStr += tmpChar;
                else uniStr += ".";
            }

            //close
            br.Close();

            //display
            richTextBox1.Text = hexStr;
            richTextBox2.Text = uniStr;

            map.EndProgress();
        }

        private void toolStripButton6_Click(object sender, EventArgs e)
        {
            map.StartProgress(1);
            map.UpdateProgress("Writing Meta...", 1);

            BinaryWriter bw = new BinaryWriter(new FileStream(map.path, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            if (parsedBtn.Checked)
            {
                //loop through nodes
                for (int x = 0; x < map.entReader.Nodes.Count; x++)
                {
                    if (map.entReader.Nodes[x].IsVisible)
                    {
                        //means this isnt in a reflexive, so pass -1 meaning just use meta offset
                        if (map.entReader.Nodes[x].Parent == null) map.entReader.Nodes[x].SaveChanges(bw, -1);
                        //pass on the translated reflexive
                        else if (map.entReader.Nodes[x].Parent.Enabled) map.entReader.Nodes[x].SaveChanges(bw, map.entReader.Nodes[x].Parent.atmChunk);
                    }
                }
            }
            else if (rawBtn.Checked)
            {

            }

            bw.Close();
            map.EndProgress();
        }

        private void entEditorToolStripMenuItem_Click(object sender, EventArgs e)
        {
            new EntEditor().ShowDialog();
        }

        private void preferencesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            new EntPreferences(this).ShowDialog();
        }

        #endregion

        #region Duplicate

        private void duplicateToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Tag tg = map.tags.DuplicateSelectedTag();

            switch (map.tags.sortTagsBy)
            {
                case SortTagsBy.TagType:
                    tv1.SelectedNode =
                        tv1.SelectedNode.Parent.Nodes.Add("", tg.Path, 2, 2);
                    break;

                case SortTagsBy.FolderHierarchy:
                    tv1.SelectedNode =
                        tv1.SelectedNode.Parent.Nodes.Add("", tg.Path + '.' + tg.Type, 2, 2);
                    break;
            }
        }

        #endregion

        #region Cloner

        public void LoadReflexivesAndChunks()
        {
            tv3.SuspendLayout();

            //clear
            tv3.Nodes.Clear();

            //add tag name
            string[] spl = map.tags.tagsList[map.tags.selectedIndex].Path.Split('\\');
            tv3.Nodes.Add(spl[spl.Length - 1]);

            BinaryReader br = new BinaryReader(new FileStream(map.path, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            int tagMagic = map.index.secondaryMagic;
            switch (map.tags.tagsList[map.tags.selectedIndex].Type)
            {
                case "sbsp":
                    tagMagic = map.sbsp.sbspInfoList[
                        map.sbsp.FindIndexByTagID(map.tags.tagsList[map.tags.selectedIndex].Ident)
                        ].Magic;
                    break;

                case "ltmp":
                    tagMagic = map.sbsp.sbspInfoList[
                        map.sbsp.FindIndexByLightmapTagID(map.tags.tagsList[map.tags.selectedIndex].Ident)
                        ].Magic;
                    break;
            }

            TreeNode atmNode = null;

            //add reflexives and chunks
            for (int x = 0; x < map.entReader.Nodes.Count; x++)
            {
                if (map.entReader.Nodes[x].NodeType == NodeType.Reflexive)
                {
                    StructNode sn = (map.entReader.Nodes[x] as StructNode);

                    if (sn.Parent == null)
                    {
                        br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset + sn.Offset;
                        sn.chunkCount = br.ReadInt32();
                        sn.translation = br.ReadInt32() - tagMagic;

                        TreeNode reflx = tv3.Nodes[0].Nodes.Add(sn.Name + " CS=" + sn.Size.ToString());
                        for (int z = 0; z < sn.chunkCount; z++) reflx.Nodes.Add("Chunk " + z.ToString());

                        atmNode = reflx;
                    }
                    else if (sn.Parent.Enabled)
                    {
                        //loop through reflexives chunks
                        for (int y = 0; y < atmNode.Nodes.Count; y++)
                        {
                            br.BaseStream.Position = sn.Parent.translation + sn.Offset + (y * sn.Parent.Size);
                            sn.chunkCount = br.ReadInt32();
                            sn.translation = br.ReadInt32() - tagMagic;

                            TreeNode reflx = atmNode.Nodes[y].Nodes.Add(sn.Name + " CS=" + sn.Size.ToString());
                            for (int z = 0; z < sn.chunkCount; z++) reflx.Nodes.Add("Chunk " + z.ToString());

                            //TreeNode reflx = tv3.Nodes[0].Nodes.Add(sn.Name
                            //    + "  Chunk Count: " + sn.chunkCount.ToString());
                            //if (sn.chunkCount > 0)
                            //{
                            //    reflx.Nodes.Add("");
                            //    reflx.Tag = sn.chunkCount;
                            //}
                        }
                    }
                }
            }

            tv3.Nodes[0].Expand();
            br.Close();

            tv3.ResumeLayout(false);
        }

        private void tv3_BeforeExpand(object sender, TreeViewCancelEventArgs e)
        {
            if (e.Node.Tag != null)
            {
                int chunkCount = (int)e.Node.Tag;
                e.Node.Nodes.RemoveAt(0);

                for (int y = 0; y < chunkCount; y++)
                    e.Node.Nodes.Add("Chunk " + y.ToString());

                e.Node.Tag = null;
            }
        }

        private void tv34_ItemDrag(object sender, ItemDragEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
                DoDragDrop(e.Item, DragDropEffects.Copy);
        }

        private void tv34_DragEnter(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent("System.Windows.Forms.TreeNode"))
                e.Effect = DragDropEffects.Copy;

            else e.Effect = DragDropEffects.None;
        }

        private void tv3_DragDrop(object sender, DragEventArgs e)
        {
            //if treenode data is present
            if (e.Data.GetDataPresent("System.Windows.Forms.TreeNode", false))
            {
                //pull out the node
                TreeNode TransferingNode = ((TreeNode)e.Data.GetData("System.Windows.Forms.TreeNode"));
                if (TransferingNode.TreeView == tv4)
                {
                    //add to map
                    if (radioButton5.Checked) //add
                    {
                        tv3.SelectedNode = tv3.SelectedNode.Nodes[
                            tv3.SelectedNode.Nodes.Add((TreeNode)TransferingNode.Clone())];
                    }
                    else //ovewrite
                    {
                        int indx = tv3.SelectedNode.Index;
                        TreeNode tnParent = tv3.SelectedNode.Parent;

                        tv3.SelectedNode.Remove();
                        tnParent.Nodes.Insert(indx, (TreeNode)TransferingNode.Clone());
                        tv3.SelectedNode = tnParent.Nodes[indx];
                    }
                }
            }
        }

        private void tv4_DragDrop(object sender, DragEventArgs e)
        {
            //if treenode data is present
            if (e.Data.GetDataPresent("System.Windows.Forms.TreeNode", false))
            {
                //pull out the node
                TreeNode TransferingNode = ((TreeNode)e.Data.GetData("System.Windows.Forms.TreeNode"));
                if (TransferingNode.TreeView == tv3)
                {
                    Globals.AddClonerMemoryNode(TransferingNode, (Form1)this.MdiParent);
                    //tv4.SelectedNode = tv4.Nodes[tv4.Nodes.Add((TreeNode)TransferingNode.Clone())];
                }
            }
        }

        private void tv3_AfterSelect(object sender, TreeViewEventArgs e)
        {
            if (e.Node != null && e.Node != tv3.Nodes[0])
            {
                tv4.Enabled =
                    //groupBox2.Enabled =
                button21.Enabled = true;
            }
            else
            {
                tv4.Enabled =
                    //groupBox2.Enabled =
                button21.Enabled = false;
            }
        }

        private void button19_Click(object sender, EventArgs e)
        {

        }

        private void tv4_PreviewKeyDown(object sender, PreviewKeyDownEventArgs e)
        {
            if (e.KeyCode == Keys.Delete && tv4.SelectedNode != null)
                Globals.RemoveClonerMemoryNode(tv4.SelectedNode.Index, (Form1)this.MdiParent);
        }

        private void textBox7_TextChanged(object sender, EventArgs e)
        {
            //int result;
            //if (int.TryParse(textBox7.Text, out result))
            //textBox7.Tag = textBox7.Text;

            //else textBox7.Text = textBox7.Tag.ToString();
        }

        #endregion

        #region Tag Drag Drop

        private void tv1_ItemDrag(object sender, ItemDragEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
                DoDragDrop(e.Item, DragDropEffects.Copy);
        }

        private void tv1_DragDrop(object sender, DragEventArgs e)
        {
            //if treenode data is present
            if (e.Data.GetDataPresent("System.Windows.Forms.TreeNode", false))
            {
                //pull out the node
                TreeNode TransferingNode = ((TreeNode)e.Data.GetData("System.Windows.Forms.TreeNode"));

                //make sure we are transfering between maps
                if (TransferingNode.TreeView != tv1)
                {
                    //pull out the transfers map
                    Map TagMap = ((MapForm)TransferingNode.TreeView.FindForm()).map;

                    //and index
                    int TagIndx = TagMap.tags.FindTagIndexByTypeAndPath(
                        TransferingNode.Parent.Text, TransferingNode.Text);

                    //if its a valid tag
                    if (TagIndx != -1)
                    {
                        //start the transfering
                        new TransferForm(TagMap, TagIndx, this).ShowDialog();
                    }
                }
            }
        }

        private void tv1_DragEnter(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent("System.Windows.Forms.TreeNode"))
                e.Effect = DragDropEffects.Copy;

            else e.Effect = DragDropEffects.None;
        }

        #endregion

        #region Specific

        /* Special tags:
         * bitm
         * lsnd?
         * map.ltmp
         * matg
         * mode
         * mulg
         * PRTM
         * sbsp
         * scnr
         * snd!
         * ugh!
         * unic?
         */

        public void LoadSpecificPanel()
        {
            if (map.rawData.rdh == null || !map.rawData.rdh.Loaded)
            {
                dockControl4.Enabled = false;
            }
            else
            {
                if (bitmpanel.Visible)
                {
                    map.bitm.LoadBitmap(map.tags.selectedIndex);
                    pictureBox1.Image = map.bitm.ReturnBitmap(0, 0, 0, 0, map.rawData.rdh.rawChunks);

                    label5.Text = "Size: " + map.bitm.BitmapInfoList[0].Width + "x" +
                        map.bitm.BitmapInfoList[0].Height;

                    label6.Text = "Format: " + map.bitm.BitmapInfoList[0].bitmFormat.ToString();
                    label9.Text = "MipMap Count: " + map.bitm.BitmapInfoList[0].MipMapCount.ToString();
                    label7.Text = "Swizzled: " + map.bitm.BitmapInfoList[0].Swizzled.ToString();
                    label8.Text = "Location: " + map.rawData.rdh.rawChunks[map.bitm.selectedChunkIndex].rawLocation.ToString();

                    button11.Enabled = (map.rawData.rdh.rawChunks[map.bitm.selectedChunkIndex].rawLocation != RawLocations.Internal);
                }
                else if (ltmppanel.Visible)
                {
                    ltmpLoaded = false;
                    map.ltmp.LoadLightmapsIntoAnArray();

                    trackint1.Value = trackint2.Value = trackint3.Value = 0;
                    textBox4.Text = textBox5.Text = textBox6.Text = "0";

                    radioButton1.Checked = true;
                    checkBox1.Checked = false;
                    radioButton3.Checked = true;

                    numUD4.Maximum = map.sbsp.sbspInfoList[map.ltmp.bspIndx].lightmap.Palettes.Count - 1;
                    map.ltmp.PositionLightmapImgs(panel8, img_Click);
                    ltmpLoaded = true;
                }
                else if (matgpanel.Visible)
                {
                    //unicode
                    comboBox1.Items.Clear();
                    lv2.Items.Clear();
                    comboBox1.Items.AddRange(Enum.GetNames(typeof(Language)));

                    Application.DoEvents();
                    comboBox1.SelectedIndex = 0;
                }
                else if (sndpanel.Visible)
                {
                    SoundData sd = (map.rawData.rdh as SoundData);

                    switch (sd.sndType)
                    {
                        case 0:
                            label10.Text = "Type:  ADPCM Mono 22050kbps";
                            break;

                        case 1:
                            label10.Text = "Type: ADPCM Stereo 44100kbps";
                            break;

                        case 2:
                            label10.Text = "Type: Windows Media Audio (WMA)";
                            break;
                    }

                    if (map.snd_.LoadSelectedSound(sd.sndType))
                    {
                        label11.Text = "Duration: " + map.snd_.aud.Duration.ToString() + " seconds";
                        headerControl6.Enabled = true;
                    }
                    else
                    {
                        label11.Text = "Duration: Unk";
                        headerControl6.Enabled = false;
                    }
                }
                else if (modepanel.Visible)
                {
                    if (headerControl7.Controls.Count == 0)
                        headerControl7.Controls.Add(map.mode);

                    map.mode.Initialize(label34, label33, label13);
                    label32.Text = "Polygon Count: " + map.mode.PolygonCount.ToString();
                    label29.Text = "Location: " + map.rawData.rdh.rawChunks[map.bitm.selectedChunkIndex].rawLocation.ToString();

                    for (int x = 0; x < map.mode.Frames.Length; x++)
                        comboBox2.Items.Add(map.mode.Frames[x].Name);
                }
                dockControl4.Enabled = true;
            }
        }

        #endregion

        #region Lightmap

        bool ltmpLoaded;

        public void img_Click(object sender, EventArgs e) //pic clicked
        {
            if (ltmpLoaded)
            {
                PictureBox pb = (PictureBox)sender;
                pb.BorderStyle = BorderStyle.FixedSingle;
                if (map.ltmp.spb != null) map.ltmp.spb.BorderStyle = BorderStyle.None;

                numUD4.Value = (int)pb.Tag;
                radioButton4.Checked = true;
                map.ltmp.spb = pb;

                panel8.Invalidate();
            }
        }

        private void panel8_Resize(object sender, EventArgs e)
        {
            if (ltmpLoaded) map.ltmp.RepositionLightmapImgs(panel8);
        }

        private void trackint123_Scroll_checkbox1_Checked(object sender, EventArgs e) //3trackbar scroll
        {
            if (ltmpLoaded)
            {
                ltmpLoaded = false;

                map.ltmp.EditLightmaps(ref panel8, trackint1.Value, trackint2.Value, trackint3.Value,
                 checkBox1.Checked, (int)numUD4.Value, radioButton1.Checked, radioButton3.Checked);

                textBox4.Text = trackint1.Value.ToString();
                textBox5.Text = trackint2.Value.ToString();
                textBox6.Text = trackint3.Value.ToString();

                ltmpLoaded = true;
                panel8.Invalidate();
            }
        }

        private void textBox456_TextChanged(object sender, EventArgs e)
        {
            trackint1.Value = int.Parse(textBox4.Text);
            trackint2.Value = int.Parse(textBox5.Text);
            trackint3.Value = int.Parse(textBox6.Text);

            if (ltmpLoaded)
                trackint123_Scroll_checkbox1_Checked(null, null);
        }

        private void radioButton3_CheckedChanged(object sender, EventArgs e)
        {
            if (ltmpLoaded)
            {
                if (radioButton3.Checked)
                {
                    numUD4.Enabled = false;
                    if (map.ltmp.spb != null)
                    {
                        map.ltmp.spb.BorderStyle = BorderStyle.None;
                        map.ltmp.spb = null;
                    }
                }
                else
                {
                    numUD4.Enabled = true;
                    img_Click((PictureBox)panel8.Controls[(int)numUD4.Value], null);
                }
                panel8.Invalidate();
            }
        }

        private void numUD4_ValueChanged(object sender, EventArgs e)
        {
            img_Click((PictureBox)panel8.Controls[(int)numUD4.Value], null);
        }

        private void radioButton1_CheckedChanged(object sender, EventArgs e)
        {
            if (ltmpLoaded)
            {
                Array.ConstrainedCopy(map.ltmp.LightMap_Array_Backup, 0, map.ltmp.LightMap_Array,
                    0, map.ltmp.LightMap_Array.Length);

                trackint1.Value = trackint2.Value = trackint3.Value = 0;
                textBox4.Text = textBox5.Text = textBox6.Text = "0";

                if (radioButton1.Checked) //hue|saturation
                {
                    checkBox1.Text = "Colorize";
                    toolTip1.SetToolTip(trackint1, "Hue");
                    toolTip1.SetToolTip(trackint2, "Saturation");
                    toolTip1.SetToolTip(trackint3, "Lightness");
                }
                else //colour balance
                {
                    checkBox1.Text = "Preserve Lightness";
                    toolTip1.SetToolTip(trackint1, "Cyan - Red");
                    toolTip1.SetToolTip(trackint2, "Magenta - Green");
                    toolTip1.SetToolTip(trackint3, "Yellow - Blue");
                }

                foreach (Control box in panel8.Controls)
                    ((PictureBox)box).Image = map.ltmp.RenderLightmap((int)box.Tag);

                panel8.Invalidate();
            }
        }

        private void button4_Click(object sender, EventArgs e)
        {
            Array.ConstrainedCopy(map.ltmp.LightMap_Array_Backup, 0, map.ltmp.LightMap_Array,
             0, map.ltmp.LightMap_Array.Length);

            trackint1.Value = trackint2.Value = trackint3.Value = 0;
            textBox4.Text = textBox5.Text = textBox6.Text = "0";

            foreach (Control box in panel8.Controls)
                ((PictureBox)box).Image = map.ltmp.RenderLightmap((int)box.Tag);

            panel8.Invalidate();
        }

        private void button5_Click(object sender, EventArgs e)
        {
            //Apply the new palettes
            Array.ConstrainedCopy(map.ltmp.LightMap_Array, 0, map.ltmp.LightMap_Array_Backup,
                0, map.ltmp.LightMap_Array.Length);

            //Save the lightmap
            map.sbsp.sbspInfoList[map.ltmp.bspIndx].lightmap.WritePalettes(map.path);

            trackint1.Value = trackint2.Value = trackint3.Value = 0;
            textBox4.Text = textBox5.Text = textBox6.Text = "0";
        }

        #endregion

        #region Unicode

        private void comboBox1_SelectedIndexChanged(object sender, EventArgs e)
        {
            lv2.Items.Clear();
            rtb3.Text = "";

            Lang lang = (map.rawData.rdh as UnicodeData).langs[comboBox1.SelectedIndex];
            int len = lang.UniStrings.Length;

            map.StartProgress(1);
            map.UpdateProgress("Populating list...", 1);

            bool bool1 = false;
            for (int x = 0; x < len; x++)
            {
                ListViewItem lv = lv2.Items.Add(lang.UniStrings[x].uString);
                lv.SubItems.Add(lang.UniStrings[x].Length.ToString());
                lv.SubItems.Add(lang.UniStrings[x].Offset.ToString());

                if (bool1) lv.BackColor = Color.Gainsboro;
                bool1 = !bool1;
            }

            map.EndProgress();
        }

        private void textBox9_TextChanged(object sender, EventArgs e)
        {
            lv2.Items.Clear();
            rtb3.Text = "";

            Lang lang = (map.rawData.rdh as UnicodeData).langs[comboBox1.SelectedIndex];
            int len = lang.UniStrings.Length;

            string txt = textBox9.Text;
            bool bool1 = false;
            for (int x = 0; x < len; x++)
            {
                if (txt != textBox9.Text) return;
                if (lang.UniStrings[x].uString.Contains(txt))
                {
                    ListViewItem lv = lv2.Items.Add(lang.UniStrings[x].uString);
                    lv.SubItems.Add(lang.UniStrings[x].Length.ToString());
                    lv.SubItems.Add(lang.UniStrings[x].Offset.ToString());

                    if (bool1) lv.BackColor = Color.Gainsboro;
                    bool1 = !bool1;
                }
            }
        }

        private void lv2_SelectedIndexChanged(object sender, EventArgs e)
        {
            rtb3.Text = lv2.FocusedItem == null ? "" : lv2.FocusedItem.Text.Replace(@" \r\n ", "\r\n");
        }

        private void richTextBox3_TextChanged(object sender, EventArgs e)
        {
            if (rtb3.TextLength == 0) button7.Enabled = false;
            else
            {
                int pos = rtb3.SelectionStart;
                int len = rtb3.SelectionLength;

                rtb3.SelectAll();
                rtb3.SelectionBackColor = rtb3.BackColor;

                for (int x = 0; x < Globals.uniTranslations.Length; x++)
                    SearchText(Globals.uniTranslations[x], 0);

                rtb3.SelectionStart = pos;
                rtb3.SelectionLength = len;
                button7.Enabled = true;
            }
        }

        public void SearchText(string txt, int indx)
        {
            int a = rtb3.Text.IndexOf(txt, indx);
            if (a != -1)
            {
                rtb3.SelectionStart = a;
                rtb3.SelectionLength = txt.Length;
                rtb3.SelectionBackColor = Color.LightSteelBlue;

                SearchText(txt, a + txt.Length);
            }
        }

        private void button7_Click(object sender, EventArgs e)
        {
            map.StartProgress(1);
            map.UpdateProgress("Writing Unicode...", 1);

            Lang lang = (map.rawData.rdh as UnicodeData).langs[comboBox1.SelectedIndex];

            for (int x = 0; x < lv2.SelectedItems.Count; x++)
            {
                map.unic.WriteUnicode(comboBox1.SelectedIndex,
                    lv2.SelectedItems[x].Index, rtb3.Text.Replace("\n", "\r\n"));

                lv2.SelectedItems[x].Text = rtb3.Text.Replace("\n", @" \r\n ");
                lv2.SelectedItems[x].SubItems[1].Text = lang.UniStrings[lv2.SelectedItems[x].Index].Length.ToString();
            }

            for (int x = lv2.SelectedItems[0].Index + 1; x < lv2.Items.Count; x++)
                lv2.Items[x].SubItems[2].Text = lang.UniStrings[x].Offset.ToString();

            map.EndProgress();
        }

        private void lv2_Resize(object sender, EventArgs e)
        {
            int w = lv2.Width / 5;
            lv2.Columns[0].Width = (w * 3) - SystemInformation.VerticalScrollBarWidth;
            lv2.Columns[1].Width = w;
            lv2.Columns[2].Width = w;
        }

        private void button8_Click(object sender, EventArgs e)
        {
            //Tmp Extrac Lang
            if (comboBox1.SelectedIndex >= 0)
            {
                map.TempExtract(
                    (map.rawData.rdh as UnicodeData).langs[comboBox1.SelectedIndex].tableOffset,
                    (map.rawData.rdh as UnicodeData).langs[comboBox1.SelectedIndex].tableSize);
            }
        }

        #endregion

        #region Bitmap

        private void SizeMode_ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ToolStripMenuItem tool = (ToolStripMenuItem)sender;
            pictureBox1.SizeMode = (PictureBoxSizeMode)contextMenuStrip2.Items.IndexOf(tool);

            foreach (ToolStripMenuItem box in contextMenuStrip2.Items)
                box.Checked = (box == tool);
        }

        private void button11_Click(object sender, EventArgs e)
        {
            map.bitm.InternalizeSelectedBitmap();

            button11.Enabled = false;
            label8.Text = "Location: " + RawLocations.Internal.ToString();
        }

        private void button9_Click(object sender, EventArgs e)
        {
            s.FileName = map.tags.tagsList[map.tags.selectedIndex].Path.Substring(
                map.tags.tagsList[map.tags.selectedIndex].Path.LastIndexOf('\\') + 1);

            s.Filter = "DDS (*.dds)|*.dds|JPEG (*.jpg, *.jpeg)|*.jpg;*.jpeg|BMP (*.bmp)|*.bmp|GIF (*.gif)|*.gif";
            if (s.ShowDialog() == DialogResult.Cancel) return;

            switch (s.FilterIndex)
            {
                case 1: //dds
                    map.bitm.dds.ExtractDDS(map.bitm.BitmapInfoList[0], s.FileName);
                    break;

                case 2: //jpg, jpeg
                    pictureBox1.Image.Save(s.FileName, System.Drawing.Imaging.ImageFormat.Jpeg);
                    break;

                case 3: //bmp
                    pictureBox1.Image.Save(s.FileName, System.Drawing.Imaging.ImageFormat.Bmp);
                    break;

                case 4: //gif
                    pictureBox1.Image.Save(s.FileName, System.Drawing.Imaging.ImageFormat.Gif);
                    break;
            }
        }

        private void button10_Click(object sender, EventArgs e)
        {
            o.FileName = map.tags.tagsList[map.tags.selectedIndex].Path.Substring(
                map.tags.tagsList[map.tags.selectedIndex].Path.LastIndexOf('\\') + 1);

            o.Filter = "DDS (*.dds)|*.dds";

            if (o.ShowDialog() == DialogResult.Cancel) return;
            map.bitm.dds.InjectDDS(map.bitm.BitmapInfoList[0], o.FileName);

            //map.rawData.LoadSelectedTag();
            //map.bitm.LoadBitmap(map.tags.selectedIndex);
            pictureBox1.Image = map.bitm.ReturnBitmap(0, 0, 0, 0, map.rawData.rdh.rawChunks);
        }

        #endregion

        #region Sound

        private void button14_Click(object sender, EventArgs e)
        {
            s.FileName = map.tags.tagsList[map.tags.selectedIndex].Path.Substring(
               map.tags.tagsList[map.tags.selectedIndex].Path.LastIndexOf('\\') + 1);

            SoundData sd = (map.rawData.rdh as SoundData);

            switch (sd.sndType)
            {
                /*
            case 0:
                s.Filter = "ADPCM Mono 22050kbps (*.adpcm)|*.adpcm|All Files (*.*)|*.*";
                break;

            case 1:
                s.Filter = "ADPCM Stereo 44100kbps (*.adpcm)|*.adpcm|All Files (*.*)|*.*";
                break;
                 */

                case 0:
                case 1:
                    s.Filter = "Windows Audio Video (*.wav)|*.wav|All Files (*.*)|*.*";
                    break;

                case 2:
                    s.Filter = "Windows Media Audio (*.wma)|*.wma|All Files (*.*)|*.*";
                    break;
            }

            if (s.ShowDialog() == DialogResult.Cancel) return;
            map.snd_.ExtractSelectedSound(s.FileName, sd.sndType);
        }

        private void button17_Click(object sender, EventArgs e)
        {
            map.snd_.aud.Stop();
            map.snd_.aud.Play();
        }

        private void button24_Click(object sender, EventArgs e)
        {
            map.snd_.aud.Pause();
        }

        private void button25_Click(object sender, EventArgs e)
        {
            map.snd_.aud.Stop();
        }

        #endregion

        #region Model

        private void button32_Click(object sender, EventArgs e)
        {
            FrameInfo f = map.mode.Frames[comboBox2.SelectedIndex];

            //AnimationController an = new AnimationController();
            //an.
        }

        #endregion
    }
}