using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;

namespace Dark
{
    public class HeaderControl : Component
    {
        private Map map;
        public HeaderControl(Map map)
        {
            this.map = map;
        }

        [Category("Misc")]
        [Description("1st four letters of the header (deah)")]
        public string Header
        {
            get { return map.header.head; }
            set
            {
                if (value.Length <= 4) map.header.head = value;
                else value = map.header.head;
            }
        }

        [Category("Misc")]
        [Description("The map version, Halo2 maps are v8")]
        public int Version
        {
            get { return map.header.version; }
            set { map.header.version = value; }
        }

        [Category("Misc")]
        [Description("The total size in bytes of the map")]
        public int Filesize
        {
            get { return map.header.filesize; }
            set { map.header.filesize = value; }
        }

        [Category("Index")]
        [Description("The offset at which the index begins")]
        public int IndexOffset
        {
            get { return map.header.indexOffset; }
            set
            {
                map.header.indexOffset = value;

                map.header.metaStart =
                    map.header.indexOffset + map.header.indexSize;
            }
        }

        [Category("Index")]
        [Description("The total size in bytes of the index")]
        public int IndexSize
        {
            get { return map.header.indexSize; }
            set
            {
                map.header.indexSize = value;

                map.header.metaStart =
                    map.header.indexOffset + map.header.indexSize;
            }
        }

        [Category("Meta")]
        [Description("The offset at which the meta begins")]
        public int MetaStart
        {
            get { return map.header.metaStart; }
        }

        [Category("Meta")]
        [Description("The total size in bytes of the meta")]
        public int MetaSize
        {
            get { return map.header.metaSize; }
            set { map.header.metaSize = value; }
        }

        [Category("Misc")]
        [Description("This is the sum of the MetaSize + Index Size + BSP Size")]
        public int NonRawSize
        {
            get { return map.header.nonRawSize; }
            set { map.header.nonRawSize = value; }
        }

        [Category("Misc")]
        [Description("The map origin")]
        public string Origin
        {
            get { return map.header.origin; }
            set
            {
                if (value.Length <= 256) map.header.origin = value;
                else value = map.header.origin;
            }
        }

        [Category("Misc")]
        [Description("Date of the map's final build")]
        public string BuildDate
        {
            get { return map.header.buildDate; }
            set
            {
                if (value.Length <= 32) map.header.buildDate = value;
                else value = map.header.buildDate;
            }
        }

        [Category("Map Type")]
        [Description("An integer representing the type of map: 0=sp 1=mp 2=mm 3=shared 4=spshared")]
        public int MapTypeInt
        {
            get { return map.header.mapTypeInt; }
            set
            {
                if (value < Enum.GetNames(typeof(MapTypes)).Length)
                {
                    map.header.mapTypeInt = value;
                    map.header.mapType = (MapTypes)value;
                }
            }
        }

        [Category("Map Type")]
        [Description("The type of map")]
        public MapTypes ParsedMapType
        {
            get { return map.header.mapType; }
            set
            {
                map.header.mapType = value;
                map.header.mapTypeInt = (int)value;
            }
        }

        [Category("Strange File Strings")]
        [Description("The total size in bytes of the strange file strings")]
        public int StrangeFileStringsSize
        {
            get { return map.header.strangeFileStringsSize; }
            set { map.header.strangeFileStringsSize = value; }
        }

        [Category("Strange File Strings")]
        [Description("The offset at which the strange file strings begins")]
        public int StrangeFileStringsOffset
        {
            get { return map.header.strangeFileStringsOffset; }
            set { map.header.strangeFileStringsOffset = value; }
        }

        [Category("Strings")]
        [Description("The offset at which the strings within their 128 byte blocks begins")]
        public int StringsOffset128
        {
            get { return map.header.stringsOffset128; }
            set { map.header.stringsOffset128 = value; }
        }

        [Category("Strings")]
        [Description("The total amount of strings")]
        public int StringCount
        {
            get { return map.header.stringCount; }
            set { map.header.stringCount = value; }
        }

        [Category("Strings")]
        [Description("The total size in bytes of all the strings")]
        public int StringsSize
        {
            get { return map.header.stringsSize; }
            set { map.header.stringsSize = value; }
        }

        [Category("Strings")]
        [Description("The offset at which the string's index begins")]
        public int StringsIndex
        {
            get { return map.header.stringsIndex; }
            set { map.header.stringsIndex = value; }
        }

        [Category("Strings")]
        [Description("The offset at which the strings begin")]
        public int StringsOffset
        {
            get { return map.header.stringsOffset; }
            set { map.header.stringsOffset = value; }
        }

        [Category("Name")]
        [Description("The internal name of the map")]
        public string InternalName
        {
            get { return map.header.internalName; }
            set
            {
                if (value.Length <= 32) map.header.internalName = value;
                else value = map.header.internalName;
            }
        }

        [Category("Name")]
        [Description("The scenario name of the map")]
        public string ScenarioName
        {
            get { return map.header.scenarioName; }
            set
            {
                if (value.Length <= 256) map.header.scenarioName = value;
                else value = map.header.scenarioName;
            }
        }

        [Category("File Table")]
        [Description("The total amount of strings in the file table")]
        public int FileTableStringsCount
        {
            get { return map.header.fileTableCount; }
            set { map.header.fileTableCount = value; }
        }

        [Category("File Table")]
        [Description("The offset at which the file table begins")]
        public int FileTableOffset
        {
            get { return map.header.fileTableOffset; }
            set { map.header.fileTableOffset = value; }
        }

        [Category("File Table")]
        [Description("The total size in bytes of the file table")]
        public int FileTableSize
        {
            get { return map.header.fileTableSize; }
            set { map.header.fileTableSize = value; }
        }

        [Category("File Table")]
        [Description("The offset at which the file table's index begins")]
        public int FileTableIndexOffset
        {
            get { return map.header.fileTableIndexOffset; }
            set { map.header.fileTableIndexOffset = value; }
        }

        [Category("Misc")]
        [Description("This is the result of every byte after the header xored")]
        public int Checksum
        {
            get { return map.header.checksum; }
            set { map.header.checksum = value; }
        }

        [Category("Misc")]
        [Description("Last four letters of the header (toof)")]
        public string Footer
        {
            get { return map.header.foot; }
            set
            {
                if (value.Length <= 4) map.header.foot = value;
                else value = map.header.foot;
            }
        }
    }
}
