using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace Dark
{
    public enum MapTypes
    {
        Single_Player_Map,
        Multi_Player_Map,
        MainMenu,
        Shared,
        Single_Player_Shared
    }

    public class Header
    {
        public Map map;

        //header info
        public string
        head,
        origin,
        internalName,
        buildDate,
        scenarioName,
        foot;

        public MapTypes mapType;

        public int
        version,
        filesize,

        indexOffset,
        indexSize,

        metaStart,
        metaSize,
        nonRawSize,

        mapTypeInt,
        
        strangeFileStringsSize,
        strangeFileStringsOffset,

        stringsOffset128,
        stringCount,
        stringsSize,
        stringsIndex,
        stringsOffset,

        fileTableCount,
        fileTableOffset,
        fileTableSize,
        fileTableIndexOffset,

        checksum;

        public Header(Map map, BinaryReader br)
        {
            this.map = map;
            char chr; //temp

            //the header (deah)
            head = map.Reverse(new string(br.ReadChars(4)));

            //the map version, halo2 maps are v8
            version = br.ReadInt32();

            //the total count of the bytes that make up the map
            filesize = br.ReadInt32();

            //+4
            br.BaseStream.Position = 16;

            //the offset of the index
            indexOffset = br.ReadInt32();

            //the total length in bytes of the index
            indexSize = br.ReadInt32();

            //the start of the meta (stragiht after the index)
            metaStart = indexOffset + indexSize;

            //the total length in bytes of the meta
            metaSize = br.ReadInt32();

            //MetaSize + index Size + BSP Size
            nonRawSize = br.ReadInt32();

            //map origin
            origin = "";
            while ((chr = br.ReadChar()) != '\0')
                origin += chr;

            br.BaseStream.Position = 288;

            //the final build date 00.00.00
            buildDate = "";
            while ((chr = br.ReadChar()) != '\0')
                buildDate += chr;

            br.BaseStream.Position = 320;

            //a number representing the type of the map
            //0=sp 1=mp 2=mm 3=shared 4=spshared
            mapTypeInt = br.ReadInt32();

            //parse the number into an understandable format
            mapType = (MapTypes)mapTypeInt;

            //+16
            br.BaseStream.Position = 340;

            //the size and offset of the strange file strings
            strangeFileStringsSize = br.ReadInt32();
            strangeFileStringsOffset = br.ReadInt32();

            //+4
            br.BaseStream.Position = 352;

            //string info
            stringsOffset128 = br.ReadInt32();
            stringCount = br.ReadInt32();
            stringsSize = br.ReadInt32();
            stringsIndex = br.ReadInt32();
            stringsOffset = br.ReadInt32();

            //+36
            br.BaseStream.Position = 408;

            //forms the internal name of the map - e.g. lockout
            internalName = "";
            while ((chr = br.ReadChar()) != '\0')
                internalName += chr;

            br.BaseStream.Position = 444;

            //forms the maps scenario name - e.g. scenario\multi\lockout\lockout
            scenarioName = "";
            while ((chr = br.ReadChar()) != '\0')
                scenarioName += chr;

            //+10
            br.BaseStream.Position = 704;

            //file strings info
            fileTableCount = br.ReadInt32();
            fileTableOffset = br.ReadInt32();
            fileTableSize = br.ReadInt32();
            fileTableIndexOffset = br.ReadInt32();

            //maps safecheck, its every byte after the header(2048bytes) xored
            checksum = br.ReadInt32();

            //+1320
            br.BaseStream.Position = 2044;

            //the footer (toof)
            foot = map.Reverse(new string(br.ReadChars(4)));
        }

        public void WriteHeader()
        {
            BinaryWriter bw = new BinaryWriter(new FileStream(map.path, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            byte nul = 0; //a nothing byte
            int diff = 0; //int for storing difference

            //begin writing
            bw.Write(map.Reverse(head).ToCharArray());
            bw.Write(version);
            bw.Write(filesize);

            bw.BaseStream.Position = 16;

            bw.Write(indexOffset);
            bw.Write(indexSize);

            bw.Write(metaSize);
            bw.Write(nonRawSize);

            bw.Write(origin.ToCharArray());
            diff = 256 - origin.Length;
            for (int x = 0; x < diff; x++) bw.Write(nul);

            bw.BaseStream.Position = 288;

            bw.Write(buildDate.ToCharArray());
            diff = 32 - buildDate.Length;
            for (int x = 0; x < diff; x++) bw.Write(nul);

            bw.BaseStream.Position = 320;
            bw.Write(mapTypeInt);

            bw.BaseStream.Position = 340;

            bw.Write(strangeFileStringsSize);
            bw.Write(strangeFileStringsOffset);

            bw.BaseStream.Position = 352;

            bw.Write(stringsOffset128);
            bw.Write(stringCount);
            bw.Write(stringsSize);
            bw.Write(stringsIndex);
            bw.Write(stringsOffset);

            bw.BaseStream.Position = 408;

            bw.Write(internalName.ToCharArray());
            diff = 32 - internalName.Length;
            for (int x = 0; x < diff; x++) bw.Write(nul);

            bw.BaseStream.Position = 444;

            bw.Write(scenarioName.ToCharArray());
            diff = 256 - scenarioName.Length;
            for (int x = 0; x < diff; x++) bw.Write(nul);

            bw.BaseStream.Position = 704;

            bw.Write(fileTableCount);
            bw.Write(fileTableOffset);
            bw.Write(fileTableSize);
            bw.Write(fileTableIndexOffset);

            bw.Write(checksum);

            bw.BaseStream.Position = 2044;
            bw.Write(map.Reverse(foot).ToCharArray());

            bw.Close();
        }
    }
}
