using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Windows.Forms;

namespace Dark
{
    public enum RefTypes
    {
        Reflexive,
        Dependency,
        LoneID,
        String
    }

    public struct Reflex
    {
        public string Description;
        public long Offset;
        public int Translation;
        public int ChunkCount;
    }

    public struct Ident
    {
        public string Description;
        public long Offset;
        public string Type;
        public string Path;
        public int ID;
        public RefTypes RefType;
    }

    public struct StringID
    {
        public string Description;
        public long Offset;
        public short Index;
        public string Name;
    }

    public class References //Reflexives, Idents and Strings
    {
        public Map map;

        public NodeType Filter = NodeType.Reflexive;
        public bool Scanned;

        public List<Reflex> rList = new List<Reflex>();
        public List<Ident> iList = new List<Ident>();
        public List<StringID> sList = new List<StringID>();

        public References(Map map)
        {
            this.map = map;
        }

        public void ManualScan(int index, int bspIndex)
        {
            Scanned = false;

            int MetaOffset = map.tags.tagsList[index].MetaOffset,
                  MetaSize = map.tags.tagsList[index].MetaSize;

            rList.Clear();
            iList.Clear();
            sList.Clear();

            BinaryReader br = new BinaryReader(new FileStream(map.path, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            map.StartProgress(MetaSize - 4);

            for (int x = 0; x < MetaSize; x += 4)
            {
                bool typeFound = false;

                #region Reflexives

                br.BaseStream.Position = MetaOffset + x;

                //read the reflexive
                int TranslatedReflexive = br.ReadInt32();
                //and then traslate it
                TranslatedReflexive -= bspIndex == -1 ? 
                    map.index.secondaryMagic : map.sbsp.sbspInfoList[bspIndex].Magic;

                //if the offset is within the current tags meta
                if (TranslatedReflexive >= MetaOffset &&
                    TranslatedReflexive <= MetaOffset + MetaSize)
                {
                    //we have a reflexive
                    Reflex refx = new Reflex();

                    refx.Offset = br.BaseStream.Position - MetaOffset - 4;

                    br.BaseStream.Position = MetaOffset + x - 4;
                    refx.ChunkCount = br.ReadInt32();
                    if (refx.ChunkCount <= 0) refx.ChunkCount = 1;

                    refx.Translation = TranslatedReflexive;
                    rList.Add(refx);

                    typeFound = true;
                }
                #endregion

                if (!typeFound)
                {
                    #region Idents

                    br.BaseStream.Position = MetaOffset + x;
                    string temptagtype = "";

                    /* the reason we dont use .readchar() here is because
                     * there might not be a char at our current position, so instead
                     * the reader would advance fwd until it found one.. this would
                     * lead to us seeing the same tag several times over */

                    for (int q = 0; q < 4; q++)
                        temptagtype += Convert.ToChar(br.ReadByte());

                    //reverse
                    temptagtype = map.Reverse(temptagtype);
                    int tempID = br.ReadInt32(); //read ID

                    //go through tags
                    for (int b = 0; b < map.tags.tagsList.Length; b++)
                    {
                        //if type match and id is null or match
                        if (temptagtype == map.tags.tagsList[b].Type &&
                            (tempID == -1 || tempID == map.tags.tagsList[b].Ident))
                        {
                            //we have a dep
                            Ident dep = new Ident();
                            dep.Offset = MetaOffset + x;
                            dep.Type = temptagtype;
                            dep.ID = tempID;
                            dep.Path = tempID == -1 ? "Nulled Out" : map.tags.tagsList[b].Path;
                            dep.RefType = RefTypes.Dependency;
                            iList.Add(dep);

                            x += 4; //so to avoid the ID
                            typeFound = true;
                            break;
                        }
                        //else if id match
                        else if (tempID == map.tags.tagsList[b].Ident)
                        {
                            //we have a loneID
                            Ident lone = new Ident();
                            lone.Offset = MetaOffset + x;
                            lone.ID = tempID;
                            lone.Type = map.tags.tagsList[b].Type;
                            lone.Path = map.tags.tagsList[b].Path;
                            lone.RefType = RefTypes.LoneID;
                            iList.Add(lone);

                            x += 4;
                            typeFound = true;
                            break;
                        }
                    }
                    #endregion

                    if (!typeFound)
                    {
                        #region Strings

                        br.BaseStream.Position = MetaOffset + x;

                        short strIndx = br.ReadInt16();
                        byte zero = br.ReadByte(), strLen = br.ReadByte();

                        if (zero == 0 &&
                            strIndx > 0 && strIndx < map.header.stringCount &&
                            strLen == map.strings.stringsList[strIndx].Length)
                        {
                            StringID str = new StringID();
                            str.Offset = MetaOffset + x;
                            str.Name = map.strings.stringsList[strIndx];
                            sList.Add(str);
                        }
                        #endregion
                    }
                }

                //update the progressbar value
                map.UpdateProgress("Reading " + map.tags.tagsList[index].Path + "...", x);
            }

            br.Close();
            Scanned = true;
            map.EndProgress();
        }

        public void EntScan()
        {
            Scanned = false;

            rList.Clear();
            iList.Clear();
            sList.Clear();

            BinaryReader br = new BinaryReader(new FileStream(map.path, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            map.StartProgress(map.entReader.Nodes.Count);

            for (int x = 0; x < map.entReader.Nodes.Count; x++)
            {
                    switch (map.entReader.Nodes[x].NodeType)
                    {
                        #region Reflexive
                        case NodeType.Reflexive:
                            //--
                            StructNode sn = (map.entReader.Nodes[x] as StructNode);

                            if (sn.Parent == null)
                            {
                                Reflex refx = new Reflex();
                                refx.Description = sn.Name;
                                refx.Offset = sn.Offset;

                                br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset
                                    + refx.Offset;

                                refx.ChunkCount = sn.chunkCount = br.ReadInt32();
                                sn.translation = br.ReadInt32();

                                refx.Translation = 
                                    sn.translation -= map.tags.tagsList[map.tags.selectedIndex].Type == "sbsp" ?
                                    map.sbsp.sbspInfoList[map.sbsp.FindIndexByTagID(map.tags.tagsList[map.tags.selectedIndex].Ident)].Magic
                                    : map.index.secondaryMagic;

                                sn.Enabled = (sn.translation >= 0);
                                rList.Add(refx);
                            }
                            else if (sn.Parent.Enabled)
                            {
                                for (int y = 0; y < sn.Parent.chunkCount; y++)
                                {
                                    Reflex refx = new Reflex();
                                    refx.Description = sn.Name + " [" + y.ToString() + "]";

                                    br.BaseStream.Position = sn.Parent.translation
                                        + (y * sn.Parent.Size) + sn.Offset;

                                    refx.Offset = br.BaseStream.Position
                                        - map.tags.tagsList[map.tags.selectedIndex].MetaOffset;

                                    refx.ChunkCount = sn.chunkCount = br.ReadInt32();
                                    sn.translation = br.ReadInt32();

                                    refx.Translation =
                                    sn.translation -= map.tags.tagsList[map.tags.selectedIndex].Type == "sbsp" ?
                                        map.sbsp.sbspInfoList[map.sbsp.FindIndexByTagID(map.tags.tagsList[map.tags.selectedIndex].Ident)].Magic
                                        : map.index.secondaryMagic;

                                    rList.Add(refx);
                                }
                            }
                            break;
                        #endregion

                        #region Dependency
                        case NodeType.Tag:
                            //--
                            TagNode tn = (map.entReader.Nodes[x] as TagNode);

                            if (tn.Parent == null)
                            {
                                Ident dep = new Ident();
                                dep.Description = tn.Name;
                                dep.Offset = tn.Offset;

                                br.BaseStream.Position =
                                    map.tags.tagsList[map.tags.selectedIndex].MetaOffset + dep.Offset;

                                dep.Type = map.Reverse(new string(br.ReadChars(4)));
                                dep.ID = br.ReadInt32();
                                int indx = map.tags.FindTagIndexByID(dep.ID);
                                dep.Path = indx == -1 ? "Nulled Out" : map.tags.tagsList[indx].Path;
                                dep.RefType = RefTypes.Dependency;
                                iList.Add(dep);
                            }
                            else if (tn.Parent.Enabled)
                            {
                                for (int y = 0; y < tn.Parent.chunkCount; y++)
                                {
                                    Ident dep = new Ident();
                                    dep.Description = tn.Name + " [" + y.ToString() + "]";

                                    br.BaseStream.Position = tn.Parent.translation
                                        + (y * tn.Parent.Size) + tn.Offset;

                                    dep.Offset = br.BaseStream.Position
                                        - map.tags.tagsList[map.tags.selectedIndex].MetaOffset;

                                    dep.Type = map.Reverse(new string(br.ReadChars(4)));
                                    dep.ID = br.ReadInt32();
                                    int indx = map.tags.FindTagIndexByID(dep.ID);
                                    dep.Path = indx == -1 ? "Nulled Out" : map.tags.tagsList[indx].Path;
                                    dep.RefType = RefTypes.Dependency;
                                    iList.Add(dep);
                                }
                            }
                            break;
                        #endregion

                        #region LoneID
                        case NodeType.ID:
                            //--
                            IDNode idn = (map.entReader.Nodes[x] as IDNode);

                            if (idn.Parent == null)
                            {
                                Ident lone = new Ident();
                                lone.Description = idn.Name;
                                lone.Offset = idn.Offset;

                                br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset
                                    + lone.Offset;

                                lone.ID = br.ReadInt32();
                                int indx = map.tags.FindTagIndexByID(lone.ID);
                                lone.Type = map.tags.tagsList[indx].Type;
                                lone.Path = map.tags.tagsList[indx].Path;
                                lone.RefType = RefTypes.LoneID;
                                iList.Add(lone);
                            }
                            else if (idn.Parent.Enabled)
                            {
                                for (int y = 0; y < idn.Parent.chunkCount; y++)
                                {
                                    Ident lone = new Ident();
                                    lone.Description = idn.Name + " [" + y.ToString() + "]";
                                    
                                    br.BaseStream.Position = idn.Parent.translation
                                        + (y * idn.Parent.Size) + idn.Offset;

                                    lone.Offset = br.BaseStream.Position
                                        - map.tags.tagsList[map.tags.selectedIndex].MetaOffset;
                                    
                                    lone.ID = br.ReadInt32();
                                    int indx = map.tags.FindTagIndexByID(lone.ID);
                                    lone.Type = map.tags.tagsList[indx].Type;
                                    lone.Path = map.tags.tagsList[indx].Path;
                                    lone.RefType = RefTypes.LoneID;
                                    iList.Add(lone);
                                }
                            }
                            break;
                        #endregion

                        #region StringID
                        case NodeType.StringID:
                            //--
                            StringIDNode sidn = (map.entReader.Nodes[x] as StringIDNode);

                            if (sidn.Parent == null)
                            {
                                StringID str = new StringID();
                                str.Description = sidn.Name;
                                str.Offset = sidn.Offset;

                                br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset
                                    + str.Offset;

                                str.Index = br.ReadInt16();
                                if (str.Index >= 0 && str.Index < map.strings.stringsList.Length)
                                {
                                    str.Name = map.strings.stringsList[str.Index];
                                    sList.Add(str);
                                }
                            }
                            else if (sidn.Parent.Enabled)
                            {
                                for (int y = 0; y < sidn.Parent.chunkCount; y++)
                                {
                                    StringID str = new StringID();
                                    str.Description = sidn.Name + " [" + y.ToString() + "]";

                                    br.BaseStream.Position = sidn.Parent.translation
                                        + (y * sidn.Parent.Size) + sidn.Offset;

                                    str.Offset = br.BaseStream.Position
                                        - map.tags.tagsList[map.tags.selectedIndex].MetaOffset;

                                    str.Index = br.ReadInt16();
                                    if (str.Index >= 0 && str.Index < map.strings.stringsList.Length)
                                    {
                                        str.Name = map.strings.stringsList[str.Index];
                                        sList.Add(str);
                                    }
                                }
                            }
                            break;
                        #endregion
                    }

                //update the progbar val
                map.UpdateProgress("Reading " + map.tags.tagsList[map.tags.selectedIndex].Path + "...", x);
            }

            br.Close();
            Scanned = true;
            map.EndProgress();
        }

        public ListViewItem[] LoadRefsIntoListviewItems()
        {
            List<ListViewItem> rslt = new List<ListViewItem>();

            switch (Filter)
            {
                case NodeType.Reflexive:
                    for (int x = 0; x < rList.Count; x++)
                        rslt.Add(MakeItem(
                            rList[x].Description,
                            rList[x].Offset.ToString(),
                            rList[x].Translation.ToString(),
                            rList[x].ChunkCount.ToString()));
                    break;

                case NodeType.ID:
                    string xType;
                    for (int x = 0; x < iList.Count; x++)
                    {
                        xType = iList[x].Type;

                        if (iList[x].Path == "Nulled Out" && !map.tags.tagTypes.Contains(xType))
                            xType = "";

                        rslt.Add(MakeItem(
                            iList[x].Description,
                            iList[x].Offset.ToString(),
                            xType,
                            iList[x].Path));
                    }
                    break;

                case NodeType.StringID:
                    for (int x = 0; x < sList.Count; x++)
                        rslt.Add(MakeItem(
                            sList[x].Description,
                            sList[x].Offset.ToString(),
                            sList[x].Name,
                            null));
                    break;
            }

            return rslt.ToArray();
        }

        public ListViewItem MakeItem(string txt1, string txt2, string txt3, string txt4)
        {
            ListViewItem a = new ListViewItem();
            a.Text = txt1;

            a.SubItems.Add(txt2);
            a.SubItems.Add(txt3);

            if (txt4 != String.Empty) a.SubItems.Add(txt4);
            return a;
        }

        public void SwapIdent(int index, string newTagType, string newTagPath)
        {
            BinaryWriter bw = new BinaryWriter(new FileStream(map.path, FileMode.Open,
               FileAccess.ReadWrite, FileShare.ReadWrite));

            if (iList[index].RefType == RefTypes.Dependency)
            {
                bw.BaseStream.Position =
                    map.tags.tagsList[map.tags.selectedIndex].MetaOffset + iList[index].Offset;

                //reverse the newtagtype and convert it into an array of chars
                char[] chr = map.Reverse(newTagType).ToCharArray();

                if (newTagPath == "Nulled Out")
                {
                    //write in the tagtype, and then -1.. -1 means null
                    bw.Write(chr);
                    bw.Write(-1);
                }
                else
                {
                    //go through all the tags
                    for (int x = 0; x < map.tags.tagsList.Length; x++)
                    {
                        //if we have found the tag
                        if (map.tags.tagsList[x].Type == newTagType && map.tags.tagsList[x].Path == newTagPath)
                        {
                            //write in the tagtype and its id
                            bw.Write(chr);
                            bw.Write(map.tags.tagsList[x].Ident);
                            break;
                        }
                    }
                }
            }
            else if (iList[index].RefType == RefTypes.LoneID)
            {
                //+4 as we only need to write in the tags id
                bw.BaseStream.Position =
                    map.tags.tagsList[map.tags.selectedIndex].MetaOffset + iList[index].Offset + 4;

                //go through all the tags
                for (int x = 0; x < map.tags.tagsList.Length; x++)
                {
                    //if we have found the tag
                    if (map.tags.tagsList[x].Type == newTagType && map.tags.tagsList[x].Path == newTagPath)
                    {
                        //write in its id
                        bw.Write(map.tags.tagsList[x].Ident);
                        break;
                    }
                }
            }
            bw.Close();
        }

        public void SwapString(int index, string newString, int newStringIndex)
        {
            BinaryWriter bw = new BinaryWriter(new FileStream(map.path, FileMode.Open,
               FileAccess.ReadWrite, FileShare.ReadWrite));

            //goto the strings offset
            bw.BaseStream.Position =
                map.tags.tagsList[map.tags.selectedIndex].MetaOffset + map.refs.sList[index].Offset;

            //write away
            bw.Write((short)newStringIndex);
            bw.Write((byte)0);
            bw.Write((byte)map.strings.stringsList[newStringIndex].Length);

            bw.Close();
        }
    }
}
