using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Windows.Forms;

namespace Dark
{
    public struct Tag
    {
        public int Ident;
        public string Type;

        public string Path;
        public int PathOffset;
        public int PathLength;

        public int RawMetaOffset;
        public int MetaOffset;
        public int MetaSize;

        public string MetaExtractPath;
    }

    public enum SortTagsBy
    {
        TagType,
        FolderHierarchy
    }

    public class Tags
    {
        public Map map;

        public Tag[] tagsList;
        public int selectedIndex = -1;

        public SortTagsBy sortTagsBy = SortTagsBy.TagType;
        public List<string> tagTypes = new List<string>();

        public Tags(Map map, BinaryReader br)
        {
            this.map = map;

            br.BaseStream.Position = map.index.offsetOfObjectIndex;

            //will store the tags of the map
            tagsList = new Tag[map.index.objectCount];
            tagTypes.Clear();

            //count through the 'said' tag amount
            for (int x = 0; x < map.index.objectCount; x++)
            {
                //this is the tagtype, they are backward in map so we reverse
                string tt = map.Reverse(new string(br.ReadChars(4)));

                //if its not in the list add it
                if (!tagTypes.Contains(tt)) tagTypes.Add(tt);

                //and this is the rest of the info
                tagsList[x].Type = tt;
                tagsList[x].Ident = br.ReadInt32();
                tagsList[x].RawMetaOffset = br.ReadInt32();
                tagsList[x].MetaOffset = tagsList[x].RawMetaOffset - map.index.secondaryMagic;
                tagsList[x].MetaSize = br.ReadInt32();
            }

            tagTypes.Sort();

            //go to the filetable which contains all the strings - globals/globals etc
            br.BaseStream.Position = (int)map.header.fileTableOffset;

            //go through the tags and add in their paths
            char chr;
            for (int i = 0; i < map.index.objectCount; i++)
            {
                tagsList[i].PathOffset = (int)br.BaseStream.Position;

                while ((chr = br.ReadChar()) != '\0')
                    tagsList[i].Path += chr;

                tagsList[i].PathLength = tagsList[i].Path.Length;
            }
        }

        public int FindTagIndexByID(int ID)
        {
            //go through tags
            for (int x = 0; x < tagsList.Length; x++)
            {
                //if theres an ID match
                if (tagsList[x].Ident == ID)
                    return x; //return the tags index in the list
            }
            return -1; //no matches, so return -1
        }

        public int FindTagIndexByTypeAndPath(string type, string path)
        {
            //go through tags
            for (int x = 0; x < tagsList.Length; x++)
            {
                //if theres a type and path match 
                if (tagsList[x].Type == type && tagsList[x].Path == path)
                    return x; //return the tags index in the list
            }
            return -1; //no matches, so return -1
        }

        public TreeNode[] LoadTagsIntoTreenodes(string tagType)
        {
            if (tagType == String.Empty) // then we load all the tags
            {
                switch (sortTagsBy)
                {
                    case SortTagsBy.TagType:
                        //the list that will contain the tagtypes that will
                        //contain the tags respectively
                        TreeNode[] tnList1 = new TreeNode[tagTypes.Count];

                        //go through tagtypes
                        for (int x = 0; x < tagTypes.Count; x++)
                        {
                            //make a new node for the type
                            tnList1[x] = new TreeNode(tagTypes[x], 1, 1);

                            //go through the tags
                            for (int y = 0; y < tagsList.Length; y++)
                            {
                                //if tag[y]'s type is a match to the current type
                                if (tagsList[y].Type == tagTypes[x])
                                    tnList1[x].Nodes.Add("", tagsList[y].Path, 2, 2); //add it
                            }
                        }
                        return tnList1; //return the now pop. list

                    case SortTagsBy.FolderHierarchy:
                        //for storing the tags
                        List<TreeNode> tnList2 = new List<TreeNode>();

                        //go through the tags
                        for (int x = 0; x < tagsList.Length; x++)
                        {
                            string[] str = tagsList[x].Path.Split('\\');

                            if (str.Length == 1) tnList2.Add(new TreeNode(str[0] + "." + tagsList[x].Type, 2, 2));
                            else
                            {
                                TreeNode node = null;
                                foreach (TreeNode tn in tnList2)
                                {
                                    if (tn.Text == str[0])
                                    {
                                        node = tn;
                                        break;
                                    }
                                }

                                if (node == null)
                                {
                                    node = new TreeNode(str[0], 1, 1);
                                    tnList2.Add(node);
                                }

                                for (int y = 1; y < str.Length - 1; y++)
                                {
                                    bool exists = false;
                                    foreach (TreeNode tn in node.Nodes)
                                    {
                                        if (tn.Text == str[y])
                                        {
                                            node = tn;
                                            exists = true;
                                            break;
                                        }
                                    }
                                    if (!exists) node = node.Nodes.Add("", str[y], 1, 1);
                                }
                                node.Nodes.Add("", str[str.Length - 1] + "." + tagsList[x].Type, 2, 2);
                            }
                        }
                        return tnList2.ToArray();

                    default:
                        return null;
                }
            }
            else // we load the tag the user wants
            {
                //size of 1 as we'll only be loading one tag
                TreeNode[] tnList3 = new TreeNode[1];

                //give the node the name that the user has given
                tnList3[0] = new TreeNode(tagType);

                //go through the tags
                for (int y = 0; y < tagsList.Length; y++)
                {
                    //if tag[y]'s type is a match to the type
                    if (tagsList[y].Type == tagType)
                        tnList3[0].Nodes.Add(tagsList[y].Path); //add it
                }
                return tnList3; //return the now pop. list
            }
        }

        public string[] LoadTagsFromTypeIntoStrings(string type)
        {
            if (type == String.Empty) // then we load all the tags
            {
                //list for tags
                string[] tmpList = new string[tagsList.Length];

                //go through and add all tags to the list
                for (int i = 0; i < tagsList.Length; i++)
                    tmpList[i] = tagsList[i].Path;

                return tmpList; //return the list

            }
            else // we load the tag the user wants
            {
                //we use List<> as we dont know how many tags
                //we'll be loading
                List<string> tmplist = new List<string>();

                //go through the tags
                for (int i = 0; i < tagsList.Length; i++)
                {
                    //if tag[i]'s type matches what we want
                    if (tagsList[i].Type == type)
                        tmplist.Add(tagsList[i].Path); //add it
                }

                return tmplist.ToArray(); //convert and return the array
            }
        }

        public void WriteSelectedTag()
        {
            BinaryWriter bw = new BinaryWriter(new FileStream(map.path, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            bw.BaseStream.Position = map.index.offsetOfObjectIndex + (selectedIndex * 16);

            bw.Write(map.Reverse(tagsList[selectedIndex].Type).ToCharArray());
            bw.Write(tagsList[selectedIndex].Ident);
            bw.Write(tagsList[selectedIndex].RawMetaOffset);
            bw.Write(tagsList[selectedIndex].MetaSize);

            //if new path length is same as old, just write it
            if (tagsList[selectedIndex].Path.Length == tagsList[selectedIndex].PathLength)
            {
                bw.BaseStream.Position = tagsList[selectedIndex].PathOffset;
                bw.Write(tagsList[selectedIndex].Path.ToCharArray());
            }
            else //we'll need todo some shifting
            {
                //create reader
                BinaryReader br = new BinaryReader(bw.BaseStream);

                //goto just after string in filetable
                br.BaseStream.Position = tagsList[selectedIndex].PathOffset + tagsList[selectedIndex].PathLength;

                //store eof filetable from position
                byte[] eofFileTable = br.ReadBytes(
                    (map.header.fileTableOffset + map.header.fileTableSize) - (int)br.BaseStream.Position);

                int oldPaddingLen = map.header.fileTableIndexOffset - (int)br.BaseStream.Position;

                //goto file index
                br.BaseStream.Position = map.header.fileTableIndexOffset;

                //store map downward from position
                byte[] mapHalf = br.ReadBytes((int)(br.BaseStream.Length - br.BaseStream.Position));

                //goto path offset
                bw.BaseStream.Position = tagsList[selectedIndex].PathOffset;

                //write in new path
                bw.Write(tagsList[selectedIndex].Path.ToCharArray());

                //and eof
                bw.Write(eofFileTable);

                //shift
                int shift = tagsList[selectedIndex].Path.Length - tagsList[selectedIndex].PathLength;
                tagsList[selectedIndex].PathLength = tagsList[selectedIndex].Path.Length;

                //--[Header]
                bw.BaseStream.Position = 712;
                bw.Write(map.header.fileTableSize += shift);

                //--[FileTable - FileIndex Padding]

                //get padding length
                int newPaddingLen = 512 - (map.header.fileTableSize % 512);

                //goto end of filetable
                bw.BaseStream.Position = map.header.fileTableOffset + map.header.fileTableSize;

                if (newPaddingLen == 512)
                {
                    map.header.fileTableIndexOffset = (int)bw.BaseStream.Position;
                    newPaddingLen = 0;

                    //write stored half
                    bw.Write(mapHalf);
                }
                else //means remainder wasnt 0
                {
                    byte[] padding = new byte[newPaddingLen];

                    //map.header.fileTableIndexOffset += shift;
                    //shift += (map.header.fileTableIndexOffset - (int)bw.BaseStream.Position) - paddingLen;
                    //shift += newPaddingLen - (map.header.fileTableIndexOffset - (int)bw.BaseStream.Position);

                    bw.Write(padding);
                    map.header.fileTableIndexOffset = (int)bw.BaseStream.Position;
                    bw.Write(mapHalf);
                }

                shift += newPaddingLen - oldPaddingLen;

                //--[Header]
                bw.BaseStream.Position = 716;
                bw.Write(map.header.fileTableIndexOffset);

                //--[FileIndex]

                //because last tag (ugh - coconuts) doesnt have a file index
                if (selectedIndex != tagsList.Length - 1)
                {
                    br.BaseStream.Position = map.header.fileTableIndexOffset + (selectedIndex * 4);
                    int len = br.ReadInt32();

                    for (int x = selectedIndex; x < tagsList.Length - 1; x++)
                    {
                        len += tagsList[x].PathLength + 1;
                        bw.Write(len);
                    }
                }

                if (shift != 0) //the padding hasnt realigned the rest of the map.. so we have to
                {
                    //--[Header]
                    bw.BaseStream.Position = 16;
                    bw.Write(map.header.indexOffset += shift);

                    map.header.metaStart += shift;

                    br.BaseStream.Position = 344;
                    bw.Write(map.header.strangeFileStringsOffset += shift);

                    //--[Index]
                    map.index.primaryMagic -= shift;
                    map.index.offsetOfObjectIndex += shift;
                    map.index.secondaryMagic -= shift;

                    //--[Unicode]
                    int matgOffset400 = tagsList[0].MetaOffset + 400 + shift;
                    Lang[] langs = new Lang[8];

                    //-Read and shift
                    for (int x = 0; x < 8; x++)
                    {
                        br.BaseStream.Position = matgOffset400 + (28 * x) + 8;
                        langs[x].indexPointerOffset = br.BaseStream.Position;

                        langs[x].indexOffset = br.ReadInt32() + shift;
                        langs[x].tableOffset = br.ReadInt32() + shift;
                    }

                    //-Write
                    for (int x = 0; x < 8; x++)
                    {
                        bw.BaseStream.Position = langs[x].indexPointerOffset;
                        bw.Write(langs[x].indexOffset);
                        bw.Write(langs[x].tableOffset);
                    }

                    //--[Bitmaps]
                    List<int> bitmTags = new List<int>();
                    for (int x = 0; x < tagsList.Length; x++)
                    {
                        if (tagsList[x].Type == "bitm")
                            bitmTags.Add(x);
                    }

                    for (int z = 0; z < bitmTags.Count; z++)
                    {
                        br.BaseStream.Position = tagsList[bitmTags[z]].MetaOffset + 68 + shift;

                        int tempc = br.ReadInt32();
                        int tempr = br.ReadInt32() - map.index.secondaryMagic;

                        for (int x = 0; x < tempc; x++)
                        {
                            for (int y = 0; y < 6; y++)
                            {
                                int PointerOffset = tempr + (x * 116) + 28 + (y * 4);
                                br.BaseStream.Position = PointerOffset;

                                int Offset = br.ReadInt32();
                                if (Offset == -1) break;

                                RawLocations rawLoc = RawLocations.Internal;
                                map.rawData.ParseSPointer(ref Offset, ref rawLoc);

                                if (rawLoc == RawLocations.Internal)
                                {
                                    bw.BaseStream.Position = PointerOffset;
                                    bw.Write(Offset + shift);
                                }
                            }
                        }
                    }

                    //--[Meta Offsets]
                    for (int x = 0; x < tagsList.Length; x++)
                    {
                        if (tagsList[x].Type != "sbsp" && tagsList[x].Type != "ltmp")
                            tagsList[x].MetaOffset += shift;
                    }

                    //--[End]
                    bw.BaseStream.Position = tagsList[tagsList.Length - 1].MetaOffset +
                        tagsList[tagsList.Length - 1].MetaSize;

                    newPaddingLen = 512 - ((int)bw.BaseStream.Position % 512);
                    newPaddingLen += newPaddingLen == 512 ? 512 : 1024;

                    int eof = (int)bw.BaseStream.Position + newPaddingLen;

                    while (bw.BaseStream.Position < eof)
                        bw.Write((byte)205); //CD

                    bw.BaseStream.SetLength(eof);

                    //--[Header]
                    bw.BaseStream.Position = 8;
                    bw.Write(eof);
                }
                br.Close();
            }
            bw.Close();
        }

        public Tag DuplicateSelectedTag()
        {
            //open map in stream
            FileStream fs = new FileStream(map.path, FileMode.Open,
              FileAccess.ReadWrite, FileShare.ReadWrite);

            //open stream in reader and writer
            BinaryReader br = new BinaryReader(fs);
            BinaryWriter bw = new BinaryWriter(fs);

            //tags
            int ughIndx = tagsList.Length - 1;
            Tag dupTag = map.tags.tagsList[map.tags.selectedIndex];
            Tag newTag = new Tag();

            //get tag meta
            br.BaseStream.Position = dupTag.MetaOffset;
            byte[] newTagMeta = br.ReadBytes(dupTag.MetaSize);
            newTag.MetaSize = dupTag.MetaSize;

            //--[Unicode]
            //we need the offset of index1 for later on
            br.BaseStream.Position = tagsList[0].MetaOffset + 408;
            int firstUnicodeIndexOffset = br.ReadInt32();

            //store map downward from file index
            br.BaseStream.Position = map.header.fileTableIndexOffset;
            int mapHalfOffset = (int)br.BaseStream.Position;
            byte[] mapHalf = br.ReadBytes((int)br.BaseStream.Length - mapHalfOffset);

            //--[FileTable]
            int oldPaddingLen = map.header.fileTableIndexOffset - (tagsList[ughIndx].PathOffset + tagsList[ughIndx].PathLength + 1);

            //go to path start of last tag
            bw.BaseStream.Position = tagsList[ughIndx].PathOffset;
            
            //make new tagPath for new tag
            newTag.Type = dupTag.Type;
            newTag.Path = dupTag.Path;
            int num = 2;
            while (FindTagIndexByTypeAndPath(newTag.Type, newTag.Path) != -1)
                newTag.Path = dupTag.Path + (num++).ToString();

            //write new path in
            newTag.PathLength = newTag.Path.Length;
            newTag.PathOffset = (int)bw.BaseStream.Position;
            bw.Write(newTag.Path.ToCharArray());
            bw.Write((byte)0);
            int newTagIndx = (int)bw.BaseStream.Position - map.header.fileTableOffset;

            //write ugh back in
            tagsList[ughIndx].PathOffset = (int)bw.BaseStream.Position;
            bw.Write(tagsList[ughIndx].Path.ToCharArray());
            bw.Write((byte)0);

            //shift
            int shift = newTag.PathLength + 1;
            map.header.fileTableSize += shift;
            map.header.fileTableCount += 1;

            //shift = (newTag.PathLength + 1) - oldPaddingLen;
            //if (shift > 0)
            //{
            //    byte[] padding = new byte[512 - shift];
            //    bw.Write(padding);
            //    shift += padding.Length;
            //}
            //else if (shift < 0)
            //{
            //    byte[] padding = new byte[512 - (map.header.fileTableSize % 512)];
            //    bw.Write(padding);
            //    shift = 0;
            //}

            //padding 
            int newPaddingLen = 512 - (map.header.fileTableSize % 512);
            if (newPaddingLen == 512) newPaddingLen = 0;
            else
            {
                byte[] padding = new byte[newPaddingLen];
                bw.Write(padding);
            }
            //shift = (map.header.fileTableSize + newPaddingLen + newTag.PathLength + 1)
            //    - (map.header.fileTableSize + oldPaddingLen);

            
            shift += newPaddingLen;
            shift -= oldPaddingLen;

            //new file index offset
            map.header.fileTableIndexOffset = (int)bw.BaseStream.Position;

            //--[FileIndex]
            //write file index in
            int mhIndx = 0;
            int mhCount = (tagsList.Length) * 4;
            bw.Write(mapHalf, mhIndx, mhCount);

            oldPaddingLen = firstUnicodeIndexOffset - (mapHalfOffset + mhCount);

            //write new path index in
            bw.Write(newTagIndx);
            shift += 4;

            //mhCount + 4 is new file index size
            newPaddingLen = 512 - ((mhCount + 4) % 512);

            if (newPaddingLen == 512) newPaddingLen = 0;
            else
            {
                byte[] padding = new byte[newPaddingLen];
                bw.Write(padding);
            }
            shift += newPaddingLen - oldPaddingLen;
            int uShift = shift;

            //-Write unicode
            mhIndx = mhCount + oldPaddingLen;
            mhCount = map.header.strangeFileStringsOffset - (mapHalfOffset + mhIndx);
            bw.Write(mapHalf, mhIndx, mhCount);

            map.header.strangeFileStringsOffset = (int)bw.BaseStream.Position;
            //-Write crazy and bitmaps
            mhIndx += mhCount;
            mhCount = map.header.indexOffset - (mapHalfOffset + mhIndx);
            bw.Write(mapHalf, mhIndx, mhCount);

            map.header.indexOffset = (int)bw.BaseStream.Position;
            //--Write index up to just before tag start
            mhIndx += mhCount;
            mhCount = map.index.offsetOfObjectIndex - (mapHalfOffset + mhIndx);
            bw.Write(mapHalf, mhIndx, mhCount);

            map.index.offsetOfObjectIndex = (int)bw.BaseStream.Position;
            //--Write tagslots in up to just before last tag
            mhIndx += mhCount;
            //mhCount = (map.index.offsetOfObjectIndex + ((map.tags.tagsList.Length - 1) * 16)) - (mapHalfOffset + mhIndx);
            mhCount = (map.tags.tagsList.Length - 1) * 16;
            bw.Write(mapHalf, mhIndx, mhCount);

            //--[Tags]
            //write new tag entry
            bw.Write(map.Reverse(newTag.Type).ToCharArray());
            bw.Write(newTag.Ident = tagsList[ughIndx].Ident);
            bw.Write(newTag.RawMetaOffset = tagsList[ughIndx].RawMetaOffset);
            newTag.MetaOffset = tagsList[ughIndx].MetaOffset;
            bw.Write(newTag.MetaSize);
            shift += 16;

            //and ugh tag
            bw.Write(map.Reverse(tagsList[ughIndx].Type).ToCharArray());
            bw.Write(tagsList[ughIndx].Ident + 65537); //each ident is 65537 apart
            bw.Write(tagsList[ughIndx].RawMetaOffset += newTag.MetaSize);
            tagsList[ughIndx].MetaOffset += newTag.MetaSize;
            bw.Write(tagsList[ughIndx].MetaSize);

            //index padding
            int indexDepth = (int)bw.BaseStream.Position - map.header.indexOffset;
            if (indexDepth > map.header.indexSize)
            {
                //hmm..
                //map.header.indexSize += 16;
                //map.header.nonRawSize += 16;
            }
            else
            {
                //byte[] padding = new byte[map.header.indexSize - depth];
                //bw.Write(padding);
                shift -= 16;
                //mhIndx += mhCount + 16 + 16 + padding.Length;
                mhIndx += mhCount + 16 + 16;
            }
            map.header.metaStart += shift;

            //write to just before last tags meta
            mhCount = newTag.MetaOffset - (mapHalfOffset + mhIndx);
            bw.Write(mapHalf, mhIndx, mhCount);

            //write in new tag meta
            bw.Write(newTagMeta);

            //write in ugh meta
            mhIndx += mhCount;
            mhCount = tagsList[ughIndx].MetaSize;
            bw.Write(mapHalf, mhIndx, mhCount);

            //map.header.metaSize += newTagMeta.Length;
            //map.header.nonRawSize += newTagMeta.Length;

            //--[End]
            newPaddingLen = 512 - ((int)bw.BaseStream.Position % 512);
            newPaddingLen += newPaddingLen == 512 ? 512 : 1024;

            int eof = (int)bw.BaseStream.Position + newPaddingLen;

            while (bw.BaseStream.Position < eof)
                bw.Write((byte)205); //CD

            bw.BaseStream.SetLength(eof);

            //--[Header]
            bw.BaseStream.Position = 8;
            bw.Write(eof);

            bw.BaseStream.Position = 16;
            bw.Write(map.header.indexOffset);
            bw.Write(map.header.indexSize);

            int diff = (eof - map.header.metaStart) - map.header.metaSize;
            bw.Write(map.header.metaSize += diff);
            bw.Write(map.header.nonRawSize += diff);

            bw.BaseStream.Position = 704;
            bw.Write(map.header.fileTableCount);
            bw.BaseStream.Position += 4;
            bw.Write(map.header.fileTableSize);
            bw.Write(map.header.fileTableIndexOffset);

            //--[Index]
            bw.BaseStream.Position = map.header.indexOffset + 24;
            bw.Write(map.index.objectCount += 1);

            //--[UGH! Reflexives]
            br.BaseStream.Position = (tagsList[ughIndx].MetaOffset+shift) + 4;
            for (int x = 4; x < tagsList[ughIndx].MetaSize; x += 4)
            {
                int rawReflx = br.ReadInt32();
                int transReflx = rawReflx - map.index.secondaryMagic;

                //if the offset is within the old ughtags meta
                if (transReflx >= newTag.MetaOffset &&
                    transReflx <= newTag.MetaOffset + tagsList[ughIndx].MetaSize)
                {
                    //we have a reflexive
                    bw.BaseStream.Position -= 4;
                    bw.Write(rawReflx + newTag.MetaSize);

                    bw.BaseStream.Position += 4;
                    x += 4;
                }
            }

            //--[NewTag Reflexives]
            br.BaseStream.Position = (newTag.MetaOffset + shift) + 4;
            int mShift = newTag.MetaOffset - dupTag.MetaOffset;
            for (int x = 4; x < newTag.MetaSize; x += 4)
            {
                int rawReflx = br.ReadInt32();
                int transReflx = rawReflx - map.index.secondaryMagic;

                //if the offset is within the old ughtags meta
                if (transReflx >= dupTag.MetaOffset &&
                    transReflx <= dupTag.MetaOffset + newTag.MetaSize)
                {
                    //we have a reflexive
                    bw.BaseStream.Position -= 4;
                    bw.Write(rawReflx + mShift);

                    bw.BaseStream.Position += 4;
                    x += 4;
                }
            }

            if (uShift != 0)
            {
                //--[Reflexives]
                for (int y = 0; y < map.tags.tagsList.Length; y++)
                {
                    //these tags are way above the norm meta
                    if (map.tags.tagsList[y].Type == "sbsp" || map.tags.tagsList[y].Type == "ltmp") continue;

                    //goto new metaoffset of tag + 4 (skip chunkcnt if ther is one)
                    br.BaseStream.Position = (map.tags.tagsList[y].MetaOffset + shift) + 4;

                    for (int x = 4; x < map.tags.tagsList[y].MetaSize; x += 4)
                    {
                        int rawReflx = br.ReadInt32();
                        int transReflx = rawReflx - map.index.secondaryMagic;

                        //if the offset is within the current tags meta
                        if (transReflx >= map.tags.tagsList[y].MetaOffset &&
                            transReflx <= map.tags.tagsList[y].MetaOffset + map.tags.tagsList[y].MetaSize)
                        {
                            //we have a reflexive
                            bw.BaseStream.Position -= 4;
                            bw.Write(rawReflx + uShift);

                            bw.BaseStream.Position += 4;
                            x += 4;
                        }
                    }
                }

                //new tag
                br.BaseStream.Position = (newTag.MetaOffset + shift) + 4;
                for (int x = 4; x < newTag.MetaSize; x += 4)
                {
                    int rawReflx = br.ReadInt32();
                    int transReflx = rawReflx - map.index.secondaryMagic;

                    //if the offset is within the current tags meta
                    if (transReflx >= newTag.MetaOffset &&
                        transReflx <= newTag.MetaOffset + newTag.MetaSize)
                    {
                        //we have a reflexive
                        bw.BaseStream.Position -= 4;
                        bw.Write(rawReflx + uShift);

                        bw.BaseStream.Position += 4;
                        x += 4;
                    }
                }
            }

            //--[Meta Offsets]
            if (shift != 0)
            {
                for (int x = 0; x < tagsList.Length; x++)
                {
                    if (tagsList[x].Type != "sbsp" && tagsList[x].Type != "ltmp")
                        tagsList[x].MetaOffset += shift;
                }
                newTag.MetaOffset += shift;
            }

            //map.index.primaryMagic -= shift;
            //map.index.secondaryMagic -= shift;

            if (uShift != 0)
            {
                //--[Unicode]
                int matgOffset400 = tagsList[0].MetaOffset + 400;
                Lang[] langs = new Lang[8];

                //-Read and shift
                for (int x = 0; x < 8; x++)
                {
                    br.BaseStream.Position = matgOffset400 + (28 * x) + 8;
                    langs[x].indexPointerOffset = br.BaseStream.Position;

                    langs[x].indexOffset = br.ReadInt32() + uShift;
                    langs[x].tableOffset = br.ReadInt32() + uShift;
                }

                //-Write
                for (int x = 0; x < 8; x++)
                {
                    bw.BaseStream.Position = langs[x].indexPointerOffset;
                    bw.Write(langs[x].indexOffset);
                    bw.Write(langs[x].tableOffset);
                }

                //--[Header]
                bw.BaseStream.Position = 344;
                bw.Write(map.header.strangeFileStringsOffset += uShift);

                //--[Bitmaps]
                List<int> bitmTags = new List<int>();
                for (int x = 0; x < tagsList.Length; x++)
                {
                    if (tagsList[x].Type == "bitm")
                        bitmTags.Add(x);
                }

                for (int z = 0; z < bitmTags.Count; z++)
                {
                    br.BaseStream.Position = tagsList[bitmTags[z]].MetaOffset + 68;

                    int tempc = br.ReadInt32();
                    int tempr = br.ReadInt32() - map.index.secondaryMagic;

                    for (int x = 0; x < tempc; x++)
                    {
                        for (int y = 0; y < 6; y++)
                        {
                            int PointerOffset = tempr + (x * 116) + 28 + (y * 4);
                            br.BaseStream.Position = PointerOffset;

                            int Offset = br.ReadInt32();
                            if (Offset == -1) break;

                            RawLocations rawLoc = RawLocations.Internal;
                            map.rawData.ParseSPointer(ref Offset, ref rawLoc);

                            if (rawLoc == RawLocations.Internal)
                            {
                                bw.BaseStream.Position = PointerOffset;
                                bw.Write(Offset + uShift);
                            }
                        }
                    }
                }

                //--[Tag Entries]
                bw.BaseStream.Position = map.index.offsetOfObjectIndex + 8;
                for (int x = 0; x < tagsList.Length-1; x++)
                {
                    if (map.tags.tagsList[x].Type == "sbsp" || map.tags.tagsList[x].Type == "ltmp") bw.BaseStream.Position += 16;
                    else
                    {
                        bw.Write(map.tags.tagsList[x].RawMetaOffset += uShift);
                        
                        //int tmp1 = map.tags.tagsList[x].MetaOffset;
                        //int tmp2 = map.tags.tagsList[x].RawMetaOffset-map.index.secondaryMagic;

                        bw.BaseStream.Position += 12;
                    }
                }
                bw.Write(newTag.RawMetaOffset += uShift);
            }
        
            //ughs new ident
            int newUghIdent = tagsList[ughIndx].Ident + 65537;

            //--[Matg Coconuts ref]
            bw.BaseStream.Position = tagsList[0].MetaOffset + 756;
            bw.Write(newUghIdent);

            //--[UGH! LoneIDs]   
            br.BaseStream.Position = tagsList[ughIndx].MetaOffset;
            for (int x = 0; x < tagsList[ughIndx].MetaSize; x += 4)
            {
                if (br.ReadInt32() == tagsList[ughIndx].Ident)
                {
                    bw.BaseStream.Position -= 4;
                    bw.Write(newUghIdent);
                }
            }
            tagsList[ughIndx].Ident = newUghIdent;

            bw.Close();
            br.Close();

            //add tag to array
            Array.Resize(ref tagsList, tagsList.Length + 1);
            tagsList[tagsList.Length - 1] = tagsList[ughIndx];
            tagsList[ughIndx] = newTag;
            return newTag;
        }
    }
}