using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Windows.Forms;

namespace Dark
{
    public class Map
    {
        public string path;

        public Header header;
        public Index index;
        public Tags tags;
        public Sbsp sbsp;
        public Strings strings;

        public References refs;
        public RawData rawData;

        public BitmapSpecific bitm;
        public LightMapSpecific ltmp;
        public UnicodeSpecific unic;
        public SoundSpecific snd_;
        public ThreeD mode;

        public EntReader entReader;

        public ToolStripProgressBar ProgressBar;
        public ToolStripStatusLabel ProgressLabel;

        public Map(ToolStripProgressBar progbar, ToolStripStatusLabel proglabel)
        {
            ProgressBar = progbar;
            ProgressLabel = proglabel;
        }

        public void Load(string mpath)
        {
            path = mpath;

            StartProgress(6);

            BinaryReader br = new BinaryReader(new FileStream(path, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            UpdateProgress("Initializing...", 1);
            refs = new References(this);
            rawData = new RawData(this);

            ltmp = new LightMapSpecific(this);
            bitm = new BitmapSpecific(this);
            unic = new UnicodeSpecific(this);
            snd_ = new SoundSpecific(this);
            mode = new ThreeD(this);

            entReader = new EntReader(this);

            UpdateProgress("Reading Header...", 1); header = new Header(this, br);
            UpdateProgress("Reading Index...", 2); index = new Index(this, br);
            UpdateProgress("Reading Tags...", 3); tags = new Tags(this, br);
            UpdateProgress("Reading Sbsp...", 3); sbsp = new Sbsp(this, br);
            UpdateProgress("Reading Strings...", 4); strings = new Strings(this, br);

            br.Close();
        }

        public string Reverse(string str)
        {
            char[] temp = str.ToCharArray();
            Array.Reverse(temp);
            return new string(temp);

            //OR
            //string rslt = "";
            //for (int x = str.Length - 1; x >= 0; x--)
            //    rslt += str.Substring(x, 1);
            //return rslt;
        }

        public void TagInjector(Tag tag, byte[] metaBytes)
        {
            //bring the map into a stream
            FileStream fs = new FileStream(path, FileMode.Open, FileAccess.ReadWrite, 
                FileShare.ReadWrite);

            //open the stream in a reader and writer
            BinaryWriter bw = new BinaryWriter(fs);
            BinaryReader br = new BinaryReader(fs);

            //now we add a number to the end of the tags path so it is unique
            //tag.Path = ReturnNewPath(tag.Path);

            /* the coconuts tag is special, it has to remain the last tag to be read.
             * so we first pull it out
             */
            Tag ughTag = tags.tagsList[tags.tagsList.Length - 1];

            /* now to continue the order of idents, we give our new tag the
             * ident that the ughTag had and give the ughTag a new one
             */
            tag.Ident = ughTag.Ident;
            ughTag.Ident += 65537; //each ident is 65537 apart

            //goto start of the file-index
            br.BaseStream.Position = header.fileTableIndexOffset;
            //and read it all into a byte array
            byte[] fileindexByteArray = br.ReadBytes(header.fileTableCount * 4);

            //get write point from the file-index
            br.BaseStream.Position -= 4;
            int writePoint = br.ReadInt32();

            //store away the map from the end of the file-index to the end of the map
            byte[] bottomHalf = br.ReadBytes((int)(br.BaseStream.Length - br.BaseStream.Position));

            //goto the writepoint within the filetable
            br.BaseStream.Position = header.fileTableOffset + writePoint;
            //store away the last string (ughstring) and the padding between the table and index
            byte[] endStringAndPadding = br.ReadBytes(header.fileTableIndexOffset - (int)br.BaseStream.Position);

            //go back to the writepoint within the filetable
            bw.BaseStream.Position = header.fileTableOffset + writePoint;

            //write in our tags path
            bw.Write(tag.Path.ToCharArray());

            //and a null termination
            bw.Write((byte)0); 

            //and the ugh string and padding
            bw.Write(endStringAndPadding); 

            //and the file index
            bw.Write(fileindexByteArray);

            //and add our tag to the index
            bw.Write(writePoint + tag.Path.Length + 1);

            //and the rest of the map
            bw.Write(bottomHalf);

            /* the shift is our tags path length + null termination length + our 
             * tags file-index entry length
             */
            int shift = tag.Path.Length + 1 + 4;

            /*
            //correct unicode
            for (int x = 0; x < 8; x++)
            {
                //fix index of language
                unicode.langs[x].indexOffset += shift;
                bw.BaseStream.Position = unicode.langs[x].indexPointerOffset;
                bw.Write(unicode.langs[x].indexOffset);

                //and table offset
                unicode.langs[x].tablePointerOffset += shift;
                bw.BaseStream.Position = unicode.langs[x].tablePointerOffset;
                bw.Write(unicode.langs[x].tablePointerOffset);

                bw.BaseStream.Position = unicode.langs[x].indexOffset;
                //now go through strings and correct their offsets
                for (int y = 0; y < unicode.langs[x].stringCount; y++)
                {
                    unicode.langs[x].UniStrings[y].Offset += shift;
                    bw.BaseStream.Position += 4;
                    bw.Write(unicode.langs[x].UniStrings[y].Offset);
                }
            }
            */

            //correct header
            header.filesize += 16 + shift + (int)metaBytes.Length;
            bw.BaseStream.Position = 8;
            bw.Write(header.filesize);

            header.indexOffset += shift;
            bw.BaseStream.Position = 16;
            bw.Write(header.indexOffset);

            header.indexSize += 16;
            bw.BaseStream.Position = 20;
            bw.Write(header.indexSize);

            header.metaStart = header.indexOffset + header.indexSize;

            header.metaSize += (int)metaBytes.Length;
            bw.BaseStream.Position = 24;
            bw.Write(header.metaSize);

            header.nonRawSize += (int)metaBytes.Length + 16;
            bw.BaseStream.Position = 28;
            bw.Write(header.nonRawSize);

            header.strangeFileStringsOffset += shift;
            bw.BaseStream.Position = 344;
            bw.Write(header.strangeFileStringsOffset);

            header.fileTableCount += 1;
            bw.BaseStream.Position = 704;
            bw.Write(header.fileTableCount);

            header.fileTableSize += shift-4; // -4 ??
            bw.BaseStream.Position = 712;
            bw.Write(header.fileTableSize);

            header.fileTableIndexOffset += shift-4;
            bw.BaseStream.Position = 716;
            bw.Write(header.fileTableIndexOffset);

            index.offsetOfObjectIndex += shift;

            //just before the ughTag in the objects
            int pos = index.offsetOfObjectIndex + ((index.objectCount-1) * 16);
            br.BaseStream.Position = pos;

            //store the ughTag info and rest of map
            bottomHalf = br.ReadBytes((int)(br.BaseStream.Length - br.BaseStream.Position));

            //go back jus before the ughTag in the objects
            bw.BaseStream.Position = pos;

            //reverse and write in our tags type
            bw.Write(Reverse(tag.Type).ToCharArray());

            //and identifier
            bw.Write(tag.Ident);

            //seeing as our tags meta will be overwriting ughTags
            tag.RawMetaOffset = ughTag.RawMetaOffset;
            //and raw meta offset (no magic applied)
            bw.Write(tag.RawMetaOffset);

            //and finally metasize
            bw.Write(tag.MetaSize);

            //the will-be-position of the ughTags ident
            int ughIdentPos = (int)bw.BaseStream.Position + 4;

            //write in the map
            bw.Write(bottomHalf);

            //type + ident + rawMO + MS
            shift += 16;
            int newSecondaryMagic = index.secondaryMagicConstant - header.metaStart;

            int oldTagMetaOffset = tag.MetaOffset + shift;
            tag.MetaOffset = tag.RawMetaOffset - newSecondaryMagic;

            //as its meta will be after our tags
            int oldUghTagMetaOffset = ughTag.MetaOffset + shift;
            ughTag.MetaOffset = tag.MetaOffset + tag.MetaSize;
            ughTag.RawMetaOffset = ughTag.MetaOffset + newSecondaryMagic;

            //goto ughTags ident and write the ident and rawmetaoffset
            bw.BaseStream.Position = ughIdentPos;
            bw.Write(ughTag.Ident);
            bw.Write(ughTag.RawMetaOffset);

            //correct index objectCount
            index.objectCount += 1;
            bw.BaseStream.Position = header.indexOffset + 24;
            bw.Write(index.objectCount);

            //goto to-be-new-meta-position
            br.BaseStream.Position = tag.MetaOffset;
            bottomHalf = br.ReadBytes((int)(br.BaseStream.Length - br.BaseStream.Position));

            bw.BaseStream.Position = tag.MetaOffset;
            bw.Write(metaBytes);
            bw.Write(bottomHalf);

            //fix new tags reflexives
            int diff = tag.MetaOffset - oldTagMetaOffset;

            for (int x = 0; x <= tag.MetaSize - 4; x += 4)
            {
                br.BaseStream.Position = tag.MetaOffset + x;

                //read the reflexive
                int TranslatedReflexive = br.ReadInt32();
                //and then translate it
                //if (bspMagic == -1) TranslatedReflexive -= map.index.secondaryMagic;
                //else TranslatedReflexive -= bspMagic;
                TranslatedReflexive -= index.secondaryMagic;

                //if the offset is within the current tags meta
                if (TranslatedReflexive >= oldTagMetaOffset &&
                    TranslatedReflexive <= oldTagMetaOffset + tag.MetaSize)
                {
                    //we have a reflexive
                    bw.BaseStream.Position = tag.MetaOffset + x;
                    bw.Write(TranslatedReflexive + diff + newSecondaryMagic);
                }
            }

            //fix ugh tags loneIDs and reflexives
            int metaEnd = ughTag.MetaOffset + ughTag.MetaSize;
            diff = ughTag.MetaOffset - oldUghTagMetaOffset;

            for (int x = ughTag.MetaOffset; x < metaEnd; x += 4)
            {
                br.BaseStream.Position = x;

                string temptagtype = "";
                for (int q = 0; q < 4; q++)
                    temptagtype += Convert.ToChar(br.ReadByte());

                //reverse
                temptagtype = Reverse(temptagtype);
                int tempID = br.ReadInt32(); //read ID

                if (temptagtype != tag.Type && tempID == tag.Ident)
                {
                    bw.BaseStream.Position = x + 4;
                    bw.Write(ughTag.Ident);
                }
                else
                {
                    br.BaseStream.Position = x;

                    //read the reflexive
                    int TranslatedReflexive = br.ReadInt32();
                    //and then translate it
                    //if (bspMagic == -1) TranslatedReflexive -= map.index.secondaryMagic;
                    //else TranslatedReflexive -= bspMagic;
                    TranslatedReflexive -= index.secondaryMagic;

                    //if the offset is within the current tags meta
                    if (TranslatedReflexive >= oldUghTagMetaOffset &&
                        TranslatedReflexive <= metaEnd)
                    {
                        //we have a reflexive
                        bw.BaseStream.Position = x;
                        bw.Write(TranslatedReflexive + diff + newSecondaryMagic);
                    }
                }
            }

            br.Close();
            bw.Close();

            // Now i need to fix sound, and bitmaps

            /* ok, so every tag has:
             * a placement in the object index (contains type, id, raw metaoffset, metasize)
             * name in the filetable d
             * meta
             * 
             * other things thatll need fixing in the header:
             * filesize d
             * index offset d
             * index size d
             * meta start d
             * meta size d
             * nonrawsize (MetaSize + index Size + BSP Size) d
             * strangeFileStringsSize d
             * strangeFileStringsOffset d
             * fileStringsCount d
             * fileStringsOffset d
             * fileStringsSize d
             * fileStringsIndex d
             * 
             * And things that need fixing in the index:
             * objectIndexPointer
             * offsetOfObjectIndex
             * objectCount
             * primaryMagicConstant
             * secondaryMagicConstant (should fix itself) 
             */
        }

        public void TempExtract(int offset, int size)
        {
            SaveFileDialog s = new SaveFileDialog();
            if (s.ShowDialog() == DialogResult.Cancel) return;

            BinaryReader br = new BinaryReader(new FileStream(path, FileMode.Open, 
                FileAccess.ReadWrite, FileShare.ReadWrite));

            BinaryWriter bw = new BinaryWriter(new FileStream(s.FileName, FileMode.Create,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            br.BaseStream.Position = offset;
            bw.Write(br.ReadBytes(size));

            br.Close();
            bw.Close();
        }

        #region Progress
        public void StartProgress(int max)
        {
            ProgressBar.Value = 0;
            ProgressBar.Maximum = max < 0 ? 0 : max;
        }

        public void UpdateProgress(string txt, int val)
        {
            ProgressLabel.Text = txt;
            ProgressBar.Value = val <= ProgressBar.Maximum ? val : ProgressBar.Maximum;
            Application.DoEvents();
        }

        public void EndProgress()
        {
            ProgressBar.Value = 0;
            ProgressLabel.Text = "Idle";
        }
        #endregion
    }
}