using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace Dark
{
    public class LightMapData : RawDataHandler
    {
        public int bspIndx = -1;

        public LightMapData(Map map)
        {
            BinaryReader br = new BinaryReader(new FileStream(map.path, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            //go thorugh all the maps bsp's
            for (int x = 0; x < map.sbsp.sbspInfoList.Length; x++)
            {
                //if we find the correct lightmap
                if (map.sbsp.sbspInfoList[x].lightmap.Ident == map.tags.tagsList[map.tags.selectedIndex].Ident)
                {
                    //store
                    bspIndx = x;

                    //goto the 1st reflexive
                    br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset + 128 + 4;

                    //goto the translation + 64
                    br.BaseStream.Position = (br.ReadInt32() - map.sbsp.sbspInfoList[x].Magic) + 64;

                    int tempc = br.ReadInt32(); //chunk count
                    int tempr = br.ReadInt32() - map.sbsp.sbspInfoList[x].Magic; //translation

                    //foreach chunk
                    for (int y = 0; y < tempc; y++)
                    {
                        //for storing info
                        RawDataChunk rdc = new RawDataChunk();

                        //chunk location within the meta
                        rdc.PointerOffset = tempr + (y * 56) + 12 - map.tags.tagsList[map.tags.selectedIndex].MetaOffset;
                        br.BaseStream.Position = tempr + (y * 56) + 12;

                        rdc.Offset = br.ReadInt32(); //offset + raw location atm
                        rdc.Size = br.ReadInt32();

                        if (rdc.Offset != -1)
                        {
                            //pull out the raw location and fix the offset
                            map.rawData.ParseSPointer(ref rdc.Offset, ref rdc.rawLocation);

                            if (rdc.rawLocation == RawLocations.Internal)
                            {
                                br.BaseStream.Position = rdc.Offset;
                                rdc.MS = new MemoryStream(br.ReadBytes(rdc.Size));
                            }
                            else
                            {
                                FileStream fs = Globals.returnStreamFromType(rdc.rawLocation);
                                if (fs == null)
                                {
                                    this.rawChunks.Clear();
                                    this.Loaded = false;
                                    Globals.PerformExternalMapNotFoundError();
                                    return;
                                }
                                else
                                {
                                    BinaryReader rawBr = new BinaryReader(fs);
                                    rawBr.BaseStream.Position = rdc.Offset;
                                    rdc.MS = new MemoryStream(rawBr.ReadBytes(rdc.Size));
                                    rawBr.Close();
                                }
                            }
                        }

                        //add the raw chunk to the list
                        this.rawChunks.Add(rdc);
                    }
                }
            }
            br.Close();
            this.Loaded = true;
        }
    }
}
