using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace Dark
{
    public class ModelData : RawDataHandler
    {
        public Map map;

        public ModelData(Map map)
        {
            this.map = map;

            BinaryReader br = new BinaryReader(new FileStream(map.path, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            //goto the 3rd reflexive
            br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset + 36;

            int tempc = br.ReadInt32(); //chunk count
            int tempr = br.ReadInt32() - map.index.secondaryMagic; //translation

            //foreach chunk
            for (int x = 0; x < tempc; x++)
            {
                //for storing info
                RawDataChunk rdc = new RawDataChunk();

                //chunk location within the meta
                rdc.PointerOffset = tempr + (x * 92) + 56 - map.tags.tagsList[map.tags.selectedIndex].MetaOffset;
                br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset + rdc.PointerOffset;

                rdc.Offset = br.ReadInt32(); //offset + raw location atm
                rdc.Size = br.ReadInt32();

                //pull out the raw location and fix the offset
                map.rawData.ParseSPointer(ref rdc.Offset, ref rdc.rawLocation);

                if (rdc.rawLocation == RawLocations.Internal)
                {
                    br.BaseStream.Position = rdc.Offset;
                    rdc.MS = new MemoryStream(br.ReadBytes(rdc.Size));
                }
                else
                {
                    FileStream fs = Globals.returnStreamFromType(rdc.rawLocation);
                    if (fs == null)
                    {
                        this.rawChunks.Clear();
                        this.Loaded = false;
                        Globals.PerformExternalMapNotFoundError();
                        return;
                    }
                    else
                    {
                        BinaryReader rawBr = new BinaryReader(fs);
                        rawBr.BaseStream.Position = rdc.Offset;
                        rdc.MS = new MemoryStream(rawBr.ReadBytes(rdc.Size));
                        rawBr.Close();
                    }
                }

                //add the raw chunk to the list
                this.rawChunks.Add(rdc);
            }

            //goto the 9th? reflexive
            br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset + 116;

            tempc = br.ReadInt32(); //chunk count
            tempr = br.ReadInt32() - map.index.secondaryMagic; //translation

            //foreach chunk
            for (int x = 0; x < tempc; x++)
            {
                //for storing info
                RawDataChunk rdc = new RawDataChunk();

                //chunk location within the meta
                rdc.PointerOffset = tempr + (x * 88) + 52 - map.tags.tagsList[map.tags.selectedIndex].MetaOffset;

                br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset + rdc.PointerOffset;
                
                //Raw.rawDataType = RawDataType.mode2;
                rdc.Offset = br.ReadInt32(); //offset + raw location atm
                rdc.Size = br.ReadInt32();

                //pull out the raw location and fix the offset
                map.rawData.ParseSPointer(ref rdc.Offset, ref rdc.rawLocation);

                if (rdc.rawLocation == RawLocations.Internal)
                {
                    br.BaseStream.Position = rdc.Offset;
                    rdc.MS = new MemoryStream(br.ReadBytes(rdc.Size));
                }
                else
                {
                    BinaryReader rawBr = new BinaryReader(Globals.returnStreamFromType(rdc.rawLocation));
                    if (rawBr == null)
                    {
                        this.rawChunks.Clear();
                        this.Loaded = false;
                        Globals.PerformExternalMapNotFoundError();
                        return;
                    }
                    else
                    {
                        rawBr.BaseStream.Position = rdc.Offset;
                        rdc.MS = new MemoryStream(rawBr.ReadBytes(rdc.Size));
                        rawBr.Close();
                    }
                }

                //add the raw chunk to the list
                this.rawChunks.Add(rdc);
            }
            br.Close();
            this.Loaded = true;
        }
    }
}
