using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace Dark
{
    public class SoundData : RawDataHandler
    {
        public int sndType, sndPermIndex, sndPermChunkCount, sndPermReflex,
                    sndChoiceIndex, sndChoiceChunkCount, sndChoiceReflex,
                    sndRawIndex, sndRawChunkCount, sndRawReflex;
                    
        public SoundData(Map map)
        {
            BinaryReader br = new BinaryReader(new FileStream(map.path, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            //pull out the ugh tag
            Tag ughTag = map.tags.tagsList[map.tags.FindTagIndexByTypeAndPath(
                "ugh!", "i've got a lovely bunch of coconuts")];

            //goto selected tag meta + 4
            br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset + 4;
            sndType = br.ReadByte();

            br.BaseStream.Position += 3;
            sndPermIndex = br.ReadInt16();
            sndPermChunkCount = br.ReadByte();

            //5th reflex in ugh tag; sound permutations
            br.BaseStream.Position = ughTag.MetaOffset + 32 + 4;
            sndPermReflex = br.ReadInt32() - map.index.secondaryMagic; //translation

            br.BaseStream.Position = sndPermReflex + (sndPermIndex * 12) + 8;
            sndChoiceIndex = br.ReadInt16();
            sndChoiceChunkCount = br.ReadInt16();

            //6th reflex in ugh tag; sound choices
            br.BaseStream.Position = ughTag.MetaOffset + 40 + 4;
            sndChoiceReflex = br.ReadInt32() - map.index.secondaryMagic; //translation

            br.BaseStream.Position = sndChoiceReflex + (sndChoiceIndex * 16) + 12;
            sndRawIndex = br.ReadInt16();
            sndRawChunkCount = br.ReadInt16();

            //9th reflex in ugh tag; sound raw
            br.BaseStream.Position = ughTag.MetaOffset + 64 + 4;
            sndRawReflex = br.ReadInt32() - map.index.secondaryMagic; //translation

            for (int x = 0; x < sndRawChunkCount; x++)
            {
                br.BaseStream.Position = sndRawReflex + (sndRawIndex * 12) + (x * 12);

                RawDataChunk srdc = new RawDataChunk();

                srdc.Offset = br.ReadInt32();
                srdc.Size = br.ReadInt32();
                if (srdc.Size <= 0) break;

                //pull out the raw location and fix the offset
                map.rawData.ParseSPointer(ref srdc.Offset, ref srdc.rawLocation);

                if (srdc.rawLocation == RawLocations.Internal)
                {
                    br.BaseStream.Position = srdc.Offset;
                    srdc.MS = new MemoryStream(br.ReadBytes(srdc.Size));
                }
                else
                {
                    FileStream fs = Globals.returnStreamFromType(srdc.rawLocation);
                    if (fs == null)
                    {
                        this.rawChunks.Clear();
                        this.Loaded = false;
                        Globals.PerformExternalMapNotFoundError();
                        return;
                    }
                    else
                    {
                        BinaryReader rawBr = new BinaryReader(fs);
                        rawBr.BaseStream.Position = srdc.Offset;
                        srdc.MS = new MemoryStream(rawBr.ReadBytes(srdc.Size));
                        rawBr.Close();
                    }
                }

                //add the raw chunk to the list
                this.rawChunks.Add(srdc);
            }

            br.Close();
            this.Loaded = true;
        }
    }
}
