using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace Dark
{
    public class UnicodeData : RawDataHandler
    {
        public Map map;
        public Lang[] langs = new Lang[8];

        public UnicodeData(Map map)
        {
            this.map = map;

            BinaryReader br = new BinaryReader(new FileStream(map.path, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            int matgOffset400 = map.tags.tagsList[0].MetaOffset + 400;

            //go through 8 times as there are 8 languages
            for (int x = 0; x < 8; x++)
            {
                //chunk size is 28 bytes
                br.BaseStream.Position = matgOffset400 + (28 * x);

                //store info
                langs[x].language = (Language)x;

                langs[x].stringCount = br.ReadInt32();
                langs[x].tableSize = br.ReadInt32(); //size of all strings along with null terminations

                langs[x].indexPointerOffset = br.BaseStream.Position;
                langs[x].indexOffset = br.ReadInt32();
                langs[x].tableOffset = br.ReadInt32();

                langs[x].IDSizes = new IDSize[langs[x].stringCount];
                langs[x].UniStrings = new UniString[langs[x].stringCount];

                br.BaseStream.Position = langs[x].indexOffset;
                for (int y = 0; y < langs[x].stringCount; y++)
                {
                    langs[x].IDSizes[y] = new IDSize(br.ReadInt16(), br.ReadByte(), br.ReadByte());
                    langs[x].UniStrings[y] = new UniString(br.ReadInt32());
                }

                br.BaseStream.Position = langs[x].tableOffset;
                //char chr;
                //int readSize;
                for (int y = 0; y < langs[x].stringCount; y++)
                {
                    langs[x].UniStrings[y].Length = y + 1 == langs[x].stringCount ?
                        langs[x].tableSize - langs[x].UniStrings[y].Offset - 1 :
                        langs[x].UniStrings[y + 1].Offset - langs[x].UniStrings[y].Offset - 1;

                    for (int z = 0; z < langs[x].UniStrings[y].Length; z++)
                    {
                        byte b = br.ReadByte();
                        switch (b)
                        {
                            case 238: //start of a sumerfin
                                {
                                    short img = br.ReadInt16();
                                    z += 2;
                                    int indx = Array.IndexOf(Globals.uniDwords, img);

                                    if (indx == -1) langs[x].UniStrings[y].uString += "[ ? ]"; // :-D
                                    else langs[x].UniStrings[y].uString += Globals.uniTranslations[indx];

                                    break;
                                }

                            case 13: //newline
                                {
                                    br.BaseStream.Position++;
                                    z++;
                                    langs[x].UniStrings[y].uString += @" \r\n ";// " [ Newline ] ";
                                    break;
                                }

                            default:
                                {
                                    langs[x].UniStrings[y].uString += (char)b;
                                    break;
                                }
                        }
                    }
                    //skip null
                    br.BaseStream.Position++;
                }
            }
            br.Close();
            this.Loaded = true;
        }
    }

    public enum Language
    {
        English,
        Japanese,
        German,
        French,
        Spanish,
        Italian,
        Korean,
        Chinese
    }

    public struct DwordTranslation
    {
        public short Dword;
        public string Translation;

        public DwordTranslation(short Dword, string Translation)
        {
            this.Dword = Dword;
            this.Translation = Translation;
        }
    }

    public struct IDSize
    {
        public int ID;
        public byte SizeMayb;

        public IDSize(int iID, byte uselessByte, byte iSize)
        {
            this.ID = iID;
            this.SizeMayb = iSize;
        }
    }

    public struct UniString
    {
        public int Offset;
        public string uString;
        public int Length;

        public UniString(int iOffset)
        {
            this.Offset = iOffset;
            this.uString = "";
            this.Length = 0;
        }
    }

    public struct Lang
    {
        public Language language;

        public int
        stringCount,
        tableSize,
        indexOffset,
        tableOffset;

        public long
        indexPointerOffset,
        tablePointerOffset;

        public IDSize[] IDSizes;
        public UniString[] UniStrings;
    }
}
