using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace Dark
{
    public class RawData
    {
        public Map map;
        public RawDataHandler rdh;

        public RawData(Map map)
        {
            this.map = map;
        }

        public void LoadSelectedTag()
        {
            rdh = null;

            switch (map.tags.tagsList[map.tags.selectedIndex].Type)
            {
                case "matg":
                    rdh = new UnicodeData(map);
                    break;

                case "PRTM":
                    //rdh = new PRTMData();
                    break;

                case "DECR":
                    //rdh = new DECRData();
                    break;

                case "jmad":
                    //rdh = new AnimationData();
                    break;

                case "mode":
                    rdh = new ModelData(map);
                    break;

                case "bitm":
                    rdh = new BitmapData(map, map.tags.selectedIndex);
                    break;

                case "sbsp":
                    //rdh = new BSPData();
                    break;

                case "weat":
                    //rdh = new WeatherData();
                    break;

                case "snd!":
                    rdh = new SoundData(map);
                    break;

                case "ugh!":
                    //rdh = new CoconutsModelData();
                    break;

                case "ltmp":
                    rdh = new LightMapData(map);
                    break;
            }
        }

        public void ParseSPointer(ref int Pointer, ref RawLocations mType)
        {
            //pull out the may type
            long tempmap = Pointer & 0xC0000000;

            //and then fix the no. so it corresponds to the offset at which
            //the data starts in the raw
            Pointer &= (int)0x3FFFFFFF;

            switch (tempmap)
            {
                case 0:
                    mType = RawLocations.Internal;
                    return;

                case 0x80000000:
                    mType = RawLocations.Shared;
                    return;

                case 0xC0000000:
                    mType = RawLocations.Single_Player_Shared;
                    return;

                case 0x40000000:
                    mType = RawLocations.MainMenu;
                    return;
            }
        }

        public void ParseUPointer(ref uint Pointer, ref RawLocations mType)
        {
            //pull out the may type
            long tempmap = Pointer & 0xC0000000;

            //and then fix the no. so it corresponds to the offset at which
            //the data starts in the raw
            Pointer &= (int)0x3FFFFFFF;

            switch (tempmap)
            {
                case 0:
                    mType = RawLocations.Internal;
                    return;

                case 0x80000000:
                    mType = RawLocations.Shared;
                    return;

                case 0xC0000000:
                    mType = RawLocations.Single_Player_Shared;
                    return;

                case 0x40000000:
                    mType = RawLocations.MainMenu;
                    return;
            }
        }
    }

    public enum RawLocations
    {
        Internal,
        MainMenu,
        Shared,
        Single_Player_Shared
    }

    public class RawDataChunk
    {
        public int Offset;
        public int Size;
        public int PointerOffset;
        public RawLocations rawLocation;
        public MemoryStream MS;
    }

    public class BitmapRawDataChunk : RawDataChunk
    {
        public int chunkNum;
        public int innerChunkNum;
    }

    public class ModelRawDataChunk : RawDataChunk
    {
        public int chunkSize;
        public int chunkCount;
    }

    public class SoundRawDataChunk : RawDataChunk
    {
        public uint uOffset;
        public uint uSize;
    }

    public class RawDataHandler
    {
        public List<RawDataChunk> rawChunks = new List<RawDataChunk>();
        public bool Loaded;

        public virtual void Load(BinaryReader br) { }
    }
}
