using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Drawing.Imaging;
using System.Windows.Forms;

namespace Dark
{
    public class BitmapSpecific
    {
        [DllImport("kernel32.dll")]
        static extern void RtlMoveMemory(IntPtr Pointer, byte[] ByteArray, int Count);

        public Map map;
        public DecodeDXT decodeDXT = new DecodeDXT();
        public Swizzler swizzler = new Swizzler();
        public DDS dds;

        public int headerSize = 0;
        public BitmapInfo[] BitmapInfoList;

        public int selectedChunkIndex;

        public ImageLib.RGBA_ByteColor[] H2Palette;
        public bool H2PaletteLoaded;

        public BitmapSpecific(Map map)
        {
            this.map = map;
            dds = new DDS(map);
        }

        public void LoadHalo2Palette()
        {
            BinaryReader br = new BinaryReader(new FileStream(Application.StartupPath + "\\H2Palette.pal", FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            H2Palette = new ImageLib.RGBA_ByteColor[256];
            for (int x = 0; x < 256; x++)
            {
                H2Palette[x].r = br.ReadByte();
                H2Palette[x].g = br.ReadByte();
                H2Palette[x].b = br.ReadByte();
                H2Palette[x].a = br.ReadByte();
            }
            
            br.Close();
            H2PaletteLoaded = true;
        }

        public void LoadBitmap(int TagIndx)
        {
             BinaryReader br = new BinaryReader(new FileStream(map.path, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite));

             //goto first reflexive
             br.BaseStream.Position = map.tags.tagsList[TagIndx].MetaOffset + 68;

             int tempc = br.ReadInt32(); //chunk count
             int tempr = br.ReadInt32() - map.index.secondaryMagic; //translation

            headerSize = tempr;
            BitmapInfoList = new BitmapInfo[tempc];

            br.BaseStream.Position = tempr;
            for (int x = 0; x < tempc; x++)
            {
                BitmapInfoList[x] = new BitmapInfo(br);
                br.BaseStream.Position += 116;
            }
            br.Close();
        }

        public Bitmap ReturnBitmap(int chunkNum, int innerChunkNum, int visualchunkindex, int bspnumber, List<RawDataChunk> rawChunks)
        {
            for (int x = 0; x < rawChunks.Count; x++)
            {
                BitmapRawDataChunk brdc = (BitmapRawDataChunk)rawChunks[x];
                if (brdc.chunkNum == chunkNum && brdc.innerChunkNum == innerChunkNum)
                {
                    selectedChunkIndex = x;

                    return DecodeByteArrayAndReturnBitmap(
                        brdc.MS.ToArray(), BitmapInfoList[chunkNum], visualchunkindex, bspnumber);
                }
            }
            return null;
        }

        public Bitmap DecodeByteArrayAndReturnBitmap(byte[] byteArray, BitmapInfo bitmapInfo, int visChunkIndx, int ID)
        {
            IntPtr ptr = new IntPtr();

            int stride = bitmapInfo.Width;
            //ushort num1 = (ushort)(8 - (bitmapInfo.Width % 8));
            //if (num1 != 8) stride += num1;

            //for storing the pix format
            PixelFormat pixFormat = new PixelFormat();

            //temp vars tht may nt always be needed
            int tmpHalf;
            byte[] tmpByteArray;

            switch (bitmapInfo.bitmFormat)
            {
                //32 bit
                case BitmapFormat.DXT1:
                    byteArray = decodeDXT.DecodeDXT1(bitmapInfo.Height, bitmapInfo.Width, byteArray);
                    stride *= 4;
                    pixFormat = PixelFormat.Format32bppArgb;
                    break;

                case BitmapFormat.DXT2_3:
                    byteArray = decodeDXT.DecodeDXT23(bitmapInfo.Height, bitmapInfo.Width, byteArray);
                    stride *= 4;
                    pixFormat = PixelFormat.Format32bppArgb;
                    break;

                case BitmapFormat.DXT4_5:
                    byteArray = decodeDXT.DecodeDXT45(bitmapInfo.Height, bitmapInfo.Width, byteArray);
                    stride *= 4;
                    pixFormat = PixelFormat.Format32bppArgb;
                    break;

                case BitmapFormat.A8R8G8B8:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.Width, bitmapInfo.Height, bitmapInfo.Depth, 32, true);

                    stride *= 4;
                    pixFormat = PixelFormat.Format32bppArgb;
                    break;

                case BitmapFormat.X8R8G8B8:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.Width, bitmapInfo.Height, -1, 32, true);

                    stride *= 4;
                    pixFormat = PixelFormat.Format32bppArgb;
                    break;

                //16 bit
                case BitmapFormat.A4R4G4B4:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.PixelOffset, bitmapInfo.Width, bitmapInfo.Height, -1, 16, true);

                    stride *= 2;
                    pixFormat = PixelFormat.Format16bppRgb565;
                    break;

                case BitmapFormat.Unknown:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        0, bitmapInfo.Width, bitmapInfo.Height, -1, 16, true);

                    stride *= 2;
                    pixFormat = PixelFormat.Format32bppArgb;

                    tmpHalf = byteArray.Length / 2;
                    tmpByteArray = new byte[tmpHalf * 4];
                    for (int x = 0; x < tmpHalf; x++)
                    {
                        int r = x * 2;
                        tmpByteArray[r * 2 + 0] = byteArray[r + 1];
                        tmpByteArray[r * 2 + 1] = byteArray[r + 1];
                        tmpByteArray[r * 2 + 2] = byteArray[r + 1];
                        tmpByteArray[r * 2 + 3] = byteArray[r + 0];
                    }
                    byteArray = tmpByteArray;
                    break;

                case BitmapFormat.A1R5G5B5:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.PixelOffset, bitmapInfo.Width, bitmapInfo.Height, -1, 16, true);

                    stride *= 2;
                    pixFormat = PixelFormat.Format16bppRgb555;
                    break;

                case BitmapFormat.R5G6B5:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.PixelOffset, bitmapInfo.Width, bitmapInfo.Height, -1, 16, true);

                    stride *= 2;
                    pixFormat = PixelFormat.Format16bppRgb565;
                    break;

                case BitmapFormat.A8Y8:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.PixelOffset, bitmapInfo.Width, bitmapInfo.Height, bitmapInfo.Depth, 16, true);

                    stride *= 4;
                    pixFormat = PixelFormat.Format32bppArgb;

                    tmpHalf = byteArray.Length / 2;
                    tmpByteArray = new byte[tmpHalf * 4];
                    for (int x = 0; x < tmpHalf; x++)
                    {
                        int r = x * 2;
                        tmpByteArray[r + 0] = byteArray[r + 1];
                        tmpByteArray[r + 1] = byteArray[r + 1];
                        tmpByteArray[r + 2] = byteArray[r + 1];
                        tmpByteArray[r + 3] = byteArray[r + 0];
                    }
                    byteArray = tmpByteArray;
                    break;

                //8 bit
                case BitmapFormat.P8:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.PixelOffset, bitmapInfo.Width, bitmapInfo.Height, bitmapInfo.Depth, 8, true);

                    stride *= 4;
                    pixFormat = PixelFormat.Format32bppArgb;

                    if (!H2PaletteLoaded) LoadHalo2Palette();

                    tmpByteArray = new byte[byteArray.Length * 4];
                    for (int x = 0; x < byteArray.Length; x++)
                    {
                        int r = x * 4;
                        tmpByteArray[r] = H2Palette[byteArray[x]].r;
                        tmpByteArray[r + 1] = H2Palette[byteArray[x]].g;
                        tmpByteArray[r + 2] = H2Palette[byteArray[x]].b;
                        tmpByteArray[r + 3] = H2Palette[byteArray[x]].a;
                    }
                    byteArray = tmpByteArray;
                    break;

                case BitmapFormat.Y8:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.PixelOffset, bitmapInfo.Width, bitmapInfo.Height, bitmapInfo.Depth, 8, true);

                    stride *= 4;
                    pixFormat = PixelFormat.Format32bppArgb;

                    tmpByteArray = new byte[byteArray.Length * 4];
                    for (int x = 0; x < byteArray.Length; x++)
                    {
                        int r = x * 4;
                        tmpByteArray[r + 0] = byteArray[x];
                        tmpByteArray[r + 1] = byteArray[x];
                        tmpByteArray[r + 2] = byteArray[x];
                        tmpByteArray[r + 3] = byteArray[255];
                    }
                    byteArray = tmpByteArray;
                    break;

                default:
                    return null;
            }

            Marshal.FreeHGlobal(ptr);
            ptr = Marshal.AllocHGlobal(byteArray.Length);
            RtlMoveMemory(ptr, byteArray, byteArray.Length);

            return new System.Drawing.Bitmap(bitmapInfo.Width, bitmapInfo.Height, stride, pixFormat, ptr);
        }

        public byte[] DecodeByteArray(byte[] byteArray, BitmapInfo bitmapInfo, int visChunkIndx, int ID)
        {
            //temp vars tht may nt always be needed
            int tmpHalf;
            byte[] tmpByteArray;

            switch (bitmapInfo.bitmFormat)
            {
                /*
                //32 bit
                case BitmapFormat.DXT1:
                    byteArray = decodeDXT.DecodeDXT1(bitmapInfo.Height, bitmapInfo.Width, byteArray);
                    break;

                case BitmapFormat.DXT2_3:
                    byteArray = decodeDXT.DecodeDXT23(bitmapInfo.Height, bitmapInfo.Width, byteArray);
                    break;

                case BitmapFormat.DXT4_5:
                    byteArray = decodeDXT.DecodeDXT45(bitmapInfo.Height, bitmapInfo.Width, byteArray);
                    break;
                */

                case BitmapFormat.A8R8G8B8:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.Width, bitmapInfo.Height, bitmapInfo.Depth, 32, true);
                    break;

                case BitmapFormat.X8R8G8B8:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.Width, bitmapInfo.Height, bitmapInfo.Depth, 32, true);
                    break;

                //16 bit
                case BitmapFormat.A4R4G4B4:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.PixelOffset, bitmapInfo.Width, bitmapInfo.Height, -1, 16, true);
                    break;

                case BitmapFormat.Unknown:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        0, bitmapInfo.Width, bitmapInfo.Height, -1, 16, true);

                    tmpHalf = byteArray.Length / 2;
                    tmpByteArray = new byte[tmpHalf * 4];
                    for (int x = 0; x < tmpHalf; x++)
                    {
                        int r = x * 2;
                        tmpByteArray[r * 2 + 0] = byteArray[r + 1];
                        tmpByteArray[r * 2 + 1] = byteArray[r + 1];
                        tmpByteArray[r * 2 + 2] = byteArray[r + 1];
                        tmpByteArray[r * 2 + 3] = byteArray[r + 0];
                    }
                    byteArray = tmpByteArray;
                    break;

                case BitmapFormat.A1R5G5B5:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.PixelOffset, bitmapInfo.Width, bitmapInfo.Height, -1, 16, true);
                    break;

                case BitmapFormat.R5G6B5:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.PixelOffset, bitmapInfo.Width, bitmapInfo.Height, -1, 16, true);
                    break;

                case BitmapFormat.A8Y8:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.PixelOffset, bitmapInfo.Width, bitmapInfo.Height, bitmapInfo.Depth, 16, true);

                    tmpHalf = byteArray.Length / 2;
                    tmpByteArray = new byte[tmpHalf * 4];
                    for (int x = 0; x < tmpHalf; x++)
                    {
                        int r = x * 2;
                        tmpByteArray[r + 0] = byteArray[r + 1];
                        tmpByteArray[r + 1] = byteArray[r + 1];
                        tmpByteArray[r + 2] = byteArray[r + 1];
                        tmpByteArray[r + 3] = byteArray[r + 0];
                    }
                    byteArray = tmpByteArray;
                    break;

                //8 bit
                case BitmapFormat.P8:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.PixelOffset, bitmapInfo.Width, bitmapInfo.Height, bitmapInfo.Depth, 8, true);

                    if (!H2PaletteLoaded) LoadHalo2Palette();

                    tmpByteArray = new byte[byteArray.Length * 4];
                    for (int x = 0; x < byteArray.Length; x++)
                    {
                        int r = x * 4;
                        tmpByteArray[r] = H2Palette[byteArray[x]].r;
                        tmpByteArray[r + 1] = H2Palette[byteArray[x]].g;
                        tmpByteArray[r + 2] = H2Palette[byteArray[x]].b;
                        tmpByteArray[r + 3] = H2Palette[byteArray[x]].a;
                    }
                    byteArray = tmpByteArray;
                    break;

                case BitmapFormat.Y8:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.PixelOffset, bitmapInfo.Width, bitmapInfo.Height, bitmapInfo.Depth, 8, true);

                    tmpByteArray = new byte[byteArray.Length * 4];
                    for (int x = 0; x < byteArray.Length; x++)
                    {
                        int r = x * 4;
                        tmpByteArray[r + 0] = byteArray[x];
                        tmpByteArray[r + 1] = byteArray[x];
                        tmpByteArray[r + 2] = byteArray[x];
                        tmpByteArray[r + 3] = byteArray[255];
                    }
                    byteArray = tmpByteArray;
                    break;
            }

            return byteArray;
        }

        public byte[] EncodeByteArray(byte[] byteArray, BitmapInfo bitmapInfo, int visChunkIndx, int ID)
        {
            //temp vars tht may nt always be needed
            int tmpHalf;
            byte[] tmpByteArray;

            switch (bitmapInfo.bitmFormat)
            {
                /*
                //32 bit
                case BitmapFormat.DXT1:
                    byteArray = decodeDXT.DecodeDXT1(bitmapInfo.Height, bitmapInfo.Width, byteArray);
                    break;

                case BitmapFormat.DXT2_3:
                    byteArray = decodeDXT.DecodeDXT23(bitmapInfo.Height, bitmapInfo.Width, byteArray);
                    break;

                case BitmapFormat.DXT4_5:
                    byteArray = decodeDXT.DecodeDXT45(bitmapInfo.Height, bitmapInfo.Width, byteArray);
                    break;
                */

                case BitmapFormat.A8R8G8B8:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.Width, bitmapInfo.Height, bitmapInfo.Depth, 32, false);
                    break;

                case BitmapFormat.X8R8G8B8:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.Width, bitmapInfo.Height, bitmapInfo.Depth, 32, false);
                    break;

                //16 bit
                case BitmapFormat.A4R4G4B4:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.PixelOffset, bitmapInfo.Width, bitmapInfo.Height, -1, 16, false);
                    break;

                case BitmapFormat.Unknown:
                    tmpHalf = byteArray.Length * 2;
                    tmpByteArray = new byte[tmpHalf / 4];
                    for (int x = 0; x < (byteArray.Length / 4); x++)
                    {
                        int r = x * 2;
                        tmpByteArray[r + 1] = byteArray[r * 2 + 0];
                        tmpByteArray[r + 1] = byteArray[r * 2 + 1];
                        tmpByteArray[r + 1] = byteArray[r * 2 + 2];
                        tmpByteArray[r + 0] = byteArray[r * 2 + 3];
                    }
                    byteArray = tmpByteArray;

                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        0, bitmapInfo.Width, bitmapInfo.Height, -1, 16, false);
                    break;

                case BitmapFormat.A1R5G5B5:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.PixelOffset, bitmapInfo.Width, bitmapInfo.Height, -1, 16, false);
                    break;

                case BitmapFormat.R5G6B5:
                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.PixelOffset, bitmapInfo.Width, bitmapInfo.Height, -1, 16, false);
                    break;

                case BitmapFormat.A8Y8:
                    tmpHalf = byteArray.Length * 2;
                    tmpByteArray = new byte[tmpHalf / 4];
                    for (int x = 0; x < (byteArray.Length / 4); x++)
                    {
                        int r = x * 2;
                        tmpByteArray[r + 1] = byteArray[r + 0];
                        tmpByteArray[r + 1] = byteArray[r + 1];
                        tmpByteArray[r + 1] = byteArray[r + 2];
                        tmpByteArray[r + 0] = byteArray[r + 3];
                    }
                    byteArray = tmpByteArray;

                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.PixelOffset, bitmapInfo.Width, bitmapInfo.Height, bitmapInfo.Depth, 16, false);
                    break;

                //8 bit
                case BitmapFormat.P8:
                    if (!H2PaletteLoaded) LoadHalo2Palette();
                    tmpByteArray = new byte[byteArray.Length / 4];

                    ImageLib.RGBA_ByteColor baColor = new ImageLib.RGBA_ByteColor();
                    for (int x = 0; x < tmpByteArray.Length; x++)
                    {
                        int r = x * 4;
                        baColor.r = byteArray[r];
                        baColor.g = byteArray[r + 1];
                        baColor.b = byteArray[r + 2];
                        baColor.a = byteArray[r + 3];

                        byte colorIndx = 0;
                        float storedDiff = 0;// 195075f;
                        ImageLib.RGBA_ByteColor palColor = new ImageLib.RGBA_ByteColor();
                        for (byte y = 0; y < 255; y++)
                        {
                            palColor.r = H2Palette[y].r;
                            palColor.g = H2Palette[y].g;
                            palColor.b = H2Palette[y].b;
                            palColor.a = H2Palette[y].a;

                            float diff =
                                (((baColor.r - palColor.r) * (baColor.r - palColor.r)) +
                                ((baColor.g - palColor.g) * (baColor.g - palColor.g))) +
                                ((baColor.b - palColor.b) * (baColor.b - palColor.b));

                            if (diff <= storedDiff)
                            {
                                colorIndx = y;
                                storedDiff = diff;
                            }
                        }
                        tmpByteArray[x] = colorIndx;
                    }
                    byteArray = tmpByteArray;

                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.PixelOffset, bitmapInfo.Width, bitmapInfo.Height, bitmapInfo.Depth, 8, false);
                    break;

                case BitmapFormat.Y8:
                    tmpByteArray = new byte[byteArray.Length / 4];
                    for (int x = 0; x < tmpByteArray.Length; x++)
                    {
                        int r = x * 4;
                        tmpByteArray[x] = byteArray[r + 0];
                        tmpByteArray[x] = byteArray[r + 1];
                        tmpByteArray[x] = byteArray[r + 2];
                        tmpByteArray[255] = byteArray[r + 3];
                    }
                    byteArray = tmpByteArray;

                    if (bitmapInfo.Swizzled) byteArray = swizzler.Swizzle(byteArray,
                        bitmapInfo.PixelOffset, bitmapInfo.Width, bitmapInfo.Height, bitmapInfo.Depth, 8, false);
                    break;
            }

            return byteArray;
        }

        public void InternalizeSelectedBitmap()
        {
            //open up rawloc in reader
            BinaryReader rawBr = new BinaryReader(Globals.returnStreamFromType(map.
                rawData.rdh.rawChunks[map.bitm.selectedChunkIndex].rawLocation));

            //open map in reader and writer
            FileStream fs = new FileStream(map.path, FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite);
            BinaryReader br = new BinaryReader(fs);
            BinaryWriter bw = new BinaryWriter(fs);

            //goto index in map (stright after bitmaps)
            br.BaseStream.Position = map.header.indexOffset;
            //store map downward from pos
            byte[] bottomHalf = br.ReadBytes((int)(br.BaseStream.Length - br.BaseStream.Position));

            //goto first reflexive of bitm tag
            br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset + 68;

            int tempc = br.ReadInt32(); //chunk cnt
            int tempr = br.ReadInt32() - map.index.secondaryMagic; //translation

            //go back to index offset
            bw.BaseStream.Position = map.header.indexOffset;

            //foreach chunk
            for (int x = 0; x < tempc; x++)
            {
                //foreach LOD
                for (int y = 0; y < map.rawData.rdh.rawChunks.Count; y++)
                {
                    //goto bitmap in rawloc
                    rawBr.BaseStream.Position = map.rawData.rdh.rawChunks[y].Offset;

                    //fix offset to internal
                    map.rawData.rdh.rawChunks[y].Offset = (int)bw.BaseStream.Position;

                    //read bitmap from rawloc and write to map
                    bw.Write(rawBr.ReadBytes(map.rawData.rdh.rawChunks[y].Size));

                    //write bitmap padding
                    byte[] padding = new byte[512 - (map.rawData.rdh.rawChunks[y].Size % 512)];
                    if (padding.Length != 512) bw.Write(padding);
                }
            }

            //bitm is now internal :P
            map.rawData.rdh.rawChunks[map.bitm.selectedChunkIndex].rawLocation = RawLocations.Internal;

            //now we need to shift
            int shift = (int)bw.BaseStream.Position - map.header.indexOffset;

            map.header.indexOffset = (int)bw.BaseStream.Position;
            bw.Write(bottomHalf);

            //--[Header]
            bw.BaseStream.Position = 16;
            bw.Write(map.header.indexOffset);

            //fix offsets
            map.header.metaStart += shift;

            map.index.primaryMagic -= shift;
            map.index.offsetOfObjectIndex += shift;
            map.index.secondaryMagic -= shift;

            //foreach chunk
            for (int x = 0; x < tempc; x++)
            {
                //foreach LOD
                for (int y = 0; y < map.rawData.rdh.rawChunks.Count; y++)
                {
                    bw.BaseStream.Position = (map.rawData.rdh.rawChunks[y].PointerOffset += shift);
                    bw.Write(map.rawData.rdh.rawChunks[y].Offset);
                }
            }

            //fix meta offsets
            for (int x = 0; x < map.tags.tagsList.Length; x++)
            {
                if (map.tags.tagsList[x].Type != "sbsp" && map.tags.tagsList[x].Type != "ltmp")
                    map.tags.tagsList[x].MetaOffset += shift;
            }

            //--[End]
            bw.BaseStream.Position = map.tags.tagsList[map.tags.tagsList.Length - 1].MetaOffset +
                map.tags.tagsList[map.tags.tagsList.Length - 1].MetaSize;

            int newPaddingLen = 512 - ((int)bw.BaseStream.Position % 512);
            newPaddingLen += newPaddingLen == 512 ? 512 : 1024;

            int eof = (int)bw.BaseStream.Position + newPaddingLen;

            while (bw.BaseStream.Position < eof)
                bw.Write((byte)205); //CD

            bw.BaseStream.SetLength(eof);

            //--[Header]
            bw.BaseStream.Position = 8;
            bw.Write(eof);

            rawBr.Close();
            br.Close();
            bw.Close();
        }
    }

    public struct BitmapInfo
    {
        public long mapOffset;

        public byte[] TagType;
        public ushort Width;
        public ushort WidthRem;
        public ushort Height;
        public ushort Depth;
        public ushort Type;
        public BitmapFormat bitmFormat;
        public bool Swizzled;
        public ushort Flags;
        public ushort RegPointX;
        public ushort RegPointY;
        public ushort MipMapCount;
        public ushort PixelOffset;

        public BitmapInfo(BinaryReader br)
        {
            mapOffset = br.BaseStream.Position;
            TagType = br.ReadBytes(4);

            Width = br.ReadUInt16();
            WidthRem = (ushort)(8 - (Width % 8));
            if (WidthRem != 8) Width += WidthRem;

            WidthRem = (ushort)(Width % 8);
            if (WidthRem != 0) Width += ((ushort)(8 - WidthRem));

            Height = br.ReadUInt16();
            Depth = br.ReadUInt16();
            Type = br.ReadUInt16();
            bitmFormat = (BitmapFormat)br.ReadUInt16();
            Flags = br.ReadUInt16();

          //Swizzled = ((flags & 0x1000) == 0x1000);
            Swizzled = ((Flags & 0x8) == 0x8);

            RegPointX = br.ReadUInt16();
            RegPointY = br.ReadUInt16();
            MipMapCount = br.ReadUInt16();
            PixelOffset = br.ReadUInt16();
        }

        public void WriteInfo(BinaryWriter bw)
        {
            bw.Write(TagType);
            bw.Write((ushort)(Width - WidthRem));
            bw.Write(Height);
            bw.Write(Depth);
            bw.Write(Type);
            bw.Write((ushort)bitmFormat);
            bw.Write(Flags);
          //bw.Write(Swizzled);
            bw.Write(RegPointX);
            bw.Write(RegPointY);
            bw.Write(MipMapCount);
            bw.Write(PixelOffset);
        }
    }

    public enum BitmapFormat : int
    {
        A8 = 0,
        Y8 = 1,
        AY8 = 2,
        A8Y8 = 3,
        R5G6B5 = 6,
        A1R5G5B5 = 8,
        A4R4G4B4 = 9,
        X8R8G8B8 = 10,
        A8R8G8B8 = 11,
        DXT1 = 14,
        DXT2_3 = 15,
        DXT4_5 = 16,
        P8 = 17,

        _2D = 0,
        _3D = 1,
        Cubemap = 2,
        Lightmap = 18,
        Unknown = 22
    }
}
