using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.IO;

namespace Dark
{
    public class LightMapSpecific
    {
        public Map map;
        public int bspIndx;

        public byte[] LightMap_Array, LightMap_Array_Backup;

        public PictureBox spb;
        public int sb = SystemInformation.VerticalScrollBarWidth;

        public LightMapSpecific(Map map)
        {
            this.map = map;
        }

        public void LoadLightmapsIntoAnArray()
        {
            this.bspIndx = (map.rawData.rdh as LightMapData).bspIndx;

            LightMap_Array = new byte[map.sbsp.sbspInfoList[bspIndx].lightmap.Palettes.Count * 1024];
            LightMap_Array_Backup = new byte[LightMap_Array.Length];

            for (int x = 0; x < map.sbsp.sbspInfoList[bspIndx].lightmap.Palettes.Count; x++)
            {
                for (int y = 0; y < 256; y++)
                {
                    PaletteColor temp = map.sbsp.sbspInfoList[bspIndx].lightmap.Palettes[x][y];

                    LightMap_Array[(x * 1024) + (y * 4) + 0] = Convert.ToByte(temp.r);
                    LightMap_Array[(x * 1024) + (y * 4) + 1] = Convert.ToByte(temp.g);
                    LightMap_Array[(x * 1024) + (y * 4) + 2] = Convert.ToByte(temp.b);
                    LightMap_Array[(x * 1024) + (y * 4) + 3] = Convert.ToByte(temp.a);
                }
            }
            Array.Copy(LightMap_Array, LightMap_Array_Backup, LightMap_Array.Length);
        }

        public void ReloadFromArray()
        {
            for (int x = 0; x < map.sbsp.sbspInfoList[bspIndx].lightmap.Palettes.Count; x++)
            {
                for (int y = 0; y < 256; y++)
                {
                    PaletteColor temp = new PaletteColor();

                    temp.r = LightMap_Array[(x * 1024) + (y * 4) + 0];
                    temp.g = LightMap_Array[(x * 1024) + (y * 4) + 1];
                    temp.b = LightMap_Array[(x * 1024) + (y * 4) + 2];
                    temp.a = LightMap_Array[(x * 1024) + (y * 4) + 3];

                    map.sbsp.sbspInfoList[bspIndx].lightmap.Palettes[x][y] = temp;
                }
            }
        }

        public Image RenderLightmap(int LightmapIndex)
        {
            MemoryStream Image_MemoryStream = new MemoryStream();
            BinaryWriter Image_BW = new BinaryWriter(Image_MemoryStream);

            //Creat the header and write it
            byte[] MyHeader = { 66, 77, 56, 3, 0, 0, 0, 0, 0, 0, 54, 0, 0, 0, 40, 0, 0, 0, 16, 0, 0, 0, 16, 0, 0, 0, 1, 0, 24, 0, 0, 0, 0, 0, 2, 3, 0, 0, 18, 11, 0, 0, 18, 11, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
            Image_BW.Write(MyHeader);

            //Write the RGB
            for (int i = 0; i < 1024; i += 4)
            {
                Image_BW.Write(LightMap_Array[(LightmapIndex * 1024) + i + 0]);
                Image_BW.Write(LightMap_Array[(LightmapIndex * 1024) + i + 1]);
                Image_BW.Write(LightMap_Array[(LightmapIndex * 1024) + i + 2]);
            }

            Image_BW.Write(0);
            Image_BW.Flush();

            Image tempimage = Image.FromStream(Image_MemoryStream);
            tempimage.RotateFlip(RotateFlipType.Rotate180FlipX);

            return tempimage;
        }

        public void PositionLightmapImgs(Panel panel, EventHandler imgclick)
        {
            panel.Controls.Clear();

            int imgX = 6, imgY = 0;
            for (int i = 0; i < map.sbsp.sbspInfoList[bspIndx].lightmap.Palettes.Count; i++)
            {
                PictureBox img = new PictureBox();
                img.Image = RenderLightmap(i);
                img.SizeMode = PictureBoxSizeMode.StretchImage;
                img.Size = new Size(64, 64);

                if (imgX + 64 > panel.Width)// - sb)
                {
                    imgX = 6;
                    imgY += 68;
                }

                img.Location = new Point(imgX, imgY);
                img.Tag = i;
                img.Click += new EventHandler(imgclick);

                panel.Controls.Add(img);
                imgX += 68;
            }
        }

        public void RepositionLightmapImgs(Panel panel)
        {
            List<PictureBox> list = new List<PictureBox>();
            foreach (Control box in panel.Controls)
                list.Add((PictureBox)box);

            panel.Controls.Clear();

            int imgX = 6, imgY = 0;
            for (int i = 0; i < list.Count; i++)
            {
                PictureBox img = list[i];

                if (imgX + 64 > panel.Width)// - sb)
                {
                    imgX = 6;
                    imgY += 68;
                }

                img.Location = new Point(imgX, imgY);
                panel.Controls.Add(img);
                imgX += 68;
            }
        }

        public void EditLightmaps(ref Panel panel1, int val1, int val2, int val3,
            bool bool1, int num, bool bool2, bool bool3)
        {
            if (bool2)
            {
                if (bool3)
                {
                    byte[][] temparray = new byte[1][];
                    for (int x = 0; x < map.sbsp.sbspInfoList[bspIndx].lightmap.Palettes.Count; x++)
                    {
                        temparray[0] = new byte[1024];
                        Array.ConstrainedCopy(LightMap_Array_Backup, x * 1024, temparray[0], 0, 1024);
                        Array.ConstrainedCopy(HueSaturation(temparray, val1, val2, val3, bool1)[0], 0, LightMap_Array, x * 1024, 1024);
                    }

                    foreach (Control box in panel1.Controls)
                        ((PictureBox)box).Image = RenderLightmap((int)box.Tag);
                }
                else
                {
                    byte[][] temparray = new byte[1][];
                    temparray[0] = new byte[1024];
                    Array.ConstrainedCopy(LightMap_Array_Backup, num * 1024, temparray[0], 0, 1024);
                    Array.ConstrainedCopy(HueSaturation(temparray, val1, val2, val3, bool1)[0], 0, LightMap_Array, num * 1024, 1024);

                    foreach (Control box in panel1.Controls)
                    {
                        if ((int)box.Tag == num)
                            ((PictureBox)box).Image = RenderLightmap((int)box.Tag);
                    }
                }
            }
            else
            {
                if (bool3)
                {
                    byte[][] temparray = new byte[1][];
                    for (int x = 0; x < map.sbsp.sbspInfoList[bspIndx].lightmap.Palettes.Count; x++)
                    {
                        temparray[0] = new byte[1024];
                        Array.ConstrainedCopy(LightMap_Array_Backup, x * 1024, temparray[0], 0, 1024);
                        Array.ConstrainedCopy(ColorBalance(temparray, val1, val2, val3, bool1)[0], 0, LightMap_Array, x * 1024, 1024);
                    }

                    foreach (Control box in panel1.Controls)
                        ((PictureBox)box).Image = RenderLightmap((int)box.Tag);
                }
                else
                {
                    byte[][] temparray = new byte[1][];
                    temparray[0] = new byte[1024];
                    Array.ConstrainedCopy(LightMap_Array_Backup, num * 1024, temparray[0], 0, 1024);
                    Array.ConstrainedCopy(ColorBalance(temparray, val1, val2, val3, bool1)[0], 0, LightMap_Array, num * 1024, 1024);

                    foreach (Control box in panel1.Controls)
                    {
                        if ((int)box.Tag == num)
                            ((PictureBox)box).Image = RenderLightmap((int)box.Tag);
                    }
                }
            }
        }

        public byte[][] HueSaturation(byte[][] source, int hueShift, int satShift, int lightShift, bool colorize)
        {
            byte[][] adjusted = new byte[source.Length][];
            double shiftH = (double)hueShift;
            double shiftS = ((double)satShift) / 100;
            double shiftL = ((double)lightShift) / 100;
            for (int i = 0; i < source.Length; i++)
            {
                adjusted[i] = new byte[1024];
                for (int x = 0; x < 1024; x += 4)
                {
                    // RGB -> HSL
                    double B = ((double)source[i][x]) / 255;
                    double G = ((double)source[i][x + 1]) / 255;
                    double R = ((double)source[i][x + 2]) / 255;
                    double max = Math.Max(Math.Max(R, G), B);
                    double min = Math.Min(Math.Min(R, G), B);
                    double L = ((double)0.5) * (max + min);
                    double S = L == 0 || L == 1 ? 0 : // Filter undefined
                        L <= 0.5 ? (max - min) / (max + min) : (max - min) / (2 - (max + min));
                    double H = colorize ? 0 : // colorize
                        S == 0 ? 0 : // saturation == 0
                        R >= G && R >= B ? 60 * (G - B) / (max - min) : // max == red
                        G >= R && G >= B ? 60 * (B - R) / (max - min) + 120 : // max == blue
                        60 * (R - G) / (max - min) + 240; // max == green

                    // Shift HSL
                    H = H + shiftH;
                    S = !colorize && S == 0 ? 0 : // Don't create saturation
                        shiftS <= 0 ? S * (1 + shiftS) : S + shiftS * (1 - S);
                    L = shiftL <= 0 ? L * (1 + shiftL) : L + shiftL * (1 - L);

                    // HSL -> RGB
                    double d = L <= 0.5 ? S * L : S * (1 - L);
                    if (L == 0 || L == 1 || S == 0 || d == 0)
                        R = G = B = L;
                    else
                    {
                        while (H >= 360)
                            H -= 360;
                        while (H < 0)
                            H += 360;
                        //double d = L <= 0.5 ? S*L : S*(1-L);
                        min = L - d;
                        max = L + d;
                        double F = H < 60 ? H :
                            H < 180 ? H - 120 :
                            H < 300 ? H - 240 :
                            H - 360;
                        double mid = min + Math.Abs(F / 60) * (max - min);
                        switch ((int)Math.Floor(H / 60))
                        {
                            case 0:
                                R = max;
                                G = mid;
                                B = min;
                                break;
                            case 1:
                                R = mid;
                                G = max;
                                B = min;
                                break;
                            case 2:
                                R = min;
                                G = max;
                                B = mid;
                                break;
                            case 3:
                                R = min;
                                G = mid;
                                B = max;
                                break;
                            case 4:
                                R = mid;
                                G = min;
                                B = max;
                                break;
                            case 5:
                                R = max;
                                G = min;
                                B = mid;
                                break;
                        }
                    }

                    adjusted[i][x] = (byte)(B * 255);
                    adjusted[i][x + 1] = (byte)(G * 255);
                    adjusted[i][x + 2] = (byte)(R * 255);
                    adjusted[i][x + 3] = (byte)255;
                }
            }
            return adjusted;
        }

        public byte[][] ColorBalance(byte[][] source, int redShift, int greenShift, int blueShift, bool preserve)
        {
            byte[][] adjusted = new byte[source.Length][];
            double shiftR = ((double)redShift) / 100d;
            double shiftG = ((double)greenShift) / 100d;
            double shiftB = ((double)blueShift) / 100d;
            if (preserve)
            {
                // Avoid Complete Desaturation
                shiftR *= 0.99d;
                shiftG *= 0.99d;
                shiftB *= 0.99d;
            }
            for (int i = 0; i < source.Length; i++)
            {
                adjusted[i] = new byte[1024];
                for (int x = 0; x < 1024; x += 4)
                {
                    double B = ((double)source[i][x]) / 255d;
                    double G = ((double)source[i][x + 1]) / 255d;
                    double R = ((double)source[i][x + 2]) / 255d;
                    if (preserve)
                    {
                        // RGB -> L
                        double max = Math.Max(Math.Max(R, G), B);
                        double min = Math.Min(Math.Min(R, G), B);
                        double L = ((double)0.5) * (max + min);
                        //double S = L == 0 || L == 1 ? 0 : // Filter undefined
                        //	L <= 0.5 ? (max-min)/(max+min) : (max-min)/(2-(max+min));

                        // RGB + Shift -> HS
                        B = shiftB <= 0 ? B * (1 + shiftB) : B + shiftB * (1 - B);
                        G = shiftG <= 0 ? G * (1 + shiftG) : G + shiftG * (1 - G);
                        R = shiftR <= 0 ? R * (1 + shiftR) : R + shiftR * (1 - R);
                        max = Math.Max(Math.Max(R, G), B);
                        min = Math.Min(Math.Min(R, G), B);
                        double tempL = 0.5d * (max + min);
                        double S = tempL == 0 || tempL == 1 ? 0 : // Filter undefined
                            tempL <= 0.5 ? (max - min) / (max + min) : (max - min) / (2 - (max + min));
                        double H = S == 0 ? 0 : // saturation == 0
                            R >= G && R >= B ? 60 * (G - B) / (max - min) : // max == red
                            G >= R && G >= B ? 60 * (B - R) / (max - min) + 120 : // max == green
                            60 * (R - G) / (max - min) + 240; // max == blue		

                        // HSL -> RGB
                        double d = L <= 0.5 ? S * L : S * (1 - L);
                        if (L == 0 || L == 1 || S == 0 || d == 0)
                            R = G = B = L;
                        else
                        {
                            while (H >= 360)
                                H -= 360;
                            while (H < 0)
                                H += 360;
                            //double d = L <= 0.5 ? S*L : S*(1-L);
                            min = L - d;
                            max = L + d;
                            double F = H < 60 ? H :
                                H < 180 ? H - 120 :
                                H < 300 ? H - 240 :
                                H - 360;
                            double mid = min + Math.Abs(F / 60) * (max - min);
                            switch ((int)Math.Floor(H / 60))
                            {
                                case 0:
                                    R = max;
                                    G = mid;
                                    B = min;
                                    break;
                                case 1:
                                    R = mid;
                                    G = max;
                                    B = min;
                                    break;
                                case 2:
                                    R = min;
                                    G = max;
                                    B = mid;
                                    break;
                                case 3:
                                    R = min;
                                    G = mid;
                                    B = max;
                                    break;
                                case 4:
                                    R = mid;
                                    G = min;
                                    B = max;
                                    break;
                                case 5:
                                    R = max;
                                    G = min;
                                    B = mid;
                                    break;
                            }
                        }
                    }
                    else
                    {
                        B = shiftB <= 0 ? B * (1 + shiftB) : B + shiftB * (1 - B);
                        G = shiftG <= 0 ? G * (1 + shiftG) : G + shiftG * (1 - G);
                        R = shiftR <= 0 ? R * (1 + shiftR) : R + shiftR * (1 - R);
                    }

                    adjusted[i][x] = (byte)(B * 255);
                    adjusted[i][x + 1] = (byte)(G * 255);
                    adjusted[i][x + 2] = (byte)(R * 255);
                    adjusted[i][x + 3] = (byte)255;
                }
            }
            return adjusted;
        }
    }
}
