using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.DirectX;

namespace Dark
{
    public class Camera
    {
        public Vector3 Pos, Target, Up;

        public float i, j, k;
        public float radius = 1.0f;

        public float radianh, radianv;
        public bool fixedrotation = false;

        public int oldx, oldy;
        public float speed = 0.08f;

        public Camera()
        {
            Pos = new Vector3(-3, 0, 0);
            Target = new Vector3(0, 0, 0);
            Up = new Vector3(0, 0, 1);
            ComputePosition();
        }

        public void Change(int x, int y)
        {
            int tmpX = oldx - x;
            int tmpY = oldy - y;

            radianh += DegreesToRadian((float)tmpX);
            radianv += DegreesToRadian((float)tmpY);

            ComputePosition();
            oldx = x;
            oldy = y;
        }

        public void ComputeStrafe(bool right)
        {
            // Keep all rotations between 0 and 2PI
            radianh = radianh > (float)Math.PI * 2 ? radianh - (float)Math.PI * 2 : radianh;
            radianh = radianh < 0 ? radianh + (float)Math.PI * 2 : radianh;

            radianv = radianv > (float)Math.PI * 2 ? radianv - (float)Math.PI * 2 : radianv;
            radianv = radianv < 0 ? radianv + (float)Math.PI * 2 : radianv;

            //Switch up-vector based on vertical rotation
            // UpVector = Position.X < 0 ? new Vector3(-1, 0, 1) : new Vector3(1, 0, 1);
            // radianv > Math.PI / 2 && radianv < Math.PI / 2 * 3 ?
            //  new Vector3(0,  1,0) : new Vector3(0,  -1,0);

            float tempi = radius * (float)(Math.Cos(radianh - 1.57) * Math.Cos(radianv)) * this.speed;
            float tempj = radius * (float)(Math.Sin(radianh - 1.57) * Math.Cos(radianv)) * this.speed;

            if (right == true)
            {
                Target.X += tempi;
                Target.Y += tempj;

                Pos.X += tempi;
                Pos.Y += tempj;
            }
            else
            {
                Target.X -= tempi;
                Target.Y -= tempj;

                Pos.X -= tempi;
                Pos.Y -= tempj;
            }
        }

        public float DegreesToRadian(float degree)
        {
            return (float)(degree * (Math.PI / 180));
        }

        public void ComputePosition()
        {
            // Keep all rotations between 0 and 2PI

            radianh = radianh > (float)Math.PI * 2 ? radianh - (float)Math.PI * 2 : radianh;
            radianh = radianh < 0 ? radianh + (float)Math.PI * 2 : radianh;

            radianv = radianv > (float)Math.PI * 2 ? radianv - (float)Math.PI * 2 : radianv;
            radianv = radianv < 0 ? radianv + (float)Math.PI * 2 : radianv;

            i = radius * (float)(Math.Cos(radianh) * Math.Cos(radianv));
            j = radius * (float)(Math.Sin(radianh) * Math.Cos(radianv));
            k = radius * (float)Math.Sin(radianv);

            if (fixedrotation)
            {
                Pos.X = i + Pos.X;
                Pos.Y = j + Pos.Y;
                Pos.Z = k + Pos.Z;
            }
            else
            {
                Target.X = i + Pos.X;
                Target.Y = j + Pos.Y;
                Target.Z = k + Pos.Z;
            }
        }
    }
}
