using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.IO;

namespace Dark
{
    public partial class ThreeD : UserControl
    {
        public Map map;
        public Effect effect;
        public bool Init;

        public ModelBoundingBox modeBox;
        public ModelChunk[] modeChunks;
        public ShaderInfo[] shad;

        public LODInfo lodInfo;
        public HlmtInfo hlmtInfo;
        public DisplayInfo displayInfo;
        public FrameInfo[] Frames;

        public Camera cam;
        public Material mat;
        public Matrix mWorld, mView, mProj;

        public int PolygonCount = 0;
        public Label xCoord, yCoord, zCoord;

        public bool mDown;
        public List<Keys> downKeys = new List<Keys>();

        public ThreeD(Map map)
        {
            InitializeComponent();
            this.SetStyle(ControlStyles.Opaque | ControlStyles.AllPaintingInWmPaint, true);

            this.Dock = DockStyle.Fill;
            this.map = map;
        }

        public void Initialize(Label xCoord, Label yCoord, Label zCoord)
        {
            timer1.Enabled = false;
            Init = false;

            this.xCoord = xCoord;
            this.yCoord = yCoord;
            this.zCoord = zCoord;

            if (Globals.DevWindow != this)
            {
                Globals.SetUpDevice(this);
                Globals.DevWindow = this;
            }

            if (Globals.devLoaded)
            {
                LoadSelectedModel();

                CreateEffect();
                SetupMatrices();

                mat = new Material();
                mat.Diffuse = Color.White;
                mat.Ambient = Color.White;
                mat.Specular = Color.White;

                Init = true;
                timer1.Enabled = true;
            }
        }

        public void RefreshMoi()
        {
            timer1.Enabled = false;
            Init = false;

            Globals.SetUpDevice(this);
            Globals.DevWindow = this;

            LoadSelectedModel();

            Init = true;
            timer1.Enabled = true;
        }

        public void CreateEffect()
        {
            effect = Effect.FromFile(Globals.dev, Application.StartupPath + "\\NormalMapping.fx",
                null, null, ShaderFlags.PartialPrecision, null);
        }

        public void SetupMatrices()
        {
            cam = new Camera();

            mWorld = Matrix.Identity;
            mView = Matrix.LookAtRH(cam.Pos, cam.Target, cam.Up);
            mProj = Matrix.PerspectiveFovRH((float)Math.PI / 4, 1.0f, 0.1f, 1000f);
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            if (Globals.DevWindow == this)
            {
                update();
                Render();
            }
            else timer1.Enabled = false;
        }

        public void update()
        {
            //UPDATE
            UpdateKeys();

            xCoord.Text = "Cam X Coord: " + cam.Pos.X.ToString();
            yCoord.Text = "Cam Y Coord: " + cam.Pos.Y.ToString();
            zCoord.Text = "Cam Z Coord: " + cam.Pos.Z.ToString();

            //set dev material
            Globals.dev.Material = mat;

            
            //--Effect.fx
            effect.Technique = "NormalMapping";

            effect.SetValue("World", mWorld);
            effect.SetValue("View", mView);
            effect.SetValue("Projection", mProj);

           // Matrix modeViewProj = mWorld * mView * mProj;
           // effect.SetValue("ModelViewProj", modeViewProj);

           // Matrix modeViewIT = mView; 
           // modeViewIT = Matrix.Invert(modeViewIT);
           // modeViewIT = Matrix.TransposeMatrix(modeViewIT);
           // effect.SetValue("ModelViewIT", modeViewIT);

           // effect.SetValue("ModelWorld", mWorld);

            //set the light position
            //float time = (float)(Environment.TickCount * 0.001);
            effect.SetValue("LightPosition", new Vector4(cam.Pos.X, cam.Pos.Y, cam.Pos.Z, 0));// new Vector4((float)(30 * Math.Sin(time)), 0.0f, (float)(30 * Math.Cos(time)), 1.0f));
            effect.SetValue("LightColor", new Vector4(1, 1, 1, 1));
            effect.SetValue("AmbientLightColor", new Vector4(.2f, .2f, .2f, 1));

            effect.SetValue("Shininess", .3f);
            effect.SetValue("SpecularPower", 4.0f);

            Globals.dev.Transform.World = mWorld;
            Globals.dev.Transform.View = mView;
            Globals.dev.Transform.Projection = mProj;
        }

        public void Render()
        {
            Globals.dev.Clear(ClearFlags.Target | ClearFlags.ZBuffer, Color.CadetBlue, 1.0f, 0);
            Globals.dev.BeginScene();

            //DRAW
            for (int x = 0; x < displayInfo.Chunk.Count; x++)
            {
                int rawIndx = displayInfo.Chunk[x];
                VertexBuffer vb = displayInfo.meshes[x].VertexBuffer;
                IndexBuffer ib = displayInfo.meshes[x].IndexBuffer;

                Globals.dev.SetStreamSource(0, vb, 0);
                Globals.dev.VertexFormat = HaloVertex.FVF;
                Globals.dev.Indices = ib;

                for (int y = 0; y < modeChunks[rawIndx].ModeSubMeshInfo.Length; y++)
                {
                    int sNum = modeChunks[rawIndx].ModeSubMeshInfo[y].ShaderNumber;
                    SetAlphaBlending(shad[sNum].Alpha);

                    /*
                    Globals.dev.SetTexture(1, shad[sNum].BaseTexture);
                    Globals.dev.Transform.Texture1 = Matrix.Scaling(shad[sNum].PrimaryDetailScale_U, shad[sNum].PrimaryDetailScale_V, shad[sNum].PrimaryDetailScale_W);
                    Globals.dev.TextureState[1].TextureTransform = TextureTransform.Count3;
                    Globals.dev.TextureState[1].ColorOperation = TextureOperation.Modulate;
                    Globals.dev.TextureState[1].ColorArgument1 = TextureArgument.TextureColor;
                    Globals.dev.TextureState[1].ColorArgument2 = TextureArgument.Diffuse;
                    Globals.dev.TextureState[1].TextureCoordinateIndex = 0;

                    if (shad[sNum].BumpTexture != null)
                    {
                        Globals.dev.SetTexture(0, shad[sNum].BumpTexture);
                        Globals.dev.Transform.Texture0 = Matrix.Scaling(shad[sNum].SecondaryDetailScale_U, shad[sNum].SecondaryDetailcale_V, shad[sNum].SecondaryDetailScale_W);
                        Globals.dev.TextureState[0].TextureTransform = TextureTransform.Count3;
                        Globals.dev.TextureState[0].ColorOperation = TextureOperation.DotProduct3;// BumpEnvironmentMap;
                        Globals.dev.TextureState[0].ColorArgument1 = TextureArgument.TextureColor;
                        Globals.dev.TextureState[0].ColorArgument2 = TextureArgument.TFactor;//.Current;
                        //Globals.dev.RenderState.TextureFactor = Color.FromArgb(0, 0, 1).ToArgb();
                        Globals.dev.TextureState[0].BumpEnvironmentMaterial00 = 1.0f;
                        Globals.dev.TextureState[0].BumpEnvironmentMaterial01 = 0.0f;
                        Globals.dev.TextureState[0].BumpEnvironmentMaterial10 = 0.0f;
                        Globals.dev.TextureState[0].BumpEnvironmentMaterial11 = 1.0f;
                        //Globals.dev.TextureState[1].BumpEnvironmentLuminanceScale = 0.5f;
                        //Globals.dev.TextureState[1].BumpEnvironmentLuminanceOffset = 0.0f;
                        Globals.dev.TextureState[0].TextureCoordinateIndex = 0;
                    }
                    else Globals.dev.SetTexture(1, null);

                    Globals.dev.RenderState.FillMode = FillMode.Solid;

                    displayInfo.meshes[x].DrawSubset(y);
                    */
               
                    if (shad[sNum].BumpTexture == null)
                    {
                        //set dev texture
                        Globals.dev.SetTexture(0, shad[sNum].BaseTexture);
                        //Globals.dev.Transform.Texture1 = Matrix.Scaling(shad[sNum].PrimaryDetailScale_U, shad[sNum].PrimaryDetailScale_V, shad[sNum].PrimaryDetailScale_W);
                        //Globals.dev.TextureState[1].TextureTransform = TextureTransform.Count3;
                        Globals.dev.TextureState[1].ColorOperation = TextureOperation.Modulate;
                        Globals.dev.TextureState[1].ColorArgument1 = TextureArgument.TextureColor;
                        Globals.dev.TextureState[1].ColorArgument2 = TextureArgument.Diffuse;

                        //draw
                        displayInfo.meshes[x].DrawSubset(y);
                    }
                    else
                    {
                        //null dev texture
                        Globals.dev.SetTexture(0, null);

                        //set the texturs
                        effect.SetValue("Texture", shad[sNum].BaseTexture);
                        effect.SetValue("NormalMap", shad[sNum].BumpTexture);

                        effect.SetValue("u", shad[sNum].DetailScaling[0].X);
                        effect.SetValue("v", shad[sNum].DetailScaling[0].Y);

                        //begin
                        effect.Begin(0);
                        effect.BeginPass(0);

                        //draw
                        displayInfo.meshes[x].DrawSubset(y);

                        //end
                        effect.EndPass();
                        effect.End();
                    }
                }
            }

            Globals.dev.EndScene();
            Globals.dev.Present();
        }

        public void LoadSelectedModel()
        {
            BinaryReader br = new BinaryReader(new FileStream(map.path, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite));

            #region BoundingBox
            br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset + 24;
            modeBox = new ModelBoundingBox(br, map.index.secondaryMagic);
            #endregion

            #region Model Chunks
            br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset + 36;
            int modeChunkCount = br.ReadInt32();
            int modeChunkStart = br.ReadInt32() - map.index.secondaryMagic;

            PolygonCount = 0;
            modeChunks = new ModelChunk[modeChunkCount];
            for (int x = 0; x < modeChunkCount; x++)
            {
                br.BaseStream.Position = modeChunkStart + (x * 92) + 4;
                modeChunks[x].VertCount = br.ReadUInt16();

                modeChunks[x].FaceCount = br.ReadUInt16();
                PolygonCount += modeChunks[x].FaceCount;

                br.BaseStream.Position += 12;
                modeChunks[x].BoneType = br.ReadByte();

                br.BaseStream.Position += 47;
                modeChunks[x].HeaderSize = map.rawData.rdh.rawChunks[x].Size - br.ReadInt32() - 4;

                int internalChunkCount = br.ReadInt32();
                int internalChunkStart = br.ReadInt32() - map.index.secondaryMagic;

                modeChunks[x].ModeRawDataChunks = new ModelRawDataChunk[internalChunkCount];
                for (int y = 0; y < internalChunkCount; y++)
                {
                    br.BaseStream.Position = internalChunkStart + (y * 16) + 6;
                    modeChunks[x].ModeRawDataChunks[y] = new ModelRawDataChunk();

                    modeChunks[x].ModeRawDataChunks[y].chunkSize = br.ReadUInt16();
                    modeChunks[x].ModeRawDataChunks[y].Size = br.ReadInt32();

                    if (modeChunks[x].ModeRawDataChunks[y].chunkSize == 0)
                        modeChunks[x].ModeRawDataChunks[y].chunkSize = modeChunks[x].ModeRawDataChunks[y].Size;

                    modeChunks[x].ModeRawDataChunks[y].Offset = br.ReadInt32();

                    modeChunks[x].ModeRawDataChunks[y].chunkCount =
                        modeChunks[x].ModeRawDataChunks[y].Size / modeChunks[x].ModeRawDataChunks[y].chunkSize;
                }

                BinaryReader rawBr = new BinaryReader(map.rawData.rdh.rawChunks[x].MS);

                modeChunks[x].ModeSubMeshInfo = new ModelSubMeshInfo[
                    modeChunks[x].ModeRawDataChunks[0].chunkCount];

                for (int y = 0; y < modeChunks[x].ModeRawDataChunks[0].chunkCount; y++)
                {
                    rawBr.BaseStream.Position = modeChunks[x].HeaderSize +
                        modeChunks[x].ModeRawDataChunks[0].Offset + (y * 72) + 4;

                    modeChunks[x].ModeSubMeshInfo[y].ShaderNumber = rawBr.ReadUInt16();
                    modeChunks[x].ModeSubMeshInfo[y].IndiceStart = rawBr.ReadUInt16();
                    modeChunks[x].ModeSubMeshInfo[y].IndiceCount = rawBr.ReadUInt16();
                }

                rawBr.BaseStream.Position = 40;
                modeChunks[x].IndiceCount = rawBr.ReadUInt16();

                //----Determine RawData Chunk Numbers
                int indiceChunk = 0;
                int vertChunk = 0;
                int uvChunk = 0;

                for (int y = 0; y < modeChunks[x].ModeRawDataChunks.Length; y++)
                {
                    if (modeChunks[x].ModeRawDataChunks[y].chunkSize == 2)
                    {
                        indiceChunk = y;
                        vertChunk = y + 2;
                        uvChunk = y + 3;
                        break;
                    }
                }

                modeChunks[x].BoneMap = new List<int>();
                int bmChunk = 0;
                rawBr.BaseStream.Position = 108;
                int bmCount = rawBr.ReadUInt16();

                if (bmCount > 0)
                {
                    for (int y = uvChunk + 1; y < modeChunks[x].ModeRawDataChunks.Length; y++)
                    {
                        if (modeChunks[x].ModeRawDataChunks[y].chunkSize == 1)
                        {
                            bmChunk = y;
                            break;
                        }
                    }

                    rawBr.BaseStream.Position = modeChunks[x].HeaderSize + modeChunks[x].ModeRawDataChunks[bmChunk].Offset;
                    for (int y = 0; y < bmCount; y++) modeChunks[x].BoneMap.Add(rawBr.ReadByte());
                }
                else modeChunks[x].BoneMap.Add(0);
                //------

                modeChunks[x].Verts = new List<Vector3>();
                modeChunks[x].ModeRawDataChunks[vertChunk].chunkSize =
                    modeChunks[x].ModeRawDataChunks[vertChunk].Size / modeChunks[x].VertCount;

                for (int y = 0; y < modeChunks[x].VertCount; y++)
                {
                    rawBr.BaseStream.Position = modeChunks[x].HeaderSize + modeChunks[x].ModeRawDataChunks[vertChunk].Offset
                        + (modeChunks[x].ModeRawDataChunks[vertChunk].chunkSize * y);

                    Vector3 vec = new Vector3();
                    vec.X = DecompressVertice(rawBr.ReadInt16(), modeBox.MinX, modeBox.MaxX);
                    vec.Y = DecompressVertice(rawBr.ReadInt16(), modeBox.MinY, modeBox.MaxY);
                    vec.Z = DecompressVertice(rawBr.ReadInt16(), modeBox.MinZ, modeBox.MaxZ);
                    modeChunks[x].Verts.Add(vec);
                    //if (tempbonemapcount == 0) continue;

                    modeChunks[x].VertBones = new List<BoneInfo>();

                    BoneInfo boneInfo = new BoneInfo();
                    boneInfo.BoneIndex = new List<int>();
                    boneInfo.Weight = new List<float>();

                    switch (modeChunks[x].ModeRawDataChunks[vertChunk].chunkSize)
                    {
                        case 6:
                            boneInfo.BoneIndex.Add(0);
                            boneInfo.Weight.Add(1.0f);
                            modeChunks[x].VertBones.Add(boneInfo);
                            break;

                        case 8:
                            boneInfo.BoneIndex.Add(rawBr.ReadByte());
                            boneInfo.Weight.Add(1.0f);

                            byte boneIndx = rawBr.ReadByte();
                            if (boneIndx == 0) modeChunks[x].VertBones.Add(boneInfo);
                            else
                            {
                                BoneInfo boneInfo2 = new BoneInfo();
                                boneInfo2.BoneIndex.Add(boneIndx);
                                boneInfo2.Weight.Add(1.0f);

                                modeChunks[x].VertBones.Add(boneInfo2);
                                modeChunks[x].VertBones.Add(boneInfo);
                            }
                            break;

                        case 12:
                            boneInfo.BoneIndex.Add(rawBr.ReadByte());
                            boneInfo.Weight.Add(0.99f);
                            modeChunks[x].VertBones.Add(boneInfo);
                            break;
                    }
                }

                modeChunks[x].UVs = new List<Vector2>();
                modeChunks[x].ModeRawDataChunks[uvChunk].chunkSize = 4;
                for (int y = 0; y < modeChunks[x].VertCount; y++)
                {
                    Vector2 vec = new Vector2();
                    rawBr.BaseStream.Position = modeChunks[x].HeaderSize + modeChunks[x].ModeRawDataChunks[uvChunk].Offset
                        + (modeChunks[x].ModeRawDataChunks[uvChunk].chunkSize * y);

                    vec.X = DecompressVertice(rawBr.ReadInt16(), modeBox.MinU, modeBox.MaxU) % 1;
                    vec.Y = DecompressVertice(rawBr.ReadInt16(), modeBox.MinV, modeBox.MaxV) % 1;

                    // if (tempuv.X > 1) { tempuv.X = tempuv.X - 1; }
                    // else 
                    if (vec.X < 0) vec.X = 1 - vec.X;

                    // if (tempuv.Y > 1) { tempuv.Y = tempuv.Y - 1; }
                    // else 
                    if (vec.Y < 0) vec.Y = 1 - vec.Y;
                    modeChunks[x].UVs.Add(vec);
                }

                modeChunks[x].Normals = new List<Vector3>();
                modeChunks[x].ModeRawDataChunks[uvChunk + 1].chunkSize = 12;
                for (int y = 0; y < modeChunks[x].VertCount; y++)
                {
                    rawBr.BaseStream.Position = modeChunks[x].HeaderSize + modeChunks[x].ModeRawDataChunks[uvChunk + 1].Offset
                        + (modeChunks[x].ModeRawDataChunks[uvChunk + 1].chunkSize * y);

                    int dword = rawBr.ReadInt32();
                    Vector3 normal = DecompressNormal(dword);

                    //int converto = ParsedModel.CompressNormal(normal);
                    //Vector3 tempnormal2 = ParsedModel.DecompressNormal(converto);
                    modeChunks[x].Normals.Add(normal);
                }

                rawBr.BaseStream.Position = modeChunks[x].HeaderSize +
                    modeChunks[x].ModeRawDataChunks[indiceChunk].Offset;

                modeChunks[x].Indices = new short[modeChunks[x].IndiceCount];
                for (int y = 0; y < modeChunks[x].IndiceCount; y++)
                {
                    //modeChunks[x].Indices[y] = (short)rawBr.ReadUInt16();
                    modeChunks[x].Indices[y] = rawBr.ReadInt16();
                }
            }
            #endregion

            #region Shaders
            br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset + 96;

            int shaderChunkCount = br.ReadInt32();
            int shaderChunkStart = br.ReadInt32() - map.index.secondaryMagic;

            shad = new ShaderInfo[shaderChunkCount];
            for (int x = 0; x < shaderChunkCount; x++)
            {
                br.BaseStream.Position = shaderChunkStart + (x * 32) + 12;
                int shadIndx = map.tags.FindTagIndexByID(br.ReadInt32());

                shad[x].Name = map.tags.tagsList[shadIndx].Path;

                br.BaseStream.Position = map.tags.tagsList[shadIndx].MetaOffset + 4;
                int stemIndx = map.tags.FindTagIndexByID(br.ReadInt32());

                //parse alpha
                if (stemIndx == -1)
                {
                    shad[x].Alpha = AlphaType.None;
                }
                else if (map.tags.tagsList[stemIndx].Path.IndexOf("alphatest") != -1)
                {
                    shad[x].Alpha = AlphaType.AlphaTest;
                }
                else if (map.tags.tagsList[stemIndx].Path.IndexOf("alpha") != -1 ||
                    map.tags.tagsList[stemIndx].Path.IndexOf("water") != -1)
                {
                    shad[x].Alpha = AlphaType.AlphaBlend;
                }
                else shad[x].Alpha = AlphaType.None;

                br.BaseStream.Position = map.tags.tagsList[shadIndx].MetaOffset + 12;

                int tempc2 = br.ReadInt32();
                int tempr2 = br.ReadInt32() - map.index.secondaryMagic;

                if (tempc2 != 0)
                {
                    //goto basemap
                    br.BaseStream.Position = tempr2 + 4;
                    int bitmIndx = map.tags.FindTagIndexByID(br.ReadInt32());

                    if (bitmIndx != -1)
                    {
                        BitmapData bd = new BitmapData(map, bitmIndx);
                        map.bitm.LoadBitmap(bitmIndx);
                        shad[x].BaseBitmap = map.bitm.ReturnBitmap(0, 0, 0, 0, bd.rawChunks);
                        shad[x].BaseName = map.tags.tagsList[bitmIndx].Path;
                        shad[x].BaseLevels = map.bitm.BitmapInfoList[0].MipMapCount;
                    }
                }

                br.BaseStream.Position = map.tags.tagsList[shadIndx].MetaOffset + 36;
                //tempc2 = br.ReadInt32();
                tempr2 = br.ReadInt32() - map.index.secondaryMagic;

                br.BaseStream.Position = tempr2 + 20;
                int DetailCount = br.ReadInt32();
                int DetailReflex = br.ReadInt32() - map.index.secondaryMagic;

                if (DetailCount > 0)
                {
                    shad[x].DetailScaling=new Vector3[DetailCount];
                    br.BaseStream.Position = DetailReflex;

                    for (int y = 0; y < DetailCount; y++)
                    {
                        shad[x].DetailScaling[y] = new Vector3(br.ReadSingle(), br.ReadSingle(), br.ReadSingle());
                        br.BaseStream.Position += 4;
                    }
                }

                br.BaseStream.Position = tempr2 + 4;
                tempc2 = br.ReadInt32();
                tempr2 = br.ReadInt32() - map.index.secondaryMagic;

                /*
                for (int x = 0; x < tempc2; x++)
                {
                    br.BaseStream.Position = tempr2 + (x * 12);
                    int tempcrap = map.tags.FindTagIndexByID(br.ReadInt32());
                    if (tempcrap == -1) { continue; }
                    Meta tempmeta = new Meta();
                    tempmeta.ReadMetaFromMap(tempcrap, mapnumber, false);
                    Raw.ParsedBitmap pm = new Raw.ParsedBitmap(ref tempmeta, mapnumber);
                    Bitmaps.Add(pm.FindChunkAndDecode(0, 0, ref tempmeta, mapnumber, 0));
                    BitmapNames.Add(Maps.map[mapnumber].FileNames.Name[tempcrap]);
                }
          
                br.BaseStream.Position = tempr + 20;
                tempc2 = br.ReadInt32();
                tempr2 = br.ReadInt32() - map.index.secondaryMagic;

                tempc2 = tempc;
                tempr2 = tempr;
                */

                if (tempc2 != 0)
                {
                    br.BaseStream.Position = tempr2;
                    int bitmIndx = map.tags.FindTagIndexByID(br.ReadInt32());

                    if (bitmIndx != -1)
                    {
                        int reflect = map.tags.tagsList[bitmIndx].Path.IndexOf("reflection_maps");

                        if (map.tags.tagsList[bitmIndx].Path.IndexOf("_bump") != -1)
                        {
                            BitmapData bd = new BitmapData(map, bitmIndx);
                            shad[x].BumpBitmap = map.bitm.ReturnBitmap(0, 0, 0, 0, bd.rawChunks);
                            shad[x].BumpName = map.tags.tagsList[bitmIndx].Path;
                        }
                        else if (map.tags.tagsList[bitmIndx].Path.IndexOf("_cube_map") != -1)
                        {
                            //BitmapData bd = new BitmapData(map, bitmIndx);
                            //shad[x].CubeBitmap = map.bitm.ReturnBitmap(0, 0, 0, 0, bd.rawChunks);
                            shad[x].CubeName = map.tags.tagsList[bitmIndx].Path;
                        }
                        else if (map.tags.tagsList[bitmIndx].Path.IndexOf("default_") == -1 && shad[x].BaseBitmap == null && reflect == -1)
                        {
                            BitmapData bd = new BitmapData(map, bitmIndx);
                            map.bitm.LoadBitmap(bitmIndx);
                            shad[x].BaseBitmap = map.bitm.ReturnBitmap(0, 0, 0, 0, bd.rawChunks);
                            shad[x].BaseName = map.tags.tagsList[bitmIndx].Path;
                            shad[x].BaseLevels = map.bitm.BitmapInfoList[0].MipMapCount;
                        }
                        else if (reflect != -1)
                        {
                            /*
                            br.BaseStream.Position += 8;
                            tempcrap = map.tags.FindTagIndexByID(br.ReadInt32());
                            Meta tempmeta = new Meta();
                            tempmeta.ReadMetaFromMap(tempcrap, mapnumber, false);
                            Raw.ParsedBitmap pm = new Raw.ParsedBitmap(ref tempmeta, mapnumber);
                            this.MainBitmap = pm.FindChunkAndDecode(0, 0, ref tempmeta, mapnumber);
                            this.MainName = Maps.map[mapnumber].FileNames.Name[tempcrap];
                            */
                        }
                    }

                    br.BaseStream.Position = tempr2 + 24;
                    bitmIndx = map.tags.FindTagIndexByID(br.ReadInt32());
                    if (bitmIndx != -1 && shad[x].BaseBitmap == null)
                    {
                        BitmapData bd = new BitmapData(map, bitmIndx);
                        map.bitm.LoadBitmap(bitmIndx);
                        shad[x].BaseBitmap = map.bitm.ReturnBitmap(0, 0, 0, 0, bd.rawChunks);
                        shad[x].BaseName = map.tags.tagsList[bitmIndx].Path;
                        shad[x].BaseLevels = map.bitm.BitmapInfoList[0].MipMapCount;
                    }
                    else if (shad[x].BaseBitmap == null)
                    {
                        br.BaseStream.Position = tempr2 + 12;
                        bitmIndx = map.tags.FindTagIndexByID(br.ReadInt32());
                        if (bitmIndx != -1)
                        {
                            BitmapData bd = new BitmapData(map, bitmIndx);
                            map.bitm.LoadBitmap(bitmIndx);
                            shad[x].BaseBitmap = map.bitm.ReturnBitmap(0, 0, 0, 0, bd.rawChunks);
                            shad[x].BaseName = map.tags.tagsList[bitmIndx].Path;
                            shad[x].BaseLevels = map.bitm.BitmapInfoList[0].MipMapCount;
                        }
                    }
                }
            }
            #endregion

            #region LOD
            br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset + 28;
            int lodChunkCount = br.ReadInt32();
            int lodChunkStart = br.ReadInt32() - map.index.secondaryMagic;

            lodInfo.lodPieceInfos = new LODPieceInfo[lodChunkCount];
            lodInfo.PermutationStrings = new List<string>();

            for (int x = 0; x < lodChunkCount; x++)
            {
                br.BaseStream.Position = lodChunkStart + (x * 16);
                lodInfo.lodPieceInfos[x].Name = map.strings.stringsList[br.ReadInt16()];

                br.BaseStream.Position = lodChunkStart + (x * 16) + 8;
                int tempc2 = br.ReadInt32();
                int tempr2 = br.ReadInt32() - map.index.secondaryMagic;

                lodInfo.lodPieceInfos[x].Permutation = new LODPermutationInfo[tempc2];
                for (int y = 0; y < tempc2; y++)
                {
                    br.BaseStream.Position = tempr2 + (y * 16);

                    string pName = map.strings.stringsList[br.ReadInt16()];
                    lodInfo.lodPieceInfos[x].Permutation[y].Name = pName;

                    if (!lodInfo.PermutationStrings.Contains(pName)) lodInfo.PermutationStrings.Add(pName);

                    br.BaseStream.Position += 2;
                    lodInfo.lodPieceInfos[x].Permutation[y].PieceNumber = new int[5];
                    for (int z = 0; z < 5; z++)
                    {
                        short pNum = br.ReadInt16();
                        lodInfo.lodPieceInfos[x].Permutation[y].PieceNumber[z] = pNum;
                        modeChunks[pNum].PieceName = lodInfo.lodPieceInfos[x].Name;
                        modeChunks[pNum].Permutation = pName;
                        modeChunks[pNum].LOD = z;
                    }
                }
            }
            #endregion

            #region Hlmt
            hlmtInfo.TagIndx = map.tags.FindTagIndexByTypeAndPath("hlmt", map.tags.tagsList[map.tags.selectedIndex].Path);
            if (hlmtInfo.TagIndx != -1)
            {
                br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset + 80;
                int tempc = br.ReadInt32();
                int tempr = br.ReadInt32() - map.index.secondaryMagic;

                hlmtInfo.Permutations = new HlmtPermutationInfo[tempc];
                for (int x = 0; x < tempc; x++)
                {
                    br.BaseStream.Position = tempr + (x * 56);
                    hlmtInfo.Permutations[x].Name = map.strings.stringsList[br.ReadInt16()];

                    br.BaseStream.Position = tempr + (x * 56) + 20;
                    int tempc2 = br.ReadInt32();
                    int tempr2 = br.ReadInt32() - map.index.secondaryMagic;

                    hlmtInfo.Permutations[x].Piece = new HlmtPermutationPiece[tempc2];
                    for (int y = 0; y < tempc2; y++)
                    {
                        br.BaseStream.Position = tempr2 + (y * 20);
                        hlmtInfo.Permutations[x].Piece[y].PieceName = map.strings.stringsList[br.ReadInt16()];

                        br.BaseStream.Position = tempr2 + (y * 20) + 8;
                        int tempc3 = br.ReadInt32();
                        int tempr3 = br.ReadInt32() - map.index.secondaryMagic;

                        hlmtInfo.Permutations[x].Piece[y].Permutation = new HlmtPermutationVariation[tempc3];
                        for (int z = 0; z < tempc3; z++)
                        {
                            br.BaseStream.Position = tempr3 + (z * 32);
                            hlmtInfo.Permutations[x].Piece[y].Permutation[z].PermutationNameX = map.strings.stringsList[br.ReadInt16()];

                            br.BaseStream.Position = tempr3 + (z * 32) + 12;
                            int tempc4 = br.ReadInt32();
                            int tempr4 = br.ReadInt32() - map.index.secondaryMagic;

                            /*
                            Permutations[x].Piece[xx].Permutation=new PermutationInfo.PermutationPiece.PermutationVariation[tempc4]; 
                            for (int w = 0; w < tempc4; w++)
                            {
                                Permutations[x].Piece[xx].Permutation[xxxx]=new entity.MetaContainers.hlmtContainer.PermutationInfo.PermutationPiece.PermutationVariation();
                                br.BaseStream.Position=tempr4+(xxxx*24);
                                string temp3=map.strings.stringsList[br.ReadInt16()];

                                PermutationPiece p=new PermutationPiece();

                                p.PermutationName=temps2;
                                p.PermutationNameX=temp3;
                                Permutations[x].Permutation.Add(p);
                            }
                            */
                        }
                    }
                }

                if (hlmtInfo.Permutations.Length > 0) hlmtInfo.PermutationString = hlmtInfo.Permutations[0].Name;
            }
            #endregion

            #region Displayed
            int lodLevel = 4;
            displayInfo.Chunk = new List<int>();

            if (hlmtInfo.TagIndx == -1 || hlmtInfo.Permutations.Length == 0)
            {
                for (int x = 0; x < lodInfo.lodPieceInfos.Length; x++)
                {
                    for (int y = 0; y < lodInfo.lodPieceInfos[x].Permutation.Length; y++)
                    {
                        if (lodInfo.lodPieceInfos[x].Permutation[y].Name == lodInfo.PermutationStrings[0])
                        {
                            if (!displayInfo.Chunk.Contains(lodInfo.lodPieceInfos[x].Permutation[y].PieceNumber[lodLevel]))
                                displayInfo.Chunk.Add(lodInfo.lodPieceInfos[x].Permutation[y].PieceNumber[lodLevel]);
                        }
                    }
                }
            }
            else
            {
                int permIndx = 0;// hlmtInfo.FindPermutation(hlmtInfo.PermutationString);
                for (int x = 0; x < hlmtInfo.Permutations[permIndx].Piece.Length; x++)
                {
                    string name1 = hlmtInfo.Permutations[permIndx].Piece[x].PieceName;

                    for (int y = 0; y < hlmtInfo.Permutations[permIndx].Piece[x].Permutation.Length; y++)
                    {
                        string name2 = hlmtInfo.Permutations[permIndx].Piece[x].Permutation[y].PermutationNameX;
                        LODPermutationInfo permInfo = lodInfo.FindPermutationInfo(name1, name2);

                        if (!displayInfo.Chunk.Contains(permInfo.PieceNumber[lodLevel]))
                            displayInfo.Chunk.Add(permInfo.PieceNumber[lodLevel]);
                    }
                }
            }

            //shaders
            List<int> tmpShaders = new List<int>();
            for (int x = 0; x < displayInfo.Chunk.Count; x++)
            {
                for (int y = 0; y < modeChunks[displayInfo.Chunk[x]].ModeSubMeshInfo.Length; y++)
                {
                    int sNum = modeChunks[displayInfo.Chunk[x]].ModeSubMeshInfo[y].ShaderNumber;
                    if (!tmpShaders.Contains(sNum)) tmpShaders.Add(sNum);
                }
            }
            displayInfo.ShaderIndex = tmpShaders.ToArray();
            #endregion

            #region Frames
            br.BaseStream.Position = map.tags.tagsList[map.tags.selectedIndex].MetaOffset + 72;
            int frameChunkCount = br.ReadInt32();
            int frameChunkStart = br.ReadInt32() - map.index.secondaryMagic;
            Frames = new FrameInfo[frameChunkCount];

            for (int x = 0; x < frameChunkCount; x++)
            {
                FrameInfo finf = new FrameInfo();

                br.BaseStream.Position = frameChunkStart + (96 * x);
                finf.Name = map.strings.stringsList[br.ReadUInt16()];

                br.BaseStream.Position += 2;
                finf.Parent = br.ReadInt16();
                finf.Child = br.ReadInt16();
                finf.Sibling = br.ReadInt16();

                br.BaseStream.Position = frameChunkStart + (96 * x) + 12;
                finf.x = br.ReadSingle();
                finf.y = br.ReadSingle();
                finf.z = br.ReadSingle();
                finf.i = br.ReadSingle();
                finf.j = br.ReadSingle();
                finf.k = br.ReadSingle();
                finf.w = br.ReadSingle();
                finf.matrix = Matrix.Identity;

                Quaternion q = new Quaternion(finf.i, finf.j, finf.k, finf.w);
                finf.matrix.Multiply(Matrix.Translation(finf.x, finf.y, finf.z));
                finf.matrix.RotateQuaternion(q);

                br.BaseStream.Position += 4;

                finf.skinweightmatrix = Matrix.Identity;
                finf.skinweightmatrix.M11 = br.ReadSingle();
                finf.skinweightmatrix.M12 = br.ReadSingle();
                finf.skinweightmatrix.M13 = br.ReadSingle();
                finf.skinweightmatrix.M14 = 0.0f;

                finf.skinweightmatrix.M21 = br.ReadSingle();
                finf.skinweightmatrix.M22 = br.ReadSingle();
                finf.skinweightmatrix.M23 = br.ReadSingle();
                finf.skinweightmatrix.M24 = 0.0f;

                finf.skinweightmatrix.M31 = br.ReadSingle();
                finf.skinweightmatrix.M32 = br.ReadSingle();
                finf.skinweightmatrix.M33 = br.ReadSingle();
                finf.skinweightmatrix.M34 = 0.0f;

                finf.skinweightmatrix.M41 = br.ReadSingle();
                finf.skinweightmatrix.M42 = br.ReadSingle();
                finf.skinweightmatrix.M43 = br.ReadSingle();
                finf.skinweightmatrix.M44 = 1.0f;

                finf.DistanceFromParent = br.ReadSingle();
                // f.matrix.MultiplyTranspose(f.skinweightmatrix);//=Matrix.Transformation(new Vector3(0, 0, 0), new Quaternion(),new Vector3(1,1,1), new Vector3(0, 0, 0), q, new Vector3(f.x, f.y, f.z));
                Frames[x] = finf;
            }

            #endregion

            br.Close();
            LoadTexturesAndBuffers();
        }

        #region Mouse
        protected override void OnMouseEnter(EventArgs e)
        {
            if (Init)
            {
                this.Focus();
                if (!timer1.Enabled && Globals.DevWindow == this)
                    timer1.Enabled = true;
            }
        }

        protected override void OnMouseDown(MouseEventArgs e)
        {
            if (Init)
            {
                cam.oldx = e.X;
                cam.oldy = e.Y;
                mDown = true;
            }
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            if (mDown)
            {
                cam.Change(e.X, e.Y);

                mView = Matrix.LookAtRH(cam.Pos, cam.Target, cam.Up);
                //Globals.dev.Transform.View = mView;
                //Render();
            }
        }

        protected override void OnMouseUp(MouseEventArgs e)
        {
            mDown = false;
        }
        #endregion

        #region Keyboard
        protected override void OnKeyDown(KeyEventArgs e)
        {
            if (!downKeys.Contains(e.KeyData)) downKeys.Add(e.KeyData);
        }

        protected override void OnKeyUp(KeyEventArgs e)
        {
            downKeys.Remove(e.KeyCode);
        }

        public void UpdateKeys()
        {
            if (downKeys.Count != 0)
            {
                for (int x = 0; x < downKeys.Count; x++)
                {
                    switch (downKeys[x])
                    {
                        case Keys.W:
                            cam.Pos.X += cam.i * cam.speed;
                            cam.Pos.Y += cam.j * cam.speed;
                            cam.Pos.Z += cam.k * cam.speed;
                            break;

                        case Keys.S:
                            cam.Pos.X -= cam.i * cam.speed;
                            cam.Pos.Y -= cam.j * cam.speed;
                            cam.Pos.Z -= cam.k * cam.speed;
                            break;

                        case Keys.A:
                            cam.ComputeStrafe(false);
                            // y += speed;
                            // Position.Y = y;
                            break;

                        case Keys.D:
                            cam.ComputeStrafe(true);
                            // y -= speed;
                            //Position.Y = y;
                            break;

                        case Keys.Z:
                            cam.Pos.Z -= cam.speed;
                            break;

                        case Keys.X:
                            cam.Pos.Z += cam.speed;
                            break;
                    }
                }

                cam.ComputePosition();
                mView = Matrix.LookAtRH(cam.Pos, cam.Target, cam.Up);
            }
        }
        #endregion

        #region Functions

        public float DecompressVertice(float input, float min, float max)
        {
            float percent = (input + 32768) / 65535;
            return (percent * (max - min)) + min;
        }

        public float CompressVertice(float input, float min, float max)
        {
            float result = input - min;
            result /= (max - min);
            result *= 65535;
            result -= 32768;
            return result;
        }

        public Vector3 DecompressNormal(int compressednormal)
        {
            int xx = compressednormal & 0x3ff;
            float xxx = Convert.ToSingle(xx) / Convert.ToSingle(0x3ff);
            int bitx = ((compressednormal >> 10) & 0x1);
            if (bitx == 1) xxx = -(1 - xxx);

            int yy = ((compressednormal >> 11) & 0x3ff);
            float yyy = Convert.ToSingle(yy) / Convert.ToSingle(0x3ff);
            int bity = ((compressednormal >> 21) & 0x1);
            if (bity == 1) yyy = -(1 - yyy);

            int zz = ((compressednormal >> 22) & 0x1ff);
            float zzz = Convert.ToSingle(zz) / Convert.ToSingle(0x1ff);
            int bitz = ((compressednormal >> 31) & 0x1);
            if (bitz == 1) zzz = -(1 - zzz);

            return new Vector3(xxx, yyy, zzz);
        }

        public short[] DecompressIndices(short[] indices, int start, int count)
        {
            bool dir = false;
            short tmpX;
            short tmpY;
            short tmpZ;
            //short[] shite = new short[50000];
            int m = start;
            //int s = 0;

            List<short> tmpShorts = new List<short>();

            while (m < start + count - 2)
            {
                tmpX = indices[m];
                tmpY = indices[m + 1];
                tmpZ = indices[m + 2];

                if (tmpX != tmpY && tmpX != tmpZ && tmpY != tmpZ)
                {
                    if (!dir)
                    {
                        tmpShorts.Add(tmpX);
                        tmpShorts.Add(tmpY);
                        tmpShorts.Add(tmpZ);
                        //shite[s] = tmpX;
                        //shite[s + 1] = tmpY;
                        //shite[s + 2] = tmpZ;
                        //s += 3;
                        dir = true;
                    }
                    else
                    {
                        tmpShorts.Add(tmpX);
                        tmpShorts.Add(tmpZ);
                        tmpShorts.Add(tmpY);
                        //shite[s] = tmpX;
                        //shite[s + 1] = tmpZ;
                        //shite[s + 2] = tmpY;
                        //s += 3;
                        dir = false;
                    }
                    m += 1;
                }
                else
                {
                    dir = !dir;
                    m += 1;
                }
            }

            //short[] uncompressedIndices = new short[s];
            //Array.Copy(shite, uncompressedIndices, s);
            //return uncompressedIndices;
            return tmpShorts.ToArray();
        }

        public void SetAlphaBlending(AlphaType alpha)
        {
            switch (alpha)
            {
                case AlphaType.AlphaBlend:
                    Globals.dev.RenderState.AlphaBlendEnable = true;
                    Globals.dev.RenderState.AlphaTestEnable = false;
                    break;

                case AlphaType.AlphaTest:
                    Globals.dev.RenderState.AlphaBlendEnable = false;
                    Globals.dev.RenderState.AlphaTestEnable = true;
                    break;

                case AlphaType.None:
                    Globals.dev.RenderState.AlphaBlendEnable = false;
                    Globals.dev.RenderState.AlphaTestEnable = false;
                    break;
            }
        }

        #endregion

        #region Creation

        public void LoadTexturesAndBuffers()
        {
            CreateVertexBuffers();
            CreateIndexBuffers();
            CreateShaderTextures();
            MakeMeshes();
        }

        public void CreateVertexBuffers()
        {
            displayInfo.vertexBuffer = new VertexBuffer[modeChunks.Length];
            for (int x = 0; x < displayInfo.Chunk.Count; x++)
            {
                int rawindex = displayInfo.Chunk[x];

                displayInfo.vertexBuffer[rawindex] = new VertexBuffer(typeof(HaloVertex),
                    modeChunks[rawindex].VertCount, Globals.dev, Usage.WriteOnly, HaloVertex.FVF, Pool.Managed);

                // Lock the buffer (which will return our structs)
                HaloVertex[] verts = (HaloVertex[])displayInfo.vertexBuffer[rawindex].Lock(0, 0);

                for (int y = 0; y < modeChunks[rawindex].VertCount; y++)
                {
                    verts[y].Position = new Vector3(
                        modeChunks[rawindex].Verts[y].X,
                        modeChunks[rawindex].Verts[y].Y,
                        modeChunks[rawindex].Verts[y].Z);

                    verts[y].Tu0 = modeChunks[rawindex].UVs[x].X;
                    verts[y].Tv0 = modeChunks[rawindex].UVs[x].Y;

                    verts[y].Normal = new Vector3(
                        modeChunks[rawindex].Verts[y].X,
                        modeChunks[rawindex].Verts[y].Y,
                        modeChunks[rawindex].Verts[y].Z);

                    verts[y].specular = 1;
                    verts[y].diffuse = 1;
                }
                // Unlock (and copy) the data
                displayInfo.vertexBuffer[rawindex].Unlock();
            }
        }

        public void CreateIndexBuffers()
        {
            // Pool pool=;
            //  if (Globals.dev.DeviceCaps..
            displayInfo.indexBuffer = new IndexBuffer[modeChunks.Length];
            for (int rawindex = 0; rawindex < displayInfo.Chunk.Count; rawindex++)
            {
                int x = displayInfo.Chunk[rawindex];

                displayInfo.indexBuffer[x] = new IndexBuffer(typeof(short),
                    modeChunks[x].IndiceCount, Globals.dev, Usage.WriteOnly, Pool.Managed);

                displayInfo.indexBuffer[x].SetData(modeChunks[x].Indices, 0, LockFlags.None);
                displayInfo.indexBuffer[x].Unlock(); //needed?
            }
        }

        public void CreateShaderTextures()
        {
            for (int x = 0; x < displayInfo.ShaderIndex.Length; x++)
            {
                /*
                BitmapTextures = new Texture[Bitmaps.Count];
                for (int x = 0; x < Bitmaps.Count; x++) BitmapTextures[x] = CreateTexture(ref Globals.dev, Bitmaps[x]);
                */

                if (shad[displayInfo.ShaderIndex[x]].BumpBitmap != null)
                {
                    Texture tex = CreateTexture(shad[displayInfo.ShaderIndex[x]].BumpBitmap);
                    //SurfaceDescription desc = tex.GetLevelDescription(0);

                    //Texture nMap = new Texture(Globals.dev, desc.Width, desc.Height,
                    //    tex.LevelCount, 0, Format.A8R8G8B8, Pool.Managed);

                    //TextureLoader.ComputeNormalMap(nMap, tex,
                    //    0, Channel.Alpha, 1.0f);
                    shad[displayInfo.ShaderIndex[x]].BumpTexture = tex;
                }
                if (shad[displayInfo.ShaderIndex[x]].PrimaryDetailBitmap != null)
                {
                    shad[displayInfo.ShaderIndex[x]].SecondaryDetailTexture = CreateTexture(shad[displayInfo.ShaderIndex[x]].PrimaryDetailBitmap);
                }
                if (shad[displayInfo.ShaderIndex[x]].SecondaryDetailBitmap != null)
                {
                    shad[displayInfo.ShaderIndex[x]].SecondaryDetailTexture = CreateTexture(shad[displayInfo.ShaderIndex[x]].SecondaryDetailBitmap);
                }
                //if (shad[displayInfo.ShaderIndex[x]].MicroDetailBitmap != null)
                //{
                //    shad[displayInfo.ShaderIndex[x]].MicroDetailTexture = CreateTexture(shad[displayInfo.ShaderIndex[x]].MicroDetailBitmap);
                //}
                //if (shad[displayInfo.ShaderIndex[x]].CubeBitmap != null)
                //{
                //    shad[displayInfo.ShaderIndex[x]].CubeMapTexture = Texture.FromBitmap(Globals.dev, CubeMapBitmap, Usage.AutoGenerateMipMap, Pool.Managed);
                //}
                if (shad[displayInfo.ShaderIndex[x]].BaseBitmap != null)
                {
                    shad[displayInfo.ShaderIndex[x]].BaseTexture = CreateTexture(shad[displayInfo.ShaderIndex[x]].BaseBitmap);
                }
            }
        }

        public Texture CreateTexture(Bitmap bitmap)
        {
            Texture tex = new Texture(Globals.dev, bitmap.Width, bitmap.Height, 1,
                Usage.AutoGenerateMipMap, Format.A8R8G8B8, Pool.Managed);

            System.Drawing.Imaging.BitmapData bitmapData = bitmap.LockBits(new Rectangle(0, 0, bitmap.Width, bitmap.Height),
                System.Drawing.Imaging.ImageLockMode.ReadOnly, System.Drawing.Imaging.PixelFormat.Format32bppArgb);

            int pitch;
            GraphicsStream textureData = tex.LockRectangle(0, LockFlags.None, out pitch);

            //Debug.Assert(bitmap.PixelFormat == System.Drawing.Imaging.PixelFormat.Format32bppArgb);
            //Debug.Assert(sizeof(int) == 4 && (bitmapData.Stride & 3) == 0 && (pitch & 3) == 0);

            unsafe
            {
                int* texturePointer = (int*)textureData.InternalDataPointer;
                for (int y = 0; y < bitmap.Height; y++)
                {
                    int* bitmapLinePointer = (int*)bitmapData.Scan0 + y * (bitmapData.Stride / sizeof(int));
                    int* textureLinePointer = texturePointer + y * (pitch / sizeof(int));
                    int length = bitmap.Width;
                    while (--length >= 0) *textureLinePointer++ = *bitmapLinePointer++;
                }
            }

            bitmap.UnlockBits(bitmapData);
            tex.UnlockRectangle(0);
            tex.GenerateMipSubLevels();
            tex.AutoGenerateFilterType = TextureFilter.Linear;
            return tex;
        }

        public void MakeMeshes()
        {
            displayInfo.meshes = new Mesh[displayInfo.Chunk.Count];

            for (int x = 0; x < displayInfo.Chunk.Count; x++)
                displayInfo.meshes[x] = MakeMesh(displayInfo.Chunk[x]);
        }

        public Mesh MakeMesh(int chunkNum)
        {
            short[] indices;
            int[] facecount = new int[modeChunks[chunkNum].ModeSubMeshInfo.Length];
            int[] facestart = new int[modeChunks[chunkNum].ModeSubMeshInfo.Length];

            if (modeChunks[chunkNum].FaceCount * 3 != modeChunks[chunkNum].Indices.Length)
            {
                short[] tempindicesx = new short[100000];
                int tempc = 0;
                for (int x = 0; x < modeChunks[chunkNum].ModeSubMeshInfo.Length; x++)
                {
                    short[] tempindices = DecompressIndices(modeChunks[chunkNum].Indices,
                        modeChunks[chunkNum].ModeSubMeshInfo[x].IndiceStart,
                        modeChunks[chunkNum].ModeSubMeshInfo[x].IndiceCount);

                    Array.ConstrainedCopy(tempindices, 0, tempindicesx, tempc, tempindices.Length);

                    facecount[x] = tempindices.Length / 3;
                    facestart[x] = tempc;
                    tempc += tempindices.Length;
                }
                indices = new short[tempindicesx.Length];
                Array.Copy(tempindicesx, indices, tempindicesx.Length);
            }
            else
            {
                indices = modeChunks[chunkNum].Indices;
                for (int x = 0; x < modeChunks[chunkNum].ModeSubMeshInfo.Length; x++)
                {
                    facecount[x] = modeChunks[chunkNum].ModeSubMeshInfo[x].IndiceCount / 3;
                    facestart[x] = modeChunks[chunkNum].ModeSubMeshInfo[x].IndiceStart / 3;
                }
            }

            Mesh m = new Mesh(indices.Length / 3, modeChunks[chunkNum].Verts.Count,
                MeshFlags.Managed, HaloVertex.FVF, Globals.dev);

            HaloVertex[] tempv = new HaloVertex[modeChunks[chunkNum].Verts.Count];
            for (int x = 0; x < modeChunks[chunkNum].Verts.Count; x++)
            {
                tempv[x].Position = modeChunks[chunkNum].Verts[x];
                tempv[x].Normal = modeChunks[chunkNum].Normals[x];
                tempv[x].Tu0 = modeChunks[chunkNum].UVs[x].X;
                tempv[x].Tv0 = modeChunks[chunkNum].UVs[x].Y;

            }
            m.VertexBuffer.SetData(tempv, 0, LockFlags.None);
            m.VertexBuffer.Unlock();

            m.IndexBuffer.SetData(indices, 0, LockFlags.None);
            return m;
        }

        #endregion
    }

    #region Structs
    public struct ModelBoundingBox
    {
        public float
        MinX, MaxX,
        MinY, MaxY,
        MinZ, MaxZ,
        MinU, MaxU,
        MinV, MaxV;

        public ModelBoundingBox(BinaryReader br, int secMagic)
        {
            br.BaseStream.Position = br.ReadInt32() - secMagic;
            MinX = br.ReadSingle();
            MaxX = br.ReadSingle();
            MinY = br.ReadSingle();
            MaxY = br.ReadSingle();
            MinZ = br.ReadSingle();
            MaxZ = br.ReadSingle();
            MinU = br.ReadSingle();
            MaxU = br.ReadSingle();
            MinV = br.ReadSingle();
            MaxV = br.ReadSingle();
        }
    }

    public struct ModelChunk
    {
        public int VertCount;
        public int FaceCount;
        public byte BoneType;

        public int HeaderSize;

        public ModelRawDataChunk[] ModeRawDataChunks;
        public ModelSubMeshInfo[] ModeSubMeshInfo;

        public int IndiceCount;

        public List<int> BoneMap;
        public List<Vector3> Verts;

        public List<BoneInfo> VertBones;

        public List<Vector2> UVs;
        public List<Vector3> Normals;

        public short[] Indices;

        public string PieceName;
        public string Permutation;
        public int LOD;
    }

    public struct ModelSubMeshInfo
    {
        public int ShaderNumber;
        public int IndiceStart;
        public int IndiceCount;
    }

    public struct BoneInfo
    {
        public List<int> BoneIndex;
        public List<float> Weight;
    }

    public struct ShaderInfo
    {
        public string Name;
        public AlphaType Alpha;

        public Bitmap BaseBitmap;
        public string BaseName;
        public int BaseLevels;
        public Texture BaseTexture;

        public Vector3[] DetailScaling;

        public Bitmap PrimaryDetailBitmap;
        public Texture PrimaryDetailTexture;

        public Bitmap SecondaryDetailBitmap;
        public Texture SecondaryDetailTexture;

        public Bitmap BumpBitmap;
        public string BumpName;
        public BitmapFormat BumpFormat;
        public Texture BumpTexture;

        public Bitmap CubeBitmap;
        public string CubeName;
        public Texture CubeTexture;
    }

    public enum AlphaType
    {
        None,
        AlphaTest,
        AlphaBlend
    }

    public struct LODInfo
    {
        public LODPieceInfo[] lodPieceInfos;
        public List<string> PermutationStrings;

        public LODPermutationInfo FindPermutationInfo(string PieceName, string PermName)
        {
            for (int x = 0; x < lodPieceInfos.Length; x++)
            {
                if (lodPieceInfos[x].Name == PieceName)
                {
                    for (int y = 0; y < lodPieceInfos[x].Permutation.Length; y++)
                    {
                        if (lodPieceInfos[x].Permutation[y].Name == PermName)
                            return lodPieceInfos[x].Permutation[y];
                    }
                    return lodPieceInfos[x].Permutation[0];
                }
            }
            return new LODPermutationInfo();
        }
    }

    public struct LODPieceInfo
    {
        public string Name;
        public LODPermutationInfo[] Permutation;
    }

    public struct LODPermutationInfo
    {
        public string Name;
        public int[] PieceNumber;
    }

    public struct HlmtInfo
    {
        public int TagIndx;
        public HlmtPermutationInfo[] Permutations;
        public string PermutationString;

        public int FindPermutation(string PermName)
        {
            for (int x = 0; x < this.Permutations.Length; x++)
            {
                if (this.Permutations[x].Name == PermName)
                    return x;
            }
            return 0;
        }
    }

    public struct HlmtPermutationInfo
    {
        public string Name;
        public HlmtPermutationPiece[] Piece;
    }

    public struct HlmtPermutationPiece
    {
        public string PieceName;
        public HlmtPermutationVariation[] Permutation;
    }

    public struct HlmtPermutationVariation
    {
        public string PermutationNameX;
    }

    public struct DisplayInfo
    {
        public List<int> Chunk;
        public int[] ShaderIndex;
        public VertexBuffer[] vertexBuffer;
        public IndexBuffer[] indexBuffer;
        public Mesh[] meshes;
    }

    public class FrameInfo
    {
        public string Name;
        public Matrix matrix;
        public Matrix skinweightmatrix;
        public int Parent = -1;
        public int Child = -1;
        public int Sibling = -1;
        public float x, y, z;
        public float i, j, k, w; // quaternion

        public float DistanceFromParent;
    }

    public struct HaloVertex
    {
        public Vector3 Position;
        public Vector3 Normal;

        public int diffuse;
        public int specular;

        public float Tu0, Tv0;
        // public float Tu1, Tv1;

        public static readonly VertexFormats FVF = VertexFormats.PositionNormal |
            VertexFormats.Diffuse | VertexFormats.Specular | VertexFormats.Texture1;
    }
    #endregion
}
