float4x4 wvp : WorldViewProjection;
float4x4 world : World;
float3 LightPosition : Direction
<
 string Object = "DirectionalLight";
    string Space = "World";
> = {1.0f, -1.0f, 1.0f};
float3 EyePosition:CAMERAPOSITION;

struct VS_OUTPUT
{
 float4 Pos : POSITION;
 float2 Tex : TEXCOORD0;
 float3 Light : TEXCOORD1;
 float3 View : TEXCOORD2;
};

VS_OUTPUT VS(float4 Pos : POSITION,float2 Tex : TEXCOORD,float3 Normal : NORMAL,float3 Tangent : TANGENT)
{
 VS_OUTPUT Out = (VS_OUTPUT)0;
  
 Out.Pos = mul(Pos,wvp);
 
 float3x3 worldToTangentSpace;
 worldToTangentSpace[0] = mul(Tangent,world);
 worldToTangentSpace[1] = mul(cross(Tangent,Normal),world);
 worldToTangentSpace[2] = mul(Normal,world);
 
 Out.Tex = Tex;
 
 float4 PosWorld = mul(Pos,world);
 
 Out.Light = mul(worldToTangentSpace,LightPosition);
 Out.View = mul(worldToTangentSpace,EyePosition - PosWorld);
 
 return Out;
}

texture ColorMap;
sampler ColorMapSampler = sampler_state
{
 Texture = <ColorMap>;
 MinFilter = Linear;
 MagFilter = Linear;
 MipFilter = Linear;
 AddressU = Clamp;
 AddressV = Clamp;
};

texture BumpMap;
sampler BumpMapSampler = sampler_state
{
 Texture = <BumpMap>;
 MinFilter = Linear;
 MagFilter = Linear;
 MipFilter = Linear;
 AddressU = Clamp;
 AddressV = Clamp;
};

float4 PS(float2 Tex : TEXCOORD0,float3 Light : TEXCOORD1,float3 View : TEXCOORD2) : COLOR
{
 float4 Color = tex2D(ColorMapSampler,Tex);
 float3 Normal = (2 * (tex2D(BumpMapSampler,Tex))) - 1.0;
 
 float3 LightDir = normalize(Light);
 float3 ViewDir = normalize(View);
 
 float Diffuse = saturate(dot(Normal,LightDir));
 float Reflect = normalize(2 * Diffuse * Normal - LightDir);
 
 float Specular = min(pow(saturate(dot(Reflect,ViewDir)),3),Color.w);
  
 return 0.2 * Color + Color * Diffuse + Specular;
}

technique BumpMapShader
{
    pass P0
    {
        Sampler[0] = (ColorMapSampler);  
        Sampler[1] = (BumpMapSampler);  
   
        VertexShader = compile vs_2_0 VS();
        PixelShader  = compile ps_2_0 PS();
    }
}