using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;
using System.IO;
using System.Drawing;
using System.Drawing.Imaging;

namespace Visual_Scenery_Mover
{
    public class ParsedBitmap
    {
        [DllImport("kernel32.dll")]
        static extern void RtlMoveMemory(IntPtr src, byte[] crap, int cb);
        public HaloBitmapHeader BitmapHeader;
        public int headersize = 0;
        public BitmapInfo[] Properties;

        public ParsedBitmap(ref Meta meta, int mapnum)
        {
            BitmapHeader = new HaloBitmapHeader(ref meta);

            BinaryReader BR = new BinaryReader(meta.MS);
            BR.BaseStream.Position = 68;
            int tempc = BR.ReadInt32();
            int tempr = BR.ReadInt32() - Map.mp[mapnum].index.secondaryMagic - meta.offset;
            headersize = tempr;
            Properties = new BitmapInfo[tempc];
            for (int x = 0; x < tempc; x++)
                Properties[x] = new BitmapInfo(tempr + (116 * x), ref meta);
        }

        public class HaloBitmapHeader
        {
            public short type;
            public short format;
            public short usage;
            public short usageflags;
            public float detailfadefactor;
            public float sharpenamount;
            public float bumpheight;
            public short spritebudgesize;
            public short spritebudgetcount;
            public short colorplatewidth;
            public short colorplateheight;
            public short compressedcolorplatedata;

            public HaloBitmapHeader(ref Meta meta)
            {
                BinaryReader BR = new BinaryReader(meta.MS);
                BR.BaseStream.Position = 0;
                type = BR.ReadInt16();
                format = BR.ReadInt16();
                usage = BR.ReadInt16();
                usageflags = BR.ReadInt16();
                detailfadefactor = BR.ReadSingle();
                sharpenamount = BR.ReadSingle();
                bumpheight = BR.ReadSingle();
                spritebudgesize = BR.ReadInt16();
                spritebudgetcount = BR.ReadInt16();
                colorplatewidth = BR.ReadInt16();
                colorplateheight = BR.ReadInt16();
                compressedcolorplatedata = BR.ReadInt16();
            }
            public void WriteHeaderHaloBitmapHeader(ref BinaryWriter BW)
            {
                BW.BaseStream.Position = 0;
                BW.Write(type);
                BW.Write(format);
                BW.Write(usage);
                BW.Write(usageflags);
                BW.Write(detailfadefactor);
                BW.Write(sharpenamount);
                BW.Write(bumpheight);
                BW.Write(spritebudgesize);
                BW.Write(spritebudgetcount);
                BW.Write(colorplatewidth);
                BW.Write(colorplateheight);
                BW.Write(compressedcolorplatedata);
            }
        }

        public class BitmapInfo
        {
            public char[] tagtype;
            public ushort width;
            public ushort height;
            public ushort depth;
            public ushort type;
            public ushort format;
            public BitmapType formatname;
            public bool swizzle;
            public ushort flags;
            public ushort regPointX;
            public ushort regPointY;
            public ushort mipMapCount;
            public ushort pixelOffset;

            public BitmapInfo(int offset, ref Meta meta)
            {
                BinaryReader BR = new BinaryReader(meta.MS);
                BR.BaseStream.Position = offset;
                tagtype = BR.ReadChars(4);
                width = BR.ReadUInt16();
                height = BR.ReadUInt16();
                depth = BR.ReadUInt16();
                type = BR.ReadUInt16();
                formatname = (BitmapType)BR.ReadUInt16();
                flags = BR.ReadUInt16();
                // if ((flags & 0x1000) == 0x1000) { swizzle = true; }
                if ((flags & 0x8) == 0x8) swizzle = true;
                //MessageBox.Show("test");
                regPointX = BR.ReadUInt16();
                regPointY = BR.ReadUInt16();
                mipMapCount = BR.ReadUInt16();
                pixelOffset = BR.ReadUInt16();
            }

            public void Write(ref BinaryWriter BW)
            {
                BW.Write(tagtype);
                BW.Write(width);
                BW.Write(height);
                BW.Write(depth);
                BW.Write(type);
                BW.Write((short)formatname);
                BW.Write(flags);

                BW.Write(regPointX);
                BW.Write(regPointY);
                BW.Write(mipMapCount);
                BW.Write(pixelOffset);
            }

            public void Generate(DDS.DDSURFACEDESC2 ddsd)
            {
                string bitmstring = "mtib";
                tagtype = bitmstring.ToCharArray();
                // width = ddsd.width; ;
                //   height = ddsd.height; ;
                //depth=ddsd.depth;
                /*
         type=ddsd.;
         format;
         formatname;
         swizzle;
         flags;
         regPointX;
         regPointY;
         mipMapCount=ddsd.MipMapCount;
         pixelOffset=d; 
              * */
            }
        }

        public enum BitmapType : int
        {
            BITM_FORMAT_A8 = 0,
            BITM_FORMAT_Y8 = 1,
            BITM_FORMAT_AY8 = 2,
            BITM_FORMAT_A8Y8 = 3,
            BITM_FORMAT_R5G6B5 = 6,
            BITM_FORMAT_A1R5G5B5 = 8,
            BITM_FORMAT_A4R4G4B4 = 9,
            BITM_FORMAT_X8R8G8B8 = 10,
            BITM_FORMAT_A8R8G8B8 = 11,
            BITM_FORMAT_DXT1 = 14,
            BITM_FORMAT_DXT2AND3 = 15,
            BITM_FORMAT_DXT4AND5 = 16,
            BITM_FORMAT_P8 = 17,
            BITM_TYPE_2D = 0,
            BITM_TYPE_3D = 1,
            BITM_TYPE_CUBEMAP = 2,
            BITM_TYPE_LIGHTMAP = 18,
            BITM_TYPE_UNK = 22
        }

        //visualchunkindex is for lightmaps only
        public Bitmap FindChunkAndDecode(int chunknumber, int numberinchunk, ref Meta meta,
            int visualchunkindex, int bspnumber, int mapnum)
        {
            for (int x = 0; x < meta.raw.rawChunks.Count; x++)
            {
                BitmapRawDataChunk tempb = (BitmapRawDataChunk)meta.raw.rawChunks[x];
                if (tempb.inchunk == chunknumber && tempb.num == numberinchunk)
                {
                    int height = this.Properties[chunknumber].height;
                    int depth = this.Properties[chunknumber].depth;
                    int width = this.Properties[chunknumber].width;
                    int pixeloff = this.Properties[chunknumber].pixelOffset;
                    int tempsize = tempb.size;//width * height * 4;
                    byte[] guh = new byte[tempsize];//];

                    Array.Copy(tempb.MS.ToArray(), 0, guh, 0, tempsize);

                    Bitmap b = DecodeBitm(guh, height, width, depth, pixeloff,
                        this.Properties[chunknumber].formatname,
                        this.Properties[chunknumber].swizzle, visualchunkindex,
                        bspnumber, mapnum);
                    return b;
                }
            }
            return null;
        }

        public Bitmap DecodeBitm(byte[] fart, int height, int width, int depth, int pixeloffset,
            BitmapType format, bool swizzle, int visualchunkindex, int ident, int mapnum)
        {

            byte[] poo = new byte[0];
            IntPtr ptr = new IntPtr();

            PixelFormat o = new PixelFormat();
            int poolength;
            byte[] fu;
            DecodeDXT decode = new DecodeDXT();


            int oldwidth = width;
            if (width % 8 != 0) width = (ushort)(width + (8 - (width % 8)));
            int stride = width;
            switch (format)
            {
                case (BitmapType)14:
                    fart = decode.DecodeDXT1(height, width, fart);
                    stride *= 4;
                    o = PixelFormat.Format32bppArgb;
                    break;

                case (BitmapType)15:
                    fart = decode.DecodeDXT23(height, width, fart);
                    stride *= 4;
                    o = PixelFormat.Format32bppArgb;
                    break;

                case (BitmapType)16:
                    fart = decode.DecodeDXT45(height, width, fart);
                    stride *= 4;
                    o = PixelFormat.Format32bppArgb;
                    break;

                case BitmapType.BITM_FORMAT_A8R8G8B8:

                    // fart = Swizzler.Swizzle(fart, width, height, -1, 32, true);
                    if (swizzle == true)
                    {
                        fart = Swizzler.Swizzle(fart, width, height, depth, 32, true);
                    }

                    stride *= 4;
                    o = PixelFormat.Format32bppArgb;
                    break;

                case BitmapType.BITM_FORMAT_X8R8G8B8:
                    if (swizzle == true)
                    {
                        fart = Swizzler.Swizzle(fart, width, height, -1, 32, true);
                    }

                    stride *= 4;
                    o = PixelFormat.Format32bppArgb;
                    break;


                //16 bit

                case BitmapType.BITM_FORMAT_A4R4G4B4:
                    if (swizzle == true)
                    {
                        fart = Swizzler.Swizzle(fart, pixeloffset, width, height, -1, 16, true);
                    }
                    stride *= 2;
                    o = PixelFormat.Format16bppRgb565;
                    break;
                case (BitmapType)22:
                    if (swizzle == true)
                    {
                        fart = Swizzler.Swizzle(fart, 0, width, height, -1, 16, true);
                    }
                    stride *= 2;
                    o = PixelFormat.Format32bppArgb;
                    poolength = fart.Length;
                    fu = new byte[poolength / 2 * 4];
                    for (int e = 0; e < poolength / 2; e++)
                    {
                        int r = e * 2;
                        fu[r * 2 + 0] = fart[r + 1];
                        fu[r * 2 + 1] = fart[r + 1];
                        fu[r * 2 + 2] = fart[r + 1];
                        fu[r * 2 + 3] = fart[r + 0];
                    }
                    fart = fu;
                    break;

                case BitmapType.BITM_FORMAT_A1R5G5B5:
                    if (swizzle == true)
                    {
                        fart = Swizzler.Swizzle(fart, pixeloffset, width, height, -1, 16, true);
                    }
                    stride *= 2;
                    o = PixelFormat.Format16bppRgb555;
                    break;
                case BitmapType.BITM_FORMAT_R5G6B5:
                    if (swizzle == true)
                    {
                        fart = Swizzler.Swizzle(fart, pixeloffset, width, height, -1, 16, true);
                    }
                    stride *= 2;
                    o = PixelFormat.Format16bppRgb565;
                    break;
                case BitmapType.BITM_FORMAT_A8Y8:
                    if (swizzle == true)
                    {
                        fart = Swizzler.Swizzle(fart, pixeloffset, width, height, depth, 16, true);
                    }
                    o = PixelFormat.Format32bppArgb;
                    poolength = fart.Length;
                    fu = new byte[poolength / 2 * 4];
                    for (int e = 0; e < poolength / 2; e++)
                    {
                        int r = e * 2;
                        fu[r + 0] = fart[r + 1];
                        fu[r + 1] = fart[r + 1];
                        fu[r + 2] = fart[r + 1];
                        fu[r + 3] = fart[r + 0];
                    }
                    fart = fu;
                    stride *= 4;
                    break;
                //8 bit
                case BitmapType.BITM_FORMAT_P8:
                    if (swizzle == true)
                    {
                        fart = Swizzler.Swizzle(fart, pixeloffset, width, height, depth, 8, true);
                    }
                    o = PixelFormat.Format32bppArgb;
                    poolength = fart.Length;
                    fu = new byte[poolength * 4];
                    for (int e = 0; e < poolength; e++)
                    {
                        int r = e * 4;
                        fu[r + 0] = fart[e];
                        fu[r + 1] = fart[e];
                        fu[r + 2] = fart[e];
                        fu[r + 3] = 255;// 255;
                        if (fart[e] == 0) { fu[r + 3] = 0; }
                    }
                    fart = fu;
                    stride *= 4;
                    break;
                case BitmapType.BITM_TYPE_LIGHTMAP:
                    if (swizzle == true)
                    {
                        fart = Swizzler.Swizzle(fart, pixeloffset, width, height, depth, 8, true);
                    }
                    o = PixelFormat.Format32bppArgb;// Format32bppArgb;
                    poolength = fart.Length;
                    fu = new byte[poolength * 4];
                    int bspnumber = ident;
                    int paletteindex;
                    if (visualchunkindex < 0)
                    {
                        int wtf = 0 - (visualchunkindex + 1);
                        paletteindex = Map.mp[mapnum].bsp.sbsp[bspnumber].SceneryChunk_LightMap_Index[wtf];
                    }
                    else
                    {

                        paletteindex = Map.mp[mapnum].bsp.sbsp[bspnumber].VisualChunk_LightMap_Index[visualchunkindex];
                    }
                    if (paletteindex == 255) return null;
                    for (int e = 0; e < poolength; e++)
                    {
                        int r = e * 4;
                        fu[r + 0] = (byte)Map.mp[mapnum].bsp.sbsp[bspnumber].LightMap_Palettes[paletteindex][fart[e]].r;
                        fu[r + 1] = (byte)Map.mp[mapnum].bsp.sbsp[bspnumber].LightMap_Palettes[paletteindex][fart[e]].g;
                        fu[r + 2] = (byte)Map.mp[mapnum].bsp.sbsp[bspnumber].LightMap_Palettes[paletteindex][fart[e]].b;
                        fu[r + 3] = (byte)Map.mp[mapnum].bsp.sbsp[bspnumber].LightMap_Palettes[paletteindex][fart[e]].a;

                    }
                    fart = fu;
                    stride *= 4;
                    break;

                case (BitmapType)1:
                    if (swizzle == true)
                    {
                        fart = Swizzler.Swizzle(fart, pixeloffset, width, height, depth, 8, true);
                    }
                    o = PixelFormat.Format32bppArgb;
                    poolength = fart.Length;
                    fu = new byte[poolength * 4];
                    for (int e = 0; e < poolength; e++)
                    {
                        int r = e * 4;
                        fu[r + 0] = fart[e];
                        fu[r + 1] = fart[e];
                        fu[r + 2] = fart[e];
                        fu[r + 3] = fart[255];
                    }
                    fart = fu;
                    stride *= 4;
                    break;


                default:
                    return null;
            }

            Marshal.FreeHGlobal(ptr);
            ptr = Marshal.AllocHGlobal(fart.Length);

            RtlMoveMemory(ptr, fart, fart.Length);

            Bitmap temp = new System.Drawing.Bitmap(width, height, stride, o, ptr);
            return temp;
        }
    }
}
