using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using DI = Microsoft.DirectX.DirectInput;
using System.IO;

namespace Visual_Scenery_Mover
{
    public partial class Form1 : Form
    {
        Map.MapInfo.DXRender dxr = Map.MapInfo.DXRender.Triangle; //we start with the triangle

        bool init = false;
        int sb = SystemInformation.VerticalScrollBarWidth,
            mapnum;

        //our keyboard Map.dev to see which buttons are being pressed
        DI.Device keyb;

        TriangleRender triangle = null;

        string type, tag;

        public Form1()
        {
            InitializeComponent();
        }

        private void openMapToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (o.ShowDialog() == DialogResult.Cancel) return;

            foreach (string str in o.FileNames)
            {
                if (Map.AddMap(str, stat, pb1))
                {
                    tv1.Nodes.Add(Map.mp[Map.mapCount-1].tags.LoadBSPTagsIntoTreenodes(false, Map.mapCount-1));
                    this.Text = "Visual Scenery Mover - [" + str + "]";
                }
                //else
                //{
                //    MessageBox.Show("Unable to load map '" + o.FileName + "'. Sorry for the inconvenience.",
                //        "", MessageBoxButtons.OK, MessageBoxIcon.Information);
                //}
            }
            stat.Text = "Maps Loaded";
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            Pref.LoadPreferences();
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        private void preferencesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            PreferencesForm pf = new PreferencesForm();
            pf.ShowDialog();
        }

        public bool InitializeGraphics()
        {
            try
            {
                // Get the ordinal for the default adapter
                int adapterOrdinal = Manager.Adapters.Default.Adapter;

                // Get our Map.dev capabilities so we can check them to set up the CreateFlags
                Caps caps = Manager.GetDeviceCaps(adapterOrdinal, DeviceType.Hardware);
                CreateFlags createFlags;

                // Check the capabilities of the graphcis card is capable of
                // performing the vertex-processing operations
                // The HardwareVertexProcessing choice is the best
                if (caps.DeviceCaps.SupportsHardwareTransformAndLight) createFlags = CreateFlags.HardwareVertexProcessing;
                else createFlags = CreateFlags.SoftwareVertexProcessing;

                // If the graphics card supports vertex processing check if the Map.dev can
                // do rasterization, matrix transformations, and lighting and shading operations
                // This combination provides the fastest game experience
                if (caps.DeviceCaps.SupportsPureDevice && createFlags == CreateFlags.HardwareVertexProcessing)
                    createFlags |= CreateFlags.PureDevice;

                // Set up the PresentParameters which determine how the Map.dev behaves
                PresentParameters pp = new PresentParameters();
                pp.Windowed = true;
                pp.SwapEffect = SwapEffect.Discard;
                pp.EnableAutoDepthStencil = true;
                pp.AutoDepthStencilFormat = DepthFormat.D16;
                pp.PresentationInterval = PresentInterval.One;

                // Now create the Map.dev
                Map.dev = new Device(adapterOrdinal, DeviceType.Hardware, vwr, createFlags, pp);
                Map.dev.DeviceReset += new EventHandler(OnDeviceReset);
                OnDeviceReset(null, null);

                init = true;
                return true;
            }
            catch
            {
                init = false;
                tv1.Nodes.Add("Failed to Initialize Graphics");
                return false;
            }
        }

        public void OnDeviceReset(object sender, EventArgs e)
        {
            Map.dev.RenderState.CullMode = Cull.None;

            //we only need lighting if we're not rendering the triangle
            Map.dev.RenderState.Lighting = (dxr != Map.MapInfo.DXRender.Triangle);

            if (Map.dev.DeviceCaps.SourceBlendCaps.SupportsDestinationAlpha &&
                Map.dev.DeviceCaps.SourceBlendCaps.SupportsSourceAlphaSat)
            {
                Map.dev.RenderState.SourceBlend = Blend.DestinationAlpha;
                Map.dev.RenderState.DestinationBlend = Blend.SourceAlphaSat;
            }

            if (Map.dev.DeviceCaps.TextureFilterCaps.SupportsMinifyAnisotropic)
            {
                for (int x = 0; x < 4; x++)
                    Map.dev.SamplerState[x].MinFilter = TextureFilter.Anisotropic;
            }
            if (Map.dev.DeviceCaps.TextureFilterCaps.SupportsMagnifyAnisotropic)
            {
                for (int x = 0; x < 4; x++)
                    Map.dev.SamplerState[x].MagFilter = TextureFilter.Anisotropic;
            }
            if (Map.dev.DeviceCaps.TextureFilterCaps.SupportsMipMapLinear)
            {
                for (int x = 0; x < 4; x++)
                    Map.dev.SamplerState[x].MipFilter = TextureFilter.Linear;
            }

            Map.dev.RenderState.Ambient = Color.FromArgb(0xAF, 0xAF, 0xAF);

            //SetUpFont();
            SetUpKeyboard();

            //if (dxr == DXRender.Triangle)
            //    LoadTriangle();

            //else if (dxr == DXRender.Model)
            //    SetUpCameraForModel();

            //else if (dxr == DXRender.PRTM)
            //    SetUpCameraForPRTM();

            //else if (dxr == DXRender.BSP)
            //SetUpCameraForBSP();
        }

        private void SetUpKeyboard()
        {
            keyb = new DI.Device(DI.SystemGuid.Keyboard);
            keyb.SetCooperativeLevel(this, DI.CooperativeLevelFlags.Foreground |
                DI.CooperativeLevelFlags.Exclusive);
        }

        public void LoadTriangle()
        {
            dxr = Map.MapInfo.DXRender.None;

            triangle = new TriangleRender();
            triangle.LoadTriangle();

            dxr = Map.MapInfo.DXRender.Triangle;
            AddRemoveTabs();
        }

        public void AddRemoveTabs()
        {
            if (dxr == Map.MapInfo.DXRender.BSP)
            {
                if (!dc1.Controls.Contains(bspEditCont))
                    dc1.AddDocument(bspEditCont);
            }
            else
            {
                if (dc1.Controls.Contains(bspEditCont))
                    dc1.RemoveDocument(bspEditCont);
            }
        }

        public void update()
        {
            if (dxr == Map.MapInfo.DXRender.Triangle)
                triangle.UpdateTriangleMatrices();

            else if (dxr == Map.MapInfo.DXRender.BSP)
            {

                Map.mp[mapnum].bsprend.update(ref keyb);

                if (textBox4.Text != Map.mp[mapnum].bsprend.cam.x.ToString())
                    textBox4.Text = Map.mp[mapnum].bsprend.cam.x.ToString();

                if (textBox6.Text != Map.mp[mapnum].bsprend.cam.y.ToString())
                    textBox6.Text = Map.mp[mapnum].bsprend.cam.y.ToString();

                if (textBox5.Text != Map.mp[mapnum].bsprend.cam.z.ToString())
                    textBox5.Text = Map.mp[mapnum].bsprend.cam.z.ToString();
            }
        }

        public void Render()
        {
            if (init && !Pref.paused && dxr != Map.MapInfo.DXRender.None)
            {
                update();

                Map.dev.Clear(ClearFlags.Target | ClearFlags.ZBuffer, Color.CornflowerBlue, 1.0f, 0);
                Map.dev.BeginScene();

                if (dxr == Map.MapInfo.DXRender.Triangle)
                    triangle.Draw();

                else if (dxr == Map.MapInfo.DXRender.BSP)
                {
                    Map.mp[mapnum].bsprend.Draw(mapnum);
                }

                Map.dev.EndScene();
                Map.dev.Present();
            }
        }

        private void vwr_MouseDown(object sender, MouseEventArgs e)
        {
            if (dxr == Map.MapInfo.DXRender.BSP)
            {
                vwr.Focus();
                Map.mp[mapnum].bsprend.MouseDown(e);
            }
        }

        private void vwr_MouseMove(object sender, MouseEventArgs e)
        {
            if (dxr == Map.MapInfo.DXRender.BSP)
            {
                vwr.Focus();
                Map.mp[mapnum].bsprend.MouseMove(e);
            }
        }

        private void vwr_MouseUp(object sender, MouseEventArgs e)
        {
            if (dxr == Map.MapInfo.DXRender.BSP)
            {
                Map.mp[mapnum].bsprend.MouseUp(e);
            }
        }

        private void tv1_AfterSelect(object sender, TreeViewEventArgs e)
        {
            if (init && e.Node != null)
            {
                if (e.Node.Tag.ToString().Substring(0, 6) == "#file#")
                {
                    this.Text = "Visual_Scenery_Mover - [" + e.Node.Tag.ToString().Substring(6) + "]";
                    dxr = Map.mp[e.Node.Index].dxr;
                    mapnum = e.Node.Index;
                }

                else if (e.Node.Tag.ToString().Substring(0, 6) == "#tagp#")
                {
                    this.Text = "Visual_Scenery_Mover - [" + e.Node.Parent.Tag.ToString().Substring(6) + "]";
                    mapnum = e.Node.Parent.Index;

                    type = e.Node.Parent.Text;
                    tag = e.Node.Text;

                    LoadMeta(Map.mp[mapnum].tags.FindByTagAndTagType("sbsp", tag), mapnum);
                    LoadBSP(mapnum);
                    dxr = Map.mp[e.Node.Index].dxr;
                }
            }
        }
        
        public void LoadMeta(int tagnum, int mapnum)
        {
            Map.mp[mapnum].meta = new Meta(tagnum);
            Map.mp[mapnum].meta.ReadMetaFromMap(false, true, false, mapnum);
            Map.mp[mapnum].meta.SortItemsByOffset();
        }

        public void LoadBSP(int mapnum)
        {
            dxr = Map.MapInfo.DXRender.None;
            stat.Text = "Loading " + tag + "...";
            Application.DoEvents();

            if (Map.mp[mapnum].bsprend != null && Map.mp[mapnum].bsprend.bspLoaded)
            {
                Map.mp[mapnum].bsprend.PopLTMP(mapnum, panel1, tabPage3, img_Click);
                Map.mp[mapnum].bsprend.PopSpawnList(true, true, tv2, clb);
            }
            else
            {
                Map.mp[mapnum].bsprend = new BSPRender();
                Map.mp[mapnum].bsprend.LoadBSP(pb1, panel1, tabPage3, tv2, clb, img_Click, mapnum);
            }

            numUD1.Value = (int)Map.mp[mapnum].bsprend.cam.speed * 100;

            //if lightmaps enabled..............
            numUD4.Maximum = Map.mp[mapnum].bsp.sbsp[Map.mp[mapnum].bsprend.bspmodel.BspNumber].LightMap_Palettes.Count - 1;

            dxr = Map.MapInfo.DXRender.BSP;
            Map.mp[mapnum].dxr = Map.MapInfo.DXRender.BSP;
            AddRemoveTabs();
            if (Pref.activateTagsTabOnSelection) dc1.ActiveDocument = bspEditCont;
            stat.Text = tag + " loaded";
        }

        private void button1_Click(object sender, EventArgs e)
        {
            for (int x = 0; x < clb.Items.Count; x++)
                clb.SetItemChecked(x, true);
        }

        private void button2_Click(object sender, EventArgs e)
        {
            for (int x = 0; x < clb.Items.Count; x++)
                clb.SetItemChecked(x, false);
        }

        private void clb_ItemCheck(object sender, ItemCheckEventArgs e)
        {
            string name = clb.Items[e.Index].ToString();

            Color clr = Color.Black;
            if (e.NewValue == CheckState.Checked) clr = Color.Green;

            for (int x = 0; x < tv2.Nodes.Count; x++)
            {
                if (tv2.Nodes[x].Text == name)
                {
                    tv2.Nodes[x].ForeColor = clr;
                    if (tv2.Nodes[x].Nodes.Count > 0)
                    {
                        for (int y = 0; y < tv2.Nodes[x].Nodes.Count; y++)
                            tv2.Nodes[x].Nodes[y].ForeColor = clr;
                    }
                    break;
                }
            }

            Type typ = typeof(SpawnInfo.SpawnType);
            SpawnInfo.SpawnType bm = (SpawnInfo.SpawnType)Enum.Parse(typ, name, true);

            Map.mp[mapnum].bsprend.visibleSpawnsBitMask ^= (int)bm;
        }

        private void tv2_DoubleClick(object sender, EventArgs e)
        {
            if (tv2.SelectedNode.Nodes.Count == 0)
            {
                int tempint = (int)tv2.SelectedNode.Tag;
                if (tempint >= 0)
                {
                    Map.mp[mapnum].bsprend.cam.radianv = 0;
                    Map.mp[mapnum].bsprend.cam.radianh = 0;

                    Map.mp[mapnum].bsprend.cam.x = Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn[tempint].X;
                    Map.mp[mapnum].bsprend.cam.y = Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn[tempint].Y;
                    Map.mp[mapnum].bsprend.cam.z = Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn[tempint].Z;

                    Map.mp[mapnum].bsprend.cam.Position = new Vector3(Map.mp[mapnum].bsprend.cam.x, Map.mp[mapnum].bsprend.cam.y, Map.mp[mapnum].bsprend.cam.z);

                    string checkeditemsname = Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn[tempint].Type.ToString();
                    int checkitemsindex = clb.Items.IndexOf(checkeditemsname);

                    Type typ = typeof(SpawnInfo.SpawnType);
                    SpawnInfo.SpawnType bm = (SpawnInfo.SpawnType)Enum.Parse(typ, checkeditemsname, true);

                    if (clb.GetItemCheckState(checkitemsindex) == CheckState.Unchecked)
                        clb.SetItemCheckState(checkitemsindex, CheckState.Checked);
                }
            }
        }

        #region Lightmaps
        private void button4_Click(object sender, EventArgs e) //reset
        {
            Array.ConstrainedCopy(Map.mp[mapnum].bsprend.ltmp.LightMap_Array_Backup, 0, Map.mp[mapnum].bsprend.ltmp.LightMap_Array,
                0, Map.mp[mapnum].bsprend.ltmp.LightMap_Array.Length);
            trackint1.Value = trackint2.Value = trackint3.Value = 0;

            foreach (Control box in panel1.Controls)
                ((PictureBox)box).Image = Map.mp[mapnum].bsprend.ltmp.RenderLightmap((int)box.Tag);
        }

        private void button3_Click(object sender, EventArgs e) //preview
        {
            Map.mp[mapnum].bsprend.ltmp.ReloadFromArray(Map.mp[mapnum].bsprend.bspmodel, mapnum);
            Map.mp[mapnum].bsprend.bspmodel.LoadLightmaps(mapnum);
            BSPModel.BSPDisplayedInfo.LoadLightmapTextures(ref Map.mp[mapnum].bsprend.bspmodel);
        }

        private void button5_Click(object sender, EventArgs e) //apply
        {
            //Apply the new palettes
            Array.ConstrainedCopy(Map.mp[mapnum].bsprend.ltmp.LightMap_Array, 0, Map.mp[mapnum].bsprend.ltmp.LightMap_Array_Backup,
                0, Map.mp[mapnum].bsprend.ltmp.LightMap_Array.Length);

            Map.mp[mapnum].bsprend.ltmp.ReloadFromArray(Map.mp[mapnum].bsprend.bspmodel, mapnum);
            Map.mp[mapnum].bsprend.bspmodel.LoadLightmaps(mapnum);
            BSPModel.BSPDisplayedInfo.LoadLightmapTextures(ref Map.mp[mapnum].bsprend.bspmodel);

            //Save the lightmap
            Map.mp[mapnum].bsp.sbsp[Map.mp[mapnum].bsprend.bspmodel.BspNumber].WritePalettes(Map.mp[mapnum].mapLoc);
            trackint1.Value = trackint2.Value = trackint3.Value = 0;
        }

        public void img_Click(object sender, EventArgs e) //pic clicked
        {
            PictureBox pb = (PictureBox)sender;
            pb.BorderStyle = BorderStyle.FixedSingle;
            if (Map.mp[mapnum].bsprend.ltmp.spb != null) Map.mp[mapnum].bsprend.ltmp.spb.BorderStyle = BorderStyle.None;

            numUD4.Value = (int)pb.Tag;
            radioButton4.Checked = true;
            Map.mp[mapnum].bsprend.ltmp.spb = pb;
        }

        private void trackint123Scroll(object sender, EventArgs e) //3trackbar scroll
        {
            Map.mp[mapnum].bsprend.ltmp.EditLightmaps(ref panel1, trackint1.Value, trackint2.Value, trackint3.Value,
                 checkBox1.Checked, Convert.ToInt32(numUD4.Value), radioButton1.Checked,
                 radioButton3.Checked, Map.mp[mapnum].bsprend.bspmodel, mapnum);
        }

        private void radioButton1_CheckedChanged(object sender, EventArgs e) //hue|saturation
        {
            Array.ConstrainedCopy(Map.mp[mapnum].bsprend.ltmp.LightMap_Array_Backup, 0, Map.mp[mapnum].bsprend.ltmp.LightMap_Array,
                0, Map.mp[mapnum].bsprend.ltmp.LightMap_Array.Length);

            trackint1.Value = trackint2.Value = trackint3.Value = 0;

            if (radioButton1.Checked) //hue|saturation
            {
                trackint1.Maximum = trackint2.Maximum = trackint3.Maximum = 180;
                trackint1.Minimum = trackint2.Minimum = trackint3.Minimum = -180;

                checkBox1.Text = "Colourize";
                toolTip1.SetToolTip(trackint1, "Hue");
                toolTip1.SetToolTip(trackint2, "Saturation");
                toolTip1.SetToolTip(trackint3, "Lightness");
            }
            else //colour balance
            {
                trackint1.Maximum = trackint2.Maximum = trackint3.Maximum = 100;
                trackint1.Minimum = trackint2.Minimum = trackint3.Minimum = -100;

                checkBox1.Text = "Preserve Lightness";
                toolTip1.SetToolTip(trackint1, "Cyan :: Red");
                toolTip1.SetToolTip(trackint2, "Magenta :: Green");
                toolTip1.SetToolTip(trackint3, "Yellow :: Blue");
            }

            foreach (Control box in panel1.Controls)
                ((PictureBox)box).Image = Map.mp[mapnum].bsprend.ltmp.RenderLightmap((int)box.Tag);
        }

        private void checkBox1_CheckedChanged(object sender, EventArgs e) //colourise/preserve
        {
            Map.mp[mapnum].bsprend.ltmp.EditLightmaps(ref panel1, trackint1.Value, trackint2.Value, trackint3.Value,
                checkBox1.Checked, Convert.ToInt32(numUD4.Value), radioButton1.Checked,
                radioButton3.Checked, Map.mp[mapnum].bsprend.bspmodel, mapnum);
        }

        private void radioButton4_CheckedChanged(object sender, EventArgs e) //edt specifc chunk
        {
            if (radioButton4.Checked)
            {
                numUD4.Enabled = true;
                Map.mp[mapnum].bsprend.ltmp.spb = (PictureBox)panel1.Controls[Convert.ToInt32(numUD4.Value)];
                Map.mp[mapnum].bsprend.ltmp.spb.BorderStyle = BorderStyle.FixedSingle;
            }
            else
            {
                numUD4.Enabled = false;
                if (Map.mp[mapnum].bsprend.ltmp.spb != null) Map.mp[mapnum].bsprend.ltmp.spb.BorderStyle = BorderStyle.None;
            }
        }

        private void numUD4_ValueChanged(object sender, EventArgs e) //chunk number
        {
            img_Click((PictureBox)panel1.Controls[
               Convert.ToInt32(numUD4.Value)], null);

            panel1.ScrollControlIntoView(panel1.Controls[
                Convert.ToInt32(numUD4.Value)]);
        }

        private void panel1_Resize(object sender, EventArgs e) //panel1 resize
        {
            if (Map.mp[mapnum].bsprend.bspmodel.renderLightmap)
                Map.mp[mapnum].bsprend.ltmp.RepositionLightmapImgs(ref panel1);
        }
        #endregion

        private void button10_Click(object sender, EventArgs e)
        {
            //clear the list of selected spawns
            Map.mp[mapnum].bsprend.SelectedSpawn.Clear();

            //open the map up
            BinaryWriter bw = new BinaryWriter(new FileStream(Map.mp[mapnum].mapLoc,
                    FileMode.Open, FileAccess.Write, FileShare.ReadWrite));

            //go through all the spawns
            for (int i = 0; i < Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn.Count; i++)
            {
                //if its a deathzone continue on to the next loop
                if (Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn[i].Type == SpawnInfo.SpawnType.DeathZone)
                    continue;

                //goto the spawns offset
                bw.BaseStream.Position = Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn[i].offset;
                //and write in the xyz coords
                bw.Write(Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn[i].X);
                bw.Write(Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn[i].Y);
                bw.Write(Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn[i].Z);

                //if the spawns rotation is just direction
                if (Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn[i] is SpawnInfo.RotateDirectionBaseSpawn)
                {
                    //then write in the direction
                    SpawnInfo.RotateDirectionBaseSpawn spwn = Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn[i] as SpawnInfo.RotateDirectionBaseSpawn;
                    bw.Write(spwn.RotationDirection);
                }

                //else if it uses yaw pitch roll
                else if (Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn[i] is SpawnInfo.RotateYawPitchRollBaseSpawn)
                {
                    //then write in the yaw pitch roll
                    SpawnInfo.RotateYawPitchRollBaseSpawn spwn = Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn[i] as SpawnInfo.RotateYawPitchRollBaseSpawn;
                    bw.Write(spwn.Roll);
                    if (spwn.isWeird == true) spwn.Pitch = -spwn.Pitch;
                    bw.Write(spwn.Pitch);
                    bw.Write(spwn.Yaw);
                }

                //if its an objective spawn
                if (Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn[i] is SpawnInfo.ObjectiveSpawn)
                {
                    //then we write in some more info
                    SpawnInfo.ObjectiveSpawn spwn = Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn[i] as SpawnInfo.ObjectiveSpawn;
                    bw.BaseStream.Position = Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn[i].offset + 16;
                    bw.Write((short)spwn.ObjectiveType);
                    bw.Write((short)spwn.Team);
                    bw.Write((short)spwn.number);
                }

                //else if its a collection spawn
                else if (Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn[i] is SpawnInfo.Collection)
                {
                    //write in what gametype it appears in
                    SpawnInfo.Collection os = Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn[i] as SpawnInfo.Collection;
                    bw.BaseStream.Position = Map.mp[mapnum].bsprend.bspmodel.Spawns.Spawn[i].offset - 60;
                    bw.Write((int)os.SpawnsInMode);
                }
            }

            //save scenery
            bw.BaseStream.Position = Map.mp[mapnum].meta.offset + 336;

            for (int x = 0; x < Map.mp[mapnum].bsprend.bspmodel.PermutationInfo.Length; x++)
            {
                bw.BaseStream.Position = Map.mp[mapnum].meta.offset + Map.mp[mapnum].bsprend.bspmodel.PermutationInfo[x].startInMeta;
                bw.BaseStream.Position += 40;
                bw.Write(Map.mp[mapnum].bsprend.bspmodel.PermutationInfo[x].mat.M41); //x
                bw.Write(Map.mp[mapnum].bsprend.bspmodel.PermutationInfo[x].mat.M42);  //y
                bw.Write(Map.mp[mapnum].bsprend.bspmodel.PermutationInfo[x].mat.M43);   //z
            }

            for (int x = 0; x < Map.mp[mapnum].bsprend.bspmodel.BSPPermutationRawDataMetaChunks.Length; x++)
            {
                BSPModel.BSPPermutationRawDataMetaChunk tmp =
                    Map.mp[mapnum].bsprend.bspmodel.BSPPermutationRawDataMetaChunks[x];

                bw.BaseStream.Position = Map.mp[mapnum].meta.offset + tmp.BoundingBox.offset;
                bw.Write(tmp.BoundingBox.MinX);
                bw.Write(tmp.BoundingBox.MaxX);

                bw.Write(tmp.BoundingBox.MinY);
                bw.Write(tmp.BoundingBox.MaxY);

                bw.Write(tmp.BoundingBox.MinZ);
                bw.Write(tmp.BoundingBox.MaxZ);

                bw.Write(tmp.BoundingBox.MinU);
                bw.Write(tmp.BoundingBox.MaxU);

                bw.Write(tmp.BoundingBox.MinV);
                bw.Write(tmp.BoundingBox.MaxV);
            }

            bw.Close();
            resign(mapnum);
            MessageBox.Show("spawns saved and map resigned successfully!");
        }

        public void resign(int mapnum)
        {
            FileStream fs = new FileStream(Map.mp[mapnum].mapLoc, FileMode.Open, FileAccess.ReadWrite,
                FileShare.ReadWrite);

            BinaryReader br = new BinaryReader(fs);
            BinaryWriter bw = new BinaryWriter(fs);

            Int32 sig = 0;
            long length = br.BaseStream.Length - 4;

            br.BaseStream.Position = 2048;
            for (long i = 2048; i <= length; i += 4)
                sig ^= br.ReadInt32();

            br.BaseStream.Position = 720;
            bw.Write(sig);

            br.Close();
            bw.Close();

            Map.mp[mapnum].header.checksum = sig;
        }

        private void button6_Click(object sender, EventArgs e)
        {
            try
            {
                Map.mp[mapnum].bsprend.cam.radianh = Map.mp[mapnum].bsprend.cam.radianv = 0;

                Map.mp[mapnum].bsprend.cam.x = Convert.ToSingle(textBox1.Text);
                Map.mp[mapnum].bsprend.cam.y = Convert.ToSingle(textBox2.Text);
                Map.mp[mapnum].bsprend.cam.z = Convert.ToSingle(textBox3.Text);

                Map.mp[mapnum].bsprend.cam.Position = new Vector3(Map.mp[mapnum].bsprend.cam.x, Map.mp[mapnum].bsprend.cam.y, Map.mp[mapnum].bsprend.cam.z);
                Map.mp[mapnum].bsprend.cam.ComputePosition();
            }
            catch { }
        }

        private void button7_Click(object sender, EventArgs e)
        {
            ListViewItem a = new ListViewItem();
            a.Text = Map.mp[mapnum].bsprend.cam.x.ToString();
            a.Tag = Map.mp[mapnum].bsprend.cam.radianh.ToString() + '#' + Map.mp[mapnum].bsprend.cam.radianv.ToString() + '#' +
                Map.mp[mapnum].bsprend.cam.i.ToString() + '#' + Map.mp[mapnum].bsprend.cam.j.ToString() + '#' + Map.mp[mapnum].bsprend.cam.k.ToString();

            ListViewItem.ListViewSubItem b = new ListViewItem.ListViewSubItem();
            b.Text = Map.mp[mapnum].bsprend.cam.y.ToString();

            ListViewItem.ListViewSubItem c = new ListViewItem.ListViewSubItem();
            c.Text = Map.mp[mapnum].bsprend.cam.z.ToString();

            a.SubItems.Add(b);
            a.SubItems.Add(c);
            listView1.Items.Add(a);
        }

        private void listView1_DoubleClick(object sender, EventArgs e)
        {
            try
            {
                if (listView1.FocusedItem != null)
                {
                    Map.mp[mapnum].bsprend.cam.x = Convert.ToSingle(listView1.FocusedItem.Text);
                    Map.mp[mapnum].bsprend.cam.y = Convert.ToSingle(listView1.FocusedItem.SubItems[1].Text);
                    Map.mp[mapnum].bsprend.cam.z = Convert.ToSingle(listView1.FocusedItem.SubItems[2].Text);

                    string[] spl = listView1.FocusedItem.Tag.ToString().Split('#');

                    Map.mp[mapnum].bsprend.cam.radianh = Convert.ToSingle(spl[0]);
                    Map.mp[mapnum].bsprend.cam.radianv = Convert.ToSingle(spl[1]);
                    Map.mp[mapnum].bsprend.cam.i = Convert.ToSingle(spl[2]);
                    Map.mp[mapnum].bsprend.cam.j = Convert.ToSingle(spl[3]);
                    Map.mp[mapnum].bsprend.cam.k = Convert.ToSingle(spl[4]);

                    Map.mp[mapnum].bsprend.cam.Position = new Vector3(Map.mp[mapnum].bsprend.cam.x, Map.mp[mapnum].bsprend.cam.y, Map.mp[mapnum].bsprend.cam.z);
                    Map.mp[mapnum].bsprend.cam.ComputePosition();
                }
            }
            catch { }
        }

        private void listView1_Resize(object sender, EventArgs e)
        {
            int w = listView1.Width / 3, gap = sb / 3;
            listView1.Columns[0].Width =
            listView1.Columns[1].Width =
            listView1.Columns[2].Width = w - gap;
        }

        private void numUD1_Scroll(object sender, EventArgs e)
        {
            Map.mp[mapnum].bsprend.cam.speed = Convert.ToSingle(numUD1.Value) / 100;
        }
    }
}