using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Windows.Forms;

namespace Visual_Scenery_Mover
{
    public class BSPContainer
    {
        public struct Palette_Color
        {
            public int a, r, g, b;
        }
        public struct BSPInfo
        {
            public Header.MapTypes location;

            public Int32
            pointerOffset,
            offset,
            size,
            magic,
            ident,
            tagnumber,

            lightmapoffset,
            lightmaptagnumber,
            lightmapident,
            lightmapsize,

            LightMap_TagNumber,

            palettesoffset;
            public List<BSPContainer.Palette_Color[]> LightMap_Palettes;

            public int[]
            VisualChunk_LightMap_Index,
            VisualChunk_Bitmap_Index,
            SceneryChunk_LightMap_Index,
            SceneryChunk_Bitmap_Index;

            public void WritePalettes(string mapLoc)
            {
                BinaryWriter bw = new BinaryWriter(new FileStream(mapLoc,
                        FileMode.Open, FileAccess.Write, FileShare.Read));

                bw.BaseStream.Position = palettesoffset;
                foreach (BSPContainer.Palette_Color[] palette in LightMap_Palettes)
                {
                    for (int x = 0; x < 256; x++)
                    {
                        bw.Write((byte)palette[x].r);
                        bw.Write((byte)palette[x].g);
                        bw.Write((byte)palette[x].b);
                        bw.Write((byte)palette[x].a);
                    }
                }
                bw.Close();
            }
        }
        public BSPInfo[] sbsp;
        public Int32 bspcount;

        public BSPContainer(BinaryReader br, Int32 secondaryMagic, ref Tags.TagInfo[] tagInfo)
        {
            //goto scnr offset + 528
            br.BaseStream.Position = tagInfo[3].MetaOffset + 528;
            //set the bspcount
            bspcount = br.ReadInt32();
            //store this value
            int tempr = br.ReadInt32() - secondaryMagic;
            //make a new bspinfo array with the size being the bsp count
            sbsp = new BSPInfo[bspcount];
            //count through the bspcount
            for (int x = 0; x < bspcount; x++)
            {
                //make a new bspinfo
                sbsp[x] = new BSPInfo();
                sbsp[x].pointerOffset = tempr + (x * 68) - tagInfo[3].MetaOffset;

                br.BaseStream.Position = tempr + (x * 68);
                sbsp[x].offset = br.ReadInt32();
                ParsePointer(ref sbsp[x].offset, ref sbsp[x].location);

                sbsp[x].size = br.ReadInt32();
                sbsp[x].magic = br.ReadInt32() - sbsp[x].offset;

                br.BaseStream.Position = tempr + (x * 68) + 20;
                sbsp[x].ident = br.ReadInt32();
                sbsp[x].tagnumber = Meta.FindMetaByID(sbsp[x].ident, tagInfo);

                tagInfo[sbsp[x].tagnumber].MetaOffset = sbsp[x].offset;
                tagInfo[sbsp[x].tagnumber].MetaSize = sbsp[x].size;

                br.BaseStream.Position = tempr + (x * 68) + 28;
                sbsp[x].lightmapident = br.ReadInt32();
                sbsp[x].lightmaptagnumber = Meta.FindMetaByID(sbsp[x].lightmapident, tagInfo);
                if (sbsp[x].lightmaptagnumber == -1) continue;

                br.BaseStream.Position = sbsp[x].offset + 8;
                sbsp[x].lightmapoffset = br.ReadInt32();
                if (sbsp[x].lightmapoffset == 0)
                {
                    sbsp[x].lightmapident = -1;
                    sbsp[x].lightmaptagnumber = -1;
                    if (MessageBox.Show("There is no lightmap for this bsp and the scenario is currently linked to a broken ID.\n Would you like Entity to fix it?",
                        "", MessageBoxButtons.YesNo) == DialogResult.Yes)
                    {
                        BinaryWriter bw = new BinaryWriter(br.BaseStream);

                        bw.BaseStream.Position = tempr + (x * 68) + 28;
                        bw.Write(int.Parse("FFFFFFFF", System.Globalization.NumberStyles.HexNumber));

                        bw.Flush();
                    }
                    continue;
                }
                sbsp[x].lightmapoffset += -sbsp[x].magic;
                sbsp[x].lightmapsize = sbsp[x].size + sbsp[x].offset - sbsp[x].lightmapoffset;

                tagInfo[sbsp[x].lightmaptagnumber].MetaOffset = sbsp[x].lightmapoffset;
                tagInfo[sbsp[x].lightmaptagnumber].MetaSize = sbsp[x].lightmapsize;

                //light map bitmap
                br.BaseStream.Position = sbsp[x].lightmapoffset + 128;
                int tempc = br.ReadInt32();
                int temprx = br.ReadInt32() - sbsp[x].magic;
                br.BaseStream.Position = temprx + 28;
                sbsp[x].LightMap_TagNumber = Meta.FindMetaByID(br.ReadInt32(), tagInfo);

                //light map palettes
                br.BaseStream.Position = temprx + 8;
                int tempc2 = br.ReadInt32();
                int tempr2 = br.ReadInt32() - sbsp[x].magic;
                sbsp[x].palettesoffset = tempr2;
                sbsp[x].LightMap_Palettes = new List<BSPContainer.Palette_Color[]>();
                for (int y = 0; y < tempc2; y++)
                {
                    br.BaseStream.Position = tempr2 + (y * 1024);
                    Palette_Color[] pc = new Palette_Color[256];
                    for (int z = 0; z < 256; z++)
                    {
                        pc[z] = new Palette_Color();
                        pc[z].r = br.ReadByte();
                        pc[z].g = br.ReadByte();
                        pc[z].b = br.ReadByte();
                        pc[z].a = br.ReadByte();
                    }
                    sbsp[x].LightMap_Palettes.Add(pc);
                }

                br.BaseStream.Position = temprx + 40;
                tempc2 = br.ReadInt32();
                tempr2 = br.ReadInt32() - sbsp[x].magic;
                sbsp[x].VisualChunk_Bitmap_Index = new int[tempc2];
                sbsp[x].VisualChunk_LightMap_Index = new int[tempc2];

                if (tempc2 != 0)
                {
                    br.BaseStream.Position = tempr2;
                    for (int y = 0; y < tempc2; y++)
                    {
                        sbsp[x].VisualChunk_Bitmap_Index[y] = br.ReadInt16();
                        sbsp[x].VisualChunk_LightMap_Index[y] = br.ReadInt16();
                    }
                }

                br.BaseStream.Position = temprx + 72;
                tempc2 = br.ReadInt32();
                tempr2 = br.ReadInt32() - sbsp[x].magic;
                sbsp[x].SceneryChunk_Bitmap_Index = new int[tempc2];
                sbsp[x].SceneryChunk_LightMap_Index = new int[tempc2];

                if (tempc2 > 0)
                {
                    br.BaseStream.Position = tempr2;
                    for (int y = 0; y < tempc2; y++)
                    {
                        sbsp[x].SceneryChunk_Bitmap_Index[y] = br.ReadInt16();
                        sbsp[x].SceneryChunk_LightMap_Index[y] = br.ReadInt16();
                    }
                }
            }
        }

        public int FindBSPNumberByBSPID(int id)
        {
            for (int x = 0; x < sbsp.Length; x++)
            {
                if (sbsp[x].ident == id)
                    return x;
            }
            return -1;
        }

        public int FindBSPNumberByLightMapID(int id)
        {
            for (int x = 0; x < sbsp.Length; x++)
            {
                if (sbsp[x].lightmapident == id)
                    return x;
            }
            return -1;
        }

        public static void ParsePointer(ref int pointer, ref Header.MapTypes type)
        {
            long tempmap = pointer & 0XC0000000;
            pointer = pointer & (int)0X3FFFFFFF;

            switch (tempmap)
            {
                case 0:
                    type = Header.MapTypes.Internal;
                    break;

                case 0X80000000:
                    type = Header.MapTypes.MPShared;
                    break;

                case 0XC0000000:
                    type = Header.MapTypes.SPShared;
                    break;

                case 0X40000000:
                    type = Header.MapTypes.MainMenu;
                    break;
            }
        }
    }
}
