using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Collections;

namespace Visual_Scenery_Mover
{
    public class Meta
    {
        public class MetaItem
        {
            public int offset;
            public ItemType type;
            public int mapOffset;
            public string description;
            public int intag;
            public string intagname;
            public string intagtype;
            public int parent = -1;
            public int sibling = -1;
            public int child = -1;
        }
        public enum ItemType
        {
            Reflexive,
            Ident,
            String
        }

        public MemoryStream MS;
        public List<MetaItem> items = new List<MetaItem>();

        public int
        itemcount,
        reflexivecount,
        headersize,
        size,
        offset,
        ident,
        tagnumber,
        magic;

        public string
        type,
        name;

        public RawDataContainer raw;
        public RawDataContainerType rawType;

        public char padding;

        public bool
        parsed = false,
        scannedwithent = false;

        public Meta(int tagnum)
        {
            tagnumber = tagnum;
        }

        public static int FindMetaByID(int ID, Tags.TagInfo[] tagInfo)
        {
            for (int x = 0; x < tagInfo.Length; x++)
            {
                if (tagInfo[x].TagID == ID)
                    return x;
            }
            return -1;
        }

        public void SortItemsByOffset()
        {
            Comparison<MetaItem> compareBySalary;
            compareBySalary = new Comparison<MetaItem>(new ItemSorter().Compare);

            items.Sort(compareBySalary);
        }

        public class ItemSorter : IComparer
        {
            public int Compare(object x, object y)
            {
                return ((MetaItem)x).offset.CompareTo(((MetaItem)y).offset);
            }
        }

        public void ReadMetaFromMap(bool manual, bool pparsed, bool dontreadraw, int mapnum)
        {
            //set meta properties
            type = Map.mp[mapnum].tags.tagInfo[tagnumber].TagType;
            name = Map.mp[mapnum].tags.tagInfo[tagnumber].TagPath;
            offset = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset;
            size = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaSize;
            ident = Map.mp[mapnum].tags.tagInfo[tagnumber].TagID;

            char[] hexCharArray = offset.ToString("X").ToCharArray();
            padding = hexCharArray[hexCharArray.Length - 1];

            //open the map in a reader
            BinaryReader br = new BinaryReader(new FileStream(Map.mp[mapnum].mapLoc,
                FileMode.Open, FileAccess.Read, FileShare.ReadWrite));

            //read meta to memorystream
            MS = new MemoryStream(size);
            br.BaseStream.Position = offset;
            MS.Write(br.ReadBytes(size), 0, size);
            br.Close();

            //checks if type has raw data
            this.rawType = RawData.CheckForRaw(this.type);
            if (!dontreadraw)
            {
                if (rawType != RawDataContainerType.Empty)
                    raw = RawData.ReadRaw(tagnumber, dontreadraw, mapnum);
            }
            if (type == "sbsp")
            {
                int h = Map.mp[mapnum].bsp.FindBSPNumberByBSPID(ident);
                magic = Map.mp[mapnum].bsp.sbsp[h].magic;
            }
            else if (type == "ltmp")
            {
                int h = Map.mp[mapnum].bsp.FindBSPNumberByLightMapID(ident);
                magic = Map.mp[mapnum].bsp.sbsp[h].magic;
            }
            else magic = Map.mp[mapnum].index.secondaryMagic;

            br.Close();
        }
    }
}
