using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Windows.Forms;

namespace Visual_Scenery_Mover
{
    public class Tags
    {
        public struct TagInfo
        {
            public Int32 TagID;
            public string TagType;
            public Int32 MetaOffset;
            public Int32 MetaSize;
            public String TagPath;
        }
        public TagInfo[] tagInfo;

        public Tags(BinaryReader br, Int32 offsetOfObjectIndex,
            Int32 objectCount, Int32 fileTableOffset, Int32 secondaryMagic)
        {
            br.BaseStream.Position = offsetOfObjectIndex;

            //will store the tags of the map
            tagInfo = new TagInfo[objectCount];
            //will store the tagtypes
            List<string> tagTypes = new List<string>();

            //go through all the tags
            for (int i = 0; i < objectCount; i++)
            {
                //this is the tagtype, they are backward in map so we reverse
                string tt = backwards(new string(br.ReadChars(4)));

                //if its not in the list add it
                if (!tagTypes.Contains(tt)) tagTypes.Add(tt);

                //and this is the rest of the info
                tagInfo[i].TagType = tt;
                tagInfo[i].TagID = br.ReadInt32();
                tagInfo[i].MetaOffset = br.ReadInt32() - secondaryMagic;
                tagInfo[i].MetaSize = br.ReadInt32();
            }

            //go to the filetable which contains all the strings - globals/globals etc
            br.BaseStream.Position = fileTableOffset;

            //go through the tags and add in their paths to their structs
            char chr;
            for (int i = 0; i < objectCount; i++)
            {
                while ((chr = br.ReadChar()) != '\0')
                    tagInfo[i].TagPath += chr;
            }
        }

        public string backwards(string str)
        {
            char[] temp = str.ToCharArray();
            Array.Reverse(temp);
            return new string(temp);

            //string rslt = "";
            //for (int x = str.Length - 1; x >= 0; x--)
            //    rslt += str.Substring(x, 1);
            //return rslt;
        }

        public TreeNode LoadBSPTagsIntoTreenodes(bool ViewShortenedTagpaths, int mapnum)
        {
            //TreeNode tnType = new TreeNode("sbsp");
            //tnType.Tag = "#type#";

            TreeNode tnName = new TreeNode(Map.mp[mapnum].mapLoc.Substring(Map.mp[mapnum].mapLoc.LastIndexOf('\\') + 1));
            tnName.Tag = "#file#" + Map.mp[mapnum].mapLoc;

            //go through all the tags
            for (int i = 0; i < tagInfo.Length; i++)
            {
                //and if the tags type is sbsp
                if (tagInfo[i].TagType == "sbsp")
                {
                    TreeNode tnTag = new TreeNode();

                    //we check to see if its null as it may be broken and therefore
                    //cause an error when we look for the \
                    if (ViewShortenedTagpaths)
                    {
                        if (tagInfo[i].TagPath != String.Empty && tagInfo[i].TagPath.LastIndexOf('\\') != -1)
                            tnTag.Text = tagInfo[i].TagPath.Substring(tagInfo[i].TagPath.LastIndexOf('\\') + 1);
                        else tnTag.Text = tagInfo[i].TagPath;
                    }
                    else tnTag.Text = tagInfo[i].TagPath;

                    tnTag.Tag = "#tagp#" + tagInfo[i].TagPath;
                    tnName.Nodes.Add(tnTag);
                }
            }

            
            //tnName.Nodes.Add(tnType);

            //return the now populated array
            return tnName;
        }

        public int FindByTagAndTagType(string type, string name)
        {
            for (int x = 0; x < tagInfo.Length; x++)
            {
                if (type == tagInfo[x].TagType && name == tagInfo[x].TagPath)
                    return x;
            }
            return -1;
        }

        public int FindModelByBaseClass(int tagnumber, string mapLoc)
        {
            if (tagnumber == -1) return -1;

            switch (tagInfo[tagnumber].TagType)
            {
                case "itmc":
                case "vehi":
                case "vehc":
                case "bloc":
                case "bipd":
                case "mach":
                case "ctrl":
                case "scen":
                case "eqip":
                case "weap":
                    break;

                default:
                    return -1;
            }

            //open the map in a reader
            BinaryReader br = new BinaryReader(new FileStream(mapLoc,
                FileMode.Open, FileAccess.Read, FileShare.ReadWrite));

            if (tagInfo[tagnumber].TagType == "vehc" | tagInfo[tagnumber].TagType == "itmc")
            {
                br.BaseStream.Position = tagInfo[tagnumber].MetaOffset + 20;
                tagnumber = Meta.FindMetaByID(br.ReadInt32(), tagInfo);
            }

            br.BaseStream.Position = tagInfo[tagnumber].MetaOffset + 56;
            int temptag = Meta.FindMetaByID(br.ReadInt32(), tagInfo);
            if (temptag == -1) return -1;

            br.BaseStream.Position = tagInfo[temptag].MetaOffset + 4;
            int temptag2 = Meta.FindMetaByID(br.ReadInt32(), tagInfo);

            br.Close();
            return temptag2;
        }
    }
}
