using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

using Microsoft.DirectX.Direct3D;
using System.Windows.Forms;

namespace Visual_Scenery_Mover
{
    public sealed class Map
    {
        public static Device dev = null;
        public static MapInfo[] mp = new MapInfo[0];
        public static int mapCount = 0;

        public static bool AddMap(string path, ToolStripLabel tsl, ToolStripProgressBar pb)
        {
            try
            {
                foreach (MapInfo m in mp)
                {
                    if (m.mapLoc == path)
                        return false;
                }

                MapInfo mi = new MapInfo();
                mi.mapLoc = path;

                BinaryReader br = new BinaryReader(new FileStream(path, FileMode.Open,
                    FileAccess.ReadWrite, FileShare.ReadWrite));

                pb.Maximum = 6;

                tsl.Text = "Loading Map Header...";
                Application.DoEvents();

                mi.header = new Header(br);

                pb.Value = 1;
                tsl.Text = "Loading Index...";
                Application.DoEvents();

                mi.index = new Index(br, mi.header.indexOffset, mi.header.indexSizeAndMetaStart);

                pb.Value = 2;
                tsl.Text = "Loading Tags...";
                Application.DoEvents();

                mi.tags = new Tags(br, mi.index.offsetOfObjectIndex, mi.index.objectCount, mi.header.fileTableOffset, mi.index.secondaryMagic);

                pb.Value = 3;
                tsl.Text = "Loading Strings...";
                Application.DoEvents();

                mi.strings = new Strings(br, mi.header.scriptReferenceCount, mi.header.offsetOfScriptReferenceIndex, mi.header.offsetOfScriptReferenceStrings);

                pb.Value = 4;
                tsl.Text = "Loading BSP Container...";
                Application.DoEvents();

                mi.bsp = new BSPContainer(br, mi.index.secondaryMagic, ref mi.tags.tagInfo);

                pb.Value = 5;
                tsl.Text = "Loading Unicode...";
                Application.DoEvents();

                mi.unicode = new Unicode(br, mi.tags.tagInfo[0].MetaOffset);

                pb.Value = 6;
                Application.DoEvents();

                br.Close();
                mi.dxr = MapInfo.DXRender.Triangle;
                pb.Value = 0;

                if (mp.Length > mapCount)
                    mp[mapCount + 1] = mi;

                else
                {
                    Array.Resize(ref mp, mapCount + 1);
                    mp[mapCount] = mi;
                }
                mapCount++;
                return true;
            }
            catch (Exception ex)
            {
                pb.Value = 0;
                tsl.Text = "";
                return false;
            }
        }

        public struct MapInfo
        {
            public string mapLoc;
            public Header header;
            public Index index;
            public Tags tags;
            public Strings strings;
            public BSPContainer bsp;
            public Meta meta;
            public Unicode unicode;
            public BSPRender bsprend;

            public enum DXRender
            {
                None,
                Triangle,
                Model,
                PRTM,
                BSP
            }
            public DXRender dxr;

            public string GetMapFromType(Header.MapTypes type)
            {
                switch (type)
                {
                    case Header.MapTypes.Internal:
                        return mapLoc;

                    default:
                        return Pref.GetMapFromType(type);
                }
            }
        }
    }
}
