using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Xml;

namespace Visual_Scenery_Mover
{
    class RawData
    {
        public static RawDataContainer ReadRaw(int tagnumber, bool dontreadraw, int mapnum)
        {
            RawDataContainer tempcontainer = new RawDataContainer();

            switch (CheckForRaw(Map.mp[mapnum].tags.tagInfo[tagnumber].TagType))
            {
                case RawDataContainerType.PRTM:
                    tempcontainer = new PRTM();
                    break;

                case RawDataContainerType.DECR:
                    tempcontainer = new DECR();
                    break;

                case RawDataContainerType.Animation:
                    tempcontainer = new Animation();
                    break;

                case RawDataContainerType.Model:
                    tempcontainer = new Model();
                    break;

                case RawDataContainerType.Bitmap:
                    tempcontainer = new BitmapRaw();
                    break;

                case RawDataContainerType.BSP:
                    tempcontainer = new BSPRaw();
                    break;

                case RawDataContainerType.Weather:
                    tempcontainer = new Weather();
                    break;

                case RawDataContainerType.CoconutsModel:
                    tempcontainer = new UghRawContainer();
                    break;

                case RawDataContainerType.LightMap:
                    tempcontainer = new LightmapRaw();
                    break;

                default:
                    break;
            }
            tempcontainer.Read(tagnumber, dontreadraw, mapnum);
            return tempcontainer;
        }

        public static RawDataContainerType CheckForRaw(string type)
        {
            switch (type)
            {
                case "PRTM":
                    return RawDataContainerType.PRTM;

                case "DECR":
                    return RawDataContainerType.DECR;

                case "ltmp":
                    return RawDataContainerType.LightMap;

                case "mod2":
                case "mode":
                    return RawDataContainerType.Model;

                case "bitm":
                    return RawDataContainerType.Bitmap;

                case "jmad":
                    return RawDataContainerType.Animation;

                case "sbsp":
                    return RawDataContainerType.BSP;

                case "weat":
                    return RawDataContainerType.Weather;

                case "snd!":
                    return RawDataContainerType.Sound;

                case "ugh!":
                    return RawDataContainerType.CoconutsModel;

                default:
                    return RawDataContainerType.Empty;
            }
        }
    }

    public enum RawDataType
    {
        bitm,
        bsp1,
        bsp2,
        bsp3,
        bsp4,
        DECR,
        jmad,
        ltmp,
        mode1,
        mode2,
        PRTM,
        snd1,
        snd2,
        weat
    }

    public enum RawDataContainerType
    {
        Bitmap,
        Header,
        Model,
        Animation,
        Sound,
        CoconutsModel,
        PRTM,
        DECR,
        BSP,
        LightMap,
        Empty,
        Weather,
        BSPMeta,
        UnicodeNamesIndex,
        UnicodeNames,
        StringsIndex,
        Strings1,
        Strings2,
        FileNamesIndex,
        FileNames,
        Crazy,
        MetaIndex,
        MetaData
    }

    public class BitmapRawDataChunk : RawDataChunk
    {
        public int inchunk;
        public int num;
    }

    public class RawDataChunk
    {
        public int shadernumber;
        public RawDataType rawDataType;
        public int offset;
        public int size;
        public int pointerMetaOffset;
        public Header.MapTypes rawLocation;
        public MemoryStream MS;
    }

    public class RawDataContainer
    {
        public RawDataContainerType containerType;
        public List<RawDataChunk> rawChunks = new List<RawDataChunk>();

        public RawDataContainer()
        {
        }

        public virtual void Read(int tagnumber, bool dontreadraw, int mapnum)
        { }

        public void SaveRawToFile(string outputFilePath, Meta meta)
        {
            int x = outputFilePath.LastIndexOf('.');
            string temp = outputFilePath.Substring(0, x + 1) + meta.type + "raw";

            XmlTextWriter xtw = new XmlTextWriter(temp + "xml", System.Text.Encoding.Default);
            xtw.Formatting = Formatting.Indented;
            xtw.WriteStartElement("RawData");
            xtw.WriteAttributeString("TagType", meta.type);
            xtw.WriteAttributeString("TagName", meta.name);
            xtw.WriteAttributeString("RawType", meta.raw.containerType.ToString());
            xtw.WriteAttributeString("RawChunkCount", this.rawChunks.Count.ToString());
            xtw.WriteAttributeString("Date", DateTime.Today.ToShortDateString());
            xtw.WriteAttributeString("Time", DateTime.Now.ToShortTimeString());
            xtw.WriteAttributeString("EntityVersion", "0.1");

            FileStream FS = new FileStream(temp, FileMode.Create);
            BinaryWriter BW = new BinaryWriter(FS);
            int loc = 0;
            for (x = 0; x < this.rawChunks.Count; x++)
            {
                RawDataChunk r = (RawDataChunk)this.rawChunks[x];
                if (r.offset == -1) { continue; }
                xtw.WriteStartElement("RawChunk");
                xtw.WriteAttributeString("RawDataType", r.rawDataType.ToString());
                xtw.WriteAttributeString("PointerMetaOffset", r.pointerMetaOffset.ToString());
                xtw.WriteAttributeString("ChunkSize", r.size.ToString());
                xtw.WriteAttributeString("RawDataOffset", loc.ToString());
                xtw.WriteAttributeString("PointsToOffset", r.offset.ToString());
                xtw.WriteAttributeString("RawLocation", r.rawLocation.ToString());
                xtw.WriteEndElement();
                BW.BaseStream.Write(r.MS.ToArray(), 0, r.size);
                loc += r.size;
            }
            BW.Close();
            FS.Close();
            xtw.WriteEndElement();
            xtw.Close();
        }

        public RawDataContainer LoadRawFromFile(string inputFilePath, Meta meta)
        {
            RawDataContainer raw = new RawDataContainer();
            int x = inputFilePath.LastIndexOf('.');
            string temp = inputFilePath.Substring(0, x + 1) + meta.type + "raw";

            XmlTextReader xtr = new XmlTextReader(temp + "xml");
            xtr.WhitespaceHandling = WhitespaceHandling.None;

            BinaryReader BR = new BinaryReader(new FileStream(temp, FileMode.Open));

            while (xtr.Read())
            {
                switch (xtr.NodeType)
                {
                    case XmlNodeType.Element:
                        if (xtr.Name == "RawData")
                        {
                            string oi = xtr.GetAttribute("RawType");

                            switch (oi)
                            {
                                case "Model":
                                    raw = new Model();
                                    break;
                                case "Bitmap":
                                    raw = new BitmapRaw();
                                    break;
                                case "Animation":
                                    raw = new Animation();
                                    break;
                                case "DECR":
                                    raw = new DECR();
                                    break;
                                case "PRTM":
                                    raw = new PRTM();
                                    break;
                                case "Weather":
                                    raw = new Weather();
                                    break;
                                case "BSP":
                                    raw = new BSPRaw();
                                    break;
                            }
                        }
                        else if (xtr.Name == "RawChunk")
                        {
                            RawDataChunk r = new RawDataChunk();
                            string temps = xtr.GetAttribute("RawDataType");
                            switch (temps)
                            {
                                case "bitm":
                                    r.rawDataType = RawDataType.bitm;
                                    break;
                                case "bsp1":
                                    r.rawDataType = RawDataType.bsp1;
                                    break;
                                case "bsp2":
                                    r.rawDataType = RawDataType.bsp2;
                                    break;
                                case "bsp3":
                                    r.rawDataType = RawDataType.bsp3;
                                    break;
                                case "bsp4":
                                    r.rawDataType = RawDataType.bsp4;
                                    break;
                                case "DECR":
                                    r.rawDataType = RawDataType.DECR;
                                    break;
                                case "jmad":
                                    r.rawDataType = RawDataType.jmad;
                                    break;
                                case "ltmp":
                                    r.rawDataType = RawDataType.ltmp;
                                    break;
                                case "mode1":
                                    r.rawDataType = RawDataType.mode1;
                                    break;
                                case "mode2":
                                    r.rawDataType = RawDataType.mode2;
                                    break;
                                case "PRTM":
                                    r.rawDataType = RawDataType.PRTM;
                                    break;
                                case "snd1":
                                    r.rawDataType = RawDataType.snd1;
                                    break;
                                case "snd2":
                                    r.rawDataType = RawDataType.snd2;
                                    break;
                            }

                            r.offset = Convert.ToInt32(xtr.GetAttribute("PointsToOffset"));
                            r.pointerMetaOffset = Convert.ToInt32(xtr.GetAttribute("PointerMetaOffset"));
                            r.size = Convert.ToInt32(xtr.GetAttribute("ChunkSize"));
                            int rawdataspot = Convert.ToInt32(xtr.GetAttribute("RawDataOffset"));
                            BR.BaseStream.Position = rawdataspot;
                            r.MS = new MemoryStream(r.size);
                            r.MS.Write(BR.ReadBytes(r.size), 0, r.size);

                            raw.rawChunks.Add(r);
                        }
                        break;
                }
            }

            BR.Close();
            xtr.Close();

            return raw;
        }
    }

    public class BitmapRaw : RawDataContainer
    {
        public BitmapRaw()
        {
            containerType = RawDataContainerType.Bitmap;
        }

        public override void Read(int tagnumber, bool dontreadraw, int mapnum)
        {
            //read the halo2 bitmap raw

            //open the Map.mp[mapnum].map in a reader
            BinaryReader br = new BinaryReader(new FileStream(Map.mp[mapnum].mapLoc,
                FileMode.Open, FileAccess.Read, FileShare.Read));

            br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset + 68;
            int tempc = br.ReadInt32();
            int tempr = br.ReadInt32() - (int)Map.mp[mapnum].index.secondaryMagic;

            for (int x = 0; x < tempc; x++)
            {
                for (int xx = 0; xx < 6; xx++)
                {
                    BitmapRawDataChunk Raw = new BitmapRawDataChunk();
                    Raw.inchunk = x;
                    Raw.num = xx;
                    Raw.pointerMetaOffset = tempr + (x * 116) + 28 + (xx * 4) -
                        (int)Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset;

                    br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset +
                        Raw.pointerMetaOffset;

                    Raw.rawDataType = RawDataType.bitm;
                    Raw.offset = br.ReadInt32();
                    if (Raw.offset == -1) break;

                    br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset +
                        Raw.pointerMetaOffset + 24;

                    Raw.size = br.ReadInt32();

                    BSPContainer.ParsePointer(ref Raw.offset, ref Raw.rawLocation);
                    if (!dontreadraw)
                    {
                        BinaryReader byr = new BinaryReader(new FileStream(
                            Map.mp[mapnum].GetMapFromType(Raw.rawLocation), FileMode.Open,
                            FileAccess.Read, FileShare.Read));

                        byr.BaseStream.Position = Raw.offset;
                        Raw.MS = new MemoryStream(Raw.size);
                        Raw.MS.Write(byr.ReadBytes(Raw.size), 0, Raw.size);

                        byr.Close();
                    }
                    rawChunks.Add(Raw);
                }
            }
            br.Close();
        }
    }

    public class Model : RawDataContainer
    {
        public Model()
        {
            containerType = RawDataContainerType.Model;
        }

        public override void Read(int tagnumber, bool dontreadraw, int mapnum)
        {
            //read halo2 model raw

            //open the Map.mp[mapnum].map in a reader
            BinaryReader br = new BinaryReader(new FileStream(Map.mp[mapnum].mapLoc,
                FileMode.Open, FileAccess.Read, FileShare.ReadWrite));

            br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset + 36;
            int tempc = br.ReadInt32();
            int tempr = br.ReadInt32() - Map.mp[mapnum].index.secondaryMagic;
            for (int x = 0; x < tempc; x++)
            {
                RawDataChunk Raw = new RawDataChunk();
                Raw.pointerMetaOffset = tempr + (x * 92) + 56 -
                    (int)Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset;

                br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset +
                    Raw.pointerMetaOffset;

                Raw.rawDataType = RawDataType.mode1;
                Raw.offset = br.ReadInt32();
                Raw.size = br.ReadInt32();
                BSPContainer.ParsePointer(ref Raw.offset, ref Raw.rawLocation);
                if (dontreadraw == false)
                {
                    BinaryReader byr = new BinaryReader(new FileStream(
                             Map.mp[mapnum].GetMapFromType(Raw.rawLocation), FileMode.Open,
                             FileAccess.Read, FileShare.Read));

                    byr.BaseStream.Position = Raw.offset;
                    Raw.MS = new MemoryStream(Raw.size);
                    Raw.MS.Write(byr.ReadBytes(Raw.size), 0, Raw.size);

                    byr.Close();
                }
                this.rawChunks.Add(Raw);
            }

            br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset + 116;
            tempc = br.ReadInt32();
            tempr = br.ReadInt32() - Map.mp[mapnum].index.secondaryMagic;
            for (int x = 0; x < tempc; x++)
            {
                RawDataChunk Raw = new RawDataChunk();
                Raw.pointerMetaOffset = tempr + (x * 88) + 52 -
                    (int)Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset;

                br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset +
                    Raw.pointerMetaOffset;

                Raw.rawDataType = RawDataType.mode2;
                Raw.offset = br.ReadInt32();
                Raw.size = br.ReadInt32();
                BSPContainer.ParsePointer(ref Raw.offset, ref Raw.rawLocation);
                if (dontreadraw == false)
                {
                    BinaryReader byr = new BinaryReader(new FileStream(
                            Map.mp[mapnum].GetMapFromType(Raw.rawLocation), FileMode.Open,
                            FileAccess.Read, FileShare.Read));

                    byr.BaseStream.Position = Raw.offset;
                    Raw.MS = new MemoryStream(Raw.size);
                    Raw.MS.Write(byr.ReadBytes(Raw.size), 0, Raw.size);

                    byr.Close();
                }
                this.rawChunks.Add(Raw);
            }
            br.Close();
        }
    }

    public class Weather : RawDataContainer
    {
        public Weather()
        {
            containerType = RawDataContainerType.Weather;
        }

        public override void Read(int tagnumber, bool dontreadraw, int mapnum)
        {
            //open the Map.mp[mapnum].map in a reader
            BinaryReader br = new BinaryReader(new FileStream(Map.mp[mapnum].mapLoc,
                FileMode.Open, FileAccess.Read, FileShare.Read));

            br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset;
            int tempc = br.ReadInt32();
            int tempr = br.ReadInt32() - Map.mp[mapnum].index.secondaryMagic;
            for (int x = 0; x < tempc; x++)
            {
                RawDataChunk Raw = new RawDataChunk();
                Raw.pointerMetaOffset = tempr + (x * 140) + 64 -
                    (int)Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset;

                br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset +
                    Raw.pointerMetaOffset;

                Raw.rawDataType = RawDataType.weat;
                Raw.offset = br.ReadInt32();
                Raw.size = br.ReadInt32();
                BSPContainer.ParsePointer(ref Raw.offset, ref Raw.rawLocation);
                if (dontreadraw == false)
                {
                    //open the Map.mp[mapnum].map in a reader
                    BinaryReader byr = new BinaryReader(new FileStream(Map.mp[mapnum].GetMapFromType(Raw.rawLocation),
                        FileMode.Open, FileAccess.Read, FileShare.Read));

                    byr.BaseStream.Position = Raw.offset;
                    Raw.MS = new MemoryStream(Raw.size);
                    Raw.MS.Write(byr.ReadBytes(Raw.size), 0, Raw.size);

                    byr.Close();
                }
                this.rawChunks.Add(Raw);
            }
        }
    }

    public class Animation : RawDataContainer
    {
        public Animation()
        {
            containerType = RawDataContainerType.Animation;
        }

        public override void Read(int tagnumber, bool dontreadraw, int mapnum)
        {
            //open the Map.mp[mapnum].map in a reader
            BinaryReader br = new BinaryReader(new FileStream(Map.mp[mapnum].mapLoc,
                FileMode.Open, FileAccess.Read, FileShare.Read));

            br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset + 172;
            int tempc = br.ReadInt32();
            int tempr = br.ReadInt32() - Map.mp[mapnum].index.secondaryMagic;
            for (int x = 0; x < tempc; x++)
            {
                RawDataChunk Raw = new RawDataChunk();
                Raw.rawDataType = RawDataType.jmad;
                Raw.pointerMetaOffset = tempr + (x * 20) + 8 -
                    (int)Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset;

                br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset +
                    Raw.pointerMetaOffset - 4;

                Raw.size = br.ReadInt32();
                Raw.offset = br.ReadInt32();

                BSPContainer.ParsePointer(ref Raw.offset, ref Raw.rawLocation);
                if (dontreadraw == false)
                {
                    BinaryReader byr = new BinaryReader(new FileStream(
                             Map.mp[mapnum].GetMapFromType(Raw.rawLocation), FileMode.Open,
                             FileAccess.Read, FileShare.Read));

                    byr.BaseStream.Position = Raw.offset;
                    Raw.MS = new MemoryStream(Raw.size);
                    Raw.MS.Write(byr.ReadBytes(Raw.size), 0, Raw.size);

                    byr.Close();
                }
                this.rawChunks.Add(Raw);
            }
        }
    }

    public class PRTM : RawDataContainer
    {
        public PRTM()
        {
            containerType = RawDataContainerType.PRTM;
        }

        public override void Read(int tagnumber, bool dontreadraw, int mapnum)
        {
            //open the map in a reader
            BinaryReader br = new BinaryReader(new FileStream(Map.mp[mapnum].mapLoc,
                FileMode.Open, FileAccess.Read, FileShare.Read));

            RawDataChunk Raw = new RawDataChunk();
            Raw.rawDataType = RawDataType.PRTM;
            Raw.pointerMetaOffset = 160;
            br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset + 160;
            Raw.offset = br.ReadInt32();
            Raw.size = br.ReadInt32();
            BSPContainer.ParsePointer(ref Raw.offset, ref Raw.rawLocation);
            if (dontreadraw == false)
            {
                BinaryReader byr = new BinaryReader(new FileStream(
                            Map.mp[mapnum].GetMapFromType(Raw.rawLocation), FileMode.Open,
                            FileAccess.Read, FileShare.Read));

                byr.BaseStream.Position = Raw.offset;
                Raw.MS = new MemoryStream(Raw.size);
                Raw.MS.Write(byr.ReadBytes(Raw.size), 0, Raw.size);

                byr.Close();
            }
            this.rawChunks.Add(Raw);
        }
    }

    public class UghRawContainer : RawDataContainer
    {
        public UghRawContainer()
        {
            containerType = RawDataContainerType.CoconutsModel;
        }

        public override void Read(int tagnumber, bool dontreadraw, int mapnum)
        {
            //open the Map.mp[mapnum].map in a reader
            BinaryReader br = new BinaryReader(new FileStream(Map.mp[mapnum].mapLoc,
                FileMode.Open, FileAccess.Read, FileShare.Read));

            br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset + 80;
            int tempc = br.ReadInt32();
            int tempr = br.ReadInt32() - Map.mp[mapnum].index.secondaryMagic;
            for (int x = 0; x < tempc; x++)
            {
                RawDataChunk Raw = new RawDataChunk();
                Raw.rawDataType = RawDataType.snd2;
                Raw.pointerMetaOffset = tempr + (x * 44) + 8 -
                    (int)Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset;

                br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset +
                    Raw.pointerMetaOffset;

                Raw.offset = br.ReadInt32();
                Raw.size = br.ReadInt32();
                BSPContainer.ParsePointer(ref Raw.offset, ref Raw.rawLocation);
                if (dontreadraw == false)
                {
                    BinaryReader byr = new BinaryReader(new FileStream(
                            Map.mp[mapnum].GetMapFromType(Raw.rawLocation), FileMode.Open,
                            FileAccess.Read, FileShare.Read));

                    byr.BaseStream.Position = Raw.offset;
                    Raw.MS = new MemoryStream(Raw.size);
                    Raw.MS.Write(byr.ReadBytes(Raw.size), 0, Raw.size);

                    byr.Close();
                }
                this.rawChunks.Add(Raw);
            }
        }
    }

    public class DECR : RawDataContainer
    {
        public DECR()
        {
            containerType = RawDataContainerType.DECR;
        }

        public override void Read(int tagnumber, bool dontreadraw, int mapnum)
        {
            //open the Map.mp[mapnum].map in a reader
            BinaryReader br = new BinaryReader(new FileStream(Map.mp[mapnum].mapLoc,
                FileMode.Open, FileAccess.Read, FileShare.Read));

            RawDataChunk Raw = new RawDataChunk();
            Raw.rawDataType = RawDataType.DECR;
            Raw.pointerMetaOffset = 56;
            br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset + 56;
            Raw.offset = br.ReadInt32();
            Raw.size = br.ReadInt32();
            BSPContainer.ParsePointer(ref Raw.offset, ref Raw.rawLocation);
            if (dontreadraw == false)
            {
                BinaryReader byr = new BinaryReader(new FileStream(
                             Map.mp[mapnum].GetMapFromType(Raw.rawLocation), FileMode.Open,
                             FileAccess.Read, FileShare.Read));

                byr.BaseStream.Position = Raw.offset;
                Raw.MS = new MemoryStream(Raw.size);
                Raw.MS.Write(byr.ReadBytes(Raw.size), 0, Raw.size);

                byr.Close();
            }
            this.rawChunks.Add(Raw);
        }
    }

    public class BSPRaw : RawDataContainer
    {
        public BSPRaw()
        {
            containerType = RawDataContainerType.BSP;
        }

        public override void Read(int tagnumber, bool dontreadraw, int mapnum)
        {
            //open the map in a reader
            BinaryReader br = new BinaryReader(new FileStream(Map.mp[mapnum].mapLoc,
                FileMode.Open, FileAccess.Read, FileShare.Read));

            int w = Map.mp[mapnum].bsp.FindBSPNumberByBSPID((int)Map.mp[mapnum].tags.tagInfo[tagnumber].TagID);

            //bsp model 1
            br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset + 172;
            int tempc = br.ReadInt32();
            int tempr = br.ReadInt32() - Map.mp[mapnum].bsp.sbsp[w].magic;
            for (int x = 0; x < tempc; x++)
            {
                RawDataChunk Raw = new RawDataChunk();
                Raw.rawDataType = RawDataType.bsp1;
                Raw.pointerMetaOffset = tempr + (x * 176) + 40 -
                    (int)Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset;

                br.BaseStream.Position = tempr + (x * 176) + 40;
                Raw.offset = br.ReadInt32();

                Raw.size = br.ReadInt32();
                if (Raw.offset == -1) { this.rawChunks.Add(Raw); continue; }
                BSPContainer.ParsePointer(ref Raw.offset, ref Raw.rawLocation);
                if (dontreadraw == false)
                {
                    BinaryReader byr = new BinaryReader(new FileStream(
                              Map.mp[mapnum].GetMapFromType(Raw.rawLocation), FileMode.Open,
                              FileAccess.Read, FileShare.Read));

                    byr.BaseStream.Position = Raw.offset;
                    Raw.MS = new MemoryStream(Raw.size);
                    Raw.MS.Write(byr.ReadBytes(Raw.size), 0, Raw.size);

                    byr.Close();
                }
                this.rawChunks.Add(Raw);
            }

            //bsp model 2
            br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset + 328;
            tempc = br.ReadInt32();
            tempr = br.ReadInt32() - Map.mp[mapnum].bsp.sbsp[w].magic;
            for (int x = 0; x < tempc; x++)
            {
                RawDataChunk Raw = new RawDataChunk();
                Raw.rawDataType = RawDataType.bsp2;
                Raw.pointerMetaOffset = tempr + (x * 200) + 40 -
                    (int)Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset;

                br.BaseStream.Position = tempr + (x * 200) + 40;
                Raw.offset = br.ReadInt32();

                Raw.size = br.ReadInt32();
                if (Raw.offset == -1) { this.rawChunks.Add(Raw); continue; }
                BSPContainer.ParsePointer(ref Raw.offset, ref Raw.rawLocation);
                if (dontreadraw == false)
                {
                    BinaryReader byr = new BinaryReader(new FileStream(
                             Map.mp[mapnum].GetMapFromType(Raw.rawLocation), FileMode.Open,
                             FileAccess.Read, FileShare.Read));

                    byr.BaseStream.Position = Raw.offset;
                    Raw.MS = new MemoryStream(Raw.size);
                    Raw.MS.Write(byr.ReadBytes(Raw.size), 0, Raw.size);

                    byr.Close();
                }
                this.rawChunks.Add(Raw);
            }

            //light map
            if (Map.mp[mapnum].bsp.sbsp[w].lightmapident != -1)
            {
                br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset + 8;
                tempr = br.ReadInt32();
                if (tempr == 0) { goto skiplightmap; }
                tempr -= Map.mp[mapnum].bsp.sbsp[w].magic;
                br.BaseStream.Position = tempr + 128;
                tempc = br.ReadInt32();
                tempr = br.ReadInt32() - Map.mp[mapnum].bsp.sbsp[w].magic;
                br.BaseStream.Position = tempr + 64;
                tempc = br.ReadInt32();
                tempr = br.ReadInt32() - Map.mp[mapnum].bsp.sbsp[w].magic;
                for (int x = 0; x < tempc; x++)
                {
                    RawDataChunk Raw = new RawDataChunk();
                    Raw.rawDataType = RawDataType.ltmp;
                    Raw.pointerMetaOffset = tempr + (x * 56) + 12 -
                        (int)Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset;

                    br.BaseStream.Position = tempr + (x * 56) + 12;
                    Raw.offset = br.ReadInt32();

                    Raw.size = br.ReadInt32();
                    if (Raw.offset == -1) { this.rawChunks.Add(Raw); continue; }
                    BSPContainer.ParsePointer(ref Raw.offset, ref Raw.rawLocation);
                    if (dontreadraw == false)
                    {
                        BinaryReader byr = new BinaryReader(new FileStream(
                             Map.mp[mapnum].GetMapFromType(Raw.rawLocation), FileMode.Open,
                             FileAccess.Read, FileShare.Read));

                        byr.BaseStream.Position = Raw.offset;
                        Raw.MS = new MemoryStream(Raw.size);
                        Raw.MS.Write(byr.ReadBytes(Raw.size), 0, Raw.size);

                        byr.Close();
                    }
                    this.rawChunks.Add(Raw);
                }
            }

        skiplightmap:

            //bsp model 3
            br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset + 580;
            tempc = br.ReadInt32();
            tempr = br.ReadInt32() - Map.mp[mapnum].bsp.sbsp[w].magic;
            //if (tempc==0){return;}
            br.BaseStream.Position = tempr + 16;
            tempc = br.ReadInt32();
            tempr = br.ReadInt32() - Map.mp[mapnum].bsp.sbsp[w].magic;
            for (int x = 0; x < tempc; x++)
            {
                RawDataChunk Raw = new RawDataChunk();
                Raw.rawDataType = RawDataType.bsp3;
                Raw.pointerMetaOffset = tempr + (x * 44) -
                    (int)Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset;

                br.BaseStream.Position = tempr + (x * 44);
                Raw.offset = br.ReadInt32();
                Raw.size = br.ReadInt32();
                if (Raw.offset == -1)
                {
                    this.rawChunks.Add(Raw);
                    continue;
                }
                BSPContainer.ParsePointer(ref Raw.offset, ref Raw.rawLocation);
                if (dontreadraw == false)
                {
                    BinaryReader byr = new BinaryReader(new FileStream(
                            Map.mp[mapnum].GetMapFromType(Raw.rawLocation), FileMode.Open,
                            FileAccess.Read, FileShare.Read));

                    byr.BaseStream.Position = Raw.offset;
                    Raw.MS = new MemoryStream(Raw.size);
                    Raw.MS.Write(byr.ReadBytes(Raw.size), 0, Raw.size);

                    byr.Close();
                }
                this.rawChunks.Add(Raw);
            }

            //bsp model 4
            br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset + 548;
            tempc = br.ReadInt32();
            tempr = br.ReadInt32() - Map.mp[mapnum].bsp.sbsp[w].magic;

            for (int x = 0; x < tempc; x++)
            {
                RawDataChunk Raw = new RawDataChunk();
                Raw.rawDataType = RawDataType.bsp4;
                Raw.pointerMetaOffset = tempr + (x * 172) + 16 -
                    (int)Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset;

                br.BaseStream.Position = tempr + (x * 172) + 16;
                Raw.offset = br.ReadInt32();
                if (Raw.offset == -1) continue;
                Raw.size = br.ReadInt32();
                BSPContainer.ParsePointer(ref Raw.offset, ref Raw.rawLocation);
                if (dontreadraw == false)
                {
                    BinaryReader byr = new BinaryReader(new FileStream(
                            Map.mp[mapnum].GetMapFromType(Raw.rawLocation), FileMode.Open,
                            FileAccess.Read, FileShare.Read));

                    byr.BaseStream.Position = Raw.offset;
                    Raw.MS = new MemoryStream(Raw.size);
                    Raw.MS.Write(byr.ReadBytes(Raw.size), 0, Raw.size);

                    byr.Close();
                }
                this.rawChunks.Add(Raw);
            }
            br.Close();
        }
    }

    public class LightmapRaw : RawDataContainer
    {
        public LightmapRaw()
        {
            containerType = RawDataContainerType.LightMap;
        }

        public override void Read(int tagnumber, bool dontreadraw, int mapnum)
        {
            //open the Map.mp[mapnum].map in a reader
            BinaryReader br = new BinaryReader(new FileStream(Map.mp[mapnum].mapLoc,
                FileMode.Open, FileAccess.Read, FileShare.Read));

            int w = Map.mp[mapnum].bsp.FindBSPNumberByLightMapID((int)Map.mp[mapnum].tags.tagInfo[tagnumber].TagID);

            //light Map.mp[mapnum].map
            if (Map.mp[mapnum].bsp.sbsp[w].lightmapident != -1)
            {
                int tempr = (int)Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset;
                br.BaseStream.Position = tempr + 128;
                int tempc = br.ReadInt32();
                tempr = br.ReadInt32() - Map.mp[mapnum].bsp.sbsp[w].magic;
                br.BaseStream.Position = tempr + 64;
                tempc = br.ReadInt32();
                tempr = br.ReadInt32() - Map.mp[mapnum].bsp.sbsp[w].magic;
                for (int x = 0; x < tempc; x++)
                {
                    RawDataChunk Raw = new RawDataChunk();
                    Raw.rawDataType = RawDataType.ltmp;
                    Raw.pointerMetaOffset = tempr + (x * 56) + 12 -
                        (int)Map.mp[mapnum].tags.tagInfo[tagnumber].MetaOffset;

                    br.BaseStream.Position = tempr + (x * 56) + 12;
                    Raw.offset = br.ReadInt32();

                    Raw.size = br.ReadInt32();
                    if (Raw.offset == -1) { this.rawChunks.Add(Raw); continue; }
                    BSPContainer.ParsePointer(ref Raw.offset, ref Raw.rawLocation);
                    if (dontreadraw == false)
                    {
                        BinaryReader byr = new BinaryReader(new FileStream(
                            Map.mp[mapnum].GetMapFromType(Raw.rawLocation), FileMode.Open,
                            FileAccess.Read, FileShare.Read));

                        byr.BaseStream.Position = Raw.offset;
                        Raw.MS = new MemoryStream(Raw.size);
                        Raw.MS.Write(byr.ReadBytes(Raw.size), 0, Raw.size);

                        byr.Close();
                    }
                    this.rawChunks.Add(Raw);
                }
            }
        }
    }
}
