using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Windows.Forms;

namespace Visual_Scenery_Mover
{
    public sealed class Pref
    {
        public static string
        mainmenuLoc, 
        MPSharedLoc, 
        SPSharedLoc;

        public static bool
        paused,
        activateTagsTabOnSelection = true,
        renderLightmap = true,
        renderSkybox = true;

        public static void SavePreferences()
        {
            string drive = Path.GetPathRoot(Application.StartupPath);
            string bunk = drive + @"Program Files\HomeBrewGames\HaloView\";
            string sav = bunk + "Preferences.cfg";

            if (!Directory.Exists(bunk))
                Directory.CreateDirectory(bunk);

            StreamWriter sw = File.CreateText(sav);
            sw.WriteLine(mainmenuLoc);
            sw.WriteLine(MPSharedLoc);
            sw.WriteLine(SPSharedLoc);
            sw.WriteLine(activateTagsTabOnSelection);
            sw.WriteLine(renderLightmap);
            sw.WriteLine(renderSkybox);
            sw.Close();
        }

        public static void LoadPreferences()
        {
            string drive = Path.GetPathRoot(Application.StartupPath);
            string bunk = drive + @"Program Files\HomeBrewGames\HaloView\";
            string sav = bunk + "Preferences.cfg";

            if (File.Exists(sav))
            {
                StreamReader sr = File.OpenText(sav);
                mainmenuLoc = sr.ReadLine();
                MPSharedLoc = sr.ReadLine();
                SPSharedLoc = sr.ReadLine();
                activateTagsTabOnSelection = Convert.ToBoolean(sr.ReadLine());
                renderLightmap = Convert.ToBoolean(sr.ReadLine());
                renderSkybox = Convert.ToBoolean(sr.ReadLine());
                sr.Close();
            }
        }

        public static string NeededExternalMapPath(Header.MapTypes type)
        {
            string txt = "";
            Header.MapTypes mtype = Header.MapTypes.Internal;
            switch (type)
            {
                case Header.MapTypes.MainMenu:
                    txt = "mainmenu.map";
                    mtype = Header.MapTypes.MainMenu;
                    break;

                case Header.MapTypes.MPShared:
                    txt = "shared.map";
                    mtype = Header.MapTypes.MPShared;
                    break;

                case Header.MapTypes.SPShared:
                    txt = "single_player_shared.map";
                    mtype = Header.MapTypes.SPShared;
                    break;
            }

            SelectExternalMap sem = new SelectExternalMap(txt, mtype);
            sem.ShowDialog();

            switch (type)
            {
                case Header.MapTypes.MainMenu:
                    return mainmenuLoc;
                case Header.MapTypes.MPShared:
                    return MPSharedLoc;
                case Header.MapTypes.SPShared:
                    return SPSharedLoc;
            }
            return null;
        }

        public static string GetMapFromType(Header.MapTypes type)
        {
            switch (type)
            {
                case Header.MapTypes.MainMenu:
                    if (mainmenuLoc == String.Empty || !File.Exists(mainmenuLoc)) 
                        return NeededExternalMapPath(Header.MapTypes.MainMenu);
                    
                    else return mainmenuLoc;

                case Header.MapTypes.MPShared:
                    if (MPSharedLoc == String.Empty || !File.Exists(MPSharedLoc)) 
                        return NeededExternalMapPath(Header.MapTypes.MPShared);
                    
                    else return MPSharedLoc;

                case Header.MapTypes.SPShared:
                    if (SPSharedLoc == String.Empty || !File.Exists(SPSharedLoc)) 
                        return NeededExternalMapPath(Header.MapTypes.SPShared);
                    
                    else return SPSharedLoc;

                default:
                    return null;
            }
        }
    }
}
