using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.IO;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using DI = Microsoft.DirectX.DirectInput;

namespace Visual_Scenery_Mover
{
    public class BSPRender
    {
        public bool bspLoaded = false;
        public BSPModel bspmodel;
        public List<ParsedModel> SpawnModel = new List<ParsedModel>();
        public int[] spawnmodelindex;
        public Mesh[] BoundingBoxModel;
        public Matrix[] TranslationMatrix;
        public int visibleSpawnsBitMask, HaloLightCount = 1, sb = SystemInformation.VerticalScrollBarWidth;

        public List<int> SelectedSpawn = new List<int>();
        public SpawnInfo.SpawnType selectedSpawnType;

        public bool itemrotate = false;

        public Material
        BoundingBoxSelected,
        BlueMaterial,
        BlackMaterial,
        GreenMaterial,
        DefaultMaterial;

        public float
        selectionHeight,
        selectionWidth,
        selectionDepth;

        public bool[] YPR = new bool[3];
        public bool selectionMulti = false, lighting = true;
        public Vector3 selectionStart;
        public Mesh selectionMesh = null;
        //DXShader shaderx;
        public Lightmap ltmp;
        public Camera cam;
        public List<DI.Key> pastList = new List<Microsoft.DirectX.DirectInput.Key>();

        public struct SelectedSpawnStruct
        {
            public int
            BaseSpawn,   //0 = obj            1 = col
            Positioning; //0 = yaw pitch roll 1 = direction

            public float X, Y, Z, Yaw, Pitch, Roll, Direction;
            public string ModelName, Type;

            public string ObjColType, ObjTeam;
            public short ObjNumber;
        }

        public void LoadBSP(ToolStripProgressBar pb1, Panel lightmappanel, TabPage tabPage3,
            TreeView tv, CheckedListBox clb, EventHandler imgclick, int mapnum)
        {
            pb1.Maximum = 8;

            pb1.Value = 1;
            Application.DoEvents();

            visibleSpawnsBitMask = 0;
            YPR[0] = YPR[1] = YPR[2] = false;

            bspmodel = new BSPModel(mapnum);

            pb1.Value += 1;
            Application.DoEvents();

            BSPModel.BSPDisplayedInfo.LoadDirectXTexturesAndBuffers(ref bspmodel);

            pb1.Value += 1;
            Application.DoEvents();

            ParsedModel.DisplayedInfo.LoadDirectXTexturesAndBuffers(ref bspmodel.SkyBox);

            pb1.Value += 1;
            Application.DoEvents();

            LoadSpawns(mapnum);

            pb1.Value += 1;
            Application.DoEvents();

            MakeMatrixes();

            pb1.Value += 1;
            Application.DoEvents();

            PopLTMP(mapnum, lightmappanel, tabPage3, imgclick);

            cam = new Camera();
            cam.speed = 0.5f;

            cam.x = 0f;
            cam.y = -1f;
            cam.z = 0.2f;
            cam.radianv = 0f;
            cam.radianh = 0f;

            cam.Position = new Vector3(cam.x, cam.y, cam.z);
            cam.ComputePosition();

            pb1.Value += 1;
            Application.DoEvents();

            BoundingBoxSelected = new Material();
            BoundingBoxSelected.Diffuse = BoundingBoxSelected.Ambient = Color.Red;

            DefaultMaterial = new Material();
            DefaultMaterial.Diffuse = DefaultMaterial.Ambient = Color.White;

            BlackMaterial = new Material();
            BlackMaterial.Diffuse = BlackMaterial.Ambient = BlackMaterial.Specular = Color.Black;

            BlueMaterial = new Material();
            BlueMaterial.Diffuse = BlueMaterial.Ambient = Color.Blue;

            GreenMaterial = new Material();
            GreenMaterial.Diffuse = GreenMaterial.Ambient = Color.GreenYellow;

            pb1.Value += 1;
            Application.DoEvents();

            Map.dev.RenderState.Lighting = true;
            PopSpawnList(true, true, tv, clb);
            SetUpCameraForBSP();

            pb1.Value = 0;
            bspLoaded = true;
        }

        public void PopLTMP(int mapnum, Panel lightmappanel, TabPage tabPage3, EventHandler imgclick)
        {
            if (bspmodel.renderLightmap)
            {
                ltmp = new Lightmap();
                ltmp.Load_Lightmaps_Into_An_Array(bspmodel, mapnum);

                int imgX = 0, imgY = 0;
                for (int i = 0; i < Map.mp[mapnum].bsp.sbsp[bspmodel.BspNumber].LightMap_Palettes.Count; i++)
                {
                    PictureBox img = new PictureBox();
                    img.Image = ltmp.RenderLightmap(i);

                    img.SizeMode = PictureBoxSizeMode.StretchImage;
                    img.Size = new Size(64, 64);

                    if (imgX + 64 > lightmappanel.Width - sb)
                    {
                        imgX = 0;
                        imgY += 68;
                    }

                    img.Location = new Point(imgX, imgY);
                    img.Tag = i;
                    img.Click += new EventHandler(imgclick);

                    lightmappanel.Controls.Add(img);
                    imgX += 68;
                }

                if (tabPage3.Tag.ToString() == "0")
                {
                    foreach (Control box in tabPage3.Controls)
                        box.Enabled = true;
                }
                tabPage3.Tag = "1";
            }
            else
            {
                if (tabPage3.Tag.ToString() == "1")
                {
                    foreach (Control box in tabPage3.Controls)
                        box.Enabled = false;
                }
                tabPage3.Tag = "0";
            }
        }

        public void PopSpawnList(bool a, bool shortpaths, TreeView tv2, CheckedListBox clb)
        {
            string[] strings = Enum.GetNames(typeof(SpawnInfo.SpawnType));

            List<TreeNode>
            Cameras = new List<TreeNode>(),
            DeathZones = new List<TreeNode>(),
            Objectives = new List<TreeNode>(),
            Players = new List<TreeNode>();

            if (a) clb.Items.Clear();
            tv2.Nodes.Clear();

            tv2.Sorted = true;

            //go through the types
            foreach (string s in strings)
            {
                if (a) clb.Items.Add(s);

                TreeNode tn = new TreeNode();
                tn.Text = s.ToString();
                tn.Tag = -1;

                //go through all the spawns
                for (int i = 0; i < bspmodel.Spawns.Spawn.Count; i++)
                {
                    //if the spawns type = our current type
                    if (s == bspmodel.Spawns.Spawn[i].Type.ToString())
                    {
                        TreeNode tn2 = new TreeNode();
                        tn2.Tag = i;

                        //if spawntype = collection
                        if (bspmodel.Spawns.Spawn[i].Type.ToString() == "Collection")
                        {
                            SpawnInfo.Collection tempspawn = (SpawnInfo.Collection)bspmodel.Spawns.Spawn[i];
                            if (shortpaths) tn2.Text = tempspawn.TagPath.Substring(tempspawn.TagPath.LastIndexOf('\\') + 1);
                            else tn2.Text = tempspawn.TagPath;
                            tn.Nodes.Add(tn2);
                        }
                        else if (bspmodel.Spawns.Spawn[i].Type.ToString() == "Obstacle")
                        {
                            SpawnInfo.ObstacleSpawn tempspawn = (SpawnInfo.ObstacleSpawn)bspmodel.Spawns.Spawn[i];
                            if (shortpaths) tn2.Text = tempspawn.TagPath.Substring(tempspawn.TagPath.LastIndexOf('\\') + 1);
                            else tn2.Text = tempspawn.TagPath;
                            tn.Nodes.Add(tn2);
                        }
                        else if (bspmodel.Spawns.Spawn[i].Type.ToString() == "Weapon")
                        {
                            SpawnInfo.WeaponSpawn tempspawn = (SpawnInfo.WeaponSpawn)bspmodel.Spawns.Spawn[i];
                            if (shortpaths) tn2.Text = tempspawn.TagPath.Substring(tempspawn.TagPath.LastIndexOf('\\') + 1);
                            else tn2.Text = tempspawn.TagPath;
                            tn.Nodes.Add(tn2);
                        }
                        else if (bspmodel.Spawns.Spawn[i].Type.ToString() == "Scenery")
                        {
                            SpawnInfo.ScenerySpawn tempspawn = (SpawnInfo.ScenerySpawn)bspmodel.Spawns.Spawn[i];
                            if (shortpaths) tn2.Text = tempspawn.TagPath.Substring(tempspawn.TagPath.LastIndexOf('\\') + 1);
                            else tn2.Text = tempspawn.TagPath;
                            tn.Nodes.Add(tn2);
                        }
                        else if (bspmodel.Spawns.Spawn[i].Type.ToString() == "Machine")
                        {
                            SpawnInfo.MachineSpawn tempspawn = (SpawnInfo.MachineSpawn)bspmodel.Spawns.Spawn[i];
                            if (shortpaths) tn2.Text = tempspawn.TagPath.Substring(tempspawn.TagPath.LastIndexOf('\\') + 1);
                            else tn2.Text = tempspawn.TagPath;
                            tn.Nodes.Add(tn2);
                        }
                        else if (bspmodel.Spawns.Spawn[i].Type.ToString() == "Control")
                        {
                            SpawnInfo.ControlSpawn tempspawn = (SpawnInfo.ControlSpawn)bspmodel.Spawns.Spawn[i];
                            if (shortpaths) tn2.Text = tempspawn.TagPath.Substring(tempspawn.TagPath.LastIndexOf('\\') + 1);
                            else tn2.Text = tempspawn.TagPath;
                            tn.Nodes.Add(tn2);
                        }
                        else if (bspmodel.Spawns.Spawn[i].Type.ToString() == "Biped")
                        {
                            SpawnInfo.BipedSpawn tempspawn = (SpawnInfo.BipedSpawn)bspmodel.Spawns.Spawn[i];
                            if (shortpaths) tn2.Text = tempspawn.TagPath.Substring(tempspawn.TagPath.LastIndexOf('\\') + 1);
                            else tn2.Text = tempspawn.TagPath;
                            tn.Nodes.Add(tn2);
                        }
                        else if (bspmodel.Spawns.Spawn[i].Type.ToString() == "Equipment")
                        {
                            SpawnInfo.EquipmentSpawn tempspawn = (SpawnInfo.EquipmentSpawn)bspmodel.Spawns.Spawn[i];
                            if (shortpaths) tn2.Text = tempspawn.TagPath.Substring(tempspawn.TagPath.LastIndexOf('\\') + 1);
                            else tn2.Text = tempspawn.TagPath;
                            tn.Nodes.Add(tn2);
                        }
                        else if (bspmodel.Spawns.Spawn[i].Type.ToString() == "Vehicle")
                        {
                            SpawnInfo.VehicleSpawn tempspawn = (SpawnInfo.VehicleSpawn)bspmodel.Spawns.Spawn[i];
                            if (shortpaths) tn2.Text = tempspawn.TagPath.Substring(tempspawn.TagPath.LastIndexOf('\\') + 1);
                            else tn2.Text = tempspawn.TagPath;
                            tn.Nodes.Add(tn2);
                        }
                        else if (bspmodel.Spawns.Spawn[i].Type.ToString() == "Light")
                        {
                            SpawnInfo.LightSpawn tempspawn = (SpawnInfo.LightSpawn)bspmodel.Spawns.Spawn[i];
                            if (shortpaths) tn2.Text = tempspawn.TagPath.Substring(tempspawn.TagPath.LastIndexOf('\\') + 1);
                            else tn2.Text = tempspawn.TagPath;
                            tn.Nodes.Add(tn2);
                        }
                        else if (bspmodel.Spawns.Spawn[i].Type.ToString() == "Camera")
                        {
                            tn2.Text = bspmodel.Spawns.Spawn[i].Type.ToString() + " " + Cameras.Count.ToString();
                            Cameras.Add(tn2);
                        }
                        else if (bspmodel.Spawns.Spawn[i].Type.ToString() == "DeathZone")
                        {
                            tn2.Text = bspmodel.Spawns.Spawn[i].Type.ToString() + " " + DeathZones.Count.ToString();
                            DeathZones.Add(tn2);
                        }
                        else if (bspmodel.Spawns.Spawn[i].Type.ToString() == "Objective")
                        {
                            tn2.Text = bspmodel.Spawns.Spawn[i].Type.ToString() + " " + Objectives.Count.ToString();
                            Objectives.Add(tn2);
                        }
                        else if (bspmodel.Spawns.Spawn[i].Type.ToString() == "Player")
                        {
                            tn2.Text = bspmodel.Spawns.Spawn[i].Type.ToString() + " " + Players.Count.ToString();
                            Players.Add(tn2);
                        }
                        else tn2.Text = bspmodel.Spawns.Spawn[i].Type.ToString();
                    }
                }
                tv2.Nodes.Add(tn);
            }

            tv2.Sorted = false;

            for (int x = 0; x < tv2.Nodes.Count; x++)
            {
                if (tv2.Nodes[x].Text == "Camera")
                    tv2.Nodes[x].Nodes.AddRange(Cameras.ToArray());

                else if (tv2.Nodes[x].Text == "DeathZone")
                    tv2.Nodes[x].Nodes.AddRange(DeathZones.ToArray());

                else if (tv2.Nodes[x].Text == "Objective")
                    tv2.Nodes[x].Nodes.AddRange(Objectives.ToArray());

                else if (tv2.Nodes[x].Text == "Player")
                    tv2.Nodes[x].Nodes.AddRange(Players.ToArray());
            }
        }

        public void MakeMatrixes()
        {
            TranslationMatrix = new Matrix[bspmodel.Spawns.Spawn.Count];
            for (int x = 0; x < bspmodel.Spawns.Spawn.Count; x++)
                TranslationMatrix[x] = MakeMatrixForSpawn(x);
        }

        public Matrix MakeMatrixForSpawn(int x)
        {
            Matrix m = Matrix.Identity;

            //SpawnWithOneRotationDirection
            if (bspmodel.Spawns.Spawn[x] is SpawnInfo.RotateDirectionBaseSpawn)
            {
                SpawnInfo.RotateDirectionBaseSpawn tempspawn;
                tempspawn = bspmodel.Spawns.Spawn[x] as SpawnInfo.RotateDirectionBaseSpawn;
                Matrix rotate = Matrix.Identity;
                rotate.RotateYawPitchRoll(0, 0, tempspawn.RotationDirection);
                m.Multiply(rotate);
            }

            //SpawnWithYawPitchRoll
            else if (bspmodel.Spawns.Spawn[x] is SpawnInfo.RotateYawPitchRollBaseSpawn)
            {
                SpawnInfo.RotateYawPitchRollBaseSpawn tempspawn;
                tempspawn = bspmodel.Spawns.Spawn[x] as SpawnInfo.RotateYawPitchRollBaseSpawn;
                Matrix rotate = Matrix.Identity;

                Matrix m1 = Matrix.Identity;
                m1.RotateX(tempspawn.Yaw);

                Matrix m2 = Matrix.Identity;
                m2.RotateY(tempspawn.Pitch);

                Matrix m3 = Matrix.Identity;
                m3.RotateZ(tempspawn.Roll);

                rotate = m1 * m2 * m3;
                m = rotate;
            }

            m.Multiply(
                Matrix.Translation(bspmodel.Spawns.Spawn[x].X, bspmodel.Spawns.Spawn[x].Y,
                bspmodel.Spawns.Spawn[x].Z));

            return m;
        }

        public void LoadSpawns(int mapnum)
        {
            SpawnModel = new List<ParsedModel>();
            spawnmodelindex = new int[bspmodel.Spawns.Spawn.Count];
            BoundingBoxModel = new Mesh[bspmodel.Spawns.Spawn.Count];

            #region Spawns
            for (int x = 0; x < bspmodel.Spawns.Spawn.Count; x++)
            {
                BinaryReader br = new BinaryReader(new FileStream(Map.mp[mapnum].mapLoc,
                   FileMode.Open, FileAccess.Read, FileShare.ReadWrite));

                // BoundingBoxOnlySpawn
                if (bspmodel.Spawns.Spawn[x] is SpawnInfo.BoundingBoxSpawn)
                {
                    SpawnInfo.BoundingBoxSpawn tempbox;
                    tempbox = bspmodel.Spawns.Spawn[x] as SpawnInfo.BoundingBoxSpawn;
                    BoundingBoxModel[x] = Mesh.Box(Map.dev, tempbox.width, tempbox.height, tempbox.length);
                    continue;
                }

                // CameraSpawn
                if (bspmodel.Spawns.Spawn[x] is SpawnInfo.CameraSpawn)
                {
                    SpawnInfo.CameraSpawn tempbox;
                    tempbox = bspmodel.Spawns.Spawn[x] as SpawnInfo.CameraSpawn;
                    BoundingBoxModel[x] = Mesh.Cylinder(Map.dev, 0.1f, 2.0f, 10f, 10, 10);
                    continue;
                }

                // LightSpawn
                if (bspmodel.Spawns.Spawn[x] is SpawnInfo.LightSpawn)
                {
                    SpawnInfo.LightSpawn tempbox;
                    tempbox = bspmodel.Spawns.Spawn[x] as SpawnInfo.LightSpawn;
                    tempbox.LightInfo = new HaloLight(tempbox.ModelTagNumber, Map.mp[mapnum].mapLoc, Map.mp[mapnum].tags.tagInfo);
                    bspmodel.Spawns.Spawn[x] = tempbox;
                    BoundingBoxModel[x] = Mesh.Cylinder(Map.dev, 0.5f, 0.0f, 1f, 10, 10);
                    if (lighting == true)
                    {
                        Map.dev.Lights[HaloLightCount].Type = LightType.Point;
                        Map.dev.Lights[HaloLightCount].Position = new Vector3(tempbox.X, tempbox.Y, -tempbox.Z);
                       Map.dev.Lights[HaloLightCount].Direction = new Vector3(-tempbox.Yaw, -tempbox.Pitch, -tempbox.Roll);
                       Map.dev.Lights[HaloLightCount].Range = 10f;
                        //  Map.Map.dev.Lights[HaloLightCount].=0.5f;
                        //  Map.Map.dev.Lights[HaloLightCount].p = 1.0f;
                       Map.dev.Lights[HaloLightCount].Falloff = 1.0f;
                        Map.dev.Lights[HaloLightCount].Attenuation0 = 1.0f;
                        Map.dev.Lights[HaloLightCount].Diffuse = Color.FromArgb(tempbox.LightInfo.r, tempbox.LightInfo.g, tempbox.LightInfo.b);
                        Map.dev.Lights[HaloLightCount].Update();
                       Map.dev.Lights[HaloLightCount].Enabled = true;
                    }
                    HaloLightCount++;
                    continue;
                }

                SpawnInfo.RotationSpawn tempspawn = bspmodel.Spawns.Spawn[x] as SpawnInfo.RotationSpawn;

                // ScanForExistingModels
                bool found = false;
                for (int y = 0; y < x; y++)
                {
                    SpawnInfo.RotationSpawn tempspawn2 = bspmodel.Spawns.Spawn[y] as SpawnInfo.RotationSpawn;
                    if (bspmodel.Spawns.Spawn[y] is SpawnInfo.BoundingBoxSpawn)
                        continue;

                    if (tempspawn.ModelTagNumber == tempspawn2.ModelTagNumber)
                    {
                        BoundingBoxModel[x] = BoundingBoxModel[y];
                        spawnmodelindex[x] = spawnmodelindex[y];
                        found = true;
                        break;
                    }
                }

                if (found) continue;

                // ReadSpawnMeta
                Meta m = new Meta(tempspawn.ModelTagNumber);
                //if (tempspawn.ModelTagNumber == -1) MessageBox.Show("Test");
                m.ReadMetaFromMap(false, true, false, mapnum);

                // DirectXModel
                ParsedModel pm = new ParsedModel(ref m, mapnum);
                //pm.PermutationString=pm.hlmt.Permutations[pm.hlmt.FindPermutationByBaseClass,
                ParsedModel.DisplayedInfo.LoadDirectXTexturesAndBuffers(ref pm);
                SpawnModel.Add(pm);
                spawnmodelindex[x] = SpawnModel.Count - 1;
                m = null;

                // BoundingBox
                float boxwidth = pm.BoundingBox.MaxX - pm.BoundingBox.MinX;
                float boxheight = pm.BoundingBox.MaxY - pm.BoundingBox.MinY;
                float boxdepth = pm.BoundingBox.MaxZ - pm.BoundingBox.MinZ;
                BoundingBoxModel[x] = Mesh.Box(Map.dev, boxwidth, boxheight, boxdepth);

                // Fog
                if (Map.dev.DeviceCaps.RasterCaps.SupportsFogTable &&
                    bspmodel.sky.fogenabled && bspmodel.sky.fog.FogThickness != 0)
                {
                    int a = (int)(bspmodel.sky.fog.A * 255);
                    int r = (int)(bspmodel.sky.fog.R * 255);
                    int g = (int)(bspmodel.sky.fog.G * 255);
                    int b = (int)(bspmodel.sky.fog.B * 255);
                    Map.dev.RenderState.FogColor = Color.FromArgb(a, r, g, b);
                    Map.dev.RenderState.FogStart = bspmodel.sky.fog.Start;
                   Map.dev.RenderState.FogEnd = bspmodel.sky.fog.End;
                   Map.dev.RenderState.FogDensity = bspmodel.sky.fog.FogThickness;

                    Map.dev.RenderState.FogTableMode = FogMode.Linear;
                   Map.dev.RenderState.FogEnable = true;
                }
               Map.dev.RenderState.FogVertexMode = FogMode.Linear; //mayb?
                br.Close();
            }
            #endregion

            //scenery
            //for (int x = 0; x < bspmodel.BSPPermutationRawDataMetaChunks.Length; x++)
            //{
            //    BSPModel.BSPPermutationRawDataMetaChunk mdl =
            //        bspmodel.BSPPermutationRawDataMetaChunks[x];

                //mesh
            //    List<short> ts = new List<short>();
            //    for (int z = 0; z < mdl.Indices.Length; z++)
            //        ts.Add(mdl.Indices[z]);
            //
            //    mdl.mesh = ParsedModel.DisplayedInfo.MakeMesh(mdl.Vertices, ts, mdl.UVs, ref Map.dev);
            //}
        }

        public void SetUpCameraForBSP()
        {
            Map.dev.Transform.World = Matrix.Identity;
            Map.dev.Transform.View = Matrix.LookAtRH(cam.Position, cam.LookAt, cam.UpVector);
            Map.dev.Transform.Projection = Matrix.PerspectiveFovRH(0.7f, 1.0f, 1.0f, 1000.0f);//(float)Math.PI / 4, 1.0f, 0.1f, 10000.0f);
        }

        public void update(ref DI.Device keyb)
        {
            ReadUserInput(keyb);

            Map.dev.Transform.View = Matrix.LookAtRH(cam.Position, cam.LookAt, cam.UpVector);
            Map.dev.Transform.Projection = Matrix.PerspectiveFovRH(0.7f, 1.0f, 1.0f, 1000.0f);
        }

        public void ReadUserInput(DI.Device keyb)
        {
            try
            { keyb.Acquire(); }
            catch { return; }

            DI.Key[] list = keyb.GetPressedKeys();
            if (list.Length > 0)
            {
                foreach (DI.Key ky in list)
                {
                    if (ky == DI.Key.W)
                    {
                        cam.x += cam.i * cam.speed;
                        cam.y += cam.j * cam.speed;
                        cam.z += cam.k * cam.speed;
                        cam.Position.X = cam.x;
                        cam.Position.Y = cam.y;
                        cam.Position.Z = cam.z;
                    }
                    else if (ky == DI.Key.S)
                    {
                        cam.x -= cam.i * cam.speed;
                        cam.y -= cam.j * cam.speed;
                        cam.z -= cam.k * cam.speed;
                        cam.Position.X = cam.x;
                        cam.Position.Y = cam.y;
                        cam.Position.Z = cam.z;
                    }
                    else if (ky == DI.Key.D)
                        cam.ComputeStrafe(true);

                    else if (ky == DI.Key.A)
                        cam.ComputeStrafe(false);

                    else if (ky == DI.Key.Z)
                    {
                        cam.z -= cam.speed;
                        cam.Position.Z = cam.z;
                    }
                    else if (ky == DI.Key.X)
                    {
                        cam.z += cam.speed;
                        cam.Position.Z = cam.z;
                    }
                    else if (ky == DI.Key.Equals || ky == DI.Key.Add)
                        cam.speed += 0.01f;

                    else if (ky == DI.Key.Minus || ky == DI.Key.NumPadMinus)
                    {
                        cam.speed -= 0.01f;
                        if (cam.speed < 0) cam.speed = 0.01f;
                    }
                }
                cam.ComputePosition();

                //move selected spawns
                for (int x = 0; x < bspmodel.Spawns.Spawn.Count; x++)
                {
                    if (!bspmodel.Spawns.Spawn[x].IsSelected) continue;
                    foreach (DI.Key ky in list)
                    {
                        if (ky == DI.Key.Up)
                            bspmodel.Spawns.Spawn[x].X += cam.speed;

                        else if (ky == DI.Key.DownArrow)
                            bspmodel.Spawns.Spawn[x].X -= cam.speed;

                        else if (ky == DI.Key.LeftArrow)
                            bspmodel.Spawns.Spawn[x].Y += cam.speed;

                        else if (ky == DI.Key.Right)
                            bspmodel.Spawns.Spawn[x].Y -= cam.speed;

                        else if (ky == DI.Key.PageDown)
                            bspmodel.Spawns.Spawn[x].Z -= cam.speed;

                        else if (ky == DI.Key.PageUp)
                            bspmodel.Spawns.Spawn[x].Z += cam.speed;

                        else
                        {
                            if (!pastList.Contains(ky))
                            {
                                if (ky == DI.Key.LeftControl || ky == DI.Key.RightControl)
                                    YPR[0] = !YPR[0];

                                else if (ky == DI.Key.LeftShift || ky == DI.Key.RightShift)
                                    YPR[1] = !YPR[1];

                                else if (ky == DI.Key.RightMenu || ky == DI.Key.LeftMenu)
                                    YPR[2] = !YPR[2];
                            }
                        }
                        TranslationMatrix[SelectedSpawn[x]] = MakeMatrixForSpawn(SelectedSpawn[x]);
                    }
                }
                for (int x = 0; x < bspmodel.PermutationInfo.Length; x++)
                {
                    if (!bspmodel.PermutationInfo[x].IsSelected) continue;
                    Matrix mat = bspmodel.PermutationInfo[x].mat;

                    foreach (DI.Key ky in list)
                    {
                        int s = bspmodel.PermutationInfo[x].sceneryIndex;
                        if (ky == DI.Key.Up)
                        {
                            bspmodel.PermutationInfo[x].x += 0.5f;
                            //bspmodel.BSPPermutationRawDataMetaChunks[s].BoundingBox.MinX += 0.5f;
                            //bspmodel.BSPPermutationRawDataMetaChunks[s].BoundingBox.MaxX += 0.5f;
                        }

                        else if (ky == DI.Key.DownArrow)
                        {
                            bspmodel.PermutationInfo[x].x -= 0.5f;
                            //bspmodel.BSPPermutationRawDataMetaChunks[s].BoundingBox.MinX -= 0.5f;
                            //bspmodel.BSPPermutationRawDataMetaChunks[s].BoundingBox.MaxX -= 0.5f;
                        }
                        else if (ky == DI.Key.LeftArrow)
                        {
                            bspmodel.PermutationInfo[x].y += 0.5f;
                            //bspmodel.BSPPermutationRawDataMetaChunks[s].BoundingBox.MinY += 0.5f;
                            //bspmodel.BSPPermutationRawDataMetaChunks[s].BoundingBox.MaxY += 0.5f;
                        }
                        else if (ky == DI.Key.Right)
                        {
                            bspmodel.PermutationInfo[x].y -= 0.5f;
                            //bspmodel.BSPPermutationRawDataMetaChunks[s].BoundingBox.MinY -= 0.5f;
                            //bspmodel.BSPPermutationRawDataMetaChunks[s].BoundingBox.MaxY -= 0.5f;
                        }
                        else if (ky == DI.Key.PageDown)
                        {
                            bspmodel.PermutationInfo[x].z += 0.5f;
                           //bspmodel.BSPPermutationRawDataMetaChunks[s].BoundingBox.MinZ += 0.5f;
                            //bspmodel.BSPPermutationRawDataMetaChunks[s].BoundingBox.MaxZ += 0.5f;
                        }
                        else if (ky == DI.Key.PageUp)
                        {
                            bspmodel.PermutationInfo[x].z -= 0.5f;
                            //bspmodel.BSPPermutationRawDataMetaChunks[s].BoundingBox.MinZ -= 0.5f;
                            //bspmodel.BSPPermutationRawDataMetaChunks[s].BoundingBox.MaxZ -= 0.5f;
                        }
                    }

                    mat.M41 = bspmodel.PermutationInfo[x].x;
                    mat.M42 = bspmodel.PermutationInfo[x].y;
                    mat.M43 = bspmodel.PermutationInfo[x].z;
                   // mat.Translate(bspmodel.PermutationInfo[x].x, 
                    //    bspmodel.PermutationInfo[x].y, bspmodel.PermutationInfo[x].z);

                    bspmodel.PermutationInfo[x].mat = mat;
                }
            }

            pastList.Clear();
            if (list.Length > 0)
                for (int x = 0; x < list.Length; x++)
                    pastList.Add(list[x]);
        }

        public void MouseMove(MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
                cam.change(e.X, e.Y);
            else
            {
                #region MultiSelect
                if (selectionMulti == true)
                {
                    Vector3 tempvec = Mark3DCursorPosition(e.X, e.Y, Matrix.Identity);
                    selectionWidth = tempvec.X - selectionStart.X;
                    selectionHeight = tempvec.Y - selectionStart.Y;
                    selectionDepth = tempvec.Z - selectionStart.Z;
                    selectionWidth *= 2;
                    selectionHeight *= 7;
                    selectionDepth *= 2;
                    float tempselectionWidth = selectionWidth;
                    float tempselectionHeight = selectionHeight;
                    float tempselectionDepth = selectionDepth;
                    if (tempselectionWidth < 0) tempselectionWidth = -tempselectionWidth;
                    if (tempselectionHeight < 0) tempselectionHeight = -tempselectionHeight;
                    if (tempselectionDepth < 0) tempselectionDepth = -tempselectionDepth;
                    //stat.Text = selectionStart.X.ToString() + "  " + selectionStart.Y.ToString() + selectionStart.Z.ToString() + "  " + selectionWidth.ToString() + "  " + selectionHeight.ToString() + "  " + selectionDepth.ToString();
                    selectionMesh = Mesh.Box(Map.dev, tempselectionWidth, tempselectionHeight, tempselectionDepth);
                }
                #endregion

                #region SelectedSpawnRotation
                if (itemrotate)
                {
                    Vector3 tempvec = Mark3DCursorPosition(e.X, e.Y, Matrix.Identity);
                    selectionWidth = tempvec.X - selectionStart.X;
                    selectionHeight = tempvec.Y - selectionStart.Y;
                    selectionDepth = tempvec.Z - selectionStart.Z;
                    selectionStart = tempvec;

                    foreach (int spawnid in SelectedSpawn)
                    {
                        if (bspmodel.Spawns.Spawn[spawnid] is SpawnInfo.BoundingBoxSpawn) continue;
                        #region ControlAndRotate

                        if (YPR[0])
                        {
                            if (bspmodel.Spawns.Spawn[spawnid] is SpawnInfo.RotateYawPitchRollBaseSpawn)
                            {
                                SpawnInfo.RotateYawPitchRollBaseSpawn temp;
                                temp = (SpawnInfo.RotateYawPitchRollBaseSpawn)bspmodel.Spawns.Spawn[spawnid];
                                temp.Yaw += selectionHeight * cam.speed;
                                if (temp.Yaw > (float)Math.PI) temp.Yaw = -(float)Math.PI;
                                else if (temp.Yaw < (float)-Math.PI) temp.Yaw = (float)Math.PI;
                            }
                            else
                            {
                                SpawnInfo.RotateDirectionBaseSpawn temp;
                                temp = (SpawnInfo.RotateDirectionBaseSpawn)bspmodel.Spawns.Spawn[spawnid];
                                temp.RotationDirection += selectionHeight * cam.speed;
                                //  if (temp.RotationDirection > (float)Math.PI) { temp.RotationDirection = temp.RotationDirection - ((float)Math.PI * 2); }
                                //  else if (temp.RotationDirection < (float)-Math.PI) { temp.RotationDirection = temp.RotationDirection + ((float)Math.PI * 2); }
                            }

                        }
                        #endregion
                        #region ShiftAndRotate
                        if (YPR[1])
                        {
                            if (bspmodel.Spawns.Spawn[spawnid] is SpawnInfo.RotateYawPitchRollBaseSpawn)
                            {
                                SpawnInfo.RotateYawPitchRollBaseSpawn temp;
                                temp = (SpawnInfo.RotateYawPitchRollBaseSpawn)bspmodel.Spawns.Spawn[spawnid];
                                temp.Pitch += selectionWidth * cam.speed; ;
                                if (temp.Pitch > (float)Math.PI) temp.Pitch = -(float)Math.PI;
                                else if (temp.Pitch < (float)-Math.PI) temp.Pitch = (float)Math.PI;
                            }
                        }
                        #endregion
                        #region AltAndRotate
                        if (YPR[2])
                        {
                            if (bspmodel.Spawns.Spawn[spawnid] is SpawnInfo.RotateYawPitchRollBaseSpawn)
                            {
                                SpawnInfo.RotateYawPitchRollBaseSpawn temp;
                                temp = (SpawnInfo.RotateYawPitchRollBaseSpawn)bspmodel.Spawns.Spawn[spawnid];
                                temp.Roll += selectionDepth * cam.speed;
                                if (temp.Roll > (float)Math.PI) temp.Roll = -(float)Math.PI;
                                else if (temp.Roll < (float)-Math.PI) temp.Roll = (float)Math.PI;
                            }
                        }
                        #endregion
                        TranslationMatrix[spawnid] = MakeMatrixForSpawn(spawnid);
                    }
                }
                #endregion
            }
        }

        public void MouseUp(MouseEventArgs e)
        {
            itemrotate = false;
        }

        public void MouseDown(MouseEventArgs e)
        {
            cam.oldx = e.X;
            cam.oldy = e.Y;

            LookforSelection(e);
        }

        public void LookforSelection(MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Middle)
            {
                selectionHeight = selectionWidth = selectionDepth = 0;
                selectionMulti = true;
                selectionStart = Mark3DCursorPosition(e.X, e.Y, Matrix.Identity);
            }
            else if (e.Button == MouseButtons.Left)
            {
                for (int x = 0; x < bspmodel.Spawns.Spawn.Count; x++)
                {
                    //check bitmask for visibility
                    if (((int)bspmodel.Spawns.Spawn[x].Type & visibleSpawnsBitMask) == 0) continue;

                    int tempcount = SpawnModel[spawnmodelindex[x]].Display.Chunk.Count;
                    bool useboundingbox = false;
                    switch (bspmodel.Spawns.Spawn[x].Type)
                    {
                        case SpawnInfo.SpawnType.Camera:
                            tempcount = 1; useboundingbox = true;
                            break;
                        case SpawnInfo.SpawnType.Light:
                            tempcount = 1; useboundingbox = true;
                            break;
                    }
                    for (int yy = 0; yy < tempcount; yy++)
                    {
                        //check for mesh intersection
                        Mesh tempm;
                        //MessageBox.Show(bsp.Display.meshes.Length.ToString());
                        if (!useboundingbox) tempm = SpawnModel[spawnmodelindex[x]].Display.meshes[yy];
                        else tempm = BoundingBoxModel[x];

                        //if true, there is an intersection
                        if (MeshPick(e.X, e.Y, tempm, TranslationMatrix[x]))
                        {
                            //look for the spawn in the list of selected spawns
                            //int tempi = SelectedSpawn.IndexOf(x);
                            //and if it is found, remove it
                            //if (tempi != -1) SelectedSpawn.RemoveAt(tempi);
                            //else
                            //{
                            //    SelectedSpawn.Add(x);
                            bspmodel.Spawns.Spawn[x].IsSelected = !bspmodel.Spawns.Spawn[x].IsSelected;
                                selectedSpawnType = bspmodel.Spawns.Spawn[x].Type;

                                SelectedSpawnStruct s = new SelectedSpawnStruct();
                                s.Type = bspmodel.Spawns.Spawn[x].Type.ToString();
                                s.X = bspmodel.Spawns.Spawn[x].X;
                                s.Y = bspmodel.Spawns.Spawn[x].Y;
                                s.Z = bspmodel.Spawns.Spawn[x].Z;

                                if (bspmodel.Spawns.Spawn[x] is SpawnInfo.RotateYawPitchRollBaseSpawn)
                                {
                                    SpawnInfo.RotateYawPitchRollBaseSpawn temprot;
                                    temprot = bspmodel.Spawns.Spawn[x] as SpawnInfo.RotateYawPitchRollBaseSpawn;
                                    s.Yaw = temprot.Yaw;
                                    s.Pitch = temprot.Pitch;
                                    s.Roll = temprot.Roll;

                                    s.ModelName = temprot.ModelName;
                                    s.Positioning = 0;
                                }
                                else if (bspmodel.Spawns.Spawn[x] is SpawnInfo.RotateDirectionBaseSpawn)
                                {
                                    SpawnInfo.RotateDirectionBaseSpawn temprot;
                                    temprot = bspmodel.Spawns.Spawn[x] as SpawnInfo.RotateDirectionBaseSpawn;
                                    s.Direction = temprot.RotationDirection;

                                    s.ModelName = temprot.ModelName;
                                    s.Positioning = 1;
                                }

                                if (bspmodel.Spawns.Spawn[x] is SpawnInfo.ObjectiveSpawn)
                                {
                                    SpawnInfo.ObjectiveSpawn os = bspmodel.Spawns.Spawn[x] as SpawnInfo.ObjectiveSpawn;
                                    s.BaseSpawn = 0;
                                    s.ObjColType = os.ObjectiveType.ToString();

                                    s.ObjTeam = os.Team.ToString();
                                    s.ObjNumber = os.number;
                                }
                                else if (bspmodel.Spawns.Spawn[x] is SpawnInfo.Collection)
                                {
                                    SpawnInfo.Collection os = bspmodel.Spawns.Spawn[x] as SpawnInfo.Collection;
                                    s.BaseSpawn = 1;
                                    s.ObjColType = os.SpawnsInMode.ToString();
                                }
                                //AddToSelectedSpawnsCombo(s);
                                break;
                            //}
                        }
                    }
                }
                for (int x = 0; x < bspmodel.PermutationInfo.Length; x++)
                {
                    if (MeshPick(e.X, e.Y, bspmodel.PermutationInfo[x].mesh, 
                        bspmodel.PermutationInfo[x].mat))
                    {
                        bspmodel.PermutationInfo[x].IsSelected = !bspmodel.PermutationInfo[x].IsSelected;

                        //MessageBox.Show(temprubbish(bspmodel.PermutationInfo[x]));




                        //int rawindex = bspmodel.PermutationInfo[x].sceneryIndex;
                        //bspmodel.BSPPermutationRawDataMetaChunks[rawindex].IsSelected =
                        //      !bspmodel.BSPPermutationRawDataMetaChunks[rawindex].IsSelected;
                    }

                    //int rawindex = bspmodel.PermutationInfo[x].sceneryIndex;
                    //if (bspmodel.BSPPermutationRawDataMetaChunks[rawindex].RawDataChunkInfo.Length == 0)
                    //    continue;

                    //for (int xx = 0; xx < bspmodel.BSPPermutationRawDataMetaChunks[rawindex].SubMeshInfo.Length; xx++)
                    //{
                    //    if (MeshPick(e.X, e.Y, bspmodel.BSPPermutationRawDataMetaChunks[rawindex].SubMeshInfo[xx].mesh, bspmodel.PermutationInfo[x].mat, Map.dev))
                    //        bspmodel.BSPPermutationRawDataMetaChunks[rawindex].IsSelected =
                    //           !bspmodel.BSPPermutationRawDataMetaChunks[rawindex].IsSelected;
                    //}
                }

                //for (int x = 0; x < bspmodel.BSPPermutationRawDataMetaChunks.Length; x++)
                //{
                //    BSPModel.BSPPermutationRawDataMetaChunk mdl = bspmodel.BSPPermutationRawDataMetaChunks[x];
                //    if (MeshPick(e.X, e.Y, mdl.mesh, TranslationMatrix[x], Map.dev) == true)
                //        bspmodel.BSPPermutationRawDataMetaChunks[x].IsSelected = 
                //           !bspmodel.BSPPermutationRawDataMetaChunks[x].IsSelected;
                //}
            }
        }

        public string temprubbish(BSPModel.PermutationPlacement pp)
        {
            Matrix mm = pp.mat;
            BSPModel.BSPPermutationRawDataMetaChunk.BSPPermutationBoundingBoxContainer bx =
                bspmodel.BSPPermutationRawDataMetaChunks[pp.sceneryIndex].BoundingBox;

            return
                "m11: " + mm.M11.ToString() + Environment.NewLine +
                "m12: " + mm.M12.ToString() + Environment.NewLine +
                "m13: " + mm.M13.ToString() + Environment.NewLine +
                "m14: " + mm.M14.ToString() + Environment.NewLine +
                "m21: " + mm.M21.ToString() + Environment.NewLine +
                "m22: " + mm.M22.ToString() + Environment.NewLine +
                "m23: " + mm.M23.ToString() + Environment.NewLine +
                "m24: " + mm.M24.ToString() + Environment.NewLine +
                "m31: " + mm.M31.ToString() + Environment.NewLine +
                "m32: " + mm.M32.ToString() + Environment.NewLine +
                "m33: " + mm.M33.ToString() + Environment.NewLine +
                "m34: " + mm.M34.ToString() + Environment.NewLine +
                "m41: " + mm.M41.ToString() + Environment.NewLine +
                "m42: " + mm.M42.ToString() + Environment.NewLine +
                "m43: " + mm.M43.ToString() + Environment.NewLine +
                "m44: " + mm.M44.ToString() + Environment.NewLine + Environment.NewLine +

                "minX " + bx.MinX.ToString() + Environment.NewLine +
                "maxX " + bx.MaxX.ToString() + Environment.NewLine + Environment.NewLine +

                "minY " + bx.MinY.ToString() + Environment.NewLine +
                "maxY " + bx.MaxY.ToString() + Environment.NewLine + Environment.NewLine +

                "minZ " + bx.MinZ.ToString() + Environment.NewLine +
                "maxZ " + bx.MaxZ.ToString() + Environment.NewLine + Environment.NewLine +

                "minU " + bx.MinU.ToString() + Environment.NewLine +
                "maxU " + bx.MaxU.ToString() + Environment.NewLine + Environment.NewLine +

                "minV " + bx.MinV.ToString() + Environment.NewLine +
                "maxV " + bx.MaxV.ToString();

        }

        public bool MeshPick(float x, float y, Mesh mesh, Matrix mat)
        {
            Vector3 posv3 = new Vector3(x, y, 0);
            Viewport vp = Map.dev.Viewport;
            Matrix proj = Matrix.PerspectiveFovRH(0.7f, 1.0f, 1.0f, 1000.0f);
            Matrix view = Matrix.LookAtRH(cam.Position, cam.LookAt, cam.UpVector);
            Vector3 s = Vector3.Unproject(posv3, vp, proj, view, mat);

            Vector3 d = Vector3.Unproject(new Vector3(x, y, 1),
                 vp, proj, view, mat);

            Vector3 rPosition = s;
            Vector3 rDirection = Vector3.Normalize(d - s);

            return mesh.Intersect(rPosition, rDirection);
        }

        public Vector3 Mark3DCursorPosition(float x, float y, Matrix m)
        {
            Vector3 tempV3 = new Vector3();

            Viewport vp = Map.dev.Viewport;
            Matrix proj = Matrix.PerspectiveFovRH(0.7f, 1.0f, 1.0f, 1000.0f);
            Matrix view = Matrix.LookAtRH(cam.Position, cam.LookAt, cam.UpVector);

            tempV3.Project(vp, proj, view, m);

            tempV3.X = x;
            tempV3.Y = y;
            tempV3.Unproject(vp, proj, view, m);

            return tempV3;
        }

        public void Draw(int mapnum)
        {
            Map.dev.Transform.World = Matrix.Identity;
            Map.dev.SamplerState[0].AddressU = Map.dev.SamplerState[0].AddressV = TextureAddress.Mirror;

            BSPModel.BSPDisplayedInfo.Draw(ref bspmodel, FillMode.Solid, ref cam);//, shaderx);

           if (bspmodel.renderSkybox)
            ParsedModel.DisplayedInfo.Draw(bspmodel.SkyBox, mapnum);

            //render spawns
            for (int x = 0; x < bspmodel.Spawns.Spawn.Count; x++)
            {
                //if the spawn is not visible then continue on to the next loop
                if (((int)bspmodel.Spawns.Spawn[x].Type & visibleSpawnsBitMask) == 0) continue;

                Map.dev.Transform.World = TranslationMatrix[x];
             
                //if the spawn is selected draw a red wire box
                if (bspmodel.Spawns.Spawn[x].IsSelected)
                {
                    Map.dev.Material = BoundingBoxSelected;
                    Map.dev.SetTexture(0, null);
                    Map.dev.RenderState.AlphaBlendEnable = false;
                    Map.dev.RenderState.AlphaTestEnable = false;
                    Map.dev.RenderState.FillMode = FillMode.WireFrame;
                    BoundingBoxModel[x].DrawSubset(0);
                }

                //draw light
                if (bspmodel.Spawns.Spawn[x] is SpawnInfo.LightSpawn)
                {
                    if (!bspmodel.Spawns.Spawn[x].IsSelected)
                    {
                        Map.dev.Material = BlueMaterial;
                        Map.dev.SetTexture(0, null);
                        Map.dev.RenderState.AlphaBlendEnable = false;
                        Map.dev.RenderState.AlphaTestEnable = false;
                        Map.dev.RenderState.FillMode = FillMode.WireFrame;
                        BoundingBoxModel[x].DrawSubset(0);
                    }
                    continue;
                }

                //draw camera
                if (bspmodel.Spawns.Spawn[x] is SpawnInfo.CameraSpawn)
                {
                    if (!bspmodel.Spawns.Spawn[x].IsSelected)
                    {
                        Map.dev.Material = BlueMaterial;
                        Map.dev.SetTexture(0, null);
                        Map.dev.RenderState.AlphaBlendEnable = false;
                        Map.dev.RenderState.AlphaTestEnable = false;
                        Map.dev.RenderState.FillMode = FillMode.WireFrame;
                        BoundingBoxModel[x].DrawSubset(0);
                    }
                    continue;
                }

                //BoundingBox
                if (bspmodel.Spawns.Spawn[x] is SpawnInfo.BoundingBoxSpawn)
                    continue;

                Map.dev.Material = DefaultMaterial;
                Map.dev.RenderState.FillMode = FillMode.Solid;

                //if its an objective spawn set its colour depending on its side
                if (bspmodel.Spawns.Spawn[x] is SpawnInfo.ObjectiveSpawn)
                {
                    SpawnInfo.ObjectiveSpawn os = bspmodel.Spawns.Spawn[x] as SpawnInfo.ObjectiveSpawn;
                    switch (os.Team)
                    {
                        case SpawnInfo.ObjectiveSpawn.TeamType.Defense:
                            Map.dev.Material = BoundingBoxSelected;
                            break;
                        case SpawnInfo.ObjectiveSpawn.TeamType.Offense:
                            Map.dev.Material = BlueMaterial;
                            break;
                        case SpawnInfo.ObjectiveSpawn.TeamType.Neutral:
                            Map.dev.Material = GreenMaterial;
                            break;
                    }
                }

                //draw the spawn
                ParsedModel.DisplayedInfo.Draw(SpawnModel[spawnmodelindex[x]], mapnum);
            }
        }
    }
}
