using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace Visual_Scenery_Mover
{
    class TriangleRender
    {
        float triAngle = 0f;
        VertexBuffer vb;

        public TriangleRender()
        {
            vb = new VertexBuffer(typeof(CustomVertex.PositionColored), 3,
                Map.dev, 0, CustomVertex.PositionColored.Format, Pool.Default);

            vb.Created += new System.EventHandler(this.OnCreateVertexBuffer);
            this.OnCreateVertexBuffer(null, null);
        }

        public void OnCreateVertexBuffer(object sender, EventArgs e)
        {
            CustomVertex.PositionColored[] verts = (CustomVertex.PositionColored[])vb.Lock(0, 0);

            //define the triangles 3points
            verts[0].X = -1.0f; verts[0].Y = -1.0f; verts[0].Z = 0.0f;
            verts[0].Color = Color.DarkGoldenrod.ToArgb();

            verts[1].X = 1.0f; verts[1].Y = -1.0f; verts[1].Z = 0.0f;
            verts[1].Color = Color.MediumOrchid.ToArgb();

            verts[2].X = 0.0f; verts[2].Y = 1.0f; verts[2].Z = 0.0f;
            verts[2].Color = Color.Cornsilk.ToArgb();

            vb.Unlock();
        }

        public void LoadTriangle()
        {
            Map.dev.SetTexture(0, null);
            Map.dev.RenderState.Lighting = false;

            SetUpCameraForTriangle();
        }

        public void SetUpCameraForTriangle()
        {
            Map.dev.Transform.View = Matrix.LookAtRH(new Vector3(0.0f, 3.0f, -5.0f),
                new Vector3(0.0f, 0.0f, 0.0f), new Vector3(0.0f, 1.0f, 0.0f));

            Map.dev.Transform.Projection = Matrix.PerspectiveFovRH((float)Math.PI / 4,
                1.0f, 1.0f, 100.0f);
        }

        public void UpdateTriangleMatrices()
        {
            triAngle += 0.01f;
            Map.dev.Transform.World = Matrix.RotationX(triAngle) * Matrix.RotationZ(triAngle);
        }

        public void Draw()
        {
            Map.dev.SetStreamSource(0, vb, 0);
            Map.dev.VertexFormat = CustomVertex.PositionColored.Format;
            Map.dev.DrawPrimitives(PrimitiveType.TriangleStrip, 0, 1);
        }
    }
}
